/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <iostream>
#include "action.h"
#include "object.h"
#include "drawing.h"
#include "action_pl.h"
#include "icons.h"


Action::Action(Drawing *drawing) : the_drawing(drawing)
{
}

Action::~Action()
{
    /* virtualize */
}

Drawing *Action::drawing() const
{
    return the_drawing;
}


NewObjectAction::NewObjectAction(Object *object)
    : Action(object->drawing()), object(object)
{
}

void NewObjectAction::apply()
{
    drawing()->rootContext()->addObject(object);
}

void NewObjectAction::unapply()
{
    drawing()->rootContext()->removeObject(object);
}


MoveHandleAction::MoveHandleAction(ActionPoint *h, const Metrum3 &p)
    : Action(h->object()->drawing()), handle(h), new_pos(p)
{
}

void MoveHandleAction::apply()
{
    old_pos = handle->pos();
    handle->move(new_pos);
}

void MoveHandleAction::unapply()
{
    handle->move(old_pos);
}


DeleteObjectAction::DeleteObjectAction(Object *ob)
    : Action(ob->drawing()), object(ob), context(0), index(0)
{
}

void DeleteObjectAction::apply()
{
    context = object->context();
    index = context->objects().indexOf(object);
    context->removeObject(object);
}

void DeleteObjectAction::unapply()
{
    context->insertObject(index, object);
}


ChangeAttributeAction::ChangeAttributeAction(const Variant &value,
					     Attribute *attribute, Object *ob)
    : Action(ob->drawing()), new_value(value), attribute(attribute), object(ob)
{
}

void ChangeAttributeAction::apply()
{
    old_value = object->attribute(attribute);
    object->setAttribute(new_value, attribute);
}

void ChangeAttributeAction::unapply()
{
    object->setAttribute(old_value, attribute);
}

/****************************************************************************/

ActionGroup::ActionGroup(const List<Action *> &actions, Icon *icon,
			 const String &name, ActionGroup *parent)
    : the_actions(actions),
      the_parent(parent),
      the_drawing(the_parent->drawing()),
      is_applied(false)
{
    the_parent->addChild(this);

    setInterior(&interior);
    setCaption(name);
    setIcon(icon);
}

ActionGroup::ActionGroup(Drawing *drawing, const String &path)
    : the_parent(0),
      the_drawing(drawing),
      is_applied(true)
{
    setInterior(&interior);
    setCaption(_("load file %1").arg(path.substr(path.rfind('/') + 1)));
    setIcon(icons::page);
    setItalic(true);
}

ActionGroup::~ActionGroup()
{
    the_actions.deleteAll();
    the_children.deleteAll();
}

ActionGroup *ActionGroup::parent() const
{
    return the_parent;
}

Drawing *ActionGroup::drawing() const
{
    return the_drawing;
}

List<ActionGroup *> ActionGroup::children() const
{
    return the_children;
}

void ActionGroup::addChild(ActionGroup *child)
{
    if (child->parent() != this)
	return;

    the_children.append(child);
    interior.addItem(child);
}

void ActionGroup::apply()
{
    if (is_applied)
	std::cout << "ActionGroup already applied" << std::endl;

    is_applied = true;

    foreach (Action *action, the_actions)
	action->apply();
}

void ActionGroup::unapply()
{
    if (!is_applied)
	std::cout << "ActionGroup not applied" << std::endl;

    is_applied = false;

    for (int i = the_actions.count() - 1; i >= 0; i--)
	the_actions[i]->unapply();
}

bool ActionGroup::isApplied() const
{
    return is_applied;
}
