/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "attribute.h"
#include "hub.h"
#include "majormode.h"
#include "object.h"
#include "action.h"
#include "icons.h"
#include "drawing.h"
#include "gui/editors.h"
#include "shared.h"


Attribute::Attribute(const String &name, Icon *icon) : name(name), icon(icon)
{
}


AttributeItem::AttributeItem(Hub &hub, Attribute &attribute)
    : hub(hub), attribute(attribute)
{
    setCaption(attribute.name);
    setIcon(attribute.icon);
}

AttributeItem::~AttributeItem()
{
    delete editor();
}

bool AttributeItem::showAttributes(const List<Attribute *> &attributes)
{
    return attributes.contains(&attribute);
}

bool AttributeItem::isSet() const
{
    foreach (Object *ob, hub.majorMode()->selected())
	if (ob->attribute(&attribute) != ob->attributeDefault(&attribute))
	    return true;

    return false;
}

bool AttributeItem::hasDefault() const
{
    foreach (Object *ob, hub.majorMode()->selected())
	if (ob->attributeDefault(&attribute).isNull())
	    return false;

    return true;
}

void AttributeItem::reset()
{
    List<Action *> actions;

    foreach (Object *ob, hub.majorMode()->selected())
	if (ob->attribute(&attribute) != ob->attributeDefault(&attribute))
	    actions << new ChangeAttributeAction(
		ob->attributeDefault(&attribute), &attribute, ob);

    if (actions.empty())
	return;

    hub.drawing()->performActions(
	actions, icons::edit_tag, _("Reset attribute"));
}

Variant AttributeItem::value() const
{
    bool first = true;
    Variant value;

    foreach (Object *ob, hub.majorMode()->selected()) {
	bool contained = false;
	foreach (Attribute *attr, ob->attributes())
	    if (attr == &attribute)
		contained = true;
	if (contained) {
	    if (first) {
		value = ob->attribute(&attribute);
		first = false;
	    } else if (ob->attribute(&attribute) != value)
		return Variant();
	}
    }

    return value;
}

void AttributeItem::setValue(const Variant &value)
{
    List<Action *> actions;

    foreach (Object *ob, hub.majorMode()->selected())
	if (ob->attribute(&attribute) != value)
	    actions << new ChangeAttributeAction(value, &attribute, ob);

    if (actions.empty())
	return;

    hub.drawing()->performActions(
	actions, icons::edit_tag, _("Change attribute"));
}

/****************************************************************************/

StringAttributeItem::StringAttributeItem(Hub &hub, Attribute &attribute)
    : AttributeItem(hub, attribute)
{
    setEditor(new StringItemEditor(*this));
}

String StringAttributeItem::text() const
{
    Variant v(value());
    return v.isNull() ? String() : v.toString();
}

void StringAttributeItem::setText(const String &value)
{
    setValue(Variant::fromString(value));
}

/****************************************************************************/

SpinboxAttributeItem::SpinboxAttributeItem(Hub &hub, Attribute &attribute)
    : AttributeItem(hub, attribute)
{
    setEditor(new SpinboxItemEditor(*this));
}


IntegerAttributeItem::IntegerAttributeItem(Hub &hub, Attribute &attribute)
    : SpinboxAttributeItem(hub, attribute)
{
}

String IntegerAttributeItem::repr() const
{
    Variant v(value());
    return v.isNull() ? String() : int_to_str(v.toInt());
}

void IntegerAttributeItem::setRepr(const String &str)
{
    bool ok;
    int value = str_to_int(str, ok);
    if (ok)
	setValue(Variant::fromInt(value));
}


MetrumAttributeItem::MetrumAttributeItem(Hub &hub, Attribute &attribute)
    : SpinboxAttributeItem(hub, attribute)
{
}

String MetrumAttributeItem::repr() const
{
    Variant v(value());
    return v.isNull() ? String() : int_to_str(v.toMetrum());
}

void MetrumAttributeItem::setRepr(const String &str)
{
    bool ok;
    int64_t value = str_to_int(str, ok);
    if (ok)
	setValue(Variant::fromMetrum(value));
}


DoubleAttributeItem::DoubleAttributeItem(Hub &hub, Attribute &attribute)
    : SpinboxAttributeItem(hub, attribute)
{
}

String DoubleAttributeItem::repr() const
{
    Variant v(value());
    return v.isNull() ? String() : double_to_str(v.toDouble());
}

void DoubleAttributeItem::setRepr(const String &str)
{
    bool ok;
    double value = str_to_double(str, ok);
    if (ok)
	setValue(Variant::fromDouble(value));
}

/****************************************************************************/

SelectorAttributeItem::SelectorAttributeItem(Hub &hub, Attribute &attribute,
					     ItemInterior &choices)
    : AttributeItem(hub, attribute), choices(choices)
{
    setEditor(new SelectorItemEditor(*this));
}


EnumAttributeItem::EnumAttributeItem(Hub &hub, Attribute &attribute,
				     ItemInterior &choices)
    : SelectorAttributeItem(hub, attribute, choices)
{
}

Item *EnumAttributeItem::selectedItem()
{
    Variant v(value());
    return v.isNull() ? 0 : choices.cachedItem(v.toInt());
}

void EnumAttributeItem::setSelectedItem(Item *value)
{
    setValue(Variant::fromInt(choices.cachedIndexOf(value)));
}


BoolAttributeItem::BoolAttributeItem(Hub &hub, Attribute &attribute)
    : SelectorAttributeItem(hub, attribute, shared.booleanEnum())
{
}

Item *BoolAttributeItem::selectedItem()
{
    Variant v(value());
    return v.isNull() ? 0 : choices.cachedItem(v.toBool());
}

void BoolAttributeItem::setSelectedItem(Item *value)
{
    setValue(Variant::fromBool(choices.cachedIndexOf(value)));
}
