/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <err.h>
#include <assert.h>

#include "dirmodel.h"
#include "util/path.h"
#include "util/c++11.h"
#include "filemonitor.h"
#include "icons.h"
#include "sch_readsch.h"
#include "read_xml.h"
#include "read_z.h"
#include "xml_writer.h"


String file_name_part(const String &path)
{
    return path.split('/').last();
}

/****************************************************************************/

void Library::addComponent(const String &name, Drawing *drawing)
{
    components[name] = drawing;
}

Drawing *Library::component(const String &name)
{
    if (components.find(name) == components.end())
	return 0;

    return components[name];
}

/****************************************************************************/

Font1::Font1() : unknown(0), space(0)
{
}

void Font1::addCharacter(Drawing *drawing)
{
    std::wstring letter = drawing->letterCharacter().decode();

    if (letter.length() != 1) {
	if (letter.empty())
	    unknown = drawing;
	return;
    }

    if (drawing->letterSpace()) {
	if (letter[0] == L'_')
	    space = drawing;
	return;
    }

    letters[letter[0]] = drawing;
}

Drawing *Font1::character(wchar_t ch)
{
    if (ch == L' ')
	return space;

    if (letters.find(ch) == letters.end())
	return unknown;

    return letters[ch];
}

/****************************************************************************/

RootInterior::~RootInterior()
{
    items.deleteAll();
}

Item *RootInterior::item(unsigned int index) const
{
    return items[index];
}

unsigned int RootInterior::indexOf(Item *item) const
{
    return items.indexOf(static_cast<ProjectDirItem *>(item));
}

void RootInterior::replacePaths(unsigned int index, unsigned int count,
				const List<String> &paths)
{
    assert(items.count() >= index + count);

    List<ProjectDirItem *> new_items;
    foreach (String path, paths)
	new_items.append(new ProjectDirItem(sanitize_path(path)));

    items.deleteRange(index, count);
    items.replace(index, count, new_items);
    replace(index, count, paths.count());
}

/****************************************************************************/

class DirectoryInteriorListener final : public FileMonitorListener
{
    DirectoryInterior *const interior;
public:
    DirectoryInteriorListener(DirectoryInterior *interior);
    void regularEvent(const String &name, FAMCodes code);
};

DirectoryInteriorListener::DirectoryInteriorListener(
    DirectoryInterior *interior) : interior(interior)
{
}

void DirectoryInteriorListener::regularEvent(const String &name, FAMCodes code)
{
    switch (code) {
    case FAMCreated:
	interior->addEntry(name);
	break;
    case FAMChanged:
	interior->updateEntry(name);
	break;
    case FAMDeleted:
	interior->removeEntry(name);
	break;
    default:
	/* ignore */;
    }
}

/****************************************************************************/

DirectoryInterior::DirectoryInterior(const String &path, bool load_all)
    : path(path)
{
    DIR *dirp;
    struct dirent *dirent;

    if ((dirp = opendir(path.data())) == NULL)
	err(EXIT_FAILURE, "opendir");

    /* TODO:	EACCES	Permission denied.
		EMFILE	Too many file descriptors in use by process.
		ENFILE	Too many files are currently open in the system.
		ENOENT	Directory does not exist, or name is an empty string.
		ENOTDIR	name is not a directory. */

    errno = 0;
    while ((dirent = readdir(dirp)) != NULL) {
	String name(dirent->d_name);
	if (!name.startsWith('.') && !name.endsWith('~'))
	    entries.append(name);
	errno = 0;
    }

    if (errno != 0)
	err(EXIT_FAILURE, "readdir");

    if (closedir(dirp) == -1)
	err(EXIT_FAILURE, "closedir");

    std::sort(entries.begin(), entries.end());

    items = List<Item *>::zero(entries.count());
    replace(0, 0, entries.count());

    listener = new DirectoryInteriorListener(this);
    shared.fileMonitor()->addListener(listener, path);

    if (load_all)
	for (unsigned int i = 0; i < entries.count(); i++)
	    item(i);
}

DirectoryInterior::~DirectoryInterior()
{
    shared.fileMonitor()->removeListener(listener);
    delete listener;
}

Item *DirectoryInterior::item(unsigned int index) const
{
    assert(index < entries.count());

    Item *&item = items[index];
    if (!item) {
	String p = path + "/" + entries[index];
	struct stat buf;

	if (lstat(p.data(), &buf) == -1)
	    err(EXIT_FAILURE, "lstat");

	/* TODO: EACCES, EFAULT, ELOOP, ENAMETOOLONG, ENOENT, ENOTDIR,
		 EOVERFLOW */

	if (S_ISDIR(buf.st_mode))
	    item = shared.directoryItem(path + "/" + entries[index]);
	else
	    item = shared.fileItem(path + "/" + entries[index]);
    }
    return item;
}

unsigned int DirectoryInterior::indexOf(Item *item) const
{
    for (unsigned int i = 0; i < entries.count(); i++)
	if (this->item(i) == item)
	    return i;

    return -1;
}

void DirectoryInterior::addEntry(const String &name)
{
    List<String>::iterator i = std::upper_bound(entries.begin(),
						entries.end(), name);
    unsigned int pos = i - entries.begin();

    entries.insert(i, name);
    items.insert(pos, 0);
    replace(pos, 0, 1);
}

void DirectoryInterior::updateEntry(const String &name)
{
    int index = entries.indexOf(name);
    if (index == -1)
	return;

    items[index] = 0;
    replace(index, 1, 1);
}

void DirectoryInterior::removeEntry(const String &name)
{
    int index = entries.indexOf(name);
    if (index == -1)
	return;

    entries.removeAt(index);
    items.removeAt(index);
    replace(index, 1, 0);
}

/****************************************************************************/

FileInterior::~FileInterior()
{
    _items.deleteAll();
}

Item *FileInterior::item(unsigned int index) const
{
    return _items[index];
}

unsigned int FileInterior::indexOf(Item *item) const
{
    return _items.indexOf(static_cast<VirtualItem *>(item));
}

List<VirtualItem *> FileInterior::items() const
{
    return _items;
}

void FileInterior::addItem(VirtualItem *item)
{
    _items.append(item);
    replace(_items.count() - 1, 0, 1);
}

/****************************************************************************/

Drawing *Item3::drawing()
{
    return 0;
}

void Item3::save(const String &path)
{
    fprintf(stderr, "Can't 'save' node at '%s' to '%s'\n",
	    this->path().data(), path.data());
}

/****************************************************************************/

ProjectDirItem::ProjectDirItem(const String &path) : _path(path)
{
    String home = home_dir();

    if (path == home)
	setCaption("~");
    else if (path.startsWith(home + "/"))
	setCaption("~" + path.mid(home.length()));
    else
	setCaption(path);

    setInterior(shared.directoryItem(path)->interior());
    setIcon(icons::folder);
}

String ProjectDirItem::path() const
{
    return _path;
}

/****************************************************************************/

DirectoryItem::DirectoryItem(const String &path, bool load_all) : _path(path)
{
    setInterior(new DirectoryInterior(path, load_all));
    setCaption(file_name_part(path));
    setIcon(icons::folder);
}

DirectoryItem::~DirectoryItem()
{
    delete interior();
}

String DirectoryItem::path() const
{
    return _path;
}

/****************************************************************************/

FileItem::FileItem(const String &path, Library *library, Font1 *font)
    : _path(path), library(library), font(font),
      virtual_item(0), backed_up(false)
{
    setCaption(file_name_part(path));
    if (path.endsWith(".sch") || path.endsWith(".sch.gz") ||
	path.endsWith(".sym") || path.endsWith(".sym.gz") ||
	path.endsWith(".xml") || path.endsWith(".xml.gz"))
	setIcon(icons::page);
    else
	setIcon(icons::page_blue);

    if (library || font)
	load();
}

FileItem::~FileItem()
{
    if (virtual_item)
	delete virtual_item;
}

void FileItem::load()
{
    if (virtual_item)
	return;

    String str = path();
    bool is_zipped = str.endsWith(".gz");

    if (is_zipped)
	str = str.substr(0, str.length() - 3);

    Shared::Sink *sink;
    if (str.endsWith(".sch"))
	sink = new_schematic_sink(Drawing::SchematicFile,
				  library, font, this, messages);
    else if (str.endsWith(".sym"))
	sink = new_schematic_sink(Drawing::SymbolFile,
				  library, font, this, messages);
    else
	sink = new_xml_sink(library, font, this, messages);

    if (is_zipped)
	sink = new_z_sink(sink, true, messages);

    shared.readFile(path(), sink, messages);
}

void FileItem::assignVirtualItem(VirtualItem *item)
{
    /* The parser does not accept multiple document elements.
     * So if there is more than one element, something is wrong. */
    assert(!virtual_item);

    virtual_item = item;
    setIcon(item->icon());
    setInterior(item->interior());
}

void FileItem::updateModificationState()
{
    setBold(virtual_item->isModified());
}

Font1 *FileItem::thisFont()
{
    if (!virtual_item)
	load();

    if (!virtual_item)
	return 0;

    return virtual_item->thisFont();
}

String FileItem::path() const
{
    return _path;
}

Drawing *FileItem::drawing()
{
    if (!virtual_item)
	load();

    if (!virtual_item)
	return 0;

    return virtual_item->drawing();
}

void FileItem::save(const String &path)
{
    if (!virtual_item)
	load();

    if (!virtual_item)
	return;

    String file_path = path.isEmpty() ? _path : path;
    String backup_path = file_path + "~";

    if (!backed_up) {
    /**	int ret = QMessageBox::Save;

	do { **/
	    struct stat buf;
	    bool exists;

	    if (lstat(backup_path.data(), &buf) != -1)
		exists = true;
	    else if (errno == ENOENT)
		exists = false;
	    else
		err(EXIT_FAILURE, "lstat");

	    /* TODO: EACCES, EFAULT, ELOOP, ENAMETOOLONG, ENOTDIR,
		     EOVERFLOW */

	    if (exists && unlink(backup_path.data()) == -1)
	    /**	ret = QMessageBox::warning(
		    0, path.isEmpty() ? _("Save") : _("Save As"),
		    _("The previous backup file could not be removed:\n%1\n\n"
		      "If you continue to save the file anyway, no backup "
		      "will be made.").arg(strerror(errno)),
		    QMessageBox::Save | QMessageBox::Retry |
		    QMessageBox::Cancel, QMessageBox::Save); **/;
	    else {
		if (lstat(file_path.data(), &buf) != -1)
		    exists = true;
		else if (errno == ENOENT)
		    exists = false;
		else
		    err(EXIT_FAILURE, "lstat");

		/* TODO: EACCES, EFAULT, ELOOP, ENAMETOOLONG, ENOTDIR,
			 EOVERFLOW */

		if (exists && rename(file_path.data(),
				     backup_path.data()) == -1)
		/** ret = QMessageBox::warning(
			0, path.isEmpty() ? _("Save") : _("Save As"),
			_("The previous file could not be renamed:\n%1\n\n"
			  "If you continue to save the file anyway, "
			  "no backup will be made.").arg(strerror(errno)),
			QMessageBox::Save | QMessageBox::Retry |
			QMessageBox::Cancel, QMessageBox::Save); **/;
	    }
    /**	} while (ret == QMessageBox::Retry);

	if (ret != QMessageBox::Save)
	    return; **/

	backed_up = true;
    }

    int fd;

    while ((fd = creat(file_path.data(), 0666)) == -1)
	err(EXIT_FAILURE, "creat: %s", file_path.data());
    /**	if (QMessageBox::critical(
		0, path.isEmpty() ? _("Save") : _("Save As"),
		_("The file could not be written:\n%1").arg(strerror(errno)),
		QMessageBox::Retry | QMessageBox::Cancel,
		QMessageBox::Cancel) == QMessageBox::Cancel)
	    return; **/

    XMLWriter writer(fd);
    virtual_item->write(&writer, file_path == _path);
}

/****************************************************************************/

VirtualItem::VirtualItem(FileItem *fileItem) : file_item(fileItem)
{
    setCaption(_("Page"));
}

FileItem *VirtualItem::fileItem() const
{
    return file_item;
}

String VirtualItem::path() const
{
    return file_item->path();
}

void VirtualItem::save(const String &path)
{
    file_item->save(path);
}

Font1 *VirtualItem::thisFont() const
{
    return 0;
}

/****************************************************************************/

ContainerItem::ContainerItem(FileItem *fileItem) : VirtualItem(fileItem)
{
    setIcon(icons::pages);
    setInterior(&interior);
}

bool ContainerItem::isModified()
{
    foreach (VirtualItem *item, interior.items())
	if (item->isModified())
	    return true;

    return false;
}

void ContainerItem::write(XMLWriter *writer, bool set_saved)
{
    writer->startElement("container");
    writer->writeAttribute("xmlns", "https://hedmen.org/xi/");
    writeInterior(writer, set_saved);
    writer->endElement();
}

void ContainerItem::writeInterior(XMLWriter *writer, bool set_saved)
{
    foreach (VirtualItem *item, interior.items())
	item->write(writer, set_saved);
}

/****************************************************************************/

FontItem::FontItem(FileItem *fileItem) : ContainerItem(fileItem),
					 font(new Font1)
{
}

FontItem::~FontItem()
{
    delete font;
}

void FontItem::write(XMLWriter *writer, bool set_saved)
{
    writer->startElement("font");
    // writer->writeAttribute("xmlns", "");
    writeInterior(writer, set_saved);
    writer->endElement();
}

Font1 *FontItem::thisFont() const
{
    return font;
}

/****************************************************************************/

DrawingItem::DrawingItem(Drawing::FileType fileType,
			 Library *library, FileItem *fileItem)
    : VirtualItem(fileItem), is_modified(false)
{
    setIcon(icons::page);

    _drawing = new Drawing(fileType, this);

    if (library) {
	if (drawing()->fileType() != Drawing::SymbolFile)
	    fprintf(stderr, "Ignoring non-symbol node in library at '%s'\n",
		    path().data());
	else {
	    String base_name = file_name_part(fileItem->path());
	    if (base_name.endsWith(".xml"))
		base_name = base_name.left(base_name.length() - 4);
	    library->addComponent(base_name, drawing());
	}
    }
}

DrawingItem::~DrawingItem()
{
    delete _drawing;
}

void DrawingItem::updateModificationState()
{
    if (drawing()->isModified() == is_modified)
	return;

    is_modified = drawing()->isModified();
    setBold(is_modified);
    fileItem()->updateModificationState();
}

Drawing *DrawingItem::drawing()
{
    return _drawing;
}

bool DrawingItem::isModified()
{
    return is_modified;
}

void DrawingItem::write(XMLWriter *writer, bool set_saved)
{
    drawing()->write(writer);
    if (set_saved)
	drawing()->setLastSaved();
}
