/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#ifndef DIRMODEL_H
#define DIRMODEL_H

#include <map>
#include "drawing.h"
#include "messages.h"

class ProjectDirItem;
class DirectoryInteriorListener; /* internal */
class VirtualItem;


class Library
{
    std::map<String, Drawing *> components;
public:
    void addComponent(const String &name, Drawing *drawing);
    Drawing *component(const String &name);
};

class Font1
{
    std::map<wchar_t, Drawing *> letters;
    Drawing *unknown, *space;
public:
    Font1();
    void addCharacter(Drawing *drawing);
    Drawing *character(wchar_t ch);
};


class RootInterior : public ItemInterior
{
    List<ProjectDirItem *> items;
public:
    ~RootInterior();
    void replacePaths(unsigned int index, unsigned int count,
		      const List<String> &paths);
protected:
    Item *item(unsigned int index) const;
    unsigned int indexOf(Item *item) const;
};

class DirectoryInterior : public ItemInterior
{
    String const path;
    List<String> entries;
    mutable List<Item *> items;
    DirectoryInteriorListener *listener;
public:
    DirectoryInterior(const String &path, bool load_all);
    ~DirectoryInterior();

    void addEntry(const String &name);
    void updateEntry(const String &name);
    void removeEntry(const String &name);
protected:
    Item *item(unsigned int index) const;
    unsigned int indexOf(Item *item) const;
};

class FileInterior : public ItemInterior
{
    List<VirtualItem *> _items;
public:
    ~FileInterior();
    List<VirtualItem *> items() const;
    void addItem(VirtualItem *item);
protected:
    Item *item(unsigned int index) const;
    unsigned int indexOf(Item *item) const;
};


class Item3 : public Item
{
public:
    virtual String path() const = 0;
    virtual Drawing *drawing();
    virtual void save(const String &path = String());
    Messages messages;
};

class ProjectDirItem : public Item3
{
    String const _path;
public:
    ProjectDirItem(const String &path);
    String path() const;
};

class DirectoryItem : public Item3
{
    String const _path;
public:
    DirectoryItem(const String &path, bool load_all);
    ~DirectoryItem();
    String path() const;
};

class FileItem : public Item3
{
    String const _path;
    Library *const library;
    Font1 *const font;
    VirtualItem *virtual_item;
    bool backed_up;
public:
    FileItem(const String &path, Library *library, Font1 *font);
    ~FileItem();
    void load();
    void assignVirtualItem(VirtualItem *item);
    void updateModificationState();
    Font1 *thisFont();

    String path() const;
    Drawing *drawing();
    void save(const String &path);
};

class VirtualItem : public Item3
{
    FileItem *const file_item;
public:
    VirtualItem(FileItem *fileItem);
    FileItem *fileItem() const;

    String path() const;
    void save(const String &path);

    virtual bool isModified() = 0;
    virtual void write(XMLWriter *writer, bool set_saved) = 0;
    virtual Font1 *thisFont() const;
};

class ContainerItem : public VirtualItem
{
public:
    ContainerItem(FileItem *fileItem);
    FileInterior interior;
    bool isModified();
    void write(XMLWriter *writer, bool set_saved);
protected:
    void writeInterior(XMLWriter *writer, bool set_saved);
};

class FontItem : public ContainerItem
{
    Font1 *const font;
public:
    FontItem(FileItem *fileItem);
    ~FontItem();
    void write(XMLWriter *writer, bool set_saved);
    Font1 *thisFont() const;
};

class DrawingItem : public VirtualItem
{
    Drawing *_drawing;
    bool is_modified;
public:
    DrawingItem(Drawing::FileType fileType, Library *library,
		FileItem *fileItem);
    ~DrawingItem();
    void updateModificationState();

    Drawing *drawing();
    bool isModified();
    void write(XMLWriter *writer, bool set_saved);
};

#endif
