/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <math.h>
#include "drawing.h"
#include "object.h"
#include "dirmodel.h"
#include "action.h"
#include "xml_writer.h"
#include "sch_attribute.h"
#include "shared.h"


Drawing::Drawing(FileType fileType, DrawingItem *item)
    : file_type(fileType), item(item), letter_width(0), letter_space(false),
      the_base(1), base_base(10), base_exponent(0),
      root_context(new ObjectContext(this)),
      the_root(new ActionGroup(this, item->path())),
      most_recent(the_root), last_saved(the_root),
      historyInterior(List<Item *>() << the_root)
{
    chain.append(the_root);
}

Drawing::~Drawing()
{
    delete the_root;
    all_objects.deleteAll();
    delete root_context;
}

void Drawing::write(XMLWriter *writer, Drawing *embedded_into)
{
    if (embedded_into)
	writer->startElement("embedded");
    else
	switch (file_type) {
	case GenericFile:
	    writer->startElement("file");
	    writer->writeAttribute("units", "metric");
	    writer->writeAttribute("base", int64_to_str(the_base));
	    writer->writeAttribute("xmlns", "https://hedmen.org/xi/");
	    break;
	case SchematicFile:
	    writer->startElement("schematic");
	    // writer->writeAttribute("xmlns", "");
	    break;
	case SymbolFile:
	    writer->startElement("symbol");
	    // writer->writeAttribute("xmlns", "");
	    break;
	case LetterFile:
	    writer->startElement("letter");
	    writer->writeAttribute("character", letter_character);
	    writer->writeAttribute("width", format(letter_width));
	    writer->writeAttribute("space",
		SchematicEnum::boolean[letter_space]);
	    // writer->writeAttribute("xmlns", "");
	    break;
	case NeverballFile:
	    writer->startElement("map");
	    writer->writeAttribute("text", map_text);
	    // writer->writeAttribute("xmlns", "");
	    break;
	}

    rootContext()->write(writer);
    writer->endElement();
}

Drawing::FileType Drawing::fileType() const
{
    return file_type;
}

String Drawing::letterCharacter() const
{
    return letter_character;
}

void Drawing::setLetterCharacter(const String &letterCharacter)
{
    letter_character = letterCharacter;
}

Metrum Drawing::letterWidth() const
{
    return letter_width;
}

void Drawing::setLetterWidth(Metrum letterWidth)
{
    letter_width = letterWidth;
}

bool Drawing::letterSpace() const
{
    return letter_space;
}

void Drawing::setLetterSpace(bool letterSpace)
{
    letter_space = letterSpace;
}

void Drawing::setMapText(const String &mapText)
{
    map_text = mapText;
}

int64_t Drawing::base() const
{
    return the_base;
}

int Drawing::baseBase() const
{
    return base_base;
}

void Drawing::setBaseBase(int baseBase)
{
    if (baseBase == base_base)
	return;

    base_base = baseBase;
    the_base = 1;
    for (int i = 0; i < base_exponent; i++)
	the_base *= base_base;
}

int Drawing::baseExponent() const
{
    return base_exponent;
}

void Drawing::setBaseExponent(int baseExponent)
{
    if (baseExponent == base_exponent)
	return;

    base_exponent = baseExponent;
    the_base = 1;
    for (int i = 0; i < base_exponent; i++)
	the_base *= base_base;
}

Metrum Drawing::parseMetrum(const String &val, bool &ok) const
{
    int i = val.indexOf(":");
    int64_t p;

    if (i == -1) {
	p = str_to_int64(val, ok);
	if (!ok)
	    p = llrint(str_to_double(val, ok) * the_base);
    } else {
	p = str_to_int64(val.left(i), ok);
	if (ok && str_to_int64(val.mid(i + 1), ok) != the_base)
	    ok = false;
    }
    return ok ? p : 0;
}

Metrum3 Drawing::parseMetrum3(const String &val, bool &ok) const
{
    int i = val.indexOf(", ");
    Metrum xp(0), yp(0), zp(0);

    if (i == -1) {
	int i = val.indexOf(" ");
	xp = parseMetrum(val.left(i), ok);
	if (!ok)
	    return Metrum3();

	String rem = val.mid(i + 1);
	int j = rem.indexOf(" ");

	if (j == -1)
	    yp = parseMetrum(rem, ok);
	else {
	    yp = parseMetrum(rem.left(j), ok);
	    if (!ok)
		return Metrum3();
	    zp = parseMetrum(rem.mid(j + 1), ok);
	}
	return ok ? Metrum3(xp, yp, zp) : Metrum3();
    }

    xp = parseMetrum(val.left(i), ok);
    if (!ok)
	return Metrum3();

    String rem = val.mid(i + 2);
    int j = rem.indexOf(", ");

    if (j == -1)
	yp = parseMetrum(rem, ok);
    else {
	yp = parseMetrum(rem.left(j), ok);
	if (!ok)
	    return Metrum3();
	zp = parseMetrum(rem.mid(j + 2), ok);
    }
    return ok ? Metrum3(xp, yp, zp) : Metrum3();
}

Metrum Drawing::scale(double eff) const
{
    return llrint(eff * the_base);
}

Metrum3 Drawing::scale(double xeff, double yeff, double zeff) const
{
    return Metrum3(llrint(xeff * the_base),
		   llrint(yeff * the_base),
		   llrint(zeff * the_base));
}

double Drawing::eff(Metrum m) const
{
    return (double)m / the_base;
}

String Drawing::format(Metrum m) const
{
    return int64_to_str(m) + ":" + int64_to_str(the_base);
}

String Drawing::format(const Metrum3 &m) const
{
    return format(m.x()) + ", " + format(m.y()) + ", " + format(m.z());
}

ObjectContext *Drawing::rootContext() const
{
    return root_context;
}

List<Object *> Drawing::objects() const
{
    return the_objects;
}

void Drawing::walk_context(ObjectContext *context)
{
    foreach (Object *ob, context->objects()) {
	the_objects.append(ob);
	if (ob->children())
	    walk_context(ob->children());
    }
}

void Drawing::updateObjectList()
{
    the_objects.clear();
    walk_context(rootContext());
}

void Drawing::registerObject(Object *object)
{
    assert(!all_objects.contains(object));
    all_objects.append(object);
}

void Drawing::unregisterObject(Object *object)
{
    assert(all_objects.contains(object));
    all_objects.removeAll(object);
}

ActionGroup *Drawing::head() const
{
    return chain.last();
}

ActionGroup *Drawing::root() const
{
    return the_root;
}

void Drawing::seek(ActionGroup *ag)
{
    List<ActionGroup *> stack;
    bool was_modified = isModified();

    while (!chain.contains(ag)) {
	stack.append(ag);
	ag = ag->parent();
    }

    while (head() != ag)
	chain.takeLast()->unapply();

    while (!stack.empty()) {
	ag = stack.takeLast();
	chain.append(ag);
	ag->apply();
    }

    shared.drawingStatusChanged(this);

    if (isModified() != was_modified)
	item->updateModificationState();
}

void Drawing::performActions(const List<Action *> &actions,
			     Icon *icon, const String &name)
{
    if (actions.isEmpty())
	return;

    most_recent = new ActionGroup(actions, icon, name, head());
    seek(most_recent);
}

bool Drawing::isModified() const
{
    return head() != last_saved;
}

ActionGroup *Drawing::mostRecent() const
{
    return most_recent;
}

ActionGroup *Drawing::lastSaved() const
{
    return last_saved;
}

void Drawing::setLastSaved()
{
    if (last_saved == head())
	return;

    last_saved = head();
    item->updateModificationState();
}
