/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <math.h>
#include "gen_rectangular.h"
#include "hub.h"
#include "icons.h"
#include "drawing.h"
#include "action.h"
#include "painter.h"
#include "view.h"


RectangularMajorMode::RectangularMajorMode(Hub &hub)
    : MajorMode(hub), startMinorMode(hub, *this),
		      endMinorMode(hub, *this)
{
    setMinorMode(&startMinorMode);
}

void RectangularMajorMode::accept(MinorMode *minor_mode)
{
    if (minor_mode == &startMinorMode)
	setMinorMode(&endMinorMode);

    if (minor_mode == &endMinorMode) {
	Object *object = buildObject(startMinorMode.point(),
				     endMinorMode.point());
	hub.drawing()->performActions(
	    List<Action *>() << new NewObjectAction(object),
	    icons::add, _("New %1").arg(object->name()));
	cancel();
    }
}

void RectangularMajorMode::cancel()
{
    setMinorMode(&startMinorMode, true);
}

void RectangularMajorMode::paint(Painter &painter)
{
    MajorMode::paint(painter);

    painter.setPen(0x7f, 0x7f, 0x7f);
    if (minorMode() == &endMinorMode && hub.pointerActive()) {
	painter.paintLine(startMinorMode.point(),
			  Metrum3(startMinorMode.point().x(),
				  endMinorMode.point().y(),
				  endMinorMode.point().z()));
	painter.paintLine(Metrum3(startMinorMode.point().x(),
				  endMinorMode.point().y(),
				  endMinorMode.point().z()),
			  endMinorMode.point());
	painter.paintLine(startMinorMode.point(),
			  Metrum3(endMinorMode.point().x(),
				  startMinorMode.point().y(),
				  startMinorMode.point().z()));
	painter.paintLine(Metrum3(endMinorMode.point().x(),
				  startMinorMode.point().y(),
				  startMinorMode.point().z()),
			  endMinorMode.point());
    }

    painter.setBlackPen();
    if (minorMode() == &endMinorMode || hub.pointerActive())
	painter.paintPoint(startMinorMode.point());
    if (minorMode() == &endMinorMode && hub.pointerActive())
	painter.paintPoint(endMinorMode.point());
}

List<ActionLine *> RectangularMajorMode::lines(ActionPoint *handle)
{
    return List<ActionLine *>();
}

/****************************************************************************/

RectangularObject::RectangularObject(const Metrum3 &p0,
				     const Metrum3 &p1, Drawing *drawing)
    : Object(drawing), p0(p0), p1(p1),
      ssHandle(_("corner"), this),
      smHandle(_("side midpoint"), this),
      seHandle(_("corner"), this),
      msHandle(_("side midpoint"), this),
      mmHandle(_("center"), this),
      meHandle(_("side midpoint"), this),
      esHandle(_("corner"), this),
      emHandle(_("side midpoint"), this),
      eeHandle(_("corner"), this)
{
}

void RectangularObject::paint(Painter &painter, const Embedding &embedding)
{
    painter.setBlackPen();
    painter.paintLine(Metrum3(embedding.transform(p0).x(),
			      embedding.transform(p1).y(),
			      embedding.transform(p0).z()),
		      embedding.transform(p1));
    painter.paintLine(Metrum3(embedding.transform(p1).x(),
			      embedding.transform(p0).y(),
			      embedding.transform(p0).z()),
		      embedding.transform(p1));
    painter.paintLine(embedding.transform(p0),
		      Metrum3(embedding.transform(p0).x(),
			      embedding.transform(p1).y(),
			      embedding.transform(p1).z()));
    painter.paintLine(embedding.transform(p0),
		      Metrum3(embedding.transform(p1).x(),
			      embedding.transform(p0).y(),
			      embedding.transform(p1).z()));
}

void RectangularObject::paintShadow(const Metrum3 &pos,
				    const ActionPoint *handle, Painter &painter)
{
    Metrum3 hp0(p0), hp1(p1);

    if (handle == &ssHandle) {
	hp0.rx() = pos.x();
	hp0.ry() = pos.y();
    }
    if (handle == &smHandle)
	hp0.rx() = pos.x();
    if (handle == &seHandle) {
	hp0.rx() = pos.x();
	hp1.ry() = pos.y();
    }
    if (handle == &msHandle)
	hp0.ry() = pos.y();
    if (handle == &mmHandle) {
	Metrum3 delta0 = p0 - (p0 + p1) / 2;
	Metrum3 delta1 = p1 - (p0 + p1) / 2;
	hp0 = pos + delta0;
	hp1 = pos + delta1;
    }
    if (handle == &meHandle)
	hp1.ry() = pos.y();
    if (handle == &esHandle) {
	hp1.rx() = pos.x();
	hp0.ry() = pos.y();
    }
    if (handle == &emHandle)
	hp1.rx() = pos.x();
    if (handle == &eeHandle) {
	hp1.rx() = pos.x();
	hp1.ry() = pos.y();
    }

    painter.setPen(0x80, 0x80, 0x80);
    painter.paintLine(Metrum3(hp0.x(), hp1.y(), hp0.z()), hp1);
    painter.paintLine(Metrum3(hp1.x(), hp0.y(), hp0.z()), hp1);
    painter.paintLine(hp0, Metrum3(hp0.x(), hp1.y(), hp1.z()));
    painter.paintLine(hp0, Metrum3(hp1.x(), hp0.y(), hp1.z()));
}

List<ActionPoint *> RectangularObject::handles()
{
    return List<ActionPoint *>() << &ssHandle << &smHandle << &seHandle
				 << &msHandle << &mmHandle << &meHandle
				 << &esHandle << &emHandle << &eeHandle;
}

Metrum3 RectangularObject::handlePos(const ActionPoint *handle) const
{
    if (handle == &ssHandle)
	return p0;
    if (handle == &smHandle)
	return Metrum3(p0.x(), (p0.y() + p1.y()) / 2, p0.z());
    if (handle == &seHandle)
	return Metrum3(p0.x(), p1.y(), p0.z());
    if (handle == &msHandle)
	return Metrum3((p0.x() + p1.x()) / 2, p0.y(), p0.z());
    if (handle == &mmHandle)
	return (p0 + p1) / 2;
    if (handle == &meHandle)
	return Metrum3((p0.x() + p1.x()) / 2, p1.y(), p0.z());
    if (handle == &esHandle)
	return Metrum3(p1.x(), p0.y(), p0.z());
    if (handle == &emHandle)
	return Metrum3(p1.x(), (p0.y() + p1.y()) / 2, p0.z());
    if (handle == &eeHandle)
	return p1;

    return Metrum3();
}

void RectangularObject::moveHandle(const Metrum3 &pos,
				   const ActionPoint *handle)
{
    if (handle == &ssHandle) {
	p0.rx() = pos.x();
	p0.ry() = pos.y();
    }
    if (handle == &smHandle)
	p0.rx() = pos.x();
    if (handle == &seHandle) {
	p0.rx() = pos.x();
	p1.ry() = pos.y();
    }
    if (handle == &msHandle)
	p0.ry() = pos.y();
    if (handle == &mmHandle) {
	Metrum3 delta0 = p0 - (p0 + p1) / 2;
	Metrum3 delta1 = p1 - (p0 + p1) / 2;
	p0 = pos + delta0;
	p1 = pos + delta1;
    }
    if (handle == &meHandle)
	p1.ry() = pos.y();
    if (handle == &esHandle) {
	p1.rx() = pos.x();
	p0.ry() = pos.y();
    }
    if (handle == &emHandle)
	p1.rx() = pos.x();
    if (handle == &eeHandle) {
	p1.rx() = pos.x();
	p1.ry() = pos.y();
    }
}

static void min_squared_point_origin_dist(
    ScreenDist &result, const ScreenPos &point, bool override)
{
    ScreenDist v = point.x() * point.x() + point.y() * point.y();
    if (v < result || override)
	result = v;
}

static void min_squared_line_origin_dist(
    ScreenDist &result, const ScreenPos &base, const ScreenPos &tangent)
{
    ScreenDist a = base.x() * tangent.x() + base.y() * tangent.y();
    ScreenDist b = tangent.x() * tangent.x() + tangent.y() * tangent.y();
    if (a < -b || a > 0)
	return;

    ScreenDist c = base.x() * tangent.y() - base.y() * tangent.x();
    ScreenDist d = c * c / b;
    if (d < result)
	result = d;
}

ScreenDist RectangularObject::proximity(const ScreenPos &point, View *view,
					const Embedding &embedding) const
{
    ScreenPos r00 = view->map(embedding.transform(p0)) - point;
    ScreenPos r01 = view->map(embedding.transform(
				  Metrum3(p0.x(), p1.y(), p1.z()))) - point;
    ScreenPos r10 = view->map(embedding.transform(
				  Metrum3(p1.x(), p0.y(), p0.z()))) - point;
    ScreenPos r11 = view->map(embedding.transform(p1)) - point;
    ScreenDist result = -1;

    min_squared_point_origin_dist(result, r00, true);
    min_squared_point_origin_dist(result, r01, false);
    min_squared_point_origin_dist(result, r10, false);
    min_squared_point_origin_dist(result, r11, false);

    min_squared_line_origin_dist(result, r00, r01 - r00);
    min_squared_line_origin_dist(result, r10, r11 - r10);
    min_squared_line_origin_dist(result, r00, r10 - r00);
    min_squared_line_origin_dist(result, r01, r11 - r01);

    return sqrt(result);
}

bool RectangularObject::within(const ScreenRect &rect, View *view,
			       bool partially, const Embedding &embedding) const
{
    ScreenPos s0 = view->map(embedding.transform(p0));
    ScreenPos s1 = view->map(embedding.transform(p1));

    if (!partially)
	return rect.contains(s0) && rect.contains(s1);

    return rect.contains(s0) || rect.contains(s1); //TODO
}
