/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "actionsets.h"
#include "util/path.h"
#include "icons.h"
#include "shared.h"
#include "hub.h"
#include "dirmodel.h"
#include "action.h"
#include "view.h"
#include "attribute.h"

static String display_path(const String &path)
{
    String name = path, home = home_dir() + "/";

    if (name.startsWith(home))
	name.replace(0, home.length(), "~/");

    return name;

/*  struct stat buf;

    if (lstat(paths[i].data(), &buf) == -1) {
	action->setEnabled(false);
	action->setStatusTip(
	    _("Can't stat %1: %2").arg(display_name, strerror(errno)));
    } */
}

FileActionSet::FileActionSet(Hub &hub, Toolkit::Connection &conn)
    : hub(hub), conn(conn),
      new_schematic_action(*this), new_neverball_action(*this),
      file_types_builder(conn), no_recent_files_action(*this),
      recent_files_builder(conn),
      newFileAction(*this), newFolderAction(*this), recentFilesAction(*this),
      saveAction(*this), saveAllAction(*this), commitAction(*this),
      // gitAction(*this), renameAction(*this), deleteAction(*this),
      printAction(*this), sendAction(*this),
      visitComponentAction(*this)
{
    newFileAction.setText(_("New file"));
    newFileAction.setIcon(icons::new_file);
    newFileAction.setMenu(file_types_builder.menu());

    newFolderAction.setText(_("New folder"));
    newFolderAction.setIcon(icons::new_folder);
    newFolderAction.disable(_("Can't %1: %2").arg(_("create folder"),
						  _("Not implemented yet")));

    recentFilesAction.setText(_("&Recent files"));
    recentFilesAction.setMenu(recent_files_builder.menu());

    saveAction.setText(_("&Save"));
    saveAction.setHotkey(_("C-s"));
    saveAction.setIcon(icons::save);
    saveAction.disable(_("Initializing..."));

    saveAllAction.setText(_("Save &all"));
    saveAllAction.setHotkey(_("C-S"));
    saveAllAction.setIcon(icons::save_all);
    saveAllAction.disable(_("Initializing..."));

    commitAction.setText(_("Commit"));
    commitAction.setIcon(icons::commit);
    commitAction.disable(_("Initializing..."));

/*  gitAction.setText(_("Initialize git"));
    gitAction.setIcon(icons::git);

    renameAction.setText(_("Rename"));
    renameAction.setIcon(icons::input_line);

    deleteAction.setText(_("Delete")); */

    printAction.setText(_("&Print"));
    printAction.setHotkey(_("C-p"));
    printAction.disable(_("Can't %1: %2").arg(_("print"),
					      _("Not implemented yet")));

    sendAction.setText(_("S&end"));
    sendAction.setIcon(icons::mail);
    sendAction.disable(_("Can't %1: %2").arg(_("send"),
					     _("Not implemented yet")));

    visitComponentAction.setText(_("Visit file"));

    new_schematic_action.setText(_("&Schematic"));
    new_schematic_action.disable(
	_("Can't %1: %2").arg(_("create schematic"),
			      _("Not implemented yet")));

    new_neverball_action.setText(_("&Neverputt/Neverball map"));
    new_neverball_action.setIcon(icons::neverball);
    new_neverball_action.disable(
	_("Can't %1: %2").arg(_("create Neverputt/Neverball map"),
			      _("Not implemented yet")));

    file_types_builder.addAction(new_schematic_action);
    file_types_builder.addAction(new_neverball_action);

    no_recent_files_action.setText(_("(empty)"));
    no_recent_files_action.disable(_("No recent files"));

    List<String> recent_files; // (settings.value("recentFiles"));

    foreach (String recent_file, recent_files)
	if (!old_recent_files.contains(recent_file)) {
	    old_recent_files.append(recent_file);

	    Toolkit::Action *action = new Toolkit::Action(*this);
	    action->setLiteralText(display_path(recent_file));
	    old_recent_file_actions.append(action);
	    recent_files_builder.addAction(*action);
	}

    if (old_recent_files.empty())
	recent_files_builder.addAction(no_recent_files_action);
}

FileActionSet::~FileActionSet()
{
    recent_files_builder.clear();
    old_recent_file_actions.deleteAll();
    new_recent_file_actions.deleteAll();
}

void FileActionSet::visit_path(const String &path)
{
    struct stat buf;

    if (lstat(path.data(), &buf) == -1)
	fprintf(stderr, "lstat: %s: %s\n", path.data(), strerror(errno));
    else if (S_ISDIR(buf.st_mode))
	fprintf(stderr, "Can't visit directory '%s'\n", path.data());
    else
	hub.setHandle(shared.fileItem(path));
}

void FileActionSet::visitSomeFile()
{
    foreach (String path, old_recent_files) {
	if (hub.handle())
	    return;
	visit_path(path);
    }

    if (!hub.handle())
	visit_path(home_dir() + "/.xi/projects/unnamed0.xml");
}

void FileActionSet::actionTriggered(Toolkit::Action *action)
{
    unsigned int i;

    if (action == &newFileAction) {
	printf("New file\n");
	return;
    }

    if (action == &newFolderAction) {
	printf("New folder\n");
	return;
    }

/*  if (action == &openAction) {
	String path = hub.handle()->path();

	String fileName = FileDialogGetOpenFileName(
	    path.left(path.rfind('/')),
	    _("XML Files (*.xml);;All Files (*)"), 0,
	    QFileDialog::DontResolveSymlinks);

	if (!fileName.isEmpty())
	    visit_path(fileName);

	return;
    } */

    if ((i = new_recent_file_actions.find(action)) != -1) {
	hub.setHandle(new_recent_files[i]);
	return;
    }

    if ((i = old_recent_file_actions.find(action)) != -1) {
	visit_path(old_recent_files[i]);
	return;
    }

    if (action == &saveAction) {
	hub.handle()->save();
	return;
    }

/*  if (action == &saveAsAction) {
	String path = hub.handle()->path();

	String fileName = FileDialogGetSaveFileName(
	    path.left(path.rfind('/')),
	    _("XML Files (*.xml);;All Files (*)"), 0,
	    QFileDialog::DontResolveSymlinks);

	if (!fileName.isEmpty())
	    hub.handle()->save(fileName);

	return;
    } */

    if (action == &saveAllAction) {
	foreach (Item3 *item, new_recent_files)
	    if (item->drawing()->isModified())
		item->save();
	return;
    }

    if (action == &commitAction) {
	printf("Commit\n");
	return;
    }

    if (action == &printAction) {
	printf("Print\n");
	return;
    }

    if (action == &sendAction) {
	printf("Send\n");
	return;
    }
}

void FileActionSet::handleChanged()
{
    unsigned int i, j;
    Toolkit::Action *action;

    // newFileAction.disable(
    //     "Can't create file: Permission denied");
    // newFolderAction.disable(
    //     "Can't create directory: Read-only file system");

    commitAction.disable(_("Can't %1: %2").arg(_("commit"),
					       _("Not a git repository")));

    if ((i = new_recent_files.find(hub.handle())) != -1) {
	action = new_recent_file_actions[i];
	new_recent_file_actions.removeAt(i);
	new_recent_files.removeAt(i);
    } else if ((i = old_recent_files.find(hub.handle()->path())) != -1) {
	action = old_recent_file_actions[i];
	old_recent_file_actions.removeAt(i);
	old_recent_files.removeAt(i);
    } else {
	action = new Toolkit::Action(*this);
	action->setLiteralText(display_path(hub.handle()->path()));
    }

    new_recent_files.prepend(hub.handle());
    new_recent_file_actions.prepend(action);

    recent_files_builder.clear();

    for (i = 0; i < new_recent_file_actions.count() - 1 && i < 25; i++)
	recent_files_builder.addAction(*new_recent_file_actions[i + 1]);
    for (j = 0; j < old_recent_file_actions.count() && i + j < 25; j++)
	recent_files_builder.addAction(*old_recent_file_actions[j]);
    if (i + j == 0)
	recent_files_builder.addAction(no_recent_files_action);
}

void FileActionSet::drawingContentsChanged()
{
    if (!hub.drawing()->isModified())
	saveAction.disable(
	    _("Can't %1: %2").arg(_("save"),
				  _("No changes need to be saved")));
    else
	saveAction.enable();

    bool unsaved_files;
    if (hub.drawing()->isModified())
	unsaved_files = true;
    else {
	unsaved_files = false;
	foreach (Item3 *recent_file, new_recent_files)
	    if (recent_file->drawing()->isModified()) {
		unsaved_files = true;
		return;
	    }
    }
    if (!unsaved_files)
	saveAllAction.disable(
	    _("Can't %1: %2").arg(_("save all"),
				  _("No files need saving")));
    else
	saveAllAction.enable();
}

/****************************************************************************/

DrawingActionSet::DrawingActionSet(Hub &hub)
    : hub(hub), undoAction(*this), redoAction(*this),
		zoomInAction(*this), zoomOutAction(*this)
{
    undoAction.setText(_("&Undo"));
    undoAction.setHotkey("C-z");
    undoAction.setIcon(icons::undo);
    undoAction.disable(_("Initializing..."));

    redoAction.setText(_("&Redo"));
    redoAction.setHotkey("C-Z");
    redoAction.setIcon(icons::redo);
    redoAction.disable(_("Initializing..."));

    zoomInAction.setText(_("Zoom &in"));
    zoomInAction.setHotkey("z");
    zoomInAction.setIcon(icons::zoom_in);
    zoomInAction.disable(_("Initializing..."));

    zoomOutAction.setText(_("Zoom &out"));
    zoomOutAction.setHotkey("Z");
    zoomOutAction.setIcon(icons::zoom_out);
    zoomOutAction.disable(_("Initializing..."));
}

void DrawingActionSet::actionTriggered(Toolkit::Action *action)
{
    Drawing *drawing = hub.drawing();
    if (!drawing)
	return;

    if (action == &undoAction) {
	if (drawing->head()->parent())
	    drawing->seek(drawing->head()->parent());
	return;
    }
    if (action == &redoAction) {
	if (!drawing->head()->children().isEmpty())
	    drawing->seek(drawing->head()->children().last());
	return;
    }
    if (action == &zoomInAction) {
	hub.setView(hub.view()->zoom(exp(0.25), hub.pointer()));
	return;
    }
    if (action == &zoomOutAction) {
	hub.setView(hub.view()->zoom(exp(-0.25), hub.pointer()));
	return;
    }
}

void DrawingActionSet::updateActions()
{
    if (!hub.drawing()) {
	undoAction.disable(
	    _("Can't %1: %2").arg(_("undo"), _("No drawing selected")));
	redoAction.disable(
	    _("Can't %1: %2").arg(_("redo"), _("No drawing selected")));
	zoomInAction.disable(
	    _("Can't %1: %2").arg(_("zoom in"), _("No drawing selected")));
	zoomOutAction.disable(
	    _("Can't %1: %2").arg(_("zoom out"), _("No drawing selected")));
	return;
    }

    ActionGroup *head = hub.drawing()->head();

    if (!head->parent())
	undoAction.disable(
	    _("Can't %1: %2").arg(
		_("undo"), _("No further undo information")));
    else
	undoAction.enable();

    if (head->children().empty())
	redoAction.disable(
	    _("Can't %1: %2").arg(
		_("redo"), _("Already at latest change")));
    else
	redoAction.enable();

    zoomInAction.enable();
    zoomOutAction.enable();
}

/****************************************************************************/

PropertiesActionSet::PropertiesActionSet(Hub &hub,
					 Toolkit::PropertiesView &view)
    : hub(hub), view(view), selectAllAction(*this), resetAction(*this)
{
    selectAllAction.setText(_("Select &all"));
    resetAction.setText(_("&Reset"));
}

void PropertiesActionSet::actionTriggered(Toolkit::Action *action)
{
    if (action == &selectAllAction) {
	return;
    }
    if (action == &resetAction) {
	if (view.selected())
	    static_cast<AttributeItem *>(view.selected()->item())->reset();
	return;
    }
}

void PropertiesActionSet::updateActions()
{
    selectAllAction.disable(
	_("Can't %1: %2").arg(_("select all"),
			      _("Not implemented yet")));
    resetAction.disable(
	_("Can't %1: %2").arg(_("reset"),
			      _("Not implemented yet")));

/** Attribute *attribute = static_cast<Attribute *>(view.selected());

    if (!attribute) {
	selectAllAction.disable(
	    _("Can't %1: %2").arg(_("select all"),
				  _("No attribute selected")));
	resetAction.disable(
	    _("Can't %1: %2").arg(_("reset"),
				  _("No attribute selected")));
    } else {
	selectAllAction.enable();

	if (!attribute->hasDefault(hub))
	    resetAction.disable(
		_("Can't %1: %2").arg(_("reset"), _("Fnord")));
	else if (!attribute->isSet(hub))
	    resetAction.disable(
		_("Can't %1: %2").arg(_("reset"), _("Fnord")));
	else
	    resetAction.enable();
    } **/
}

/****************************************************************************/

PopupActionSet::PopupActionSet(Hub &hub) : hub(hub), deselectModeAction(*this)
{
    deselectModeAction.setText(_("&Exit mode"));
    deselectModeAction.setHotkey("Escape");
    deselectModeAction.setIcon(icons::cursor);
}

void PopupActionSet::actionTriggered(Toolkit::Action *action)
{
    if (action == &deselectModeAction) {
	hub.deselectMajorMode();
	return;
    }
}
