/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <toolkit/toolkit.h>
#include "editors.h"
#include "attribute.h"

#define EDITOR_PADDING 4
#define EDITOR_ICON_PADDING 4


StringItemEditor::StringItemEditor(StringAttributeItem &item)
    : item(item), input(0)
{
}

void StringItemEditor::paint(Toolkit::Connection &conn, X::GC gc,
			     const Toolkit::Region &region,
			     unsigned long background, X::Pixmap pixmap)
{
    gc.setForeground(conn.colors().viewItemText.pixel);
    pixmap.mbDrawString(conn.fontSet(), gc,
			region.x + EDITOR_PADDING,
			region.y + (region.height + conn.fontSet().ascent()
				      - conn.fontSet().descent()) / 2,
			item.text());
}

void StringItemEditor::createWidget(Toolkit::Connection &conn)
{
    input = new Toolkit::Input(conn);
    input->setText(item.text());
    input->selectNone();
    setWidget(input);
}

void StringItemEditor::deleteWidget()
{
    delete input;
    input = 0;
    setWidget(0);
}

void StringItemEditor::invoke(Time time, int x, int y, int x_root, int y_root)
{
    input->selectAll();
}

void StringItemEditor::accept()
{
    item.setText(input->text());
    input->selectNone();
}

void StringItemEditor::reset()
{
    input->setText(item.text());
    input->selectNone();
}

/****************************************************************************/

SpinboxItemEditor::SpinboxItemEditor(SpinboxAttributeItem &item)
    : item(item), input(0)
{
}

void SpinboxItemEditor::paint(Toolkit::Connection &conn, X::GC gc,
			      const Toolkit::Region &region,
			      unsigned long background, X::Pixmap pixmap)
{
    gc.setForeground(conn.colors().viewItemText.pixel);
    pixmap.mbDrawString(conn.fontSet(), gc,
			region.x + EDITOR_PADDING,
			region.y + (region.height + conn.fontSet().ascent()
				      - conn.fontSet().descent()) / 2,
			item.repr());
}

void SpinboxItemEditor::createWidget(Toolkit::Connection &conn)
{
    input = new Toolkit::Input(conn);
    input->setText(item.repr());
    input->selectNone();
    setWidget(input);
}

void SpinboxItemEditor::deleteWidget()
{
    delete input;
    input = 0;
    setWidget(0);
}

void SpinboxItemEditor::invoke(Time time, int x, int y, int x_root, int y_root)
{
    input->selectAll();
}

void SpinboxItemEditor::accept()
{
    item.setRepr(input->text());
    input->setText(item.repr());
    input->selectNone();
}

void SpinboxItemEditor::reset()
{
    input->setText(item.repr());
    input->selectNone();
}

/****************************************************************************/

SelectorItemEditor::SelectorItemEditor(SelectorAttributeItem &item)
    : item(item), selector(0)
{
}

void SelectorItemEditor::paint(Toolkit::Connection &conn, X::GC gc,
			       const Toolkit::Region &region,
			       unsigned long background, X::Pixmap pixmap)
{
    //if (!item.hasUniqueValue())
    //return;

    Item *selected = item.selectedItem();

    gc.setForeground(conn.colors().viewItemText.pixel);
    pixmap.wcDrawString(conn.fontSet(), gc,
			region.x + EDITOR_PADDING +
			    selected->icon()->width() + EDITOR_ICON_PADDING,
			region.y +
			    (region.height + conn.fontSet().ascent()
					   - conn.fontSet().descent()) / 2,
			selected->caption());

    X::Drawable::copyArea(selected->icon()->pixmap(conn, background),
			  pixmap, gc, 0, 0, selected->icon()->width(),
					    selected->icon()->height(),
			  region.x + EDITOR_PADDING,
			  region.y + (region.height -
					  selected->icon()->height()) / 2);
}

void SelectorItemEditor::createWidget(Toolkit::Connection &conn)
{
    selector = new Toolkit::Selector(&item.choices, conn);
    selector->setSelectedItem(item.selectedItem());
    setWidget(selector);
}

void SelectorItemEditor::deleteWidget()
{
    delete selector;
    selector = 0;
    setWidget(0);
}

void SelectorItemEditor::invoke(Time time, int x, int y, int x_root, int y_root)
{
    selector->openMenu(time, x_root - x + selector->x(),
			     y_root - y + selector->y());
}

void SelectorItemEditor::accept()
{
    selector->closeMenu(CurrentTime, true);
    item.setSelectedItem(selector->selectedItem());
    selector->setSelectedItem(item.selectedItem());
}

void SelectorItemEditor::reset()
{
    selector->closeMenu(CurrentTime, false);
    selector->setSelectedItem(item.selectedItem());
}
