/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "views.h"
#include "util/path.h"
#include "dirmodel.h"
#include "shared.h"
#include "hub.h"
#include "action.h"
#include "objectmodel.h"
#include "majormode.h"


ProjectDirView::ProjectDirView(Toolkit::Connection &conn, Hub &hub)
    : DirectoryView(conn), hub(hub)
{
}

static std::vector<Item *> locate(Item3 *item, ItemInterior *interior)
{
    String path = item->path();
    std::vector<Item *> shortest;

    for (unsigned int i = 0; i < interior->count(); i++) {
	Item3 *pd_item = static_cast<Item3 *>(interior->cachedItem(i));
	String pd_path = pd_item->path();
	std::vector<Item *> l;

	if (path == pd_path)
	    l.push_back(pd_item);
	else if (path_contains(pd_path, path)) {
	    List<String> tokens = path.mid(pd_path.length()).split('/');

	    l.push_back(pd_item);
	    for (unsigned int i = 2; i < tokens.count(); i++)
		l.push_back(shared.directoryItem(
		    pd_path + join_strings(tokens.mid(0, i), '/')));
	    l.push_back(item);
	}

	if (!l.empty() && (shortest.empty() || l.size() < shortest.size()))
	    shortest = l;
    }

    return shortest;
}

Toolkit::ViewShadow *ProjectDirView::selected() const
{
    Toolkit::ViewShadow *&shadow = recent[hub.handle()];
    if (!shadow)
	shadow = static_cast<Toolkit::ViewShadow *>(
	    rootShadow()->lookup(locate(hub.handle(), rootInterior())));
    assert(!shadow || shadow->item() == hub.handle());
    return shadow;
}

void ProjectDirView::setSelected(Toolkit::ViewShadow *selected,
				 bool /*toggle*/)
{
    if (!selected || selected->item() == hub.handle())
	return;

    Item3 *item = static_cast<Item3 *>(selected->item());

    if (item->drawing()) {
	recent[item] = selected;
	hub.setHandle(item);
    } else if (selected->count() != 0)
	selected->setExpanded(!selected->isExpanded());
    // TODO: copy from librarymodule.cpp
}

bool ProjectDirView::isSelected(Toolkit::ViewShadow *shadow) const
{
    return shadow->item() == hub.handle();
}

/*  treeView.setAcceptDrops(true);
    treeView.setDragEnabled(true);
    treeView.setRootIsDecorated(false);
    treeView.setDragDropMode(QAbstractItemView::DragDrop);
    treeView.setEditTriggers(QAbstractItemView::EditKeyPressed);
    treeView.setSelectionBehavior(QAbstractItemView::SelectItems);
    treeView.setSelectionMode(QAbstractItemView::NoSelection);
    treeView.setContextMenuPolicy(Qt::CustomContextMenu); */

/****************************************************************************/

LibraryView::LibraryView(Toolkit::Connection &conn, Hub &hub,
			 Toolkit::Menu *menu)
    : DirectoryView(conn), hub(hub), _selected(0), menu(menu)
{
}

Toolkit::ViewShadow *LibraryView::selected() const
{
    return _selected;
}

void LibraryView::setSelected(Toolkit::ViewShadow *selected, bool /*toggle*/)
{
    if (selected) {
	if (!static_cast<Item3 *>(selected->item())->drawing()) {
	    if (selected->count() != 0)
		selected->setExpanded(!selected->isExpanded());
	    return;
	}
	hub.setMajorMode(hub.componentMajorMode());
    }

    if (selected == _selected)
	return;

    _selected = selected;
    invalidate();
}

bool LibraryView::isSelected(Toolkit::ViewShadow *shadow) const
{
    return shadow == _selected && hub.majorMode() == hub.componentMajorMode();
}

void LibraryView::highlightedShadowChanged()
{
    /* TODO: show information about highlighted item */
}

void LibraryView::buttonPressEvent(Time time, int x, int y,
				   int x_root, int y_root,
				   unsigned int state, unsigned int button)
{
    if (button != Button3) {
	DirectoryView::buttonPressEvent(
	    time, x, y, x_root, y_root, state, button);
	return;
    }
    Toolkit::ViewShadow *shadow =
	findShadow(x, y, Toolkit::ViewShadow::SensitiveRegion);
    if (!shadow || !menu)
	return;
    conn.enterMenuMode(time, menu);
    menu->show(time, x_root, y_root, hook());
}

/*  treeView.setAcceptDrops(true);
    treeView.setDragEnabled(true);
    treeView.setRootIsDecorated(false);
    treeView.setDragDropMode(QAbstractItemView::DragDrop);
    treeView.setEditTriggers(QAbstractItemView::EditKeyPressed);
    treeView.setSelectionBehavior(QAbstractItemView::SelectItems);
    treeView.setSelectionMode(QAbstractItemView::NoSelection);
    treeView.setContextMenuPolicy(Qt::CustomContextMenu); */

/****************************************************************************/

HistoryView1::HistoryView1(Toolkit::Connection &conn, Hub &hub)
    : HistoryView(conn), hub(hub)
{
}

Toolkit::HistoryViewShadow *HistoryView1::selected() const
{
    std::vector<Item *> path;
    for (ActionGroup *ag = hub.drawing()->head(); ag; ag = ag->parent())
	path.insert(path.begin(), ag);
    return static_cast<Toolkit::HistoryViewShadow *>(
	rootShadow()->lookup(path));
}

void HistoryView1::setSelected(Toolkit::HistoryViewShadow *selected)
{
    if (!selected || selected->item() == hub.drawing()->head())
	return;

    hub.drawing()->seek(static_cast<ActionGroup *>(selected->item()));
    invalidate();
}

bool HistoryView1::isSelected(Toolkit::HistoryViewShadow *shadow) const
{
    return shadow->item() == hub.drawing()->head();
}

/****************************************************************************/

ObjectsView::ObjectsView(Toolkit::Connection &conn, Hub &hub)
    : DirectoryView(conn), hub(hub)
{
}

void ObjectsView::setSelected(Toolkit::ViewShadow *selected, bool toggle)
{
    List<Object *> objects;

    if (!selected) {
	if (toggle)
	    return;
    } else {
	Object *object = static_cast<ObjectItem *>(selected->item())->object();

	if (!toggle)
	    objects.append(object);
	else {
	    objects = hub.majorMode()->selected();
	    if (objects.contains(object))
		objects.removeAll(object);
	    else
		objects.append(object);
	}
    }

    hub.selectObjects(objects);
}

bool ObjectsView::isSelected(Toolkit::ViewShadow *shadow) const
{
    return hub.majorMode() && hub.majorMode()->selected().contains(
	static_cast<ObjectItem *>(shadow->item())->object());
}

/*
//dragEnabled
//dragDropMode
//treeView.setSelectionBehavior(QAbstractItemView::SelectRows);
treeView.setSelectionMode(QAbstractItemView::ExtendedSelection);
//showDropIndicator
*/

/****************************************************************************/

ToolboxView::ToolboxView(Toolkit::Connection &conn, Hub &hub)
    : ButtonView(conn), hub(hub)
{
}

static Toolkit::ViewShadow *traverse_shadow(Shadow *shadow, Item *item)
{
    for (unsigned int i = 0; i < shadow->count(); i++) {
	Shadow *s = shadow->shadow(i);
	if (s->item() == item || (s = traverse_shadow(s, item)))
	    return static_cast<Toolkit::ViewShadow *>(s);
    }
    return 0;
}

Toolkit::ViewShadow *ToolboxView::selected() const
{
    return traverse_shadow(rootShadow(), hub.majorMode());
}

void ToolboxView::setSelected(Toolkit::ViewShadow *selected)
{
    hub.setMajorMode(static_cast<MajorMode *>(selected->item()));
}

bool ToolboxView::isSelected(Toolkit::ViewShadow *shadow) const
{
    return shadow->item() == hub.majorMode();
}

/****************************************************************************/

MyPropertiesView::MyPropertiesView(Toolkit::Connection &conn,
				   Toolkit::Menu *menu)
    : PropertiesView(conn), menu(menu)
{
}

void MyPropertiesView::buttonPressEvent(
    Time time, int x, int y, int x_root, int y_root,
    unsigned int state, unsigned int button)
{
    if (button != Button3) {
	PropertiesView::buttonPressEvent(
	    time, x, y, x_root, y_root, state, button);
	return;
    }

    focusThis();

    Toolkit::ViewShadow *shadow =
	findShadow(x, y, Toolkit::ViewShadow::SensitiveRegion);
    if (!shadow || !shadow->item() || !shadow->item()->editor())
	return;
    setSelected(shadow);

    if (!menu)
	return;
    conn.enterMenuMode(time, menu);
    menu->show(time, x_root, y_root, hook());
}
