/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "minormode.h"
#include "majormode.h"
#include "hub.h"
#include "drawing.h"
#include "object.h"
#include "view.h"
#include "action_pl.h"
#include "painter.h"

/* Button names from /usr/include/X11/X.h
   Used in ButtonPress and ButtonRelease events */

enum { Button1 = 1, Button2 = 2, Button3 = 3, Button4 = 4, Button5 = 5 };


MinorMode::MinorMode(Hub &hub, MajorMode &parent) : hub(hub), parent(parent)
{
}

void MinorMode::invoke()
{
}

void MinorMode::accept()
{
    parent.accept(this);
}

void MinorMode::cancel()
{
    parent.cancel();
}

void MinorMode::paint(Painter &/*painter*/)
{
}

void MinorMode::mousePressEvent(const ScreenPos &p, unsigned int button,
				unsigned int buttons, unsigned int state)
{
}

void MinorMode::mouseMoveEvent(const ScreenPos &p,
			       unsigned int buttons, unsigned int state)
{
}

void MinorMode::mouseReleaseEvent(const ScreenPos &p, unsigned int button,
				  unsigned int buttons, unsigned int state,
				  bool significantlyMoved)
{
}

bool MinorMode::snapToPoint()
{
    return false;
}

bool MinorMode::snapToLine()
{
    return false;
}

bool MinorMode::showPoint()
{
    return false;
}

String MinorMode::statusMessage() const
{
    return status_message;
}

void MinorMode::setStatusMessage(const String &message)
{
    if (message == status_message)
	return;

    status_message = message;
    hub.statusMessageChanged();
}

/****************************************************************************/

PointSelectingMinorMode::PointSelectingMinorMode(Hub &hub, MajorMode &parent)
    : MinorMode(hub, parent)
{
}

void PointSelectingMinorMode::mouseMoveEvent(
    const ScreenPos &p, unsigned int buttons, unsigned int state)
{
    ActionPoint *snappedPoint = 0;

    snappedLines.clear();

    if (/*hub.minorMode()->snapToPoint()*/ true) {
	ScreenDist least_distance = 50.0;

	foreach (Object *ob, hub.drawing()->objects()) {
	    foreach (ActionPoint *h, ob->handles()) {
		ScreenPos delta = p - hub.view()->map(h->pos());
		ScreenDist dist = delta.x() * delta.x() +
				  delta.y() * delta.y();
		if (dist < least_distance) {
		    least_distance = dist;
		    snappedPoint = h;
		}
	    }
	}
    }

    parent.notifySnapped(snappedPoint);

    if (snappedPoint) {
	snapPoint(snappedPoint);
	return;
    }

    if (/*hub.minorMode()->snapToLine()*/ true) {
	ActionPoint *point0 = 0;
	ActionLine *line0 = 0, *line1 = 0;
	ScreenDist dist_limit = 7.0;
	ScreenDist distance0 = dist_limit, distance1 = dist_limit;
	List<ActionLine *> cand;

	foreach (ActionPoint *_p, parent.activePoints())
	    foreach (ActionLine *l, _p->lines()) {
		ScreenDist dist = l->distance(p, hub.view());
		if (dist < 0)
		    dist = -dist;
		if (dist > dist_limit)
		    continue;
		cand.append(l);
		if (dist < distance0) {
		    point0 = _p;
		    line0 = l;
		    distance0 = dist;
		}
	    }

	if (line0) {
	    foreach (ActionLine *l, cand) {
		ScreenDist dist = l->distance(p, hub.view());
		if (dist >= distance1 || !line0->doesIntersect(l))
		    continue;
		bool abort = false;
		/* is there an action point where the lines cross?
		dist_limit /= 10;
		foreach (Object *ob, hub.drawing()->objects())
		    foreach (ActionPoint *p, ob->handles()) {
			qreal dist0 = line0->distance(p->pos(), hub.view());
			if (dist0 < 0)
			    dist0 = -dist0;
			qreal dist1 = l->distance(p->pos(), hub.view());
			if (dist1 < 0)
			    dist1 = -dist1;
			if (dist0 < dist_limit && dist1 < dist_limit) {
			    abort = true;
			    break;
			}
		    }
		*/
		if (!abort) {
		    line1 = l;
		    distance1 = dist;
		}
	    }

	    snappedLines.append(line0);
	    if (line1) {
		snappedLines.append(line1);
		snapIntersection(line0, line1);
	    } else
		snapLine(line0, point0, p);
	    return;
	}
    }

    freeSelect(hub.view()->unmap(p));
}

void PointSelectingMinorMode::paint(Painter &painter)
{
    foreach (ActionLine *l, snappedLines) {
	painter.setDashedPen(5, 0, 0xcf, 0xcf, 0xcf);
	painter.paintLine(l->base(), l->base() + l->tangent(), true);
    }
}

/****************************************************************************/

PointSelectMinorMode::PointSelectMinorMode(Hub &hub, MajorMode &parent)
    : PointSelectingMinorMode(hub, parent)
{
/** int left, top, right, bottom;

    freeRadioButton = new QRadioButton(_("Free select"));

    freeDeltaCheckBox = new QCheckBox(_("Relative to cursor"));
    freeDeltaCheckBox->setChecked(true);

    QHBoxLayout *freeLayout0 = new QHBoxLayout;
    freeLayout0->getContentsMargins(&left, &top, &right, &bottom);
    freeLayout0->setContentsMargins(left + 20, top, right, bottom);
    freeLayout0->addWidget(freeDeltaCheckBox);

    freeStepCheckBox = new QCheckBox(_("Quantize"));

    freeXStepSpinBox = new MetricSpinBox;
    //freeXStepSpinBox->setValue(Metrum(1.0, hub.drawing()));
    freeXStepSpinBox->setEnabled(false);
    freeYStepSpinBox = new MetricSpinBox;
    //freeYStepSpinBox->setValue(Metrum(1.0, hub.drawing()));
    freeYStepSpinBox->setEnabled(false);

    connect(freeStepCheckBox, SIGNAL(toggled(bool)),
	    freeXStepSpinBox, SLOT(setEnabled(bool)));
    connect(freeStepCheckBox, SIGNAL(toggled(bool)),
	    freeYStepSpinBox, SLOT(setEnabled(bool)));

    QHBoxLayout *freeLayout1 = new QHBoxLayout;
    freeLayout1->getContentsMargins(&left, &top, &right, &bottom);
    freeLayout1->setContentsMargins(left + 20, top, right, bottom);
    freeLayout1->addWidget(freeStepCheckBox);
    freeLayout1->addWidget(freeXStepSpinBox, 1);
    freeLayout1->addWidget(freeYStepSpinBox, 1);

    pointRadioButton = new QRadioButton(_("Snapped point"));

    lineRadioButton = new QRadioButton(_("Use guide"));

    lineDeltaCheckBox = new QCheckBox(_("Relative to cursor"));

    QHBoxLayout *lineLayout0 = new QHBoxLayout;
    lineLayout0->getContentsMargins(&left, &top, &right, &bottom);
    lineLayout0->setContentsMargins(left + 20, top, right, bottom);
    lineLayout0->addWidget(lineDeltaCheckBox);

    lineStepCheckBox = new QCheckBox(_("Quantize"));
    lineStepCheckBox->setChecked(true);

    lineStepSpinBox = new MetricSpinBox;

    connect(lineStepCheckBox, SIGNAL(toggled(bool)),
	    lineStepSpinBox, SLOT(setEnabled(bool)));

    QHBoxLayout *lineLayout1 = new QHBoxLayout;
    lineLayout1->getContentsMargins(&left, &top, &right, &bottom);
    lineLayout1->setContentsMargins(left + 20, top, right, bottom);
    lineLayout1->addWidget(lineStepCheckBox);
    lineLayout1->addWidget(lineStepSpinBox, 1);

    //QLabel *debug = new QLabel("debug");

    for (int i = 0; i < 4; i++) {
	spinbox[i] = new MetricSpinBox(widget());
	//connect(spinbox[i], SIGNAL(valueChanged(const String &)),
	//	debug, SLOT(setText(const String &)));
	connect(spinbox[i], SIGNAL(valueChanged(const String &)),
		this, SLOT(setSpinBoxValue(const String &)));
	connect(spinbox[i], SIGNAL(accepted()),
		this, SLOT(accept()));
    }

    phi = new AngularSpinBox(widget());
    connect(phi, SIGNAL(valueChanged(const String &)),
	    this, SLOT(setSpinBoxValue(const String &)));
    connect(phi, SIGNAL(accepted()),
	    this, SLOT(accept()));
    theta = new AngularSpinBox(widget());
    connect(theta, SIGNAL(valueChanged(const String &)),
	    this, SLOT(setSpinBoxValue(const String &)));
    connect(theta, SIGNAL(accepted()),
	    this, SLOT(accept()));

    QLabel *label0 = new QLabel(_("X coord"));
    QLabel *label1 = new QLabel(_("Y coord"));
    QLabel *label2 = new QLabel(_("Z coord"));
    QLabel *label3 = new QLabel(_("Distance"));
    QLabel *label4 = new QLabel(_("Azimuth"));
    QLabel *label5 = new QLabel(_("Elevation"));

    label0->setBuddy(spinbox[0]);
    label1->setBuddy(spinbox[1]);
    label2->setBuddy(spinbox[2]);
    label3->setBuddy(spinbox[3]);
    label4->setBuddy(phi);
    label5->setBuddy(theta);

    QGridLayout *punktauswahl = new QGridLayout;
    punktauswahl->addWidget(label0, 0, 0);
    punktauswahl->addWidget(label1, 0, 1);
    punktauswahl->addWidget(label2, 0, 2);
    punktauswahl->addWidget(spinbox[0], 1, 0);
    punktauswahl->addWidget(spinbox[1], 1, 1);
    punktauswahl->addWidget(spinbox[2], 1, 2);
    punktauswahl->addWidget(label3, 2, 0);
    punktauswahl->addWidget(label4, 2, 1);
    punktauswahl->addWidget(label5, 2, 2);
    punktauswahl->addWidget(spinbox[3], 3, 0);
    punktauswahl->addWidget(phi, 3, 1);
    punktauswahl->addWidget(theta, 3, 2);

    QVBoxLayout *layout = new QVBoxLayout(widget());
    layout->addWidget(freeRadioButton);
    layout->addLayout(freeLayout0);
    layout->addLayout(freeLayout1);
    layout->addWidget(pointRadioButton);
    layout->addWidget(lineRadioButton);
    layout->addLayout(lineLayout0);
    layout->addLayout(lineLayout1);
    layout->addLayout(punktauswahl);
    //layout->addWidget(debug);
    layout->addSpacing(4);
    layout->addStretch(); **/
}

/**
void PointSelectMinorMode::setSpinBoxValue(const String &/ *m* /)
{
    spinbox[0]->setEnabled(!spinbox[3]->isSet() && !phi->isSet() &&
			   !theta->isSet());
    spinbox[1]->setEnabled(!spinbox[3]->isSet() && !phi->isSet() &&
			   !theta->isSet());
    spinbox[2]->setEnabled(!theta->isSet());
    spinbox[3]->setEnabled(!spinbox[0]->isSet() && !spinbox[1]->isSet());
    phi->setEnabled(!spinbox[0]->isSet() && !spinbox[1]->isSet());
    theta->setEnabled(!spinbox[0]->isSet() && !spinbox[1]->isSet() &&
		      !spinbox[2]->isSet());
}
**/

void PointSelectMinorMode::invoke()
{
/** spinbox[0]->clear();
    spinbox[1]->clear();
    spinbox[2]->clear();
    spinbox[3]->clear();
    phi->clear();
    theta->clear();

    spinbox[0]->setFocus(Qt::TabFocusReason); **/
    released = false;
}

void PointSelectMinorMode::mousePressEvent(
    const ScreenPos &p, unsigned int button,
    unsigned int buttons, unsigned int state)
{
    if (button == Button1)
	accept();
    if (button == Button3)
	cancel();
}

void PointSelectMinorMode::mouseReleaseEvent(
    const ScreenPos &p, unsigned int button,
    unsigned int buttons, unsigned int state, bool significantlyMoved)
{
    if (button == Button1 && significantlyMoved) {
	released = true;
	accept();
    }
}

bool PointSelectMinorMode::showPoint()
{
    return true;
}

Metrum3 PointSelectMinorMode::point()
{
    return the_point;
}

void PointSelectMinorMode::setPoint(const Metrum3 &point)
{
    the_point = point;
}

void PointSelectMinorMode::snapPoint(ActionPoint *point)
{
    the_point = point->pos();
    applyOffset();

/** pointRadioButton->setChecked(true); **/
    setStatusMessage(_("%1 %2").arg(point->object()->name(),
				    point->name()).capitalize());
}

void PointSelectMinorMode::snapLine(ActionLine *line, ActionPoint *point,
				    const ScreenPos &p)
{
    the_point = line->nearest(p, hub.view());

/** Metrum3 delta = the_point - point->pos();
    char sign = delta.x() * line->tangent().x() +
		delta.y() * line->tangent().y() +
		delta.z() * line->tangent().z() >= 0 ? 1 : -1;

    double b0[3] = {
	hub.drawing()->eff(line->tangent().x()) * sign,
	hub.drawing()->eff(line->tangent().y()) * sign,
	hub.drawing()->eff(line->tangent().z()) * sign
    };
    double l = sqrt(b0[0] * b0[0] + b0[1] * b0[1] + b0[2] * b0[2]);
    b0[0] /= l; b0[1] /= l; b0[2] /= l;
    double b1[3] = { -b0[1], b0[0], 0 };
    l = sqrt(b0[0] * b0[0] + b0[1] * b0[1] + b0[2] * b0[2]);
    if (l == 0)
	b1[1] = 1.0;
    else {
	b0[0] /= l; b0[1] /= l; b0[2] /= l;
    }
    double b2[3] = {
	b0[1] * b1[2] - b0[2] * b1[1],
	b0[2] * b1[0] - b0[0] * b1[2],
	b0[0] * b1[1] - b0[1] * b1[0]
    };

    if (!lineDeltaCheckBox->isChecked() && (
	    spinbox[0]->isSet() || spinbox[1]->isSet() ||
	    spinbox[2]->isSet() || spinbox[3]->isSet() ||
	    phi->isSet() || theta->isSet()))
	the_point = point->pos();
    else {
	if (lineStepCheckBox->isChecked() && lineStepSpinBox->isSet()) {
	    Metrum3 delta = the_point - point->pos();
	    double length = sqrt(delta.x() * delta.x() +
				 delta.y() * delta.y() +
				 delta.z() * delta.z());
	    double result = nearbyint(length / lineStepSpinBox->value())
				* hub.drawing()->eff(lineStepSpinBox->value());
	    the_point = point->pos() + Metrum3::scale(result * b0[0],
						      result * b0[1],
						      result * b0[2],
						      hub.drawing());
	}
    }

    if (spinbox[0]->isSet() || spinbox[1]->isSet()) {
	the_point.rx() += spinbox[0]->value() * b0[0];
	the_point.ry() += spinbox[0]->value() * b0[1];
	the_point.rz() += spinbox[0]->value() * b0[2];
	the_point.rx() += spinbox[1]->value() * b1[0];
	the_point.ry() += spinbox[1]->value() * b1[1];
	the_point.rz() += spinbox[1]->value() * b1[2];
    }

    if (spinbox[2]->isSet()) {
	the_point.rx() += spinbox[2]->value() * b2[0];
	the_point.ry() += spinbox[2]->value() * b2[1];
	the_point.rz() += spinbox[2]->value() * b2[2];
    }

    if (spinbox[3]->isSet() || phi->isSet()) {
	the_point.rx() += spinbox[3]->value()
	    * phi->sin() * theta->cos() * b0[0];
	the_point.ry() += spinbox[3]->value()
	    * phi->sin() * theta->cos() * b0[1];
	the_point.rz() += spinbox[3]->value()
	    * phi->sin() * theta->cos() * b0[2];
	the_point.rx() += spinbox[3]->value()
	    * phi->cos() * theta->cos() * b1[0];
	the_point.ry() += spinbox[3]->value()
	    * phi->cos() * theta->cos() * b1[1];
	the_point.rz() += spinbox[3]->value()
	    * phi->cos() * theta->cos() * b1[2];
    }

    if (theta->isSet()) {
	the_point.rx() += spinbox[3]->value() * theta->sin() * b2[0];
	the_point.ry() += spinbox[3]->value() * theta->sin() * b2[1];
	the_point.rz() += spinbox[3]->value() * theta->sin() * b2[2];
    }

    lineRadioButton->setChecked(true); **/
    setStatusMessage(
	_("%1 %2").arg(line->name(), hub.drawing()->format(
			   (the_point - line->base()).length()))
		  .capitalize());
}

void PointSelectMinorMode::snapIntersection(ActionLine *line0,
					    ActionLine *line1)
{
    the_point = line0->intersection(line1);
    applyOffset();
/** pointRadioButton->setChecked(true); **/
    setStatusMessage(_("%1 x %2").arg(line0->name(),
				      line1->name()).capitalize());
}

void PointSelectMinorMode::freeSelect(const Metrum3 &m)
{
/** if (!freeDeltaCheckBox->isChecked() && (
	    spinbox[0]->isSet() || spinbox[1]->isSet() ||
	    spinbox[2]->isSet() || spinbox[3]->isSet() ||
	    phi->isSet() || theta->isSet()))
	the_point = Metrum3(hub.drawing());
    else { **/
	the_point = m;
/**	if (freeStepCheckBox->isChecked() && freeXStepSpinBox->isSet()) {
	    int64_t step = freeXStepSpinBox->value().pos();
	    if (the_point.x().pos() >= 0)
		the_point.rx() -= Metrum((the_point.x().pos() + step / 2)
					     % step - step / 2,
					 hub.drawing());
	    else
		the_point.rx() -= Metrum((the_point.x().pos() - step / 2)
					     % step + step / 2,
					 hub.drawing());
	}
	if (freeStepCheckBox->isChecked() && freeYStepSpinBox->isSet()) {
	    int64_t step = freeYStepSpinBox->value().pos();
	    if (the_point.y().pos() >= 0)
		the_point.ry() -= Metrum((the_point.y().pos() + step / 2)
					     % step - step / 2,
					 hub.drawing());
	    else
		the_point.ry() -= Metrum((the_point.y().pos() - step / 2)
					     % step + step / 2,
					 hub.drawing());
	}
    } **/

    applyOffset();
/** freeRadioButton->setChecked(true); **/
    setStatusMessage(_("Free select"));
}

void PointSelectMinorMode::applyOffset()
{
/** if (spinbox[0]->isSet() || spinbox[1]->isSet()) {
	the_point.rx() += spinbox[0]->value();
	the_point.ry() += spinbox[1]->value();
    }

    if (spinbox[2]->isSet()) {
	the_point.rz() += spinbox[2]->value();
    }

    if (spinbox[3]->isSet() || phi->isSet()) {
	the_point.rx() += spinbox[3]->value() * phi->cos()
	    * theta->cos();
	the_point.ry() += spinbox[3]->value() * phi->sin()
	    * theta->cos();
    }

    if (theta->isSet()) {
	the_point.rz() += spinbox[3]->value() * theta->sin();
    } **/
}

/****************************************************************************/

SelectMinorMode::SelectMinorMode(Hub &hub, MajorMode &parent)
    : MinorMode(hub, parent),
      rubber_box(false),
      select_partially_covered(false)
{
/** QCheckBox *checkBox = new QCheckBox(_("Select partially covered objects"));
    connect(checkBox, SIGNAL(stateChanged(int)),
	    this, SLOT(changeSelectPartiallyCovered(int)));

    QVBoxLayout *layout = new QVBoxLayout(widget());
    layout->addWidget(checkBox);
    layout->addStretch(); **/

/*
    QLineEdit *lineEdit = new QLineEdit("Foo");
    lineEdit->setCursor(Qt::ArrowCursor);
    lineEdit->setAlignment(Qt::AlignCenter);
    lineEdit->setReadOnly(true);
    lineEdit->setFocusPolicy(Qt::NoFocus);

    QHBoxLayout *l = new QHBoxLayout;
    l->addWidget(lineEdit);
*/
}

void SelectMinorMode::paint(Painter &painter)
{
    if (!rubber_box)
	return;

    ScreenRect rect(start, stop);

    // (stop.x() >= start.x()) == (stop.y() >= start.y())
    if (!select_partially_covered) {
	painter.setBlackPen();
	painter.paintRectangle(rect, 0xdf, 0xdf, 0xff);
    } else {
	painter.setDashedBlackPen(5, 0);
	painter.paintRectangle(rect, 0xcf, 0xff, 0xcf);
    }
}

void SelectMinorMode::mousePressEvent(const ScreenPos &p, unsigned int button,
				      unsigned int buttons, unsigned int state)
{
    if (button == Button1) {
	the_objects.clear();
	the_handle = 0;

	ScreenDist least_distance = 3.0;

	foreach (Object *ob, parent.selected())
	    foreach (ActionPoint *h, ob->handles()) {
		ScreenPos delta = p - hub.view()->map(h->pos());
		if (delta.x() < 0)
		    delta.setX(-delta.x());
		if (delta.y() < 0)
		    delta.setY(-delta.y());
		ScreenDist dist = delta.x() > delta.y() ? delta.x() : delta.y();
		if (dist < least_distance) {
		    least_distance = dist;
		    the_handle = h;
		}
	    }

	if (handle())
	    accept();
	else {
	    start = stop = p;
	    rubber_box = true;
	}
    }
}

void SelectMinorMode::mouseMoveEvent(const ScreenPos &p,
				     unsigned int buttons, unsigned int state)
{
    if (rubber_box)
	stop = p;
}

void SelectMinorMode::mouseReleaseEvent(
    const ScreenPos &p, unsigned int button,
    unsigned int buttons, unsigned int state, bool significantlyMoved)
{
    if (button != Button1 || !rubber_box)
	return;

    if (significantlyMoved) {
	foreach (Object *ob, hub.drawing()->rootContext()->objects())
	    if (ob->within(ScreenRect(start, stop), hub.view(),
			   select_partially_covered))
		the_objects.append(ob);
    } else {
	ScreenDist least_distance = 5.0;
	Object *object = 0;

	foreach (Object *ob, hub.drawing()->objects()) {
	    ScreenDist dist = ob->proximity(stop, hub.view());
	    if (dist < least_distance && dist >= 0) {
		least_distance = dist;
		object = ob;
	    }
	}

	if (object) {
	    while (object->parent() &&
		   !parent.selected().contains(object) &&
		   !parent.selected().contains(object->parent()))
		object = object->parent();

	    the_objects.append(object);
	}
    }

    accept();
}

void SelectMinorMode::accept()
{
    rubber_box = false;
    MinorMode::accept();
}

List<Object *> SelectMinorMode::objects()
{
    return the_objects;
}

ActionPoint *SelectMinorMode::handle()
{
    return the_handle;
}

/****************************************************************************/

GrabHandleMinorMode::GrabHandleMinorMode(Hub &hub, MajorMode &parent)
    : PointSelectingMinorMode(hub, parent)
{
}

void GrabHandleMinorMode::accept()
{
    MinorMode::accept();
}

void GrabHandleMinorMode::paint(Painter &painter)
{
    PointSelectingMinorMode::paint(painter);
    handle()->object()->paintShadow(point(), handle(), painter);
}

void GrabHandleMinorMode::mousePressEvent(
    const ScreenPos &p, unsigned int button,
    unsigned int buttons, unsigned int state)
{
    if (button == Button1)
	accept();
    if (button == Button3)
	cancel();
}

void GrabHandleMinorMode::mouseReleaseEvent(
    const ScreenPos &p, unsigned int button,
    unsigned int buttons, unsigned int state, bool significantlyMoved)
{
    if (button == Button1 && significantlyMoved)
	accept();
}

ActionPoint *GrabHandleMinorMode::handle()
{
    return the_handle;
}

void GrabHandleMinorMode::setHandle(ActionPoint *handle)
{
    the_handle = handle;
    the_point = handle->pos();
}

Metrum3 GrabHandleMinorMode::point()
{
    return the_point;
}

void GrabHandleMinorMode::snapPoint(ActionPoint *point)
{
    the_point = point->pos();
    setStatusMessage(_("%1 %2").arg(point->object()->name(),
				    point->name()).capitalize());
}

void GrabHandleMinorMode::snapLine(ActionLine *line, ActionPoint *point,
				   const ScreenPos &p)
{
    the_point = line->nearest(p, hub.view());
    setStatusMessage(
	_("%1 %2").arg(line->name(), hub.drawing()->format(
			   (the_point - line->base()).length()))
		  .capitalize());
}

void GrabHandleMinorMode::snapIntersection(ActionLine *line0, ActionLine *line1)
{
    the_point = line0->intersection(line1);
    setStatusMessage(_("%1 x %2").arg(line0->name(),
				      line1->name()).capitalize());
}

void GrabHandleMinorMode::freeSelect(const Metrum3 &m)
{
    the_point = m;
    setStatusMessage(_("Free select"));
}
