/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "nb_object.h"
#include "xml.h"
#include "painter.h"
#include "view.h"
#include "xml_writer.h"
#include "drawing.h"
#include "nb_attribute.h"
#include "variant.h"
#include "nb_lump.h"
#include "nb_matrix.h"
#include "icons.h"

#define SMALL 0.0005f

class BodyXMLHandler : public XMLHandler
{
public:
    BodyXMLHandler(ObjectContext *context, Drawing *drawing);
    void element(XMLElement *element, Messages &messages);

private:
    Drawing *const drawing;
    ObjectContext *const context;
};

class LumpXMLHandler : public XMLHandler
{
public:
    LumpXMLHandler(LumpObject *lump, Drawing *drawing);
    void element(XMLElement *element, Messages &messages);
    void done(Messages &messages);

private:
    Drawing *const drawing;
    LumpObject *const lump;
};

PathObject::PathObject(Metrum3 pos, Drawing *drawing)
    : Object(drawing), the_pos(pos),
      travelTime(1.0),
      the_path(this),
      enable(true),
      handle(_("node"), this),
      tangent(_("tangent"), this),
      normal(_("normal"), this)
{
}

PathObject::PathObject(XMLElement *element, Drawing *drawing)
    : Object(drawing),
      travelTime(1.0),
      the_path(this),
      enable(true),
      handle(_("node"), this),
      tangent(_("tangent"), this),
      normal(_("normal"), this)
{
    if (element->selectAttribute("symbol"))
	the_symbol = element->value();
    if (element->selectAttribute("starting-position"))
	the_pos = element->toMetrum3(drawing);
    if (element->selectAttribute("travel-time"))
	travelTime = element->toDouble();
    if (element->selectAttribute("path")) {
	/* Todo */
    }
    if (element->selectAttribute("enable-flag"))
	enable = element->value() != "0";
}

String PathObject::symbol() const
{
    return the_symbol;
}

Metrum3 PathObject::pos() const
{
    return the_pos;
}

PathObject *PathObject::path() const
{
    return the_path;
}

void PathObject::setPath(PathObject *path)
{
    the_path = path;
}

String PathObject::name() const
{
    return _("path");
}

void PathObject::paint(Painter &painter, const Embedding &embedding)
{
    painter.setPen(0xa0, 0xa0, 0xa0);
    painter.paintPoint(pos());

    if (path() != this)
	painter.paintLine(pos(), path()->pos());
}

void PathObject::paintShadow(const Metrum3 &pos, const ActionPoint *handle,
			     Painter &painter)
{
    Metrum3 hpos = this->pos();

    if (handle == &this->handle)
	hpos = pos;

    painter.setPen(0x80, 0x80, 0x80);
    painter.paintPoint(hpos);

    if (path() != this)
	painter.paintLine(hpos, path()->pos());
}

List<ActionPoint *> PathObject::handles()
{
    return List<ActionPoint *>() << &handle;
}

Metrum3 PathObject::handlePos(const ActionPoint *handle) const
{
    if (handle == &this->handle)
	return pos();

    return Metrum3();
}

List<ActionLine *> PathObject::handleLines(const ActionPoint *handle)
{
    if (handle == &this->handle && path()->pos() != pos())
	return List<ActionLine *>() << &tangent << &normal;

    return List<ActionLine *>();
}

void PathObject::moveHandle(const Metrum3 &pos, const ActionPoint *handle)
{
    if (handle == &this->handle)
	the_pos = pos;
}

Metrum3 PathObject::lineBase(const ActionLine *line) const
{
    if (line == &tangent || line == &normal)
	return pos();

    return Object::lineBase(line);
}

Metrum3 PathObject::lineTangent(const ActionLine *line) const
{
    if (line == &tangent)
	return path()->pos() - pos();

    if (line == &normal)
	return Metrum3((path()->pos() - pos()).y(),
		       (pos() - path()->pos()).x(), 0);

    return Object::lineTangent(line);
}

ScreenDist PathObject::proximity(const ScreenPos &point, View *view,
				 const Embedding &embedding) const
{
    if (path()->pos() == pos())
	return (point - view->map(pos())).length();

    ScreenPos s0 = view->map(pos());
    ScreenPos s1 = view->map(path()->pos());
    ScreenDist parameter, distance;

    point.computeParameters(s0, s1 - s0, parameter, distance);

    if (parameter < 0)
	return (point - s0).length();

    if (parameter > (s1 - s0).length())
	return (point - s1).length();

    return distance;
}

bool PathObject::within(const ScreenRect &rect, View *view,
			bool partially, const Embedding &embedding) const
{
    if (path()->pos() == pos())
	return rect.contains(view->map(pos()));

    ScreenPos s0 = view->map(pos());
    ScreenPos s1 = view->map(path()->pos());

    if (!partially)
	return rect.contains(s0) && rect.contains(s1);

    return rect.contains(s0) || rect.contains(s1); //TODO
}

void PathObject::write(XMLWriter *writer)
{
    writer->startElement("path");
    writer->writeAttribute("symbol", symbol());
    writer->writeAttribute("starting-position", drawing()->format(pos()));
    writer->writeAttribute("travel-time", double_to_str(travelTime));
    writer->writeAttribute("path", path() ? path()->symbol() : String());
    writer->writeAttribute("enable-flag", enable ? "yes" : "no");
    writer->endElement();
}

List<Attribute *> PathObject::attributes() const
{
    return List<Attribute *>() << &NeverballAttributes::symbol
			       << &NeverballAttributes::travelTime
			       << &NeverballAttributes::path
			       << &NeverballAttributes::enableFlag;
}

Variant PathObject::attribute(const Attribute *attribute) const
{
    if (attribute == &NeverballAttributes::symbol)
	return Variant::fromString(symbol());

    if (attribute == &NeverballAttributes::travelTime)
	return Variant::fromDouble(travelTime);

    if (attribute == &NeverballAttributes::path)
	return Variant::fromPointer<PathObject *>(path());

    if (attribute == &NeverballAttributes::enableFlag)
	return Variant::fromBool(enable);

    return Variant();
}

Variant PathObject::attributeDefault(const Attribute *attribute) const
{
    if (attribute == &NeverballAttributes::symbol)
	return Variant(); /* no default */

    if (attribute == &NeverballAttributes::travelTime)
	return Variant::fromDouble(1.);

    if (attribute == &NeverballAttributes::path)
	return Variant::fromPointer<PathObject *>(
	    const_cast<PathObject *>(this));

    if (attribute == &NeverballAttributes::enableFlag)
	return Variant::fromBool(true);

    return Variant();
}

void PathObject::setAttribute(const Variant &value,
			      const Attribute *attribute)
{
    if (attribute == &NeverballAttributes::symbol)
	the_symbol = value.toString();

    if (attribute == &NeverballAttributes::travelTime)
	travelTime = value.toDouble();

    if (attribute == &NeverballAttributes::path) {
	PathObject *new_path = value.toPointer<PathObject *>();
	if (new_path)
	    the_path = new_path;
    }

    if (attribute == &NeverballAttributes::enableFlag)
	enable = value.toBool();
}

/****************************************************************************/

BodyXMLHandler::BodyXMLHandler(ObjectContext *context, Drawing *drawing)
    : XMLHandler(), drawing(drawing), context(context)
{
}

void BodyXMLHandler::element(XMLElement *element, Messages &messages)
{
    if (element->name() == "lump") {
	context->addObject(new LumpObject(element, drawing));
	return;
    }

    XMLHandler::element(element, messages);
}

BodyObject::BodyObject(Metrum3 pos, Drawing *drawing)
    : Object(drawing), pos(pos), path(0), context(this),
      handle(_("origin"), this)
{
}

BodyObject::BodyObject(XMLElement *element, Drawing *drawing)
    : Object(drawing), path(0), context(this),
      handle(_("origin"), this)
{
    if (element->selectAttribute("position"))
	pos = element->toMetrum3(drawing);
    if (element->selectAttribute("model"))
	model = element->value();
    if (element->selectAttribute("path")) {
	/* Todo */
    }

    element->setHandler(new BodyXMLHandler(&context, drawing));
}

String BodyObject::name() const
{
    return _("body");
}

ObjectContext *BodyObject::children() const
{
    return const_cast<ObjectContext *>(&context);
}

void BodyObject::paint(Painter &painter, const Embedding &embedding)
{
    painter.setBlackPen();
    painter.paintPoint(pos);
}

void BodyObject::paintShadow(const Metrum3 &pos, const ActionPoint *handle,
			     Painter &painter)
{
    Metrum3 hpos = this->pos;

    if (handle == &this->handle)
	hpos = pos;

    painter.setPen(0x80, 0x80, 0x80);
    painter.paintPoint(hpos);
}

List<ActionPoint *> BodyObject::handles()
{
    return List<ActionPoint *>() << &handle;
}

Metrum3 BodyObject::handlePos(const ActionPoint *handle) const
{
    if (handle == &this->handle)
	return pos;

    return Metrum3();
}

void BodyObject::moveHandle(const Metrum3 &pos, const ActionPoint *handle)
{
    if (handle == &this->handle)
	this->pos = pos;
}

ScreenDist BodyObject::proximity(const ScreenPos &point, View *view,
				 const Embedding &embedding) const
{
    return (point - view->map(pos)).length();
}

bool BodyObject::within(const ScreenRect &rect, View *view,
			bool partially, const Embedding &embedding) const
{
    return rect.contains(view->map(pos));
}

void BodyObject::write(XMLWriter *writer)
{
    writer->startElement("body");
    writer->writeAttribute("position", drawing()->format(pos));
    writer->writeAttribute("model", model);
    writer->writeAttribute("path", path ? path->symbol() : String());
    writer->endElement();
}

List<Attribute *> BodyObject::attributes() const
{
    return List<Attribute *>() << &NeverballAttributes::model
			       << &NeverballAttributes::path;
}

Variant BodyObject::attribute(const Attribute *attribute) const
{
    if (attribute == &NeverballAttributes::model)
	return Variant::fromString(model);

    if (attribute == &NeverballAttributes::path)
	return Variant::fromPointer<PathObject *>(path);

    return Variant();
}

Variant BodyObject::attributeDefault(const Attribute *attribute) const
{
    if (attribute == &NeverballAttributes::model)
	return Variant::fromString(String());

    if (attribute == &NeverballAttributes::path)
	return Variant::fromPointer<PathObject *>(0);

    return Variant();
}

void BodyObject::setAttribute(const Variant &value,
			      const Attribute *attribute)
{
    if (attribute == &NeverballAttributes::model)
	model = value.toString();

    if (attribute == &NeverballAttributes::path)
	path = value.toPointer<PathObject *>();
}

/****************************************************************************/

static Vector3 m3_to_v3(const Metrum3 &p, Drawing *drawing)
{
    return Vector3(drawing->eff(p.x()),
		   drawing->eff(p.y()),
		   drawing->eff(p.z()));
}

static Metrum3 v3_to_m3(const Vector3 &v, Drawing *drawing)
{
    return drawing->scale(v.x(), v.y(), v.z());
}

LumpXMLHandler::LumpXMLHandler(LumpObject *lump, Drawing *drawing)
    : XMLHandler(), drawing(drawing), lump(lump)
{
}

void LumpXMLHandler::element(XMLElement *element, Messages &messages)
{
    if (element->name() == "side" || element->name() == "plane") {
	Metrum3 p0, p1, p2;
	short tu = 0, tv = 0, r = 0;
	float su = 0.5, sv = 0.5;
	String s;

	if (element->selectAttribute("p0"))
	    p0 = element->toMetrum3(drawing);
	if (element->selectAttribute("p1"))
	    p1 = element->toMetrum3(drawing);
	if (element->selectAttribute("p2"))
	    p2 = element->toMetrum3(drawing);
	if (element->selectAttribute("tu"))
	    tu = element->toInt();
	if (element->selectAttribute("tv"))
	    tv = element->toInt();
	if (element->selectAttribute("r"))
	    r = element->toInt();
	if (element->selectAttribute("su"))
	    su = element->toFloat();
	if (element->selectAttribute("sv"))
	    sv = element->toFloat();
	if (element->selectAttribute("key"))
	    s = element->value();

	lump->s.append(new Side(m3_to_v3(p0, drawing),
				m3_to_v3(p1, drawing),
				m3_to_v3(p2, drawing), tu, tv, r, su, sv, s));
	return;
    }

    XMLHandler::element(element, messages);
}

void LumpXMLHandler::done(Messages &messages)
{
    lump->clip();
}

LumpHandle::LumpHandle(Side *s0, Side *s1, Side *s2, Object *object)
    : handle(_("vertex"), object), s0(s0), s1(s1), s2(s2)
{
}

bool LumpHandle::matches(Side *t0, Side *t1, Side *t2)
{
    return (s0 == t0 && s1 == t1 && s2 == t2) ||
	   (s0 == t0 && s1 == t2 && s2 == t1) ||
	   (s0 == t1 && s1 == t2 && s2 == t0) ||
	   (s0 == t1 && s1 == t0 && s2 == t2) ||
	   (s0 == t2 && s1 == t0 && s2 == t1) ||
	   (s0 == t2 && s1 == t1 && s2 == t0);
}


LumpObject::LumpObject(bool detail, Drawing *drawing)
    : Object(drawing), is_detail(detail)
{
}

LumpObject::LumpObject(XMLElement *element, Drawing *drawing) : Object(drawing)
{
    is_detail = element->selectAttribute("detail") && element->value() == "yes";
    element->setHandler(new LumpXMLHandler(this, drawing));
}

bool LumpObject::detail() const
{
    return is_detail;
}

void LumpObject::setDetail(bool detail)
{
    is_detail = detail;
}

String LumpObject::name() const
{
    return _("lump");
}

void LumpObject::paint(Painter &painter, const Embedding &embedding)
{
    unsigned int red, green, blue;

    if (painter.isSelected(this)) {
	red = 0xff;
	green = 0x00;
	blue = 0x00;

	foreach (Vertex *vertex, v) {
	    painter.setPen(red, green, blue);
	    painter.paintHandle(v3_to_m3(vertex->p, drawing()),
				0xff, 0x80, 0x80);
	}
    } else {
	red = 0x00;
	green = 0x80;
	blue = 0x00;
    }

    foreach (Side *side, s) {
	if (side->plane_m->d[3] > 0)
	    painter.setPen(red, green, blue);
	else
	    painter.setPen(0xef + red * 0x10 / 0xff,
			   0xef + green * 0x10 / 0xff,
			   0xef + blue * 0x10 / 0xff);

	for (unsigned int i = 0; i < side->vertices.count(); i++) {
	    Vertex *vi = side->vertices[i];
	    Vertex *vj = side->vertices[(i + 1) % side->vertices.count()];
	    painter.paintLine(v3_to_m3(vi->p, drawing()),
			      v3_to_m3(vj->p, drawing()));
	}
    }
}

void LumpObject::paintShadow(const Metrum3 &pos, const ActionPoint *handle,
			     Painter &painter)
{
}

List<ActionPoint *> LumpObject::handles()
{
    List<ActionPoint *> h;

    foreach (Vertex *vertex, v) {
	LumpHandle *x = 0;

	foreach (LumpHandle *y, handle_buf)
	    if (y->matches(vertex->s0, vertex->s1, vertex->s2)) {
		x = y;
		break;
	    }

	if (!x) {
	    x = new LumpHandle(vertex->s0, vertex->s1, vertex->s2, this);
	    handle_buf.append(x);
	}

	h.append(&x->handle);
    }

    return h;
}

Metrum3 LumpObject::handlePos(const ActionPoint *handle) const
{
    foreach (LumpHandle *y, handle_buf)
	if (handle == &y->handle) {
	    return where_do_these_thingies_meet(y->s0, y->s1, y->s2,
						drawing());
	}

    return Metrum3();
}

void LumpObject::moveHandle(const Metrum3 &pos, const ActionPoint *handle)
{
    foreach (LumpHandle *y, handle_buf)
	if (handle == &y->handle) {
	    Vector3 p(m3_to_v3(pos, drawing()));

	    y->s0->d = v_dot(y->s0->n, p);
	    y->s1->d = v_dot(y->s1->n, p);
	    y->s2->d = v_dot(y->s2->n, p);

	    clip();
	    return;
	}
}

ScreenDist LumpObject::proximity(const ScreenPos &point, View *view,
				 const Embedding &embedding) const
{
    ScreenDist result;

    result = 10000;

    foreach (Vertex *vertex, v) {
	ScreenDist p = (point - view->map(
			    v3_to_m3(vertex->p, drawing()))).length();

	if (p < result)
	    result = p;
    }

    return result;
}

bool LumpObject::within(const ScreenRect &rect, View *view,
			bool partially, const Embedding &embedding) const
{
    return false;
}

void LumpObject::write(XMLWriter *writer)
{
    writer->startElement("lump");
    writer->writeAttribute("detail", is_detail ? "yes" : "no");
    writer->endElement();
}

List<Attribute *> LumpObject::attributes() const
{
    return List<Attribute *>() << &NeverballAttributes::detail;
}

Variant LumpObject::attribute(const Attribute *attribute) const
{
    if (attribute == &NeverballAttributes::detail)
	return Variant::fromBool(is_detail);

    return Variant();
}

Variant LumpObject::attributeDefault(const Attribute *attribute) const
{
    if (attribute == &NeverballAttributes::detail)
	return Variant::fromBool(false);

    return Variant();
}

void LumpObject::setAttribute(const Variant &value,
			      const Attribute *attribute)
{
    if (attribute == &NeverballAttributes::detail)
	is_detail = value.toBool();
}

/* All bodies with an associated path are assumed to be positioned at
 * the beginning of that path. These bodies must be moved to the
 * origin in order for their path transforms to behave correctly.
 * This is how we get away with defining func_trains with no origin
 * specification. */

void LumpObject::move(const Vector3 &p)
{
    foreach (Side *side, s)
	side->move(p);
    foreach (Vertex *vertex, v)
	vertex->move(p);
}

/* Confirm that the addition of a vert would not result in degenerate
 * geometry. */

bool LumpObject::ok_vert(const Vector3 &p) const
{
    foreach (Vertex *vertex, v)
	if (v_len(p - vertex->p) < SMALL)
	    return false;

    return true;
}

/* The following functions take the set of planes defining a lump and
 * find the verts, edges, and geoms that describe its boundaries. To
 * do this, they first find the verts, and then search these verts for
 * valid edges and geoms. It may be more efficient to compute edges
 * and geoms directly by clipping down infinite line segments and
 * planes, but this would be more complex and prone to numerical
 * error. */

/* Iterate the sides of the lump, attemping to generate a new vert for
 * each trio of planes, a new edge for each pair of planes, and a new
 * set of geom for each visible plane. */

void LumpObject::clip()
{
    unsigned int i, j, k;

    v.deleteAll(); v.clear();
    e.deleteAll(); e.clear();
    g.deleteAll(); g.clear();

    for (i = 2; i < s.count(); i++)
	for (j = 1; j < i; j++)
	    for (k = 0; k < j; k++)
		clip_vert(s[i], s[j], s[k]);

    for (i = 1; i < s.count(); i++)
	for (j = 0; j < i; j++)
	    clip_edge(s[i], s[j]);

    for (i = 0; i < s.count(); i++)
	clip_geom(s[i]);
}

/* Given 3 side planes, compute the point of intersection, if any.
 * Confirm that this point falls within the current lump, and that it
 * is unique. Add it as a vert of the solid. */

Metrum3 LumpObject::where_do_these_thingies_meet(Side *si, Side *sj, Side *sk,
						 Drawing *drawing) const
{
    float M[16], X[16], I[16];
    Vector3 d(si->d, sj->d, sk->d), p;

    m_basis(M, si->n, sj->n, sk->n);
    m_xps(X, M);

    if (m_inv(I, X))
	return v3_to_m3(m_vxfm(I, d), drawing);

    return Metrum3();
}

void LumpObject::clip_vert(Side *si, Side *sj, Side *sk)
{
    float M[16], X[16], I[16];
    Vector3 d(si->d, sj->d, sk->d), p;

    m_basis(M, si->n, sj->n, sk->n);
    m_xps(X, M);

    if (m_inv(I, X)) {
	p = m_vxfm(I, d);

	foreach (Side *side, s)
	    if (side->fore_side(p))
		return;

	if (ok_vert(p))
	    v.append(new Vertex(p, si, sj, sk));
    }
}

/* Given two side planes, find an edge along their intersection by
 * finding a pair of vertices that fall on both planes. Add it to the
 * solid. */

void LumpObject::clip_edge(Side *si, Side *sj)
{
    unsigned int i, j;

    for (i = 1; i < v.count(); i++)
	for (j = 0; j < i; j++)
	    if (si->on_side(v[i]->p) && si->on_side(v[j]->p) &&
		sj->on_side(v[i]->p) && sj->on_side(v[j]->p))
		e.append(new Edge(v[i], v[j]));
}

/* Find all verts that lie on the given side of the lump. Sort these
 * verts to have a counter-clockwise winding about the plane normal.
 * Create geoms to tessalate the resulting convex polygon. */

void LumpObject::clip_geom(Side *sp)
{
    Vertex *m[256];
    TextureCoord *t[256];
    short i, j, n = 0;
    Vector3 w;

    /* Find em. */

    foreach (Vertex *vertex, v)
	if (sp->on_side(vertex->p)) {
	    w = vertex->p + sp->plane_p;

	    m[n] = vertex;
	    t[n] = new TextureCoord(v_dot(w, sp->plane_u),
				    v_dot(w, sp->plane_v));
	    n++;
	}

    /* Sort em. */

    for (i = 1; i < n; i++)
	for (j = i + 1; j < n; j++) {
	    if (v_dot(v_crs(m[i]->p - m[0]->p,
			    m[j]->p - m[0]->p), sp->n) < 0.f) {
		Vertex *_m;
		TextureCoord *_t;

		_m    = m[i];
		m[i]  = m[j];
		m[j]  =   _m;

		_t    = t[i];
		t[i]  = t[j];
		t[j]  =   _t;
	    }
	}

    /* Index em. */

    for (i = 0; i < n - 2; i++)
	g.append(new Geometry(t[0],     sp, m[0],
			      t[i + 1], sp, m[i + 1],
			      t[i + 2], sp, m[i + 2], sp->plane_m));

    sp->vertices.clear();
    sp->textureCoords.clear();

    for (i = 0; i < n; i++) {
	sp->vertices.append(m[i]);
	sp->textureCoords.append(t[i]);
    }
}

/****************************************************************************/

CoinObject::CoinObject(Metrum3 pos, Drawing *drawing)
    : Object(drawing), pos(pos), value(1),
      handle(_("position"), this)
{
}

CoinObject::CoinObject(XMLElement *element, Drawing *drawing)
    : Object(drawing), value(1),
      handle(_("position"), this)
{
    if (element->selectAttribute("position"))
	pos = element->toMetrum3(drawing);
    if (element->selectAttribute("value"))
	value = element->toInt();
}

String CoinObject::name() const
{
    return _("coin");
}

void CoinObject::paint(Painter &painter, const Embedding &embedding)
{
    painter.setPen(0x80, 0x80, 0x00);
    painter.paintPoint(pos);
}

void CoinObject::paintShadow(const Metrum3 &pos, const ActionPoint *handle,
			     Painter &painter)
{
    Metrum3 hpos = this->pos;

    if (handle == &this->handle)
	hpos = pos;

    painter.setPen(0x80, 0x80, 0x80);
    painter.paintPoint(hpos);
}

List<ActionPoint *> CoinObject::handles()
{
    return List<ActionPoint *>() << &handle;
}

Metrum3 CoinObject::handlePos(const ActionPoint *handle) const
{
    if (handle == &this->handle)
	return pos;

    return Metrum3();
}

void CoinObject::moveHandle(const Metrum3 &pos, const ActionPoint *handle)
{
    if (handle == &this->handle)
	this->pos = pos;
}

ScreenDist CoinObject::proximity(const ScreenPos &point, View *view,
				 const Embedding &embedding) const
{
    return (point - view->map(pos)).length();
}

bool CoinObject::within(const ScreenRect &rect, View *view,
			bool partially, const Embedding &embedding) const
{
    return rect.contains(view->map(pos));
}

void CoinObject::write(XMLWriter *writer)
{
    writer->startElement("coin");
    writer->writeAttribute("position", drawing()->format(pos));
    writer->writeAttribute("value", int_to_str(value));
    writer->endElement();
}

List<Attribute *> CoinObject::attributes() const
{
    return List<Attribute *>() << &NeverballAttributes::value;
}

Variant CoinObject::attribute(const Attribute *attribute) const
{
    if (attribute == &NeverballAttributes::value)
	return Variant::fromInt(value);

    return Variant();
}

Variant CoinObject::attributeDefault(const Attribute *attribute) const
{
    if (attribute == &NeverballAttributes::value)
	return Variant::fromInt(1);

    return Variant();
}

void CoinObject::setAttribute(const Variant &value,
			      const Attribute *attribute)
{
    if (attribute == &NeverballAttributes::value)
	this->value = value.toInt();
}

/****************************************************************************/

GoalObject::GoalObject(const Metrum3 &pos, Drawing *drawing)
    : Object(drawing), pos(pos), radius(drawing->scale(0.75)),
      centerHandle(_("center"), this),
      perimeterHandle(_("perimeter"), this)
{
}

GoalObject::GoalObject(XMLElement *element, Drawing *drawing)
    : Object(drawing), radius(drawing->scale(0.75)),
      centerHandle(_("center"), this),
      perimeterHandle(_("perimeter"), this)
{
    if (element->selectAttribute("position"))
	pos = element->toMetrum3(drawing);
    if (element->selectAttribute("radius"))
	radius = element->toMetrum(drawing);
}

String GoalObject::name() const
{
    return _("goal");
}

void GoalObject::paint(Painter &painter, const Embedding &embedding)
{
    painter.setPen(0xff, 0x00, 0x00);
    painter.paintCircle(pos, radius);
    if (painter.isSelected(this))
	painter.paintLine(pos, pos + Metrum3(radius, 0, 0));
}

void GoalObject::paintShadow(const Metrum3 &pos, const ActionPoint *handle,
			     Painter &painter)
{
    Metrum3 shadow_pos = this->pos;
    Metrum3 shadow_per = this->pos + Metrum3(radius, 0, 0);
    Metrum shadow_radius = radius;

    if (handle == &centerHandle) {
	shadow_pos = pos;
	shadow_per = pos + Metrum3(radius, 0, 0);
    }
    if (handle == &perimeterHandle) {
	shadow_per = pos;
	shadow_radius = (pos - this->pos).length();
    }

    painter.setPen(0x80, 0x80, 0x80);
    painter.paintCircle(shadow_pos, shadow_radius);
    painter.paintLine(shadow_pos, shadow_per);
}

List<ActionPoint *> GoalObject::handles()
{
    return List<ActionPoint *>() << &centerHandle
				 << &perimeterHandle;
}

Metrum3 GoalObject::handlePos(const ActionPoint *handle) const
{
    if (handle == &centerHandle)
	return pos;
    if (handle == &perimeterHandle)
	return pos + Metrum3(radius, 0, 0);

    return Metrum3();
}

void GoalObject::moveHandle(const Metrum3 &pos, const ActionPoint *handle)
{
    if (handle == &centerHandle)
	this->pos = pos;
    if (handle == &perimeterHandle)
	radius = (pos - this->pos).length();
}

ScreenDist GoalObject::proximity(const ScreenPos &point, View *view,
				 const Embedding &embedding) const
{
    ScreenDist r = (point - view->map(pos)).length() -
		   (view->map(Metrum3(radius, 0, 0)) -
		    view->map(Metrum3())).length();
    if (r < 0)
	return 0;
    return r;
}

bool GoalObject::within(const ScreenRect &rect, View *view,
			bool partially, const Embedding &embedding) const
{
    // TODO

    ScreenPos s = view->map(pos);
    ScreenPos rx = view->map(Metrum3(radius, 0, 0));
    ScreenPos ry = view->map(Metrum3(0, radius, 0));

    if (!partially)
	return rect.contains(ScreenRect(s - rx - ry, s + rx + ry));

    ScreenDist r = rx.length();

    return r >= (ScreenPos(rect.x0(), rect.y0()) - s).length() ||
	   r >= (ScreenPos(rect.x1(), rect.y0()) - s).length() ||
	   r >= (ScreenPos(rect.x0(), rect.y1()) - s).length() ||
	   r >= (ScreenPos(rect.x1(), rect.y1()) - s).length();
}

void GoalObject::write(XMLWriter *writer)
{
    writer->startElement("goal");
    writer->writeAttribute("position", drawing()->format(pos));
    writer->writeAttribute("radius", drawing()->format(radius));
    writer->endElement();
}

List<Attribute *> GoalObject::attributes() const
{
    return List<Attribute *>() << &NeverballAttributes::radius;
}

Variant GoalObject::attribute(const Attribute *attribute) const
{
    if (attribute == &NeverballAttributes::radius)
	return Variant::fromMetrum(radius);

    return Variant();
}

Variant GoalObject::attributeDefault(const Attribute *attribute) const
{
    if (attribute == &NeverballAttributes::radius)
	return Variant::fromMetrum(drawing()->scale(0.75));

    return Variant();
}

void GoalObject::setAttribute(const Variant &value,
			      const Attribute *attribute)
{
    if (attribute == &NeverballAttributes::radius)
	radius = value.toMetrum();
}

/****************************************************************************/

JumpObject::JumpObject(const Metrum3 &pos, const Metrum3 &target,
		       Drawing *drawing)
    : Object(drawing), pos(pos), target(target),
      radius(drawing->scale(0.5)),
      centerHandle(_("center"), this),
      perimeterHandle(_("perimeter"), this),
      targetHandle(_("target"), this),
      tangent(_("tangent"), this),
      startNormal(_("center normal"), this),
      targetNormal(_("target normal"), this)
{
}

JumpObject::JumpObject(XMLElement *element, Drawing *drawing)
    : Object(drawing),
      radius(drawing->scale(0.5)),
      centerHandle(_("center"), this),
      perimeterHandle(_("perimeter"), this),
      targetHandle(_("target"), this),
      tangent(_("tangent"), this),
      startNormal(_("center normal"), this),
      targetNormal(_("target normal"), this)
{
    if (element->selectAttribute("position"))
	pos = element->toMetrum3(drawing);
    if (element->selectAttribute("target-position"))
	target = element->toMetrum3(drawing);
    if (element->selectAttribute("radius"))
	radius = element->toMetrum(drawing);
}

String JumpObject::name() const
{
    return _("jump");
}

void JumpObject::paint(Painter &painter, const Embedding &embedding)
{
    painter.setPen(0xa0, 0xa0, 0xa0);
    painter.paintCircle(pos, radius);
    if (painter.isSelected(this))
	painter.paintLine(pos, pos + Metrum3(radius, 0, 0));

    painter.paintLine(pos, target);
    painter.paintPoint(target);
}

void JumpObject::paintShadow(const Metrum3 &pos, const ActionPoint *handle,
			     Painter &painter)
{
    Metrum3 shadow_pos = this->pos;
    Metrum3 shadow_per = this->pos + Metrum3(radius, 0, 0);
    Metrum shadow_radius = radius;
    Metrum3 shadow_target = this->target;

    if (handle == &centerHandle) {
	shadow_pos = pos;
	shadow_per = pos + Metrum3(radius, 0, 0);
    }
    if (handle == &perimeterHandle) {
	shadow_per = pos;
	shadow_radius = (pos - this->pos).length();
    }
    if (handle == &targetHandle)
	shadow_target = pos;

    painter.setPen(0x80, 0x80, 0x80);
    painter.paintCircle(shadow_pos, shadow_radius);
    painter.paintLine(shadow_pos, shadow_per);

    painter.paintLine(shadow_pos, shadow_target);
    painter.paintPoint(shadow_target);
}

List<ActionPoint *> JumpObject::handles()
{
    return List<ActionPoint *>() << &centerHandle
				 << &perimeterHandle
				 << &targetHandle;
}

Metrum3 JumpObject::handlePos(const ActionPoint *handle) const
{
    if (handle == &this->centerHandle)
	return pos;
    if (handle == &this->perimeterHandle)
	return pos + Metrum3(radius, 0, 0);
    if (handle == &this->targetHandle)
	return target;

    return Metrum3();
}

List<ActionLine *> JumpObject::handleLines(const ActionPoint *handle)
{
    if (handle == &this->centerHandle)
	return List<ActionLine *>() << &tangent << &startNormal;
    if (handle == &this->targetHandle)
	return List<ActionLine *>() << &tangent << &targetNormal;

    return List<ActionLine *>();
}

void JumpObject::moveHandle(const Metrum3 &pos, const ActionPoint *handle)
{
    if (handle == &this->centerHandle)
	this->pos = pos;
    if (handle == &this->perimeterHandle)
	radius = (pos - this->pos).length();
    if (handle == &this->targetHandle)
	target = pos;
}

Metrum3 JumpObject::lineBase(const ActionLine *line) const
{
    if (line == &tangent || line == &startNormal)
	return pos;
    if (line == &targetNormal)
	return target;

    return Object::lineBase(line);
}

Metrum3 JumpObject::lineTangent(const ActionLine *line) const
{
    if (line == &tangent)
	return target - pos;

    if (line == &startNormal || line == &targetNormal)
	return Metrum3((target - pos).y(), (pos - target).x(), 0);

    return Object::lineTangent(line);
}

ScreenDist JumpObject::proximity(const ScreenPos &point, View *view,
				 const Embedding &embedding) const
{
    /* TODO */
    return (point - view->map(pos)).length();
}

bool JumpObject::within(const ScreenRect &rect, View *view,
			bool partially, const Embedding &embedding) const
{
    /* TODO */
    return rect.contains(view->map(pos));
}

void JumpObject::write(XMLWriter *writer)
{
    writer->startElement("jump");
    writer->writeAttribute("position", drawing()->format(pos));
    writer->writeAttribute("target-position", drawing()->format(target));
    writer->writeAttribute("radius", drawing()->format(radius));
    writer->endElement();
}

List<Attribute *> JumpObject::attributes() const
{
    return List<Attribute *>() << &NeverballAttributes::radius;
}

Variant JumpObject::attribute(const Attribute *attribute) const
{
    if (attribute == &NeverballAttributes::radius)
	return Variant::fromMetrum(radius);

    return Variant();
}

Variant JumpObject::attributeDefault(const Attribute *attribute) const
{
    if (attribute == &NeverballAttributes::radius)
	return Variant::fromMetrum(drawing()->scale(0.5));

    return Variant();
}

void JumpObject::setAttribute(const Variant &value,
			      const Attribute *attribute)
{
    if (attribute == &NeverballAttributes::radius)
	radius = value.toMetrum();
}

/****************************************************************************/

SwitchObject::SwitchObject(const Metrum3 &pos, Drawing *drawing)
    : Object(drawing), pos(pos), radius(drawing->scale(0.5)),
      path(0), timer(0), state(0),
      centerHandle(_("center"), this),
      perimeterHandle(_("perimeter"), this)
{
}

SwitchObject::SwitchObject(XMLElement *element, Drawing *drawing)
    : Object(drawing), radius(drawing->scale(0.5)),
      path(0), timer(0), state(0),
      centerHandle(_("center"), this),
      perimeterHandle(_("perimeter"), this)
{
    if (element->selectAttribute("position"))
	pos = element->toMetrum3(drawing);
    if (element->selectAttribute("radius"))
	radius = element->toMetrum(drawing);
    if (element->selectAttribute("path")) {
	/* Todo */
    }
    if (element->selectAttribute("timer"))
	timer = element->toDouble();
    if (element->selectAttribute("state"))
	state = element->toInt();
}

String SwitchObject::name() const
{
    return _("switch");
}

void SwitchObject::paint(Painter &painter, const Embedding &embedding)
{
    painter.setPen(0xa0, 0xa0, 0xa0);
    painter.paintCircle(pos, radius);
    if (painter.isSelected(this))
	painter.paintLine(pos, pos + Metrum3(radius, 0, 0));

    if (path)
	painter.paintLine(pos, path->pos());
}

void SwitchObject::paintShadow(const Metrum3 &pos, const ActionPoint *handle,
			       Painter &painter)
{
    Metrum3 shadow_pos = this->pos;
    Metrum3 shadow_per = this->pos + Metrum3(radius, 0, 0);
    Metrum shadow_radius = radius;

    if (handle == &centerHandle) {
	shadow_pos = pos;
	shadow_per = pos + Metrum3(radius, 0, 0);
    }
    if (handle == &perimeterHandle) {
	shadow_per = pos;
	shadow_radius = (pos - this->pos).length();
    }

    painter.setPen(0x80, 0x80, 0x80);
    painter.paintCircle(shadow_pos, shadow_radius);
    painter.paintLine(shadow_pos, shadow_per);

    if (path)
	painter.paintLine(pos, path->pos());
}

List<ActionPoint *> SwitchObject::handles()
{
    return List<ActionPoint *>() << &centerHandle
				 << &perimeterHandle;
}

Metrum3 SwitchObject::handlePos(const ActionPoint *handle) const
{
    if (handle == &this->centerHandle)
	return pos;
    if (handle == &this->perimeterHandle)
	return pos + Metrum3(radius, 0, 0);

    return Metrum3();
}

void SwitchObject::moveHandle(const Metrum3 &pos, const ActionPoint *handle)
{
    if (handle == &this->centerHandle)
	this->pos = pos;
    if (handle == &this->perimeterHandle)
	radius = (pos - this->pos).length();
}

ScreenDist SwitchObject::proximity(const ScreenPos &point, View *view,
				   const Embedding &embedding) const
{
    /* TODO */
    return (point - view->map(pos)).length();
}

bool SwitchObject::within(const ScreenRect &rect, View *view,
			  bool partially, const Embedding &embedding) const
{
    /* TODO */
    return rect.contains(view->map(pos));
}

void SwitchObject::write(XMLWriter *writer)
{
    writer->startElement("switch");
    writer->writeAttribute("position", drawing()->format(pos));
    writer->writeAttribute("radius", drawing()->format(radius));
    writer->writeAttribute("path", path ? path->symbol() : String());
    writer->writeAttribute("timer", double_to_str(timer));
    writer->writeAttribute("state", int_to_str(state));
    writer->endElement();
}

List<Attribute *> SwitchObject::attributes() const
{
    return List<Attribute *>() << &NeverballAttributes::radius
			       << &NeverballAttributes::path
			       << &NeverballAttributes::timer
			       << &NeverballAttributes::state;
}

Variant SwitchObject::attribute(const Attribute *attribute) const
{
    if (attribute == &NeverballAttributes::radius)
	return Variant::fromMetrum(radius);
    if (attribute == &NeverballAttributes::path)
	return Variant::fromPointer<PathObject *>(path);
    if (attribute == &NeverballAttributes::timer)
	return Variant::fromDouble(timer);
    if (attribute == &NeverballAttributes::state)
	return Variant::fromInt(state);

    return Variant();
}

Variant SwitchObject::attributeDefault(const Attribute *attribute) const
{
    if (attribute == &NeverballAttributes::radius)
	return Variant::fromMetrum(drawing()->scale(0.5));
    if (attribute == &NeverballAttributes::path)
	return Variant::fromPointer<PathObject *>(0);
    if (attribute == &NeverballAttributes::timer)
	return Variant::fromDouble(0.);
    if (attribute == &NeverballAttributes::state)
	return Variant::fromInt(0);

    return Variant();
}

void SwitchObject::setAttribute(const Variant &value,
				const Attribute *attribute)
{
    if (attribute == &NeverballAttributes::radius)
	radius = value.toMetrum();
    if (attribute == &NeverballAttributes::path)
	path = value.toPointer<PathObject *>();
    if (attribute == &NeverballAttributes::timer)
	timer = value.toDouble();
    if (attribute == &NeverballAttributes::state)
	state = value.toInt();
}

/****************************************************************************/

BillboardObject::BillboardObject(const Metrum3 &pos, Drawing *drawing)
    : Object(drawing), fl(0), repeatTimeInterval(1.0), pos(pos)
{
}

BillboardObject::BillboardObject(XMLElement *element, Drawing *drawing)
    : Object(drawing), fl(0), repeatTimeInterval(1.0)
{
}

String BillboardObject::name() const
{
    return _("billboard");
}

void BillboardObject::paint(Painter &painter, const Embedding &embedding)
{
    /* TODO */
}

void BillboardObject::paintShadow(const Metrum3 &pos, const ActionPoint *handle,
				  Painter &painter)
{
    /* TODO */
}

List<ActionPoint *> BillboardObject::handles()
{
    /* TODO */

    return List<ActionPoint *>();
}

Metrum3 BillboardObject::handlePos(const ActionPoint *handle) const
{
    /* TODO */

    return Metrum3();
}

List<ActionLine *> BillboardObject::handleLines(const ActionPoint *handle)
{
    /* TODO */

    return List<ActionLine *>();
}

void BillboardObject::moveHandle(const Metrum3 &pos, const ActionPoint *handle)
{
    /* TODO */
}

Metrum3 BillboardObject::lineBase(const ActionLine *line) const
{
    /* TODO */

    return Object::lineBase(line);
}

Metrum3 BillboardObject::lineTangent(const ActionLine *line) const
{
    /* TODO */

    return Object::lineTangent(line);
}

ScreenDist BillboardObject::proximity(const ScreenPos &point, View *view,
				      const Embedding &embedding) const
{
    /* TODO */
    return 1000;
}

bool BillboardObject::within(const ScreenRect &rect, View *view,
			     bool partially, const Embedding &embedding) const
{
    /* TODO */
    return false;
}

void BillboardObject::write(XMLWriter *writer)
{
    /* TODO */
}

List<Attribute *> BillboardObject::attributes() const
{
    /* TODO */

    return List<Attribute *>();
}

Variant BillboardObject::attribute(const Attribute *attribute) const
{
    /* TODO */

    return Variant();
}

Variant BillboardObject::attributeDefault(const Attribute *attribute) const
{
    /* TODO */

    return Variant();
}

void BillboardObject::setAttribute(const Variant &value,
				   const Attribute *attribute)
{
    /* TODO */
}

/****************************************************************************/

UserObject::UserObject(const Metrum3 &pos, Drawing *drawing)
    : Object(drawing), pos(pos), radius(drawing->scale(0.25)),
      centerHandle(_("center"), this),
      perimeterHandle(_("perimeter"), this)
{
}

UserObject::UserObject(XMLElement *element, Drawing *drawing)
    : Object(drawing), radius(drawing->scale(0.25)),
      centerHandle(_("center"), this),
      perimeterHandle(_("perimeter"), this)
{
    if (element->selectAttribute("position"))
	pos = element->toMetrum3(drawing);
    if (element->selectAttribute("radius"))
	radius = element->toMetrum(drawing);
}

String UserObject::name() const
{
    return _("user");
}

void UserObject::paint(Painter &painter, const Embedding &embedding)
{
    painter.setBlackPen();
    painter.paintCircle(pos, radius);
    if (painter.isSelected(this))
	painter.paintLine(pos, pos + Metrum3(radius, 0, 0));
}

void UserObject::paintShadow(const Metrum3 &pos, const ActionPoint *handle,
			     Painter &painter)
{
    Metrum3 shadow_pos = this->pos;
    Metrum3 shadow_per = this->pos + Metrum3(radius, 0, 0);
    Metrum shadow_radius = radius;

    if (handle == &centerHandle) {
	shadow_pos = pos;
	shadow_per = pos + Metrum3(radius, 0, 0);
    }
    if (handle == &perimeterHandle) {
	shadow_per = pos;
	shadow_radius = (pos - this->pos).length();
    }

    painter.setPen(0x80, 0x80, 0x80);
    painter.paintCircle(shadow_pos, shadow_radius);
    painter.paintLine(shadow_pos, shadow_per);
}

List<ActionPoint *> UserObject::handles()
{
    return List<ActionPoint *>() << &centerHandle
				 << &perimeterHandle;
}

Metrum3 UserObject::handlePos(const ActionPoint *handle) const
{
    if (handle == &this->centerHandle)
	return pos;
    if (handle == &this->perimeterHandle)
	return pos + Metrum3(radius, 0, 0);

    return Metrum3();
}

void UserObject::moveHandle(const Metrum3 &pos, const ActionPoint *handle)
{
    if (handle == &centerHandle)
	this->pos = pos;
    if (handle == &perimeterHandle)
	radius = (pos - this->pos).length();
}

ScreenDist UserObject::proximity(const ScreenPos &point, View *view,
				 const Embedding &embedding) const
{
    /* TODO */
    return (point - view->map(pos)).length();
}

bool UserObject::within(const ScreenRect &rect, View *view,
			bool partially, const Embedding &embedding) const
{
    /* TODO */
    return rect.contains(view->map(pos));
}

void UserObject::write(XMLWriter *writer)
{
    writer->startElement("ball");
    writer->writeAttribute("position", drawing()->format(pos));
    writer->writeAttribute("radius", drawing()->format(radius));
    writer->endElement();
}

List<Attribute *> UserObject::attributes() const
{
    return List<Attribute *>() << &NeverballAttributes::radius;
}

Variant UserObject::attribute(const Attribute *attribute) const
{
    if (attribute == &NeverballAttributes::radius)
	return Variant::fromMetrum(radius);

    return Variant();
}

Variant UserObject::attributeDefault(const Attribute *attribute) const
{
    if (attribute == &NeverballAttributes::radius)
	return Variant::fromMetrum(drawing()->scale(0.25));

    return Variant();
}

void UserObject::setAttribute(const Variant &value,
			      const Attribute *attribute)
{
    if (attribute == &NeverballAttributes::radius)
	radius = value.toMetrum();
}

/****************************************************************************/

ViewObject::ViewObject(const Metrum3 &pos, const Metrum3 &target,
		       Drawing *drawing)
    : Object(drawing), pos(pos), target(target),
      posHandle(_("camera position"), this),
      targetHandle(_("camera target"), this),
      tangent(_("tangent"), this),
      posNormal(_("camera position normal"), this),
      targetNormal(_("camera target normal"), this)
{
}

ViewObject::ViewObject(XMLElement *element, Drawing *drawing)
    : Object(drawing),
      posHandle(_("camera position"), this),
      targetHandle(_("camera target"), this),
      tangent(_("tangent"), this),
      posNormal(_("camera position normal"), this),
      targetNormal(_("camera target normal"), this)
{
    if (element->selectAttribute("position"))
	pos = element->toMetrum3(drawing);
    if (element->selectAttribute("target-position"))
	target = element->toMetrum3(drawing);
}

String ViewObject::name() const
{
    return _("view");
}

Icon *ViewObject::icon() const
{
    return icons::eye;
}

void ViewObject::paint(Painter &painter, const Embedding &embedding)
{
    painter.setPen(0xa0, 0xa0, 0xa0);
    painter.paintPoint(pos);
    painter.paintLine(pos, target);
    painter.paintPoint(target);
}

void ViewObject::paintShadow(const Metrum3 &pos, const ActionPoint *handle,
			     Painter &painter)
{
    Metrum3 shadow_pos(this->pos), shadow_target(target);

    if (handle == &posHandle)
	shadow_pos = pos;
    /*
    if (handle == &midpointHandle) {
	Metrum3 delta = p1 - p0;
	hp0 = pos - delta / 2;
	hp1 = hp0 + delta;
    }
    */
    if (handle == &targetHandle)
	shadow_target = pos;

    painter.setPen(0x80, 0x80, 0x80);
    painter.paintLine(shadow_pos, shadow_target);
}

List<ActionPoint *> ViewObject::handles()
{
    return List<ActionPoint *>() << &posHandle
				 << &targetHandle;
}

Metrum3 ViewObject::handlePos(const ActionPoint *handle) const
{
    if (handle == &this->posHandle)
	return pos;
    if (handle == &this->targetHandle)
	return target;

    return Metrum3();
}

List<ActionLine *> ViewObject::handleLines(const ActionPoint *handle)
{
    if (handle == &this->posHandle)
	return List<ActionLine *>() << &tangent << &posNormal;
    if (handle == &this->targetHandle)
	return List<ActionLine *>() << &tangent << &targetNormal;

    return List<ActionLine *>();
}

void ViewObject::moveHandle(const Metrum3 &pos, const ActionPoint *handle)
{
    if (handle == &this->posHandle)
	this->pos = pos;
    if (handle == &this->targetHandle)
	target = pos;
}

Metrum3 ViewObject::lineBase(const ActionLine *line) const
{
    if (line == &tangent || line == &posNormal)
	return pos;
    if (line == &targetNormal)
	return target;

    return Object::lineBase(line);
}

Metrum3 ViewObject::lineTangent(const ActionLine *line) const
{
    if (line == &tangent)
	return target - pos;

    if (line == &posNormal || line == &targetNormal)
	return Metrum3((target - pos).y(), (pos - target).x(), 0);

    return Object::lineTangent(line);
}

ScreenDist ViewObject::proximity(const ScreenPos &point, View *view,
				 const Embedding &embedding) const
{
    /* TODO */
    return (point - view->map(pos)).length();
}

bool ViewObject::within(const ScreenRect &rect, View *view,
			bool partially, const Embedding &embedding) const
{
    ScreenPos s0 = view->map(pos);
    ScreenPos s1 = view->map(target);

    if (!partially)
	return rect.contains(s0) && rect.contains(s1);

    return rect.contains(s0) || rect.contains(s1); //TODO
}

void ViewObject::write(XMLWriter *writer)
{
    writer->startElement("view");
    writer->writeAttribute("position", drawing()->format(pos));
    writer->writeAttribute("target-position", drawing()->format(target));
    writer->endElement();
}

List<Attribute *> ViewObject::attributes() const
{
    return List<Attribute *>();
}

Variant ViewObject::attribute(const Attribute *attribute) const
{
    return Variant();
}

Variant ViewObject::attributeDefault(const Attribute *attribute) const
{
    return Variant();
}

void ViewObject::setAttribute(const Variant &value,
			      const Attribute *attribute)
{
}
