/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <assert.h>
#include "object.h"
#include "util/metrum3.h"
#include "drawing.h"
#include "objectmodel.h"
#include "icons.h"


int EmbeddingLayer::angle() const
{
    return 0;
}

bool EmbeddingLayer::mirror() const
{
    return false;
}

Metrum EmbeddingLayer::transform(Metrum m) const
{
    return m;
}

Metrum3 EmbeddingLayer::transform(const Metrum3 &pos) const
{
    return pos;
}

/****************************************************************************/

Embedding::Embedding() : List<EmbeddingLayer *>()
{
}

Embedding::Embedding(const Embedding &embedding, EmbeddingLayer *l)
    : List<EmbeddingLayer *>(embedding)
{
    prepend(l);
}

int Embedding::angle() const
{
    int result = 0;

    foreach (EmbeddingLayer *l, *this)
	result += l->angle();

    return result % 4;
}

bool Embedding::mirror() const
{
    bool result = false;

    foreach (EmbeddingLayer *l, *this)
	result = result != l->mirror();

    return result;
}

Metrum Embedding::transform(Metrum m) const
{
    Metrum result(m);

    foreach (EmbeddingLayer *l, *this)
	result = l->transform(result);

    return result;
}

Metrum3 Embedding::transform(const Metrum3 &pos) const
{
    Metrum3 result(pos);

    foreach (EmbeddingLayer *l, *this)
	result = l->transform(result);

    return result;
}

/****************************************************************************/

Object::Object(Drawing *drawing) : the_drawing(drawing), the_context(0),
				   _item(0)
{
    the_drawing->registerObject(this);
}

Object::~Object()
{
    the_drawing->unregisterObject(this);
    if (_item)
	delete _item;
}

void Object::construct_item()
{
    if (_item)
	return;
    _item = new ObjectItem(this);
}

Drawing *Object::drawing() const
{
    return the_drawing;
}

ObjectContext *Object::context() const
{
    return the_context;
}

void Object::setContext(ObjectContext *context)
{
    if (context == the_context)
	return;

    assert(!the_context || !the_context->objects().contains(this));
    assert(!context || context->objects().contains(this));
    the_context = context;

    drawing()->updateObjectList();
}

Object *Object::parent() const
{
    return context() ? context()->parent() : 0;
}

ObjectItem *Object::item() const
{
    const_cast<Object *>(this)->construct_item();
    return _item;
}

Icon *Object::icon() const
{
    return icons::piece;
}

ObjectContext *Object::children() const
{
    return 0;
}

void Object::paint(Painter &/*painter*/, const Embedding &/*embedding*/)
{
}

void Object::paintShadow(const Metrum3 &/*pos*/, const ActionPoint */*handle*/,
			 Painter &/*painter*/)
{
}

List<ActionPoint *> Object::handles()
{
    return List<ActionPoint *>();
}

Metrum3 Object::handlePos(const ActionPoint */*handle*/) const
{
    return Metrum3();
}

List<ActionLine *> Object::handleLines(const ActionPoint */*handle*/)
{
    return List<ActionLine *>();
}

void Object::moveHandle(const Metrum3 &/*pos*/, const ActionPoint */*handle*/)
{
}

Metrum3 Object::lineBase(const ActionLine */*line*/) const
{
    return Metrum3();
}

Metrum3 Object::lineTangent(const ActionLine */*line*/) const
{
    return Metrum3(1, 0, 0);
}

Metrum3 Object::relativePos() const
{
    return Metrum3();
}

Metrum3 Object::absolutePos() const
{
    assert(context());
    return context()->origin() + relativePos();
}

/****************************************************************************/

ObjectContext::ObjectContext(Object *parent)
    : _drawing(parent->drawing()), _parent(parent),
      _interior(new ObjectInterior(this))
{
}

ObjectContext::ObjectContext(Drawing *drawing)
    : _drawing(drawing), _parent(0), _interior(new ObjectInterior(this))
{
}

ObjectContext::~ObjectContext()
{
    delete _interior;
}

Drawing *ObjectContext::drawing() const
{
    return _drawing;
}

Object *ObjectContext::parent() const
{
    return _parent;
}

ItemInterior *ObjectContext::interior() const
{
    return _interior;
}

List<Object *> ObjectContext::objects() const
{
    return _objects;
}

void ObjectContext::addObject(Object *object)
{
    assert(!_objects.contains(object));
    _objects.append(object);
    object->setContext(this);
    _interior->replace(objects().count() - 1, 0,
		       List<Item *>() << object->item());
}

void ObjectContext::insertObject(int index, Object *object)
{
    assert(!_objects.contains(object));
    _objects.insert(index, object);
    object->setContext(this);
    _interior->replace(index, 0, List<Item *>() << object->item());
}

void ObjectContext::removeObject(Object *object)
{
    assert(_objects.contains(object));
    _objects.removeAll(object);
    object->setContext(0);
    _interior->replace(interior()->cachedIndexOf(object->item()), 1,
		       List<Item *>());
}

void ObjectContext::write(XMLWriter *writer)
{
    foreach (Object *ob, objects())
	ob->write(writer);
}

Metrum3 ObjectContext::origin() const
{
    return parent() ? parent()->absolutePos() : Metrum3();
}
