/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <math.h>
#include "painter.h"
#include "hub.h"
#include "majormode.h"
#include "view.h"


Painter::Painter(Hub &hub) : hub(hub)
{
}

bool Painter::isSelected(Object *object) const
{
    return hub.majorMode()->selected().contains(object);
}

void Painter::paintPoint(const Metrum3 &point)
{
    ScreenPos p = hub.view()->map(point);
    int x = llrint(p.x()), y = llrint(p.y());
    drawLine(x - 10, y, x + 10, y);
    drawLine(x, y - 10, x, y + 10);
}

void Painter::paintActivePoint(const Metrum3 &point)
{
    ScreenPos p = hub.view()->map(point);
    int x = llrint(p.x()), y = llrint(p.y());
    drawLine(x - 10, y, x + 10, y);
    drawLine(x, y - 10, x, y + 10);
    drawArc(x - 3, y - 3, 6, 6);
}

void Painter::paintHandle(const Metrum3 &point, unsigned int red,
			  unsigned int green, unsigned int blue)
{
    ScreenPos pos = hub.view()->map(point);
    int x = llrint(pos.x()), y = llrint(pos.y());
    drawRectangle(x - 2, y - 2, 4, 4);
    setPen(red, green, blue);
    fillRectangle(x - 1, y - 1, 3, 3);
}

void Painter::paintCircle(const Metrum3 &point, Metrum radius)
{
    Metrum3 rx(radius, 0, 0);
    Metrum3 ry(0, radius, 0);

    ScreenPos ppx = hub.view()->map(point + rx);
    ScreenPos pnx = hub.view()->map(point - rx);
    ScreenPos ppy = hub.view()->map(point + ry);
    ScreenPos pny = hub.view()->map(point - ry);

    drawLine(ppx, ppy);
    drawLine(ppy, pnx);
    drawLine(pnx, pny);
    drawLine(pny, ppx);
}

void Painter::paintLine(const Metrum3 &point0,
			const Metrum3 &point1, bool extend)
{
    ScreenPos p0 = hub.view()->map(point0);
    ScreenPos p1 = hub.view()->map(point1);

    if (p0 == p1)
	return;

    ScreenDist minX = visibleRect().x0();
    ScreenDist maxX = visibleRect().x1();
    ScreenDist minY = visibleRect().y0();
    ScreenDist maxY = visibleRect().y1();

    if (p0.y() == p1.y()) {
	if (p1.x() < p0.x()) {
	    ScreenPos tmp = p1;
	    p1 = p0;
	    p0 = tmp;
	}

	if (!extend && (p1.x() < minX || p0.x() > maxX))
	    return;
	if (p0.y() < minY || p0.y() > maxY)
	    return;

	if (extend || p0.x() < minX)
	    p0.setX(minX);
	if (extend || p1.x() > maxX)
	    p1.setX(maxX);
    } else if (p0.x() == p1.x()) {
	if (p1.y() < p0.y()) {
	    ScreenPos tmp = p1;
	    p1 = p0;
	    p0 = tmp;
	}

	if (!extend && (p1.y() < minY || p0.y() > maxY))
	    return;
	if (p0.x() < minX || p0.x() > maxX)
	    return;

	if (extend || p0.y() < minY)
	    p0.setY(minY);
	if (extend || p1.y() > maxY)
	    p1.setY(maxY);
    } else {
	if (p1.x() < p0.x()) {
	    ScreenPos tmp = p1;
	    p1 = p0;
	    p0 = tmp;
	}

	ScreenPos delta = p1 - p0;
	ScreenDist mx = delta.y() / delta.x();
	ScreenDist my = delta.x() / delta.y();

	if (extend || p0.x() < minX)
	    p0 = ScreenPos(minX, p0.y() - mx * (p0.x() - minX));
	if (extend || p1.x() > maxX)
	    p1 = ScreenPos(maxX, p1.y() - mx * (p1.x() - maxX));

	if (p0.y() < minY)
	    p0 = ScreenPos(p0.x() - my * (p0.y() - minY), minY);
	if (p0.y() > maxY)
	    p0 = ScreenPos(p0.x() - my * (p0.y() - maxY), maxY);

	if (p1.y() < minY)
	    p1 = ScreenPos(p1.x() - my * (p1.y() - minY), minY);
	if (p1.y() > maxY)
	    p1 = ScreenPos(p1.x() - my * (p1.y() - maxY), maxY);

	if (p0 == p1 || p1.x() < minX || p0.x() >= maxX)
	    return;
    }

    drawLine(p0, p1);
}

void Painter::paintRectangle(const ScreenRect &rect, unsigned int red,
			     unsigned int green, unsigned int blue)
{
    int x = llrint(rect.x0()), y = llrint(rect.y0());
    unsigned int width = llrint(rect.x1()) - x;
    unsigned int height = llrint(rect.y1()) - y;

    drawRectangle(x, y, width, height);
    setPen(red, green, blue);
    if (width >= 1 && height >= 1)
	fillRectangle(x + 1, y + 1, width - 1, height - 1);
}
