/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "read_xml.h"
#include "dirmodel.h"

#include "sch_readxml.h"
#include "sch_attribute.h"
#include "nb_readxml.h"

static VirtualItem *build_virtual_item(
    XMLElement *element, Messages &messages,
    Library *library, Font1 *font, FileItem *fileItem);


class FileXMLHandler : public XMLHandler
{
    Library *const library;
    Font1 *const font;
    FileItem *const file_item;
public:
    FileXMLHandler(Library *library, Font1 *font, FileItem *fileItem)
	: library(library), font(font), file_item(fileItem) {
    }
    void element(XMLElement *element, Messages &messages) {
	file_item->assignVirtualItem(
	    build_virtual_item(element, messages, library, font, file_item));
    }
    /* If there is no document element, the parser will complain,
     * so we don't need to check this here. */
};

class ContainerXMLHandler : public XMLHandler
{
    Library *const library;
    Font1 *const font;
    FileItem *const file_item;
    FileInterior &interior;
public:
    ContainerXMLHandler(Library *library, Font1 *font, FileItem *fileItem,
			FileInterior &interior)
	: library(library), font(font), file_item(fileItem),
	  interior(interior) {
    }
    void element(XMLElement *element, Messages &messages) {
	interior.addItem(build_virtual_item(element, messages,
					    library, font, file_item));
    }
};

static VirtualItem *build_virtual_item(
    XMLElement *element, Messages &messages,
    Library *library, Font1 *font, FileItem *fileItem)
{
    String xi_ns = "https://hedmen.org/xi/";

    if (element->name() == xi_ns + NSSEP + "container") {
	ContainerItem *item = new ContainerItem(fileItem);
	element->setHandler(new ContainerXMLHandler(
				library, font, fileItem, item->interior));
	return item;
    }

    if (element->name() == "font") {
	FontItem *item = new FontItem(fileItem);
	element->setHandler(new ContainerXMLHandler(
				0, item->thisFont(), fileItem, item->interior));
	return item;
    }

    if (element->name() == xi_ns + NSSEP + "file") {
	DrawingItem *item = new DrawingItem(Drawing::GenericFile,
					    library, fileItem);
	Drawing *drawing = item->drawing();

	element->setNamespaceURI(xi_ns);
	element->setHandler(new SchematicXMLHandler(drawing,
						    drawing->rootContext()));

	drawing->setBaseBase(10);
	drawing->setBaseExponent(0);

	if (element->selectAttribute("base"))
	    switch (element->toInt64()) {
	    case 1000:
		drawing->setBaseExponent(3);
		break;
	    case 100:
		drawing->setBaseExponent(2);
		break;
	    default:
		messages.error(_("invalid coordinate base"));
	    }
	else
	    messages.error(_("Coordinate base not set"));

	if (element->selectAttribute("granularity")) {
	}

	if (element->selectAttribute("units"))
	    if (element->value() != "metric")
		messages.error(_("Only metric units supported"));

	return item;
    }

    if (element->name() == "schematic" || element->name() == "symbol") {
	DrawingItem *item = new DrawingItem(
	    element->name() == "symbol" ? Drawing::SymbolFile
					: Drawing::SchematicFile,
	    library, fileItem);
	Drawing *drawing = item->drawing();

	element->setHandler(new SchematicXMLHandler(drawing,
						    drawing->rootContext()));

	drawing->setBaseBase(10);
	drawing->setBaseExponent(3);

	if (element->selectAttribute("version")) {
	}
	if (element->selectAttribute("fileformat_version")) {
	}

	return item;
    }

    if (element->name() == "letter") {
	DrawingItem *item = new DrawingItem(Drawing::LetterFile,
					    library, fileItem);
	Drawing *drawing = item->drawing();

	element->setHandler(new SchematicXMLHandler(drawing,
						    drawing->rootContext()));

	drawing->setBaseBase(10);
	drawing->setBaseExponent(3);

	if (element->selectAttribute("version")) {
	}
	if (element->selectAttribute("fileformat_version")) {
	}
	if (element->selectAttribute("character"))
	    drawing->setLetterCharacter(element->value());
	if (element->selectAttribute("width"))
	    drawing->setLetterWidth(element->toMetrum(drawing));
	if (element->selectAttribute("space"))
	    drawing->setLetterSpace(element->lookup(SchematicEnum::boolean));

	if (font)
	    font->addCharacter(drawing);

	return item;
    }

    if (element->name() == "map")
	return neverball_read_xml(element, messages, library, fileItem);

    messages.error(_("Invalid drawing root: %1").arg(element->name()));
    return new DrawingItem(Drawing::GenericFile, library, fileItem);
}

Shared::Sink *new_xml_sink(Library *library, Font1 *font, FileItem *fileItem,
			   Messages &messages)
{
    return new XMLSink(new FileXMLHandler(library, font, fileItem), messages);
}
