/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "sch_circle.h"
#include "hub.h"
#include "drawing.h"
#include "action.h"
#include "icons.h"
#include "painter.h"
#include "view.h"
#include "xml_writer.h"


CircleMajorMode::CircleMajorMode(Hub &hub) : MajorMode(hub),
					     posMinorMode(hub, *this),
					     radiusMinorMode(hub, *this)
{
    setCaption(_("Circle"));
    setMinorMode(&posMinorMode);

/** QVBoxLayout *layout = new QVBoxLayout(widget());
    layout->addStretch(); **/
}

void CircleMajorMode::accept(MinorMode *minor_mode)
{
    if (minor_mode == &posMinorMode)
	setMinorMode(&radiusMinorMode);

    if (minor_mode == &radiusMinorMode) {
	Metrum3 delta = radiusMinorMode.point() - posMinorMode.point();
	delta.rz() = 0;

	hub.drawing()->performActions(
	    List<Action *>() << new NewObjectAction(
		new CircleObject(posMinorMode.point(),
				 delta.length(), hub.drawing())),
	    icons::add, _("New circle"));

	cancel();
    }
}

void CircleMajorMode::cancel()
{
    setMinorMode(&posMinorMode, true);
}

void CircleMajorMode::paint(Painter &painter)
{
    MajorMode::paint(painter);

    painter.setPen(0x7f, 0x7f, 0x7f);
    if (minorMode() == &radiusMinorMode && hub.pointerActive()) {
	Metrum3 delta = radiusMinorMode.point() - posMinorMode.point();
	delta.rz() = 0;
	painter.paintCircle(posMinorMode.point(), delta.length());
	painter.paintLine(posMinorMode.point(), radiusMinorMode.point());
    }

    painter.setBlackPen();
    if (minorMode() == &radiusMinorMode || hub.pointerActive())
	painter.paintPoint(posMinorMode.point());
    if (minorMode() == &radiusMinorMode && hub.pointerActive())
	painter.paintPoint(radiusMinorMode.point());
}

List<ActionLine *> CircleMajorMode::lines(ActionPoint *handle)
{
    return List<ActionLine *>();
}

/****************************************************************************/

CircleObject::CircleObject(const Metrum3 &pos, Metrum radius, Drawing *drawing)
    : Object(drawing), pos(pos), radius(radius),
      is_arc(false), startangle(0), sweepangle(0),
      color(3), line(drawing), fill(drawing),
      posHandle(_("center"), this),
      perimeterHandle(_("radius"), this)
{
}

String CircleObject::name() const
{
    return is_arc ? _("arc") : _("circle");
}

void CircleObject::paint(Painter &painter, const Embedding &embedding)
{
    painter.setBlackPen();
    painter.paintCircle(embedding.transform(pos),
			embedding.transform(radius));
    if (painter.isSelected(this))
	painter.paintLine(embedding.transform(pos),
			  embedding.transform(pos + Metrum3(radius, 0, 0)));
}

void CircleObject::paintShadow(const Metrum3 &pos, const ActionPoint *handle,
			       Painter &painter)
{
    Metrum3 shadow_pos = this->pos;
    Metrum3 shadow_per = this->pos + Metrum3(radius, 0, 0);
    Metrum shadow_radius = radius;

    if (handle == &posHandle) {
	shadow_pos = pos;
	shadow_per = pos + Metrum3(radius, 0, 0);
    }
    if (handle == &perimeterHandle) {
	shadow_per = pos;
	shadow_radius = (pos - this->pos).length();
    }

    painter.setPen(0x80, 0x80, 0x80);
    painter.paintCircle(shadow_pos, shadow_radius);
    painter.paintLine(shadow_pos, shadow_per);
}

List<ActionPoint *> CircleObject::handles()
{
    return List<ActionPoint *>() << &posHandle << &perimeterHandle;
}

Metrum3 CircleObject::handlePos(const ActionPoint *handle) const
{
    if (handle == &posHandle)
	return pos;
    if (handle == &perimeterHandle)
	return pos + Metrum3(radius, 0, 0);

    return Metrum3();
}

void CircleObject::moveHandle(const Metrum3 &pos, const ActionPoint *handle)
{
    if (handle == &posHandle)
	this->pos = pos;
    if (handle == &perimeterHandle)
	radius = (pos - this->pos).length();
}

ScreenDist CircleObject::proximity(const ScreenPos &point, View *view,
				   const Embedding &embedding) const
{
    ScreenDist r =
	(point - view->map(embedding.transform(pos))).length() -
	(view->map(embedding.transform(Metrum3(radius, 0, 0))) -
	 view->map(embedding.transform(Metrum3(0, 0, 0)))).length();
    if (r < 0)
	return 0;
    return r;
}

bool CircleObject::within(const ScreenRect &rect, View *view,
			  bool partially, const Embedding &embedding) const
{
    // TODO

    ScreenPos s = view->map(embedding.transform(pos));
    ScreenPos rx = view->map(embedding.transform(Metrum3(radius, 0, 0)));
    ScreenPos ry = view->map(embedding.transform(Metrum3(0, radius, 0)));

    if (!partially)
	return rect.contains(ScreenRect(s - rx - ry, s + rx + ry));

    ScreenDist r = rx.length();

    return r >= (ScreenPos(rect.x0(), rect.y0()) - s).length() ||
	   r >= (ScreenPos(rect.x1(), rect.y0()) - s).length() ||
	   r >= (ScreenPos(rect.x0(), rect.y1()) - s).length() ||
	   r >= (ScreenPos(rect.x1(), rect.y1()) - s).length();
}

void CircleObject::write(XMLWriter *writer)
{
    writer->startElement(is_arc ? "arc" : "circle");
    writer->writeAttribute("position", drawing()->format(pos));
    writer->writeAttribute("radius", drawing()->format(radius));
    if (is_arc) {
	writer->writeAttribute("startangle", int_to_str(startangle));
	writer->writeAttribute("sweepangle", int_to_str(sweepangle));
    }
    writer->writeAttribute("color", SchematicEnum::color[color]);
    line.write(writer);
    if (!is_arc)
	fill.write(writer);
    writer->endElement();
}

List<Attribute *> CircleObject::attributes() const
{
    if (!is_arc)
	return List<Attribute *>() << schematicAttributes->color
				   << line.attributes()
				   << fill.attributes();
    else
	return List<Attribute *>() << schematicAttributes->color
				   << line.attributes();
}

Variant CircleObject::attribute(const Attribute *attribute) const
{
    if (attribute == schematicAttributes->color)
	return Variant::fromInt(color);

    foreach (Attribute *attr, line.attributes())
	if (attr == attribute)
	    return line.attribute(attribute);

    if (!is_arc)
	foreach (Attribute *attr, fill.attributes())
	    if (attr == attribute)
		return fill.attribute(attribute);

    return Variant();
}

Variant CircleObject::attributeDefault(const Attribute *attribute) const
{
    if (attribute == schematicAttributes->color)
	return Variant::fromInt(3);

    foreach (Attribute *attr, line.attributes())
	if (attr == attribute)
	    return line.attributeDefault(attribute);

    if (!is_arc)
	foreach (Attribute *attr, fill.attributes())
	    if (attr == attribute)
		return fill.attributeDefault(attribute);

    return Variant();
}

void CircleObject::setAttribute(const Variant &value,
				const Attribute *attribute)
{
    if (attribute == schematicAttributes->color)
	color = value.toInt();

    foreach (Attribute *attr, line.attributes())
	if (attr == attribute)
	    line.setAttribute(value, attribute);

    if (!is_arc)
	foreach (Attribute *attr, fill.attributes())
	    if (attr == attribute)
		fill.setAttribute(value, attribute);
}
