/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "sch_component.h"
#include "hub.h"
#include "shared.h"
#include "icons.h"
#include "drawing.h"
#include "action.h"
#include "painter.h"
#include "xml_writer.h"
#include "sch_attribute.h"


ComponentMajorMode::ComponentMajorMode(Hub &hub) : MajorMode(hub),
						   minorMode(hub, *this)
{
    setCaption(_("Component"));
    setMinorMode(&minorMode);

/** comboBox = new QComboBox;
    comboBox->addItem(_("Reference component"));
    comboBox->addItem(_("Embed component in schematic"));
    comboBox->addItem(_("Include component as individual objects"));

    QVBoxLayout *layout = new QVBoxLayout(widget());
    layout->addWidget(comboBox);
    layout->addStretch(); **/
}

void ComponentMajorMode::accept(MinorMode *minor_mode)
{
    String path = hub.selectedLibraryComponent();
    if (path.empty())
	return;

    ComponentObject *component = new ComponentObject(minorMode.point(),
						     hub.drawing());
    component->basename = path.split('/').last();
    component->contents = shared.component(component->basename);

    if (!component->contents) {
	delete component;
	return;
    }

    hub.drawing()->performActions(
	List<Action *>() << new NewObjectAction(component),
	icons::add, _("New component"));

    // if (comboBox->currentIndex() == 0) ...
}

void ComponentMajorMode::cancel()
{
    setMinorMode(&minorMode, true);
}

void ComponentMajorMode::paint(Painter &painter)
{
    MajorMode::paint(painter);

    if (hub.pointerActive()) {
	painter.setBlackPen();
	painter.paintPoint(minorMode.point());
    }
}

List<ActionLine *> ComponentMajorMode::lines(ActionPoint *handle)
{
    return List<ActionLine *>();
}

/****************************************************************************/

class ComponentEmbeddingLayer : public EmbeddingLayer
{
public:
    ComponentEmbeddingLayer(const Metrum3 &pos, int angle, bool mirror,
			    Drawing *drawing);
    int angle() const;
    bool mirror() const;
    Metrum transform(Metrum m) const;
    Metrum3 transform(const Metrum3 &pos) const;

private:
    Metrum3 const pos;
    int const the_angle;
    bool const the_mirror;
    Drawing *const drawing;
};

ComponentEmbeddingLayer::ComponentEmbeddingLayer(
    const Metrum3 &pos, int angle, bool mirror, Drawing *drawing)
    : EmbeddingLayer(),
      pos(pos), the_angle(angle), the_mirror(mirror), drawing(drawing)
{
}

int ComponentEmbeddingLayer::angle() const
{
    return the_angle;
}

bool ComponentEmbeddingLayer::mirror() const
{
    return the_mirror;
}

Metrum ComponentEmbeddingLayer::transform(Metrum m) const
{
    return m;
}

Metrum3 ComponentEmbeddingLayer::transform(const Metrum3 &pos) const
{
    Metrum3 res = mirror() ? Metrum3(-pos.x(), pos.y(), pos.z()) : pos;

    switch (the_angle) {
    case 1:
	res = Metrum3(-res.y(), res.x(), res.z());
	break;
    case 2:
	res = Metrum3(-res.x(), -res.y(), res.z());
	break;
    case 3:
	res = Metrum3(res.y(), -res.x(), res.z());
	break;
    }

    return this->pos + res;
}

/****************************************************************************/

ComponentObject::ComponentObject(const Metrum3 &pos, Drawing *drawing)
    : Object(drawing), pos(pos), selectable(true), angle(0), mirror(false),
      contents(0), is_embedded(0), context(this),
      handle(_("position"), this)
{
}

Metrum3 ComponentObject::relativePos() const
{
    return pos;
}

String ComponentObject::name() const
{
    return _("component");
}

ObjectContext *ComponentObject::children() const
{
    return const_cast<ObjectContext *>(&context);
}

void ComponentObject::paint(Painter &painter, const Embedding &embedding)
{
    ComponentEmbeddingLayer el(pos, angle, mirror, drawing());
    Embedding emb(embedding, &el);

    if (contents)
	foreach (Object *ob, contents->objects())
	    ob->paint(painter, emb);
}

void ComponentObject::paintShadow(const Metrum3 &pos,
				  const ActionPoint *handle,
				  Painter &painter)
{
}

List<ActionPoint *> ComponentObject::handles()
{
    return List<ActionPoint *>() << &handle;
}

Metrum3 ComponentObject::handlePos(const ActionPoint *handle) const
{
    if (handle == &this->handle)
	return pos;

    return Metrum3();
}

List<ActionLine *> ComponentObject::handleLines(const ActionPoint *handle)
{
    return List<ActionLine *>();
}

void ComponentObject::moveHandle(const Metrum3 &pos, const ActionPoint *handle)
{
    if (handle == &this->handle)
	this->pos = pos;
}

Metrum3 ComponentObject::lineBase(const ActionLine *line) const
{
    return Metrum3();
}

Metrum3 ComponentObject::lineTangent(const ActionLine *line) const
{
    return Metrum3(1, 0, 0);
}

ScreenDist ComponentObject::proximity(const ScreenPos &point, View *view,
				      const Embedding &embedding) const
{
    ScreenDist least_dist = -1;
    ComponentEmbeddingLayer el(pos, angle, mirror, drawing());
    Embedding emb(embedding, &el);

    if (contents)
	foreach (Object *ob, contents->objects()) {
	    ScreenDist dist = ob->proximity(point, view, emb);
	    if (dist < least_dist || least_dist < 0)
		least_dist = dist;
	}

    return least_dist;
}

bool ComponentObject::within(const ScreenRect &rect, View *view,
			     bool partially, const Embedding &embedding) const
{
    bool res = !partially;
    ComponentEmbeddingLayer el(pos, angle, mirror, drawing());
    Embedding emb(embedding, &el);

    if (contents)
	foreach (Object *ob, contents->objects())
	    if (res != partially &&
		ob->within(rect, view, partially, emb) == partially)
		res = partially;

    foreach (Object *ob, context.objects())
	if (res != partially &&
	    ob->within(rect, view, partially, embedding) == partially)
	    res = partially;

    return res;
}

void ComponentObject::write(XMLWriter *writer)
{
    writer->startElement("component");
    writer->writeAttribute("position", drawing()->format(pos));
    writer->writeAttribute("selectable", SchematicEnum::boolean[selectable]);
    writer->writeAttribute("angle", int_to_str(angle));
    writer->writeAttribute("mirror", SchematicEnum::boolean[mirror]);
    writer->writeAttribute("basename", basename);

    if (is_embedded)
	contents->write(writer, drawing());

    context.write(writer);
    writer->endElement();
}

List<Attribute *> ComponentObject::attributes() const
{
    return List<Attribute *>() << schematicAttributes->selectable
			       << schematicAttributes->angle
			       << schematicAttributes->mirror;
}

Variant ComponentObject::attribute(const Attribute *attribute) const
{
    if (attribute == schematicAttributes->selectable)
	return Variant::fromBool(selectable);

    if (attribute == schematicAttributes->angle)
	return Variant::fromInt(angle);

    if (attribute == schematicAttributes->mirror)
	return Variant::fromBool(mirror);

    return Variant();
}

Variant ComponentObject::attributeDefault(const Attribute *attribute) const
{
    if (attribute == schematicAttributes->selectable)
	return Variant::fromBool(true);

    if (attribute == schematicAttributes->angle)
	return Variant::fromInt(0);

    if (attribute == schematicAttributes->mirror)
	return Variant::fromBool(false);

    return Variant();
}

void ComponentObject::setAttribute(const Variant &value,
				   const Attribute *attribute)
{
    if (attribute == schematicAttributes->selectable)
	selectable = value.toBool();

    if (attribute == schematicAttributes->angle)
	angle = value.toInt();

    if (attribute == schematicAttributes->mirror)
	mirror = value.toBool();
}
