/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "sch_net.h"
#include "hub.h"
#include "xml_writer.h"
#include "drawing.h"
#include "sch_attribute.h"
#include "variant.h"


NetMajorMode::NetMajorMode(Hub &hub) : LinearMajorMode(hub)
{
    setCaption(_("Net"));
}

Object *NetMajorMode::buildObject(const Metrum3 &p0, const Metrum3 &p1)
{
    return new NetObject(p0, p1, hub.drawing());
}

/****************************************************************************/

NetObject::NetObject(const Metrum3 &p0, const Metrum3 &p1, Drawing *drawing)
    : LinearObject(p0, p1, drawing),
      color(1), is_bus(false), is_pin(false), is_inverted(false), context(this)
{
}

String NetObject::name() const
{
    return _("net");
}

ObjectContext *NetObject::children() const
{
    return const_cast<ObjectContext *>(&context);
}

void NetObject::write(XMLWriter *writer)
{
    writer->startElement(is_pin ? "pin" : is_bus ? "bus" : "net");
    writer->writeAttribute("p0", drawing()->format(p0));
    writer->writeAttribute("p1", drawing()->format(p1));
    writer->writeAttribute("color", SchematicEnum::color[color]);
    if (is_pin)
	writer->writeAttribute("pintype", SchematicEnum::pintype[is_bus]);
    writer->writeAttribute("inverted", SchematicEnum::boolean[is_inverted]);
    context.write(writer);
    writer->endElement();
}

bool NetObject::within(const ScreenRect &rect, View *view,
		       bool partially, const Embedding &embedding) const
{
    bool res = LinearObject::within(rect, view, partially, embedding);

    foreach (Object *ob, context.objects())
	if (res != partially && ob->within(rect, view,
					   partially, embedding) == partially)
	    res = partially;

    return res;
}

List<Attribute *> NetObject::attributes() const
{
    return List<Attribute *>() << schematicAttributes->color
			       << schematicAttributes->isBus
			       << schematicAttributes->isPin
			       << schematicAttributes->isInverted;
}

Variant NetObject::attribute(const Attribute *attribute) const
{
    if (attribute == schematicAttributes->color)
	return Variant::fromInt(color);

    if (attribute == schematicAttributes->isBus)
	return Variant::fromBool(is_bus);

    if (attribute == schematicAttributes->isPin)
	return Variant::fromBool(is_pin);

    if (attribute == schematicAttributes->isInverted)
	return Variant::fromBool(is_inverted);

    return Variant();
}

Variant NetObject::attributeDefault(const Attribute *attribute) const
{
    if (attribute == schematicAttributes->color)
	return Variant::fromInt(1);

    if (attribute == schematicAttributes->isBus)
	return Variant::fromBool(false);

    if (attribute == schematicAttributes->isPin)
	return Variant::fromBool(false);

    if (attribute == schematicAttributes->isInverted)
	return Variant::fromBool(false);

    return Variant();
}

void NetObject::setAttribute(const Variant &value, const Attribute *attribute)
{
    if (attribute == schematicAttributes->color)
	color = value.toInt();

    if (attribute == schematicAttributes->isBus)
	is_bus = value.toBool();

    if (attribute == schematicAttributes->isPin)
	is_pin = value.toBool();

    if (attribute == schematicAttributes->isInverted)
	is_inverted = value.toBool();
}
