/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "sch_readxml.h"
#include "drawing.h"
#include "messages.h"
#include "sch_box.h"
#include "sch_circle.h"
#include "sch_component.h"
#include "sch_line.h"
#include "sch_net.h"
#include "sch_path.h"
#include "sch_picture.h"
#include "sch_text.h"


class SchematicXMLHandler::LineXMLHandler : public XMLHandler
{
public:
    LineXMLHandler(List<String> &text) : text(text) {
    }
    void characterData(const String &data, Messages &messages) {
	line += data;
    }
    void done(Messages &messages) {
	text.append(line);
    }
private:
    List<String> &text;
    String line;
};

class SchematicXMLHandler::TextXMLHandler : public XMLHandler
{
public:
    TextXMLHandler(List<String> &text, TextObject *ob) : text(text), ob(ob) {
    }
    void element(XMLElement *element, Messages &messages) {
	if (element->name() == "line")
	    element->setHandler(new LineXMLHandler(text));
	else
	    XMLHandler::element(element, messages);
    }
    void done(Messages &messages) {
	ob->update();
    }
private:
    List<String> &text;
    TextObject *const ob;
};

class SchematicXMLHandler::ComponentXMLHandler : public SchematicXMLHandler
{
public:
    ComponentXMLHandler(Drawing *drawing, ObjectContext *context,
			Drawing *&contents, bool &is_embedded,
			const String &base_name)
	: SchematicXMLHandler(drawing, context),
	  contents(contents), is_embedded(is_embedded), base_name(base_name) {
    }
    void element(XMLElement *element, Messages &messages) {
	if (element->name() == "embedded" && !contents) {
	    contents = new Drawing(drawing->fileType(), 0);
	    contents->setBaseBase(drawing->baseBase());
	    contents->setBaseExponent(drawing->baseExponent());

	    element->setHandler(
		new SchematicXMLHandler(contents,
					contents->rootContext()));
	    is_embedded = true;
	    return;
	}

	if (element->name() == "text")
	    read_text_object(element);
	else
	    XMLHandler::element(element, messages);
    }
    void done(Messages &messages) {
	if (is_embedded)
	    return;

	contents = shared.component(base_name);
	if (!contents)
	    messages.warning(_("Component not found: %1").arg(base_name));
    }
private:
    Drawing *&contents;
    bool &is_embedded;
    String const base_name;
};

/****************************************************************************/

SchematicXMLHandler::SchematicXMLHandler(Drawing *drawing,
					 ObjectContext *context)
    : drawing(drawing), context(context)
{
}

void SchematicXMLHandler::read_line(SchematicLine &line, XMLElement *element)
{
    if (element->selectAttribute("width"))
	line.width = element->toMetrum(drawing);
    if (element->selectAttribute("capstyle"))
	line.cap_style = element->lookup(SchematicEnum::capstyle);
    if (element->selectAttribute("dashstyle"))
	line.dash_style = element->lookup(SchematicEnum::dashstyle);
    if (element->selectAttribute("dashlength"))
	line.dash_length = element->toMetrum(drawing);
    if (element->selectAttribute("dashspace"))
	line.dash_space = element->toMetrum(drawing);
}

void SchematicXMLHandler::read_fill(SchematicFill &fill, XMLElement *element)
{
    if (element->selectAttribute("filltype"))
	fill.type = element->lookup(SchematicEnum::filltype);
    if (element->selectAttribute("fillwidth"))
	fill.width = element->toMetrum(drawing);
    if (element->selectAttribute("angle0"))
	fill.angle0 = element->toInt();
    if (element->selectAttribute("pitch0"))
	fill.pitch0 = element->toMetrum(drawing);
    if (element->selectAttribute("angle1"))
	fill.angle1 = element->toInt();
    if (element->selectAttribute("pitch1"))
	fill.pitch1 = element->toMetrum(drawing);
}

void SchematicXMLHandler::read_arc_object(XMLElement *element)
{
    CircleObject *arc = new CircleObject(Metrum3(), 0, drawing);
    arc->is_arc = true;

    if (element->selectAttribute("position"))
	arc->pos = element->toMetrum3(drawing);
    if (element->selectAttribute("radius"))
	arc->radius = element->toMetrum(drawing);
    if (element->selectAttribute("startangle"))
	arc->startangle = element->toInt();
    if (element->selectAttribute("sweepangle"))
	arc->sweepangle = element->toInt();
    if (element->selectAttribute("color"))
	arc->color = element->lookup(SchematicEnum::color);

    read_line(arc->line, element);
    context->addObject(arc);
}

void SchematicXMLHandler::read_box_object(XMLElement *element)
{
    BoxObject *box = new BoxObject(Metrum3(), Metrum3(), drawing);

    if (element->selectAttribute("p0") || element->selectAttribute("start"))
	box->p0 = element->toMetrum3(drawing);
    if (element->selectAttribute("p1") || element->selectAttribute("end"))
	box->p1 = element->toMetrum3(drawing);
    if (element->selectAttribute("color"))
	box->color = element->lookup(SchematicEnum::color);

    read_line(box->line, element);
    read_fill(box->fill, element);
    context->addObject(box);
}

void SchematicXMLHandler::read_circle_object(XMLElement *element)
{
    CircleObject *circle = new CircleObject(Metrum3(), 0, drawing);

    if (element->selectAttribute("position"))
	circle->pos = element->toMetrum3(drawing);
    if (element->selectAttribute("radius"))
	circle->radius = element->toMetrum(drawing);
    if (element->selectAttribute("color"))
	circle->color = element->lookup(SchematicEnum::color);

    read_line(circle->line, element);
    read_fill(circle->fill, element);
    context->addObject(circle);
}

void SchematicXMLHandler::read_component_object(XMLElement *element)
{
    ComponentObject *component =
	new ComponentObject(Metrum3(), drawing);

    if (element->selectAttribute("position"))
	component->pos = element->toMetrum3(drawing);
    if (element->selectAttribute("selectable"))
	component->selectable = element->lookup(SchematicEnum::boolean);
    if (element->selectAttribute("angle"))
	component->angle = element->toInt();
    if (element->selectAttribute("mirror"))
	component->mirror = element->lookup(SchematicEnum::boolean);
    if (element->selectAttribute("basename"))
	component->basename = element->value();

    element->setHandler(new ComponentXMLHandler(drawing, &component->context,
						component->contents,
						component->is_embedded,
						component->basename));
    context->addObject(component);
}

void SchematicXMLHandler::read_line_object(XMLElement *element)
{
    LineObject *line = new LineObject(Metrum3(), Metrum3(), drawing);

    if (element->selectAttribute("p0") || element->selectAttribute("start"))
	line->p0 = element->toMetrum3(drawing);
    if (element->selectAttribute("p1") || element->selectAttribute("end"))
	line->p1 = element->toMetrum3(drawing);
    if (element->selectAttribute("color"))
	line->color = element->lookup(SchematicEnum::color);

    read_line(line->line, element);
    context->addObject(line);
}

void SchematicXMLHandler::read_net_object(XMLElement *element)
{
    NetObject *net = new NetObject(Metrum3(), Metrum3(), drawing);
    net->is_bus = element->name() == "bus";

    if (element->selectAttribute("p0"))
	net->p0 = element->toMetrum3(drawing);
    if (element->selectAttribute("p1"))
	net->p1 = element->toMetrum3(drawing);
    if (element->selectAttribute("color"))
	net->color = element->lookup(SchematicEnum::color);
    if (element->selectAttribute("inverted"))
	net->is_inverted = element->lookup(SchematicEnum::boolean);

    context->addObject(net);
}

void SchematicXMLHandler::read_path_object(XMLElement *element)
{
    Path1Object *path = new Path1Object(drawing);

    if (element->selectAttribute("d"))
	path->data = List<String>() << element->value();
    if (element->selectAttribute("color"))
	path->color = element->lookup(SchematicEnum::color);

    read_line(path->line, element);
    read_fill(path->fill, element);
    context->addObject(path);
}

void SchematicXMLHandler::read_picture_object(XMLElement *element)
{
    PictureObject *picture = new PictureObject(Metrum3(), Metrum3(), drawing);

    if (element->selectAttribute("p0"))
	picture->p0 = element->toMetrum3(drawing);
    if (element->selectAttribute("p1"))
	picture->p1 = element->toMetrum3(drawing);
    if (element->selectAttribute("angle"))
	picture->angle = element->toInt();
    if (element->selectAttribute("mirrored"))
	picture->mirrored = element->lookup(SchematicEnum::boolean);
    if (element->selectAttribute("embedded"))
	picture->embedded = element->lookup(SchematicEnum::boolean);
    if (element->selectAttribute("filename"))
	picture->filename = element->value();
    if (element->selectAttribute("data"))
	picture->data = element->value();

    context->addObject(picture);
}

void SchematicXMLHandler::read_pin_object(XMLElement *element)
{
    NetObject *pin = new NetObject(Metrum3(), Metrum3(), drawing);
    pin->is_pin = true;

    if (element->selectAttribute("p0"))
	pin->p0 = element->toMetrum3(drawing);
    if (element->selectAttribute("p1"))
	pin->p1 = element->toMetrum3(drawing);
    if (element->selectAttribute("color"))
	pin->color = element->lookup(SchematicEnum::color);
    if (element->selectAttribute("pintype"))
	pin->is_bus = element->lookup(SchematicEnum::pintype);
    if (element->selectAttribute("inverted"))
	pin->is_inverted = element->lookup(SchematicEnum::boolean);

    context->addObject(pin);
}

void SchematicXMLHandler::read_text_object(XMLElement *element)
{
    TextObject *text = new TextObject(Metrum3(), List<String>(), drawing);

    if (element->selectAttribute("position"))
	text->pos = element->toMetrum3(drawing);
    if (element->selectAttribute("color"))
	text->color = element->lookup(SchematicEnum::color);
    if (element->selectAttribute("size"))
	text->size = element->toInt();
    if (element->selectAttribute("visible"))
	text->visibility = element->lookup(SchematicEnum::boolean);
    if (element->selectAttribute("show"))
	text->show_name_value = element->lookup(SchematicEnum::show_name_value);
    if (element->selectAttribute("angle"))
	text->angle = element->toInt();
    if (element->selectAttribute("alignment"))
	text->alignment = element->lookup(SchematicEnum::alignment);

    element->setHandler(new TextXMLHandler(text->text, text));
    context->addObject(text);
}

void SchematicXMLHandler::element(XMLElement *element, Messages &messages)
{
    if (element->name() == "line")
	read_line_object(element);
    else if (element->name() == "picture")
	read_picture_object(element);
    else if (element->name() == "box")
	read_box_object(element);
    else if (element->name() == "circle")
	read_circle_object(element);
    else if (element->name() == "arc")
	read_arc_object(element);
    else if (element->name() == "text")
	read_text_object(element);
    else if (element->name() == "net" || element->name() == "bus")
	read_net_object(element);
    else if (element->name() == "pin")
	read_pin_object(element);
    else if (element->name() == "component")
	read_component_object(element);
    else if (element->name() == "path")
	read_path_object(element);
    else
	XMLHandler::element(element, messages);
}
