/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#ifndef UTIL_LIST_H
#define UTIL_LIST_H

#include <deque>
#include <algorithm>

template<typename T> class ForeachHelper
{
public:
    ForeachHelper(const T &t) : c(t), i(c.begin()), e(c.end()), f(false) {
    }
    const T c;
    typename T::const_iterator i, e;
    bool f;
};

#define foreach(X, Y) \
    for (ForeachHelper<__typeof__(Y)> _h(Y); \
	 !_h.f && _h.i != _h.e; ++_h.i, _h.f = !_h.f) \
	for (X = *_h.i; !_h.f; _h.f = true)

template<typename T> class List : public std::deque<T>
{
public:
    typedef typename std::deque<T>::value_type value_type;
    typedef typename std::deque<T>::const_reference const_reference;
    typedef typename std::deque<T>::iterator iterator;
    typedef typename std::deque<T>::const_iterator const_iterator;
    typedef typename std::deque<T>::size_type size_type;

    using std::deque<T>::back;
    using std::deque<T>::begin;
    using std::deque<T>::empty;
    using std::deque<T>::end;
    using std::deque<T>::front;
    using std::deque<T>::insert;
    using std::deque<T>::pop_back;
    using std::deque<T>::pop_front;
    using std::deque<T>::size;

    static const size_type npos = static_cast<size_type>(-1);

    List() {
    }
    explicit List(size_type n) : std::deque<T>(n) {
    }
    List(size_type n, const value_type &value) : std::deque<T>(n, value) {
    }
    List(const std::deque<T> &x) : std::deque<T>(x) {
    }
    template<typename I> List(I first, I last) : std::deque<T>(first, last) {
    }

    /* accessing */

    bool contains(const value_type &value) const {
	return std::find(begin(), end(), value) != end();
    }
    size_type count() const {
	return size();
    }
    size_type find(const value_type &value, size_type from = 0) const {
	const_iterator i = std::find(begin() + from, end(), value);
	if (i == end())
	    return npos;
	return i - begin();
    }
    size_type indexOf(const value_type &value, size_type from = 0) const {
	const_iterator i = std::find(begin() + from, end(), value);
	if (i == end())
	    return npos;
	return i - begin();
    }
    bool isEmpty() const {
	return empty();
    }
    const_reference last() const {
	return back();
    }
    List mid(size_type pos, size_type n = npos) const {
	if (n == npos)
	    return List(begin() + pos, end());
	return List(begin() + pos, begin() + pos + n);
    }

    /* modifying */

    List &operator+=(const value_type &x) {
	this->push_back(x);
	return *this;
    }
    List &operator+=(const std::deque<T> &l) {
	insert(end(), l.begin(), l.end());
	return *this;
    }
    List &operator<<(const value_type &x) {
	this->push_back(x);
	return *this;
    }
    List &operator<<(const std::deque<T> &l) {
	insert(end(), l.begin(), l.end());
	return *this;
    }

    void append(const value_type &x) {
	this->push_back(x);
    }
    void append(const std::deque<T> &l) {
	insert(end(), l.begin(), l.end());
    }
    void deleteAll() const {
	std::deque<T> l = *this;
	const_iterator i = l.begin();
	while (i != l.end())
	    delete *i++;
    }
    void deleteRange(size_type pos, size_type n = npos) const {
	std::deque<T> l = *this;
	const_iterator i = l.begin() + pos;
	const_iterator j = n == npos ? l.end() : l.begin() + pos + n;
	while (i != j)
	    delete *i++;
    }
    void insert(size_type pos, const value_type &x) {
	insert(begin() + pos, x);
    }
    void insert(size_type pos, const std::deque<T> &l) {
	insert(begin() + pos, l.begin(), l.end());
    }
    void prepend(const value_type &x) {
	this->push_front(x);
    }
    size_type removeAll(const value_type &value) {
	iterator i = begin();
	size_type n = 0;
	while ((i = std::find(i, end(), value)) != end()) {
	    i = this->erase(i);
	    n++;
	}
	return n;
    }
    void removeAt(size_type pos) {
	this->erase(begin() + pos);
    }
    void removeLast() {
	pop_back();
    }
    void removeRange(size_type pos, size_type n = npos) {
	if (n == npos)
	    this->erase(begin() + pos, end());
	else
	    this->erase(begin() + pos, begin() + pos + n);
    }
    void replace(size_type pos, size_type n, const std::deque<T> &l) {
	iterator i;
	if (n == npos)
	    i = this->erase(begin() + pos, end());
	else
	    i = this->erase(begin() + pos, begin() + pos + n);
	insert(i, l.begin(), l.end());
    }
    value_type takeAt(size_type pos) {
	value_type x = this->operator[](pos);
	this->erase(begin() + pos);
	return x;
    }
    value_type takeFirst() {
	value_type x = front();
	pop_front();
	return x;
    }
    value_type takeLast() {
	value_type x = back();
	pop_back();
	return x;
    }

    /* static */

    static List<T> zero(size_type n) {
	return List<T>(n, 0);
    }
};

template<typename T> List<T> operator+(const T &lhs, const List<T> &rhs)
{
    List<T> l = rhs;
    l.push_front(lhs);
    return l;
}
template<typename T> List<T> operator+(const List<T> &lhs, const T &rhs)
{
    List<T> l = lhs;
    l.push_back(rhs);
    return l;
}
template<typename T> List<T> operator+(const List<T> &lhs, const List<T> &rhs)
{
    List<T> l = lhs;
    l.insert(l.end(), rhs.begin(), rhs.end());
    return l;
}

template<typename S, typename T> inline List<S> list_cast(const List<T> &list)
{
    List<S> result;

    foreach (T item, list)
	result.push_back(item);

    return result;
}

#endif
