/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <math.h>
#include <assert.h>
#include "view.h"
#include "drawing.h"


View::View(Drawing *drawing) : the_drawing(drawing)
{
}

Drawing *View::drawing() const
{
    return the_drawing;
}

View::ScrollInformation::ScrollInformation()
    : horzRange(0), horzPageSize(0), horzPosition(0),
      vertRange(0), vertPageSize(0), vertPosition(0)
{
}

/****************************************************************************/

SimpleView::SimpleView(const Metrum3 &origin, double scale,
		       double rotate, Drawing *drawing)
    : View(drawing), the_origin(origin), the_scale(scale), the_rotate(rotate),
      r_cos(cos(rotate)), r_sin(sin(rotate)), x_size(0), x_start(0), x_stop(0),
					      y_size(0), y_start(0), y_stop(0)
{
}

Metrum3 SimpleView::origin() const
{
    return the_origin;
}

double SimpleView::scale() const
{
    return the_scale;
}

double SimpleView::rotate() const
{
    return the_rotate;
}

ScreenPos SimpleView::map(const Metrum3 &point) const
{
    Metrum3 rel_point = point - origin();
    double x = drawing()->eff(rel_point.x());
    double y = drawing()->eff(rel_point.y());
    return ScreenPos((x * r_cos - y * r_sin) * scale(),
		     (y * r_cos + x * r_sin) * scale());
}

Metrum3 SimpleView::unmap(const ScreenPos &point) const
{
    return drawing()->scale((point.x() * r_cos + point.y() * r_sin) / scale(),
			    (point.y() * r_cos - point.x() * r_sin) / scale(),
			    0.0) + origin();
}

void SimpleView::setViewportDimensions(ScreenDist width, ScreenDist height)
{
    x_size = drawing()->scale(width) / scale();
    x_start = drawing()->scale(-15) - x_size;
    x_stop = drawing()->scale(15) + x_size;

    y_size = drawing()->scale(height) / scale();
    y_start = drawing()->scale(-15) - y_size;
    y_stop = drawing()->scale(15) + y_size;
}

View::ScrollInformation SimpleView::getScrollInformation() const
{
    ScrollInformation scroll_info;
    scroll_info.horzRange = x_stop - x_start;
    scroll_info.horzPageSize = x_size;
    scroll_info.horzPosition = origin().x() - x_start - x_size / 2;
    scroll_info.vertRange = y_stop - y_start;
    scroll_info.vertPageSize = y_size;
    scroll_info.vertPosition = y_stop - origin().y() - y_size / 2;
    return scroll_info;
}

View *SimpleView::zoom(double factor, const ScreenPos &center) const
{
    return new SimpleView(origin() / factor + unmap(center) * (1 - 1 / factor),
			  scale() * factor, rotate(), drawing());
}

View *SimpleView::move(const ScreenPos &delta) const
{
    return new SimpleView(unmap(map(origin()) - delta),
			  scale(), rotate(), drawing());
}

View *SimpleView::scrollTo(unsigned int x, unsigned int y) const
{
    assert(x_size && y_size);

    return new SimpleView(Metrum3(x_start + x_size / 2 + (int)x,
				  y_stop - y_size / 2 - (int)y, origin().z()),
			  scale(), rotate(), drawing());
}

/****************************************************************************/

OrthogonalView::OrthogonalView(const Metrum3 &origin, const ScreenPos &x,
			       const ScreenPos &y, const ScreenPos &z,
			       Drawing *drawing)
    : View(drawing), the_origin(origin), x(x), y(y), z(z)
{
}

Metrum3 OrthogonalView::origin() const
{
    return the_origin;
}

ScreenPos OrthogonalView::map(const Metrum3 &point) const
{
    Metrum3 rel_point = point - origin();
    return x * drawing()->eff(rel_point.x()) +
	   y * drawing()->eff(rel_point.y()) +
	   z * drawing()->eff(rel_point.z());
}

Metrum3 OrthogonalView::unmap(const ScreenPos &point) const
{
    return drawing()->scale(0.0, 0.0, 0.0) + origin(); //TODO
}

void OrthogonalView::setViewportDimensions(ScreenDist width, ScreenDist height)
{
}

View::ScrollInformation OrthogonalView::getScrollInformation() const
{
    return ScrollInformation();
}

View *OrthogonalView::zoom(double factor, const ScreenPos &center) const
{
    return new OrthogonalView(
	origin() / factor + unmap(center) * (1 - 1 / factor),
	x * factor, y * factor, z * factor, drawing());
}

View *OrthogonalView::move(const ScreenPos &delta) const
{
    return new OrthogonalView(unmap(map(origin()) - delta), x, y, z, drawing());
}

View *OrthogonalView::scrollTo(unsigned int x, unsigned int y) const
{
    return new OrthogonalView(*this);
}
