/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <err.h>
#include <assert.h>
#include "xml_writer.h"
#include "util/string.h"


XMLWriter::XMLWriter(int fd) : fp(fdopen(fd, "w")),
			       is_initialized(false), has_root_element(false),
			       pre_depth(-1), state(idle)
{
    if (fp == NULL)
	err(EXIT_FAILURE, "fdopen");
}

XMLWriter::~XMLWriter()
{
    if (fclose(fp) == EOF)
	err(EXIT_FAILURE, "fclose");
    assert(has_root_element);
    assert(stack.empty());
}

void XMLWriter::write(const String &data)
{
/*  String::size_type pos = data.rfind('\n');

    if (pos != String::npos)
	column = data.length() - pos - 1;
    else
	column += data.length();  */

    String data_utf8(String::encodeUTF8(data.decode()));

    if (fputs(data_utf8.data(), fp) == EOF)
	err(EXIT_FAILURE, "fputs");
}

void XMLWriter::escape(const String &data)
{
    String::size_type from = 0, to;

    do {
	String::size_type to_lt = data.find('<', from);
	String::size_type to_gt = data.find('>', from);
	String::size_type to_amp = data.find('&', from);
	String::size_type to_quot = data.find('"', from);

	to = to_lt;
	if (to_gt != String::npos && (to == String::npos || to_gt < to))
	    to = to_gt;
	if (to_amp != String::npos && (to == String::npos || to_amp < to))
	    to = to_amp;
	if (to_quot != String::npos && (to == String::npos || to_quot < to))
	    to = to_quot;

	write(data.substr(from, to - from));

	if (to != String::npos) {
	    if (to == to_lt)
		write("&lt;");
	    if (to == to_gt)
		write("&gt;");
	    if (to == to_amp)
		write("&amp;");
	    if (to == to_quot)
		write("&quot;");
	    from = to + 1;
	}
    } while (to != String::npos);
}

void XMLWriter::prepare_for_data()
{
    if (!is_initialized) {
	write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	is_initialized = true;
    }

    if (state == start_tag) {
	write(">");
	if (pre_depth == -1)
	    write("\n");
	current_attrs.clear();
	state = idle;
    }
}

static void validate_name(const String &name)
{
    std::wstring wstr = name.decode();
    wchar_t c;

    assert(!name.empty());
    c = wstr[0];
    assert(c != '-' && c != '.' && !(c >= '0' && c <= '9') && c != 0xb7 &&
	   !(c >= 0x0300 && c <= 0x036f) && !(c >= 0x203f && c <= 0x2040));

    for (unsigned int i = 1; i < wstr.length(); i++) {
	c = wstr[i];
	assert(c == '-' || c == '.' ||
	       (c >= '0' && c <= '9') || c == ':' ||
	       (c >= 'A' && c <= 'Z') || c == '_' ||
	       (c >= 'a' && c <= 'z') || c == 0xb7 ||
	       (c >= 0xc0 && c <= 0xd6) ||
	       (c >= 0xd8 && c <= 0xf6) ||
	       (c >= 0xf8 && c <= 0x2ff) ||
	       (c >= 0x300 && c <= 0x36f) ||
	       (c >= 0x370 && c <= 0x37d) ||
	       (c >= 0x37f && c <= 0x1fff) ||
	       (c >= 0x200c && c <= 0x200d) ||
	       (c >= 0x203f && c <= 0x2040) ||
	       (c >= 0x2070 && c <= 0x218f) ||
	       (c >= 0x2c00 && c <= 0x2fef) ||
	       (c >= 0x3001 && c <= 0xd7ff) ||
	       (c >= 0xf900 && c <= 0xfdcf) ||
	       (c >= 0xfdf0 && c <= 0xfffd) ||
	       (c >= 0x10000 && c <= 0xeffff));
    }
}

void XMLWriter::startElement(const String &name, bool preserve_whitespace)
{
    validate_name(name);

    if (stack.empty()) {
	assert(!has_root_element);
	has_root_element = true;
    }

    prepare_for_data();
    if (pre_depth == -1) {
	if (state == character_data)
	    write("\n");
	write(String(stack.count() * 2, ' '));
    }
    write("<" + name);
    if (preserve_whitespace && pre_depth == -1)
	pre_depth = stack.count();
    stack.push_back(name);
    state = start_tag;
}

void XMLWriter::endElement()
{
    switch (state) {
    case character_data:
	if (pre_depth == -1)
	    write("\n");
	/* fallthrough */
    case idle:
	if (pre_depth == -1)
	    write(String((stack.count() - 1) * 2, ' '));
	write("</" + stack.back() + ">");
	break;
    case start_tag:
	write("/>");
	break;
    }
    state = idle;

    stack.pop_back();
    if (pre_depth == stack.count())
	pre_depth = -1;
    if (pre_depth == -1)
	write("\n");
    current_attrs.clear();
}

void XMLWriter::writeAttribute(const String &name, const String &value)
{
    assert(state == start_tag);
    validate_name(name);
    // validate_characters(value);

    assert(!current_attrs.contains(name));
    current_attrs.append(name);

    write(" " + name + "=\"");
    escape(value);
    write("\"");
}

void XMLWriter::writeCharacterData(const String &data)
{
    // validate_characters(data);

    prepare_for_data();
    if (state == idle && pre_depth == -1)
	write(String(stack.count() * 2, ' '));
    state = character_data;

    escape(data);
}

void XMLWriter::writeCDATASection(const String &data)
{
    String::size_type from = 0, to;

    prepare_for_data();
    if (state == idle && pre_depth == -1)
	write(String(stack.count() * 2, ' '));
    state = character_data;

    write("<![CDATA[");

    while ((to = data.find("]]>", from)) != -1) {
	write(data.substr(from, to));
	write("]]>]]&gt;<![CDATA[");
	from = to + 3;
    }

    write(data.substr(from));
    write("]]>");
}
