/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <stdio.h>
#include <assert.h>
#include <errno.h>
#include <err.h>
#include <algorithm>
#include "connection.h"
#include "hooks.h"
#include "event_loop.h"
#include "menus.h"

using namespace Toolkit;


class Connection::Impl
{
    void check_pixmap_format();
    void choose_visual();
    X::FontSet load_font_set(const char *base_font_name_list);
    void open_input_method();
public:
    Impl(char *display_name);
    ~Impl();
    WindowHook *find_hook(Window window);
    void grab_pointer(Time time, X::Window window, bool owner_events);
    void ungrab_pointer(Time time);
    void leave_menu_mode(Time time);

    X::Display display;
    X::Screen screen;
    std::vector<WindowHook *> hooks;
    XVisualInfo visual_info;
    ConnectionAtoms atoms;
    ConnectionColors colors;
    X::FontSet font_set, font_set_bold;
    std::map<std::pair<XFontSet, std::wstring>, int> escapement_cache;
    AbstractMenu *root_menu;
    bool emulates_implicit_grab;
    bool disable_pointer_grab;
    X::InputMethod input_method;
    unsigned long xim_style;
};


Connection::Connection(char *display_name) : p(new Impl(display_name))
{
    EventLoop::addConnection(this);
}

Connection::~Connection()
{
    EventLoop::removeConnection(this);
    delete p;
}

X::Display Connection::display() const
{
    return p->display;
}

X::Screen Connection::screen() const
{
    return p->screen;
}

void Connection::addHook(WindowHook *hook)
{
    assert(&hook->conn == this);
    assert(std::find(p->hooks.begin(),
		     p->hooks.end(), hook) == p->hooks.end());
    p->hooks.push_back(hook);
}

void Connection::removeHook(WindowHook *hook)
{
    std::vector<WindowHook *>::iterator i =
	std::find(p->hooks.begin(), p->hooks.end(), hook);
    assert(i != p->hooks.end());
    p->hooks.erase(i);
}

void Connection::update()
{
    std::vector<WindowHook *> cached_hooks = p->hooks;

    for (std::vector<WindowHook *>::const_iterator i = cached_hooks.begin();
	 i != cached_hooks.end(); i++)
	(*i)->update();
}

bool Connection::abandoned() const
{
    return p->hooks.empty();
}

Visual *Connection::visual() const
{
    return p->visual_info.visual;
}

XVisualInfo const &Connection::visualInfo() const
{
    return p->visual_info;
}

ConnectionAtoms const &Connection::atoms() const
{
    return p->atoms;
}

ConnectionColors const &Connection::colors() const
{
    return p->colors;
}

X::FontSet Connection::fontSet() const
{
    return p->font_set;
}

X::FontSet Connection::fontSetBold() const
{
    return p->font_set_bold;
}

int Connection::cachedEscapement(X::FontSet fontSet,
				 const std::wstring &str) const
{
    std::pair<XFontSet, std::wstring> key(fontSet, str);
    std::map<std::pair<XFontSet, std::wstring>, int>::const_iterator i =
	p->escapement_cache.find(key);

    if (i != p->escapement_cache.end())
	return (*i).second;

    return p->escapement_cache[key] = fontSet.wcTextEscapement(str);
}

std::wstring Connection::ellipsis(X::FontSet fontSet,
				  const std::wstring &str, int width) const
{
    static std::wstring const ellipsis = L"...";

    if (width >= cachedEscapement(fontSet, str))
	return str;
    if (width < cachedEscapement(fontSet, ellipsis))
	return std::wstring();

    int lower_bound = 0;
    int upper_bound = str.length();

    while (upper_bound - lower_bound >= 2) {
	int probe = (upper_bound + lower_bound) / 2;
	if (width < cachedEscapement(fontSet, str.substr(0, probe) + ellipsis))
	    upper_bound = probe;
	else
	    lower_bound = probe;
    }

    return str.substr(0, lower_bound) + ellipsis;
}

AbstractMenu *Connection::rootMenu() const
{
    return p->root_menu;
}

void Connection::enterMenuMode(Time time, AbstractMenu *menu)
{
    assert(&menu->conn == this);

    if (p->root_menu)
	p->leave_menu_mode(time);

    p->root_menu = menu;

    p->grab_pointer(time, p->screen.rootWindow(), true);
    p->emulates_implicit_grab = false;
}

void Connection::leaveMenuMode(Time time)
{
    p->leave_menu_mode(time);
}

void Connection::notifyAutomaticGrab(Time time, Widget *widget,
				     WidgetWindowHook *hook)
{
    /* Since the pointer is explicitly grabbed in menu mode, the implicit
       pointer grab mechanics no longer apply, and all pointer events are, as
       requested by the owner_events flag passed to XGrabPointer, reported to
       the window in which they happened. This is the correct bevaviour for
       menus; however, for drop-down selector scroll bars, it is not, so we
       have to emulate the implicit grab behaviour there. */

    if (p->root_menu && widget && !widget->isMenu()) {
	p->grab_pointer(time, hook->window(), false);
	p->emulates_implicit_grab = true;
    }

    hook->autograbbed_widget = widget;
}

void Connection::notifyAutomaticUngrab(Time time, WidgetWindowHook *hook)
{
    if (p->emulates_implicit_grab) {
	if (p->root_menu)
	    p->grab_pointer(time, p->screen.rootWindow(), true);
	else
	    p->ungrab_pointer(time);
	p->emulates_implicit_grab = false;
    }
    hook->autograbbed_widget = 0;
}

void Connection::transferAutomaticGrab(Time time, WidgetWindowHook *from_hook,
				       Widget *to_widget)
{
    if (from_hook == to_widget->hook()) {
	from_hook->autograbbed_widget = to_widget;
	return;
    }

    p->grab_pointer(time, to_widget->window(), false);
    p->emulates_implicit_grab = true;

    from_hook->autograbbed_widget = 0;
    to_widget->hook()->autograbbed_widget = to_widget;
}

void Connection::disablePointerGrab()
{
    /* only called by initialization code,
       so there's no need to check for currently active grab */

    p->disable_pointer_grab = true;
}

X::InputMethod Connection::inputMethod() const
{
    return p->input_method;
}

unsigned long Connection::ximStyle() const
{
    return p->xim_style;
}

ConnectionAtoms::ConnectionAtoms()
{
}

void ConnectionAtoms::internAtoms(X::Display display)
{
    char *atom_names[] = { "UTF8_STRING",
			   "TEXT",
			   "COMPOUND_TEXT",
			   "CLIPBOARD",
			   "WM_PROTOCOLS",
			   "WM_DELETE_WINDOW",
			   "_NET_WM_STATE",
			   "_NET_WM_STATE_FULLSCREEN" };

    assert(sizeof atom_names / sizeof(char *) ==
	   sizeof *this / sizeof(Atom));

    if (display.internAtoms(atom_names, sizeof *this / sizeof(Atom),
			    false, (Atom *)this) == 0)
	errx(EXIT_FAILURE, "XInternAtoms: Could not get all atoms");
}

ConnectionColors::ConnectionColors()
{
}

static void alloc_color(
    unsigned short red, unsigned short green, unsigned short blue,
    XColor &color, X::Colormap colormap)
{
    color.red = red;
    color.green = green;
    color.blue = blue;
    colormap.allocColor(&color);
}

void ConnectionColors::allocColors(X::Colormap colormap)
{
    alloc_color(0xc0c0, 0xc0c0, 0xc0c0, iconBackground, colormap);
    alloc_color(0xe8e8, 0xe8e8, 0xe8e8, background, colormap);
    alloc_color(0x4040, 0x4040, 0x8080, selected, colormap);
    alloc_color(0x9595, 0x9595, 0x9999, buttonBorder, colormap);
    alloc_color(0xc5c5, 0xc5c5, 0xc7c7, buttonBorderPixel, colormap);
    alloc_color(0xf8f8, 0xf8f8, 0xffff, buttonRaised, colormap);
    alloc_color(0xd0d0, 0xd0d0, 0xd6d6, buttonLowered, colormap);
    alloc_color(0xe8e8, 0xe8e8, 0xe8e8, buttonHighlight, colormap);
    alloc_color(0xd8d8, 0xd8d8, 0xd8d8, buttonDown, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, buttonText, colormap);
    alloc_color(0xf8f8, 0xf8f8, 0xf8f8, scrollbar, colormap);
    alloc_color(0xf0f0, 0xf0f0, 0xf0f0, scrollbarBorder, colormap);
    alloc_color(0xa0a0, 0xa0a0, 0xa0a0, scrollbarRipple0, colormap);
    alloc_color(0xfcfc, 0xfcfc, 0xfcfc, scrollbarRipple1, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, scrollbarArrow, colormap);
    alloc_color(0xa8a8, 0xa8a8, 0xa8a8, scrollbarInsensitiveArrow, colormap);
    alloc_color(0x2020, 0x4040, 0x8080, navigatorSymbol, colormap);
    alloc_color(0xe8e8, 0xe8e8, 0xe8e8, navigatorBackground, colormap);
    alloc_color(0x8080, 0x8080, 0x8080, navigatorBorder, colormap);

    alloc_color(0x0000, 0x0000, 0x0000, inputText, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, inputCursor, colormap);
    alloc_color(0xe8e8, 0xe8e8, 0xe8e8, inputBackground, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, inputSelectionText, colormap);
    alloc_color(0xa0a0, 0xa0a0, 0xa0a0, inputSelectionBackground, colormap);
    alloc_color(0x0000, 0x0000, 0x0000,
		inputSelectionUnfocussedText, colormap);
    alloc_color(0xd0d0, 0xd0d0, 0xd0d0,
		inputSelectionUnfocussedBackground, colormap);

    alloc_color(0x0000, 0x0000, 0x0000, selectorText, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, selectorArrow, colormap);
    alloc_color(0xe8e8, 0xe8e8, 0xe8e8, selectorBackground, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, selectorFocussedText, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, selectorFocussedArrow, colormap);
    alloc_color(0xe8e8, 0xe8e8, 0xe8e8, selectorFocussedBackground, colormap);

    alloc_color(0x7070, 0x7070, 0x7070, tabbarSelected, colormap);
    alloc_color(0xffff, 0xffff, 0xffff, tabbarSelectedText, colormap);
    alloc_color(0x3c3c, 0x3c3c, 0x3c3c, tabbarUnselected, colormap);
    alloc_color(0xffff, 0xffff, 0xffff, tabbarUnselectedText, colormap);
    alloc_color(0x3c3c, 0x3c3c, 0x3c3c, tabbarPlasma, colormap);

    alloc_color(0x3c3c, 0x3c3c, 0x3c3c, dockPlasma, colormap);
    alloc_color(0xffff, 0xffff, 0xffff, dockPlasmaText, colormap);
    alloc_color(0x6464, 0x6464, 0x6464, dockTitle, colormap);
    alloc_color(0xffff, 0xffff, 0xffff, dockTitleText, colormap);
    alloc_color(0x8080, 0x8080, 0x8080, dockFocussedTitle, colormap);
    alloc_color(0xffff, 0xffff, 0xffff, dockFocussedTitleText, colormap);
    alloc_color(0xc8c8, 0xc8c8, 0xc8c8, dockToolbar, colormap);
    alloc_color(0x9494, 0x9494, 0x9494, dockToolbarSeparator, colormap);
    alloc_color(0x4040, 0x4040, 0x4040, statusbar, colormap);
    alloc_color(0xffff, 0xffff, 0xffff, statusbarText, colormap);
    alloc_color(0x8080, 0x8080, 0x8080, statusbarProgress, colormap);

    alloc_color(0x0000, 0x0000, 0x0000, menuLabel, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, menuMnemonic, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, menuHotkey, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, menuArrow, colormap);
    alloc_color(0xf0f0, 0xf0f0, 0xf0f0, menuBackground, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, menuHighlightLabel, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, menuHighlightMnemonic, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, menuHighlightHotkey, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, menuHighlightArrow, colormap);
    alloc_color(0xc0c0, 0xc0c0, 0xc0c0, menuHighlight, colormap);
    alloc_color(0xa8a8, 0xa8a8, 0xa8a8, menuDisabledLabel, colormap);
    alloc_color(0xa8a8, 0xa8a8, 0xa8a8, menuDisabledMnemonic, colormap);
    alloc_color(0xa8a8, 0xa8a8, 0xa8a8, menuDisabledHotkey, colormap);
    alloc_color(0xa8a8, 0xa8a8, 0xa8a8, menuDisabledArrow, colormap);
    alloc_color(0xf0f0, 0xf0f0, 0xf0f0, menuDisabled, colormap);
    alloc_color(0xa0a0, 0xa0a0, 0xa0a0, menuBorder, colormap);
    alloc_color(0xc0c0, 0xc0c0, 0xc0c0, menuSeparator, colormap);

    alloc_color(0xe8e8, 0xe8e8, 0xe8e8, toolbar, colormap);
    alloc_color(0xe8e8, 0xe8e8, 0xe8e8, toolbarButton, colormap);
    alloc_color(0xe8e8, 0xe8e8, 0xe8e8, toolbarButtonBorder, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, toolbarButtonArrow, colormap);
    alloc_color(0xd0d0, 0xd0d0, 0xd0d0, toolbarActive, colormap);
    alloc_color(0xb0b0, 0xb0b0, 0xb0b0, toolbarActiveBorder, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, toolbarActiveArrow, colormap);
    alloc_color(0xf0f0, 0xf0f0, 0xf0f0, toolbarHighlight, colormap);
    alloc_color(0xc0c0, 0xc0c0, 0xc0c0, toolbarHighlightBorder, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, toolbarHighlightArrow, colormap);
    alloc_color(0xd4d4, 0xd4d4, 0xd4d4, toolbarHighlightActive, colormap);
    alloc_color(0xb0b0, 0xb0b0, 0xb0b0, toolbarHighlightActiveBorder, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, toolbarHighlightActiveArrow, colormap);
    alloc_color(0xe8e8, 0xe8e8, 0xe8e8, toolbarDisabled, colormap);
    alloc_color(0xe8e8, 0xe8e8, 0xe8e8, toolbarDisabledBorder, colormap);
    alloc_color(0x8080, 0x8080, 0x8080, toolbarDisabledArrow, colormap);
    alloc_color(0xd0d0, 0xd0d0, 0xd0d0, toolbarDisabledActive, colormap);
    alloc_color(0xb0b0, 0xb0b0, 0xb0b0, toolbarDisabledActiveBorder, colormap);
    alloc_color(0x8080, 0x8080, 0x8080, toolbarDisabledActiveArrow, colormap);
    alloc_color(0xc0c0, 0xc0c0, 0xc0c0, toolbarPressed, colormap);
    alloc_color(0xa0a0, 0xa0a0, 0xa0a0, toolbarPressedBorder, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, toolbarPressedArrow, colormap);
    alloc_color(0xc0c0, 0xc0c0, 0xc0c0, toolbarSeparator, colormap);
    alloc_color(0xa0a0, 0xa0a0, 0xa0a0, toolbarOverflow, colormap);

    alloc_color(0xffff, 0xffff, 0xffff, viewBackground, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, viewItemText, colormap);
    alloc_color(0xffff, 0xffff, 0xffff, viewItemBackground0, colormap);
    alloc_color(0xf8f8, 0xf8f8, 0xf8f8, viewItemBackground1, colormap);
    alloc_color(0xa0a0, 0xa0a0, 0xa0a0, viewSelected, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, viewSelectedText, colormap);
    alloc_color(0xffff, 0xffff, 0xffff, viewHighlight, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, viewHighlightText, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, viewLines, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, viewPlupForeground, colormap);
    alloc_color(0xffff, 0xffff, 0xffff, viewPlupBackground, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, viewPlupBorder, colormap);

    alloc_color(0xffff, 0xffff, 0xffff, btnviewBackground, colormap);
    alloc_color(0xffff, 0xffff, 0xffff, btnviewItemBorder, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, btnviewItemText, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, btnviewItemHotkey, colormap);
    alloc_color(0xffff, 0xffff, 0xffff, btnviewItemBackground, colormap);
    alloc_color(0xbbbb, 0xbbbb, 0xbbbb, btnviewCategoryBorder, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, btnviewCategoryText, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, btnviewCategoryHotkey, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, btnviewCategoryArrow, colormap);
    alloc_color(0xdddd, 0xdddd, 0xdddd, btnviewCategoryBackground, colormap);
    alloc_color(0xdddd, 0xdddd, 0xdddd, btnviewHighlightBorder, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, btnviewHighlightText, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, btnviewHighlightHotkey, colormap);
    alloc_color(0xf8f8, 0xf8f8, 0xf8f8, btnviewHighlightBackground, colormap);
    alloc_color(0xbbbb, 0xbbbb, 0xbbbb, btnviewCatHghlghtBorder, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, btnviewCatHghlghtText, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, btnviewCatHghlghtHotkey, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, btnviewCatHghlghtArrow, colormap);
    alloc_color(0xdddd, 0xdddd, 0xdddd, btnviewCatHghlghtBackground, colormap);
    alloc_color(0x8888, 0x8888, 0x8888, btnviewSelectedBorder, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, btnviewSelectedText, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, btnviewSelectedHotkey, colormap);
    alloc_color(0xaaaa, 0xaaaa, 0xaaaa, btnviewSelectedBackground, colormap);

    alloc_color(0xc0c0, 0xc0c0, 0xc0c0, propertiesFixed, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, propertiesFixedText, colormap);
    alloc_color(0xe8e8, 0xe8e8, 0xe8e8, propertiesSelected, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, propertiesSelectedText, colormap);
    alloc_color(0xa0a0, 0xa0a0, 0xa0a0, propertiesSelectedFixed, colormap);
    alloc_color(0x0000, 0x0000, 0x0000, propertiesSelectedFixedText, colormap);
}

/****************************************************************************/

Connection::Impl::Impl(char *display_name)
    : root_menu(0),
      emulates_implicit_grab(false), disable_pointer_grab(false), xim_style(0)
{
    display = X::Display::open(display_name);
    screen = display.defaultScreen();

    if (display.isNull())
	errx(EXIT_FAILURE, "Cannot connect to X server");

    check_pixmap_format();
    choose_visual();

    atoms.internAtoms(display);
    colors.allocColors(screen.defaultColormap());

    font_set = load_font_set(
	"-*-helvetica-medium-r-*-*-12-*-*-*-*-*-*-*,"
	"-*-liberation.sans-medium-r-*-*-12-*-*-*-*-*-*-*");
    font_set_bold = load_font_set(
	"-*-helvetica-bold-r-*-*-12-*-*-*-*-*-*-*,"
	"-*-liberation.sans-bold-r-*-*-12-*-*-*-*-*-*-*");

    open_input_method();
}

Connection::Impl::~Impl()
{
    assert(hooks.empty());

    input_method.close();
    font_set.free();
    font_set_bold.free();
    display.close();
}

void Connection::Impl::check_pixmap_format()
{
    std::vector<XPixmapFormatValues> pfv = display.listPixmapFormats();

    for (unsigned int i = 0; i < pfv.size(); i++)
	if (pfv[i].depth == 24 && pfv[i].bits_per_pixel == 32 &&
	    pfv[i].scanline_pad == 32)
	    return;

    fprintf(stderr, "%s: XListPixmapFormats: No usable pixmap format found\n",
	    program_invocation_short_name);
    fprintf(stderr, "%s: Available pixmap formats are:\n",
	    program_invocation_short_name);

    for (unsigned int i = 0; i < pfv.size(); i++)
	fprintf(stderr,
		"\tdepth = %d, bits_per_pixel = %d, scanline_pad = %d\n",
		pfv[i].depth, pfv[i].bits_per_pixel, pfv[i].scanline_pad);

    exit(EXIT_FAILURE);
}

void Connection::Impl::choose_visual()
{
    XVisualInfo vinfo_template = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

    vinfo_template.screen = screen;
    vinfo_template.depth = 24;
    vinfo_template.c_class = TrueColor;
    vinfo_template.bits_per_rgb = 8;

    std::vector<XVisualInfo> vi = display.getVisualInfo(
	VisualScreenMask | VisualDepthMask |
	VisualClassMask | VisualBitsPerRGBMask, &vinfo_template);

    if (!vi.empty()) {
	visual_info = vi.front();
	return;
    }

    fprintf(stderr, "%s: XGetVisualInfo: No 24-bit TrueColor visual found\n",
	    program_invocation_short_name);
    fprintf(stderr, "%s: Available visuals are:\n",
	    program_invocation_short_name);

    vi = display.getVisualInfo(0, NULL);

    for (unsigned int i = 0; i < vi.size(); i++) {
	const char *class_str = "?";
	switch (vi[i].c_class) {
	case StaticGray: class_str = "StaticGray"; break;
	case GrayScale: class_str = "GrayScale"; break;
	case StaticColor: class_str = "StaticColor"; break;
	case PseudoColor: class_str = "PseudoColor"; break;
	case TrueColor: class_str = "TrueColor"; break;
	case DirectColor: class_str = "DirectColor"; break;
	}
	if (vi[i].visual == screen.defaultVisual())
	    fprintf(stderr, "    =>");
	fprintf(stderr, "\tvisualid = %ld, screen = %d, depth = %d, "
			"class = %s,\n\t    masks = %lx/%lx/%lx, "
			"colormap_size = %d, bits_per_rgb = %d\n",
		vi[i].visualid, vi[i].screen, vi[i].depth, class_str,
		vi[i].red_mask, vi[i].green_mask, vi[i].blue_mask,
		vi[i].colormap_size, vi[i].bits_per_rgb);
    }

    exit(EXIT_FAILURE);
}

X::FontSet Connection::Impl::load_font_set(const char *base_font_name_list)
{
    X::FontSet font_set;
    char **missing_charset_list_return;
    int missing_charset_count_return;
    char *def_string_return;

    font_set = display.createFontSet(
	base_font_name_list,
	&missing_charset_list_return,
	&missing_charset_count_return,
	&def_string_return);

    if (font_set == NULL)
	errx(EXIT_FAILURE, "XCreateFontSet: Unable to create font set");

    if (missing_charset_count_return != 0)
	XFreeStringList(missing_charset_list_return);

    return font_set;
}

void Connection::Impl::open_input_method()
{
    XIMStyles *xim_styles = NULL;

    if ((input_method = display.openIM(NULL, NULL, NULL)) == NULL) {
	fprintf(stderr, "%s: XOpenIM: No input method could be opened\n",
		program_invocation_short_name);
	if (getenv("XMODIFIERS") != NULL)
	    fprintf(stderr, "  (Try unsetting the 'XMODIFIERS' environment "
			    "variable)\n");
	exit(EXIT_FAILURE);
    }

    if (XGetIMValues(input_method,
		     XNQueryInputStyle, &xim_styles, NULL) != NULL)
	errx(EXIT_FAILURE, "XGetIMValues: Could not obtain input styles");

    if (xim_styles == NULL)
	errx(EXIT_FAILURE, "Input method doesn't support any styles");

    xim_style = 0;

    for (unsigned int i = 0; i < xim_styles->count_styles; i++)
	if (xim_styles->supported_styles[i] ==
	    (XIMPreeditNothing | XIMStatusNothing)) {
	    xim_style = xim_styles->supported_styles[i];
	    break;
	}

    if (xim_style == 0)
	errx(EXIT_FAILURE, "Input method doesn't support root window style");

    XFree(xim_styles);
}

WindowHook *Connection::Impl::find_hook(Window window)
{
    for (std::vector<WindowHook *>::const_iterator i = hooks.begin();
	 i != hooks.end(); i++)
	if ((*i)->window() == window)
	    return *i;

    return 0;
}

void Connection::Impl::grab_pointer(Time time, X::Window window,
				    bool owner_events)
{
    if (disable_pointer_grab)
	return;

    window.grabPointer(
	owner_events,
	ButtonPressMask | ButtonReleaseMask |
	EnterWindowMask | LeaveWindowMask |
	PointerMotionMask | StructureNotifyMask | FocusChangeMask,
	GrabModeAsync, GrabModeAsync, None, None, time);
}

void Connection::Impl::ungrab_pointer(Time time)
{
    if (disable_pointer_grab)
	return;

    display.ungrabPointer(time);
}

void Connection::Impl::leave_menu_mode(Time time)
{
    if (!root_menu)
	return;

    root_menu->close();
    root_menu = 0;

    ungrab_pointer(time);
    emulates_implicit_grab = false;
}

/****************************************************************************/

void Connection::handleEvent(const XEvent &event)
{
    if (XFilterEvent(const_cast<XEvent *>(&event), None))
	return;

    if (event.type == ButtonPress &&
	event.xbutton.window == p->screen.rootWindow())
	leaveMenuMode(event.xbutton.time);

    switch (event.type) {
	WindowHook *hook;

    case KeyPress:
	/* Display *display;
	   Window window;
	   Window root;
	   Window subwindow;
	   Time time;
	   int x, y;
	   int x_root, y_root;
	   unsigned int state;
	   unsigned int keycode;
	   Bool same_screen; */
	hook = p->find_hook(event.xkey.window);
	if (hook)
	    hook->keyPress(
		X::Window(event.xkey.display, event.xkey.root),
		X::Window(event.xkey.display, event.xkey.subwindow),
		event.xkey);
	break;

    case KeyRelease:
	hook = p->find_hook(event.xkey.window);
	if (hook)
	    hook->keyRelease(
		X::Window(event.xkey.display, event.xkey.root),
		X::Window(event.xkey.display, event.xkey.subwindow),
		event.xkey);
	break;

    case ButtonPress:
	/* Display *display;
	   Window window;
	   Window root;
	   Window subwindow;
	   Time time;
	   int x, y;
	   int x_root, y_root;
	   unsigned int state;
	   unsigned int button;
	   Bool same_screen; */
	hook = p->find_hook(event.xbutton.window);
	if (hook)
	    hook->buttonPress(
		X::Window(event.xbutton.display, event.xbutton.root),
		X::Window(event.xbutton.display, event.xbutton.subwindow),
		event.xbutton);
	break;

    case ButtonRelease:
	hook = p->find_hook(event.xbutton.window);
	if (hook)
	    hook->buttonRelease(
		X::Window(event.xbutton.display, event.xbutton.root),
		X::Window(event.xbutton.display, event.xbutton.subwindow),
		event.xbutton);
	break;

    case MotionNotify:
	/* Display *display;
	   Window window;
	   Window root;
	   Window subwindow;
	   Time time;
	   int x, y;
	   int x_root, y_root;
	   unsigned int state;
	   char is_hint;
	   Bool same_screen; */
	hook = p->find_hook(event.xmotion.window);
	if (hook)
	    hook->motionNotify(
		X::Window(event.xmotion.display, event.xmotion.root),
		X::Window(event.xmotion.display, event.xmotion.subwindow),
		event.xmotion);
	break;

    case EnterNotify:
	/* Display *display;
	   Window window;
	   Window root;
	   Window subwindow;
	   Time time;
	   int x, y;
	   int x_root, y_root;
	   int mode;
	     * NotifyNormal, NotifyGrab, NotifyUngrab
	   int detail;
	     * NotifyAncestor, NotifyVirtual, NotifyInferior,
	     * NotifyNonlinear, NotifyNonlinearVirtual
	   Bool same_screen;
	   Bool focus;
	   unsigned int state; */
	hook = p->find_hook(event.xcrossing.window);
	if (hook)
	    hook->enterNotify(
		X::Window(event.xcrossing.display, event.xcrossing.root),
		X::Window(event.xcrossing.display, event.xcrossing.subwindow),
		event.xcrossing);
	break;

    case LeaveNotify:
	hook = p->find_hook(event.xcrossing.window);
	if (hook)
	    hook->leaveNotify(
		X::Window(event.xcrossing.display, event.xcrossing.root),
		X::Window(event.xcrossing.display, event.xcrossing.subwindow),
		event.xcrossing);
	break;

    case FocusIn:
	/* Display *display;
	   Window window;
	   int mode;
	     * NotifyNormal, NotifyWhileGrabbed,
	     * NotifyGrab, NotifyUngrab
	   int detail;
	     * NotifyAncestor, NotifyVirtual, NotifyInferior,
	     * NotifyNonlinear, NotifyNonlinearVirtual, NotifyPointer,
	     * NotifyPointerRoot, NotifyDetailNone */
	hook = p->find_hook(event.xfocus.window);
	if (hook)
	    hook->focusIn(event.xfocus);
	break;

    case FocusOut:
	hook = p->find_hook(event.xfocus.window);
	if (hook)
	    hook->focusOut(event.xfocus);
	break;

    case Expose:
	/* Display *display;
	   Window window;
	   int x, y;
	   int width, height;
	   int count; */
	hook = p->find_hook(event.xexpose.window);
	if (hook)
	    hook->expose(event.xexpose);
	break;

    case VisibilityNotify:
	/* Display *display;
	   Window window;
	   int state; */
	hook = p->find_hook(event.xvisibility.window);
	if (hook)
	    hook->visibilityNotify(event.xvisibility);
	break;

    case DestroyNotify:
	/* Display *display;
	   Window event;
	   Window window; */
	hook = p->find_hook(event.xdestroywindow.window);
	if (hook)
	    hook->destroyNotify(event.xdestroywindow);
	break;

    case UnmapNotify:
	/* Display *display;
	   Window event;
	   Window window;
	   Bool from_configure; */
	hook = p->find_hook(event.xunmap.window);
	if (hook)
	    hook->unmapNotify(event.xunmap);
	break;

    case MapNotify:
	/* Display *display;
	   Window event;
	   Window window;
	   Bool override_redirect; */
	hook = p->find_hook(event.xmap.window);
	if (hook)
	    hook->mapNotify(event.xmap);
	break;

    case ReparentNotify:
	/* Display *display;
	   Window event;
	   Window window;
	   Window parent;
	   int x, y;
	   Bool override_redirect; */
	hook = p->find_hook(event.xreparent.window);
	if (hook)
	    hook->reparentNotify(event.xreparent);
	break;

    case ConfigureNotify:
	/* Display *display;
	   Window event;
	   Window window;
	   int x, y;
	   int width, height;
	   int border_width;
	   Window above;
	   Bool override_redirect; */
	hook = p->find_hook(event.xconfigure.window);
	if (hook)
	    hook->configureNotify(event.xconfigure);
	break;

    case SelectionClear:
	/* Display *display;
	   Window window;
	   Atom selection;
	   Time time; */
	hook = p->find_hook(event.xselectionclear.window);
	if (hook)
	    hook->selectionClear(event.xselectionclear);
	break;

    case SelectionRequest:
	/* Display *display;
	   Window owner;
	   Window requestor;
	   Atom selection;
	   Atom target;
	   Atom property;
	   Time time; */
	hook = p->find_hook(event.xselectionrequest.owner);
	if (hook)
	    hook->selectionRequest(
		X::Window(event.xselectionrequest.display,
			  event.xselectionrequest.requestor),
		event.xselectionrequest);
	break;

    case SelectionNotify:
	/* Display *display;
	   Window requestor;
	   Atom selection;
	   Atom target;
	   Atom property;
	   Time time; */
	hook = p->find_hook(event.xselection.requestor);
	if (hook)
	    hook->selectionNotify(event.xselection);

    case ClientMessage:
	/* Display *display;
	   Window window;
	   Atom message_type;
	   int format;
	   union {
	     char b[20];
	     short s[10];
	     long l[5];
	   } data; */
	if (event.xclient.message_type != p->atoms.WM_PROTOCOLS) {
	    char *atom = p->display.getAtomName(event.xclient.message_type);
	    warnx("Received ClientMessage event, message_type = %ld ('%s')",
		  event.xclient.message_type, atom);
	    XFree(atom);
	} else if ((Atom)event.xclient.data.l[0] != p->atoms.WM_DELETE_WINDOW) {
	    char *atom = p->display.getAtomName(event.xclient.data.l[0]);
	    warnx("Received WM_PROTOCOLS event, protocol = %ld ('%s')",
		  event.xclient.data.l[0], atom);
	    XFree(atom);
	} else {
	    hook = p->find_hook(event.xclient.window);
	    if (hook)
		hook->deleteWindow(event.xclient);
	}
	break;
    }
}
