/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <assert.h>
#include <algorithm>
#include "container0.h"

#define CEND(x) static_cast<std::vector<Control *>::const_iterator>((x).end())

using namespace Toolkit;


ContainerWidget::Control::Control(ContainerWidget &container)
    : WidgetControl(container.conn), horz_scroller(0), vert_scroller(0),
      is_enabled(true), container(container)
{
    container.add_control(this);
}

ContainerWidget::Control::~Control()
{
    container.remove_control(this);
}

WidgetControlScroller *ContainerWidget::Control::horzScroller() const
{
    return horz_scroller;
}

void ContainerWidget::Control::setHorzScroller(
    WidgetControlScroller *horzScroller)
{
    horz_scroller = horzScroller;
}

WidgetControlScroller *ContainerWidget::Control::vertScroller() const
{
    return vert_scroller;
}

void ContainerWidget::Control::setVertScroller(
    WidgetControlScroller *vertScroller)
{
    vert_scroller = vertScroller;
}

bool ContainerWidget::Control::isEnabled() const
{
    return is_enabled;
}

void ContainerWidget::Control::setEnabled(bool enable)
{
    if (enable == is_enabled)
	return;

    is_enabled = enable;

    if (!enable)
	container.disable_control(this);
    else
	container.enable_control(this);
}

X::Window ContainerWidget::Control::window() const
{
    if (!container.control())
	return X::Window();

    return container.control()->window();
}

X::GC ContainerWidget::Control::gc() const
{
    if (!container.control())
	return X::GC();

    return container.control()->gc();
}

WidgetWindowHook *ContainerWidget::Control::hook() const
{
    if (!container.control())
	return 0;

    return container.control()->hook();
}

void ContainerWidget::Control::sizeRequirementsChanged()
{
    container.sizeRequirementsChanged(this);
}

void ContainerWidget::Control::propagateChildLayoutRequest()
{
    if (container.is_updating_layout)
	return;

    if (container.child_update_iterator &&
	std::find(*container.child_update_iterator + 1,
		  CEND(container.visible_controls), this)
	    != container.visible_controls.end())
	return;

    if (container.control())
	container.control()->requestChildLayout();
}

void ContainerWidget::Control::focus(Focussable *focussable)
{
    if (container.control())
	container.control()->focus(focussable);
}

void ContainerWidget::Control::focussablesChanged()
{
    std::vector<Control *> controls = container.visible_controls;
    std::vector<Focussable *> focussables;

    for (std::vector<Control *>::const_iterator i = controls.begin();
	 i != controls.end(); i++) {
	Widget *widget = (*i)->widget();
	if (!widget)
	    continue;

	std::vector<Focussable *> f = widget->focussables();
	focussables.reserve(focussables.size() + f.size());
	focussables.insert(focussables.end(), f.begin(), f.end());
    }

    container.setFocussables(focussables);
}

void ContainerWidget::Control::setDimensions(const Region &dimensions)
{
    WidgetControl::setDimensions(dimensions);
}

/****************************************************************************/

ContainerWidget::ContainerWidget(Connection &conn)
    : Widget(conn), is_updating_layout(false), child_update_iterator(0)
{
}

ContainerWidget::~ContainerWidget()
{
    assert(registered_controls.empty());
}

std::vector<Widget *> ContainerWidget::children() const
{
    std::vector<Widget *> widgets;

    for (std::vector<Control *>::const_iterator i = visible_controls.begin();
	 i != visible_controls.end(); i++) {
	Widget *widget = (*i)->widget();
	if (widget)
	    widgets.push_back(widget);
    }

    return widgets;
}

void ContainerWidget::recursivelyUpdateLayout()
{
    std::vector<Control *>::const_iterator i;

    is_updating_layout = true;
    Widget::recursivelyUpdateLayout();
    is_updating_layout = false;

    child_update_iterator = &i;
    for (i = visible_controls.begin(); i != visible_controls.end(); i++)
	(*i)->updateChildLayout();
    child_update_iterator = 0;
}

void ContainerWidget::setMnemonicVisibility(bool visible)
{
    for (std::vector<Control *>::const_iterator i = visible_controls.begin();
	 i != visible_controls.end(); i++) {
	Widget *widget = (*i)->widget();
	if (widget)
	    widget->setMnemonicVisibility(visible);
    }
}

void ContainerWidget::windowPositionChanged()
{
    for (std::vector<Control *>::const_iterator i = visible_controls.begin();
	 i != visible_controls.end(); i++) {
	Widget *widget = (*i)->widget();
	if (widget)
	    widget->windowPositionChanged();
    }
}

void ContainerWidget::paint(const Region &region, X::Pixmap pixmap)
{
    for (std::vector<Control *>::const_iterator i = visible_controls.begin();
	 i != visible_controls.end(); i++) {
	Widget *widget = (*i)->widget();
	if (widget)
	    widget->paint(region, pixmap);
    }
}

std::vector<ContainerWidget::Control *>
    ContainerWidget::registeredControls() const
{
    return registered_controls;
}

std::vector<ContainerWidget::Control *>
    ContainerWidget::visibleControls() const
{
    return visible_controls;
}

void ContainerWidget::add_control(Control *control)
{
    assert(std::find(registered_controls.begin(),
		     registered_controls.end(), control)
	   == registered_controls.end());
    registered_controls.push_back(control);

    if (control->isEnabled())
	enable_control(control);
}

void ContainerWidget::remove_control(Control *control)
{
    if (control->isEnabled())
	disable_control(control);

    std::vector<Control *>::iterator i =
	std::find(registered_controls.begin(),
		  registered_controls.end(), control);
    assert(i != registered_controls.end());
    registered_controls.erase(i);
}

void ContainerWidget::enable_control(Control *control)
{
    assert(!child_update_iterator);
    assert(std::find(visible_controls.begin(),
		     visible_controls.end(), control)
	   == visible_controls.end());
    visible_controls.push_back(control);
}

void ContainerWidget::disable_control(Control *control)
{
    assert(!child_update_iterator);
    std::vector<Control *>::iterator i =
	std::find(visible_controls.begin(),
		  visible_controls.end(), control);
    assert(i != visible_controls.end());
    visible_controls.erase(i);
}

void ContainerWidget::sizeRequirementsChanged(WidgetControl *control)
{
    /* Relax. */
}
