/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <math.h>
#include <assert.h>
#include <sstream>
#include <algorithm>
#include "columnw.h"
#include "dockboxw.h"
#include "configuration.h"
#include "logic.h"

#define INNER_PLASMA_X 6
#define INNER_PLASMA_Y 6

using namespace Toolkit::Docks;


static int fixup_split_pos_y(int splitPos, int height,
			     const Toolkit::SizeRequirements &top,
			     const Toolkit::SizeRequirements &bottom)
{
    if (bottom.maxHeight != -1 &&
	splitPos < height - bottom.maxHeight)
	splitPos = height - bottom.maxHeight;
    if (splitPos > height - bottom.minHeight)
	splitPos = height - bottom.minHeight;

    if (top.maxHeight != -1 &&
	splitPos > top.maxHeight)
	splitPos = top.maxHeight;
    if (splitPos < top.minHeight)
	splitPos = top.minHeight;

    return splitPos;
}

static std::vector<unsigned int> distribute_height(
    const std::vector<DockboxWidget *> &docks,
    const std::map<DockboxWidget *, double> &share, unsigned int height)
{
    std::vector<unsigned int> heights;
    double shares_left = 0.;

    height -= (docks.size() - 1) * INNER_PLASMA_Y;

    for (std::vector<DockboxWidget *>::const_iterator i = docks.begin();
	 i != docks.end(); i++)
	shares_left += share.at(*i);

    for (std::vector<DockboxWidget *>::const_iterator i = docks.begin();
	 i != docks.end(); i++) {
	unsigned int dock_height = height;
	if (i != docks.end() - 1)
	    dock_height = lrint(dock_height * share.at(*i) / shares_left);
	heights.push_back(dock_height);
	shares_left -= share.at(*i);
	height -= dock_height;
    }

    /* Make distribution more fair. */

    std::vector<Toolkit::SizeRequirements> reqs;
    for (std::vector<DockboxWidget *>::const_iterator i = docks.begin();
	 i != docks.end(); i++)
	reqs.push_back((*i)->sizeRequirements());

    int to_be_distributed = 0;
    for (unsigned int i = 0; i < docks.size(); i++) {
	if (heights[i] > reqs[i].maxHeight) {
	    to_be_distributed += heights[i] - reqs[i].maxHeight;
	    heights[i] = reqs[i].maxHeight;
	}
	if (heights[i] < reqs[i].minHeight) {
	    to_be_distributed += heights[i] - reqs[i].minHeight;
	    heights[i] = reqs[i].minHeight;
	}
    }

    while (to_be_distributed != 0) {
	unsigned int enlargeable_dock_count = 0;
	unsigned int shrinkable_dock_count = 0;
	int max_enlargement = -1;
	int max_shrinking = -1;

	for (unsigned int i = 0; i < docks.size(); i++) {
	    if (heights[i] > reqs[i].minHeight) {
		shrinkable_dock_count++;
		if (max_shrinking == -1 ||
		    max_shrinking > heights[i] - reqs[i].minHeight)
		    max_shrinking = heights[i] - reqs[i].minHeight;
	    }
	    if (reqs[i].maxHeight == -1)
		enlargeable_dock_count++;
	    else if (heights[i] < reqs[i].maxHeight) {
		enlargeable_dock_count++;
		if (max_enlargement == -1 ||
		    max_enlargement > reqs[i].maxHeight - heights[i])
		    max_enlargement = reqs[i].maxHeight - heights[i];
	    }
	}

	if (to_be_distributed < 0) {
	    assert(shrinkable_dock_count != 0);
	    for (unsigned int i = 0; i < docks.size(); i++) {
		if (heights[i] <= reqs[i].minHeight)
		    continue;
		int amount = -to_be_distributed / shrinkable_dock_count;
		if (max_shrinking != -1 && amount > max_shrinking)
		    amount = max_shrinking;
		heights[i] -= amount;
		to_be_distributed += amount;
		shrinkable_dock_count--;
	    }
	} else {
	    assert(enlargeable_dock_count != 0);
	    for (unsigned int i = 0; i < docks.size(); i++) {
		if (reqs[i].maxHeight != -1 &&
		    heights[i] >= reqs[i].maxHeight)
		    continue;
		int amount = to_be_distributed / enlargeable_dock_count;
		if (max_enlargement != 1 && amount > max_enlargement)
		    amount = max_enlargement;
		heights[i] += amount;
		to_be_distributed -= amount;
		enlargeable_dock_count--;
	    }
	}
    }

    return heights;
}

/****************************************************************************/

class ColumnWidget::ResizeOperation
{
    ColumnWidget &column;
    DockboxWidget *const dockbox;
    double const top_ratio;
    int const grip;
    std::map<DockboxWidget *, double> const share;

    /* cached computations */
    std::vector<DockboxWidget *>::const_iterator pos;
    bool invalid_configuration;
    std::vector<DockboxWidget *> top_docks, bottom_docks;
    SizeRequirements top_reqs, gross_top_reqs, bottom_reqs;
public:
    ResizeOperation(ColumnWidget &column, DockboxWidget *dockbox,
		    double top_ratio, int grip,
		    std::map<DockboxWidget *, double> share)
	: column(column), dockbox(dockbox), top_ratio(top_ratio),
	  grip(grip), share(share), invalid_configuration(false) {
	update();
    }
    void update();
    void motionEvent(int y);
};

static Toolkit::SizeRequirements sum_size_requirements(
    std::vector<DockboxWidget *>::const_iterator begin,
    std::vector<DockboxWidget *>::const_iterator end)
{
    if (begin == end)
	return Toolkit::SizeRequirements();

    Toolkit::SizeRequirements reqs((*begin)->sizeRequirements());

    while (++begin != end) {
	reqs.addY(INNER_PLASMA_Y);
	reqs.addY((*begin)->sizeRequirements());
    }
    return reqs;
}

void ColumnWidget::ResizeOperation::update()
{
    /* Cache all computations that depend only on the dock configuration and
       the docks' size requirements. */

    pos = std::find(column.docks.begin(), column.docks.end(), dockbox);
    invalid_configuration = pos == column.docks.end() ||
			    pos == column.docks.end() - 1;
    if (invalid_configuration)
	return;

    top_docks.assign(column.docks.begin(), pos);
    bottom_docks.assign(pos + 1, column.docks.end());

    top_reqs = sum_size_requirements(column.docks.begin(), pos);
    gross_top_reqs = sum_size_requirements(column.docks.begin(), pos + 1);
    bottom_reqs = sum_size_requirements(pos + 1, column.docks.end());
}

void ColumnWidget::ResizeOperation::motionEvent(int y)
{
    if (invalid_configuration)
	return;

    int height_to_distribute = column.height() - INNER_PLASMA_Y;
    int gross_top_height = fixup_split_pos_y(y - grip - column.y(),
	height_to_distribute, gross_top_reqs, bottom_reqs);

    std::vector<unsigned int> heights;

    if (pos == column.docks.begin())
	heights.push_back(gross_top_height);
    else {
	int top_height = fixup_split_pos_y(
	    top_ratio * height_to_distribute,
	    gross_top_height - INNER_PLASMA_Y,
	    top_reqs, dockbox->sizeRequirements());

	heights = distribute_height(top_docks, share, top_height);
	heights.push_back(gross_top_height - top_height - INNER_PLASMA_Y);
    }

    std::vector<unsigned int> bottom_heights = distribute_height(
	bottom_docks, share, height_to_distribute - gross_top_height);
    heights.insert(heights.end(), bottom_heights.begin(),
				  bottom_heights.end());
    assert(heights.size() == column.docks.size());

/*  unsigned int total_height = 0;
    for (std::vector<unsigned int>::const_iterator i = heights.begin();
	 i != heights.end(); i++)
	total_height += *i;
    assert(total_height == column.height()
	       - (column.docks.size() - 1) * INNER_PLASMA_Y); */

    for (unsigned int i = 0; i < column.docks.size(); i++)
	column.share[column.docks[i]] = heights[i];

    column.requestLayout();
}

/****************************************************************************/

ColumnWidget::ColumnWidget(Connection &conn, Logic &logic)
    : ContainerWidget(conn), logic(logic), resize(0), docks(_docks)
{
}

ColumnWidget::~ColumnWidget()
{
    delete resize;

    for (std::vector<DockboxWidget *>::const_iterator i = _docks.begin();
	 i != _docks.end(); i++)
	delete *i;

    for (std::vector<Control *>::const_iterator i = controls.begin();
	 i != controls.end(); i++)
	delete *i;
}

void ColumnWidget::readConfiguration(Configuration &config)
{
    if (config.listFirst())
	do {
	    DockboxWidget *dockbox = new DockboxWidget(conn, logic);
	    try {
		dockbox->readConfiguration(config);
		config.consumeSpace();
		share[dockbox] =
		    Configuration::toDouble(config.readLastToken());
	    } catch (Configuration::Error &e) {
		delete dockbox;
		throw;
	    };

	    _docks.push_back(dockbox);

	    Control *control = new FocussableControl(
		*this, logic.focusHandler);
	    control->setWidget(dockbox);
	    controls.push_back(control);

	    update_size_requirements();
	    requestLayout();
	} while (config.listNext());
}

void ColumnWidget::writeConfiguration(std::ostringstream &stream) const
{
    double total_share = 0.;
    for (std::vector<DockboxWidget *>::const_iterator i = docks.begin();
	 i != docks.end(); i++)
	total_share += share.at(*i);

    stream << "(";
    for (std::vector<DockboxWidget *>::const_iterator i = docks.begin();
	 i != docks.end(); i++) {
	if (i != docks.begin())
	    stream << ", ";
	(*i)->writeConfiguration(stream);
	stream << " " << share.at(*i) / total_share;
    }
    stream << ")";
}

void ColumnWidget::resetConfiguration()
{
    for (std::vector<DockboxWidget *>::const_iterator i = _docks.begin();
	 i != _docks.end(); i++)
	delete *i;
    _docks.clear();

    for (std::vector<Control *>::const_iterator i = controls.begin();
	 i != controls.end(); i++)
	delete *i;
    controls.clear();

    share.clear();

    update_size_requirements();
    requestLayout();
}

void ColumnWidget::sizeRequirementsChanged(WidgetControl *control)
{
    update_size_requirements();
}

void ColumnWidget::update_size_requirements()
{
    SizeRequirements reqs;

    if (!docks.empty()) {
	std::vector<DockboxWidget *>::const_iterator i = docks.begin();
	reqs = (*i)->sizeRequirements();

	while (++i != docks.end()) {
	    reqs.addY(INNER_PLASMA_Y);
	    reqs.addY((*i)->sizeRequirements());
	}
    }

    setSizeRequirements(reqs);
}

void ColumnWidget::updateLayout()
{
    Region dim(control()->dimensions());
    std::vector<unsigned int> heights =
	distribute_height(docks, share, dim.height);
    unsigned int ypos = 0;

    for (unsigned int i = 0; i < docks.size(); i++) {
	controls[i]->setDimensions(Region(dim.x, dim.y + ypos,
					  dim.width, heights[i]));
	ypos += heights[i] + INNER_PLASMA_Y;
    }
    assert(ypos - INNER_PLASMA_Y == dim.height);

    if (resize)
	resize->update();
}

void ColumnWidget::motionEvent(
    Time time, int x, int y, int x_root, int y_root,
    unsigned int state, char is_hint)
{
    if (resize)
	resize->motionEvent(y);
}

void ColumnWidget::buttonPressEvent(
    Time time, int x, int y, int x_root, int y_root,
    unsigned int state, unsigned int button)
{
    if (button != Button1)
	return;

    Region dim(control()->dimensions());
    if (x < dim.x || x >= dim.x + dim.width || resize || docks.empty())
	return;

    for (std::vector<DockboxWidget *>::const_iterator i = docks.begin();
	 i != docks.end() - 1; i++) {
	Region region = (*i)->control()->dimensions();

	if (y >= region.y + region.height &&
	    y < region.y + region.height + INNER_PLASMA_Y) {
	    resize = new ResizeOperation(
		*this, *i,
		(double)(region.y - INNER_PLASMA_Y - dim.y) /
		(double)(dim.height - INNER_PLASMA_Y),
		y - (region.y + region.height), share);
	    window().defineCursor(logic.vertResizeCursor());
	    break;
	}
    }
}

void ColumnWidget::buttonReleaseEvent(
    Time time, int x, int y, int x_root, int y_root,
    unsigned int state, unsigned int button)
{
    if (button != Button1 || !resize)
	return;

    delete resize;
    resize = 0;
    window().undefineCursor();
}
