/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <assert.h>
#include <algorithm>
#include "focus.h"

using namespace Toolkit::Docks;


FocussableControl::FocussableControl(ContainerWidget &container,
				     FocusHandler &focusHandler)
    : Control(container), focus_handler(focusHandler)
{
}

void FocussableControl::focus(Focussable *focussable)
{
    focus_handler.focus(this, focussable);
}

void FocussableControl::focussablesChanged()
{
    if (focus_handler.focussedControl() == this)
	focus_handler.focussablesChanged();
}


FocusHandler::FocusHandler(FocussableControl &defaultControl)
    : focussed_control(&defaultControl), default_control(&defaultControl),
      focussed(0), has_focus(false)
{
}

FocussableControl *FocusHandler::focussedControl() const
{
    return focussed_control;
}

void FocusHandler::setFocussedControl(FocussableControl *control)
{
    focus(control, 0);
}

void FocusHandler::focus(FocussableControl *control, Focussable *focussable)
{
    if (control != focussed_control) {
	if (focussed_control &&
	    focussed_control->widget())
	    focussed_control->widget()->invalidate();
	focussed_control = control;
	if (focussed_control &&
	    focussed_control->widget()) {
	    focussed_control->widget()->invalidate();
	    focussables = focussed_control->widget()->focussables();
	} else
	    focussables.clear();
    } else if (!focussable || focussable == focussed)
	return;

    if (focussed && has_focus)
	focussed->focusOutEvent();

    if (focussable) {
	assert(std::find(focussables.begin(),
			 focussables.end(), focussable) != focussables.end());
	focussed = focussable;
    } else
	focussed = focussables.empty() ? 0 : focussables.front();

    if (focussed && has_focus)
	focussed->focusInEvent();
}

void FocusHandler::focussablesChanged()
{
    std::vector<Focussable *> f;
    if (focussed_control->widget())
	f = focussed_control->widget()->focussables();

    if (f == focussables)
	return;

    focussables = f;

    if (focussed &&
	std::find(focussables.begin(),
		  focussables.end(), focussed) == focussables.end()) {
	if (has_focus)
	    focussed->focusOutEvent();
	focussed = 0;
    }

    if (!focussed && !focussables.empty()) {
	focussed = focussables.front();
	if (has_focus)
	    focussed->focusInEvent();
    }
}

void FocusHandler::keyPressEvent(
    Time time, int x, int y, int x_root, int y_root,
    unsigned int state, unsigned int keycode,
    const std::string &text, KeySym keysym)
{
    if (keysym == XK_Tab || keysym == XK_ISO_Left_Tab) {
	if (focussables.empty())
	    return;

	std::vector<Focussable *>::const_iterator i =
	    std::find(focussables.begin(), focussables.end(), focussed);
	assert(i != focussables.end());

	switch (keysym) {
	case XK_Tab:
	    if (i == focussables.end() - 1)
		i = focussables.begin();
	    else
		i++;
	    break;
	case XK_ISO_Left_Tab:
	    if (i == focussables.begin())
		i = focussables.end() - 1;
	    else
		i--;
	    break;
	}

	if (*i == focussed)
	    return;

	if (has_focus)
	    focussed->focusOutEvent();
	focussed = *i;
	if (has_focus)
	    focussed->focusInEvent();
	return;
    }

    if (focussed)
	focussed->keyPressEvent(
	    time, x, y, x_root, y_root, state, keycode, text, keysym);

    if (keysym == XK_Escape)
	setFocussedControl(default_control);
}

void FocusHandler::keyReleaseEvent(
    Time time, int x, int y, int x_root, int y_root,
    unsigned int state, unsigned int keycode)
{
    if (focussed)
	focussed->keyReleaseEvent(
	    time, x, y, x_root, y_root, state, keycode);
}

void FocusHandler::focusInEvent()
{
    has_focus = true;
    if (focussed)
	focussed->focusInEvent();
}

void FocusHandler::focusOutEvent()
{
    has_focus = false;
    if (focussed)
	focussed->focusOutEvent();
}
