/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <err.h>
#include <assert.h>
#include <algorithm>
#include "event_loop.h"
#include "connection.h"
#include "timer.h"

using namespace Toolkit;

std::vector<Connection *> EventLoop::connections;
std::vector<EventLoop::Reader *> EventLoop::readers;
std::vector<EventLoop::Idler *> EventLoop::idlers;


EventLoop::Reader::Reader(int fd) : fd(fd)
{
    readers.push_back(this);
}

EventLoop::Reader::~Reader()
{
    readers.erase(std::find(readers.begin(), readers.end(), this));
}

EventLoop::Idler::Idler()
{
    idlers.push_back(this);
}

EventLoop::Idler::~Idler()
{
    idlers.erase(std::find(idlers.begin(), idlers.end(), this));
}


void EventLoop::addConnection(Connection *connection)
{
    assert(std::find(connections.begin(),
		     connections.end(), connection) == connections.end());
    connections.push_back(connection);
}

void EventLoop::removeConnection(Connection *connection)
{
    std::vector<Connection *>::iterator i = std::find(
	connections.begin(), connections.end(), connection);
    assert(i != connections.end());
    connections.erase(i);
}

void EventLoop::run()
{
    XEvent event;

    struct timespec idle_timespec;
    idle_timespec.tv_sec = 0;
    idle_timespec.tv_nsec = 0;

    for (std::vector<Connection *>::const_iterator i = connections.begin();
	 i != connections.end(); i++)
	while ((*i)->display().checkMaskEvent(~0, &event))
	    (*i)->handleEvent(event);

    while (!connections.empty() || !readers.empty() || !idlers.empty()) {
	int nfds = 0;
	fd_set readfds, writefds, exceptfds;
	const struct timespec *timeout;
	// sigset_t sigmask;

	FD_ZERO(&readfds);
	for (std::vector<Connection *>::const_iterator i = connections.begin();
	     i != connections.end(); i++) {
	    int fd = (*i)->display().connectionNumber();
	    if (fd >= (int)sizeof readfds * 8)
		errx(EXIT_FAILURE, "Connection number exceeds limit (is %d, "
				   "limit %ld)", fd, sizeof readfds * 8);
	    if (nfds < fd + 1)
		nfds = fd + 1;
	    FD_SET(fd, &readfds);
	}

	for (std::vector<Reader *>::const_iterator i = readers.begin();
	     i != readers.end(); i++) {
	    int fd = (*i)->fd;
	    if (fd >= (int)sizeof readfds * 8)
		errx(EXIT_FAILURE, "File descriptor exceeds limit (is %d, "
				   "limit %ld)", fd, sizeof readfds * 8);
	    if (nfds < fd + 1)
		nfds = fd + 1;
	    FD_SET(fd, &readfds);
	}

	FD_ZERO(&writefds);
	FD_ZERO(&exceptfds);

	timeout = process_timers();

	// sigemptyset(&sigmask);

	for (std::vector<Connection *>::const_iterator i = connections.begin();
	     i != connections.end(); i++) {
	    (*i)->update();
	    (*i)->display().flush();
	    if ((*i)->display().eventsQueued(QueuedAlready) != 0)
		timeout = &idle_timespec;
	}

	if (!idlers.empty())
	    timeout = &idle_timespec;

	if (pselect(nfds, &readfds, &writefds, &exceptfds,
		    timeout, NULL) == -1)
	    err(EXIT_FAILURE, "pselect");

	std::vector<Connection *> cached_connections = connections;
	for (std::vector<Connection *>::const_iterator i
		 = cached_connections.begin();
	     i != cached_connections.end(); i++) {
	    if (FD_ISSET((*i)->display().connectionNumber(), &readfds))
		(*i)->display().flush();

	    while ((*i)->display().eventsQueued(QueuedAlready) != 0) {
		(*i)->display().nextEvent(&event);
		(*i)->handleEvent(event);
	    }

	    if ((*i)->abandoned())
		delete *i;
	}

	std::vector<Reader *> cached_readers = readers;
	for (std::vector<Reader *>::const_iterator i = cached_readers.begin();
	     i != cached_readers.end(); i++)
	    if (FD_ISSET((*i)->fd, &readfds))
		(*i)->read();

	std::vector<Idler *> cached_idlers = idlers;
	for (std::vector<Idler *>::const_iterator i = cached_idlers.begin();
	     i != cached_idlers.end(); i++)
	    (*i)->idle();
    }
}
