/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <wctype.h>
#include <err.h>

#define XK_MISCELLANY
#define XK_XKB_KEYS
#include <X11/keysymdef.h>
#undef XK_XKB_KEYS
#undef XK_MISCELLANY

#include "hotkey.h"
#include "util/encoding.h"
#include "keysym_glyph.h"
#include "keysym_nonglyph.h"

using namespace Toolkit;


Hotkey::Step::Step(unsigned int state, KeySym keysym, KeySym rawKeysym)
    : state(0), keysym(0), is_glyph(false), ch(0)
{
    state &= ~Mod2Mask; /* not interested in Num Lock */

    if (keysym == XK_ISO_Left_Tab && rawKeysym == XK_Tab) {
	/* Special case: Shift+Tab is translated to Shift+ISO_Left_Tab */
	/* We prefer Shift+Tab instead, so undo the translation */
	keysym = XK_Tab;
	state |= ShiftMask; /* just in case there is an actual left tab key */
    }

    /* Ctrl+Pause is translated to Ctrl+Break,
       Alt+Print is translated to Alt+Sys_Req */

    wchar_t ch = reverse_lookup_glyph_keysym(keysym);
    if (ch) {
	if (rawKeysym != keysym)
	    state &= ~ShiftMask & ~Mod5Mask;
	this->state = state;
	this->keysym = keysym;
	this->is_glyph = true;
	this->ch = ch;
	return;
    }

    const char *name = reverse_lookup_nonglyph_keysym(keysym);
    if (name) {
	this->state = state;
	this->keysym = keysym;
	this->is_glyph = false;
	this->name = name;
	return;
    }

    throw Invalid();
}

Hotkey::Step::Step(std::string const &repr)
    : state(0), keysym(0), is_glyph(false), ch(0)
{
    std::size_t p0 = 0, p1;

    while ((p1 = repr.find('-', p0)) != repr.npos) {
	     if (repr.compare(p0, p1 - p0, "S") == 0) state |= ShiftMask;
	else if (repr.compare(p0, p1 - p0, "L") == 0) state |= LockMask;
	else if (repr.compare(p0, p1 - p0, "C") == 0) state |= ControlMask;
	else if (repr.compare(p0, p1 - p0, "A") == 0) state |= Mod1Mask;
	else if (repr.compare(p0, p1 - p0, "M3") == 0) state |= Mod3Mask;
	else if (repr.compare(p0, p1 - p0, "M4") == 0) state |= Mod4Mask;
	else if (repr.compare(p0, p1 - p0, "M5") == 0) state |= Mod5Mask;
	else throw Invalid();

	p0 = p1 + 1;
    }

    std::string name = repr.substr(p0);
    std::wstring w;

    if (!decode(name, w))
	throw Invalid();

    if (w.length() == 1) {
	this->keysym = lookup_glyph_keysym(w[0]);
	this->is_glyph = true;
	this->ch = w[0];
    } else {
	this->keysym = lookup_nonglyph_keysym(name.data());
	this->is_glyph = false;
	this->name = name;
    }

    if (this->keysym == 0)
	throw Invalid();
}

bool Hotkey::Step::equals(const Step &step) const
{
    return state == step.state && keysym == step.keysym;
}

std::string Hotkey::Step::format() const
{
    bool longFormat = false;
    std::string result;

    if (!is_glyph)
	result = name;
    else {
	result = assert_encode(std::wstring(1, ch));

	if (longFormat) {
	    wchar_t lower = towlower(ch);
	    wchar_t upper = towupper(ch);

	    if (lower != upper) {
		if (ch == lower)
		    result = assert_encode(std::wstring(1, upper));
		if (ch == upper)
		    result = "Shift+" + result;
	    }
	}
    }

    unsigned int state = this->state;
    int modifier = 0;

    while (state) {
	if (state & 1)
	    switch (modifier) {
	    case 0:
		result = (longFormat ? "Shift+" : "S-") + result;
		break;
	    case 1:
		result = (longFormat ? "Lock+" : "L-") + result;
		break;
	    case 2:
		result = (longFormat ? "Ctrl+" : "C-") + result;
		break;
	    case 3:
		result = (longFormat ? "Alt+" : "A-") + result;
		break;
	    default:
		std::string mod = longFormat ? "ModX+" : "MX-";
		mod[longFormat ? 3 : 1] = '0' + modifier - 2;
		result = mod + result;
	    }

	state >>= 1;
	modifier++;
    }

    return result;
}

bool Toolkit::operator==(const Hotkey::Step &p0, const Hotkey::Step &p1)
{
    return p0.equals(p1);
}

bool Toolkit::operator!=(const Hotkey::Step &p0, const Hotkey::Step &p1)
{
    return !p0.equals(p1);
}

/****************************************************************************/

Hotkey::Hotkey() : is_invalid(false)
{
}

Hotkey::Hotkey(const Hotkey &continued,
	       unsigned int state, KeySym keysym, KeySym rawKeysym)
    : steps(continued.steps), is_invalid(continued.is_invalid)
{
    if (is_invalid)
	return;

    switch (keysym) {
    case XK_Shift_L:
    case XK_Shift_R:
    case XK_Control_L:
    case XK_Control_R:
    case XK_Caps_Lock:
    case XK_Shift_Lock:

    case XK_Meta_L:
    case XK_Meta_R:
    case XK_Alt_L:
    case XK_Alt_R:
    case XK_Super_L:
    case XK_Super_R:
    case XK_Hyper_L:
    case XK_Hyper_R:

    case XK_ISO_Level3_Shift:
	return;
    }

    try {
	steps.push_back(Step(state, keysym, rawKeysym));
    } catch (Step::Invalid &e) {
	steps.clear();
	is_invalid = true;
	warnx("Received invalid hotkey (state 0x%02x, keysym 0x%lx <- 0x%lx)",
	      state, keysym, rawKeysym);
    }
}

Hotkey::Hotkey(const char *repr) : is_invalid(false)
{
    parse(repr);
}

Hotkey::Hotkey(std::string const &repr) : is_invalid(false)
{
    parse(repr);
}

void Hotkey::parse(std::string const &repr)
{
    if (repr.empty())
	return;

    try {
	std::size_t p0 = 0, p1;
	while ((p1 = repr.find(' ', p0)) != repr.npos) {
	    steps.push_back(Step(repr.substr(p0, p1 - p0)));
	    p0 = p1 + 1;
	}
	steps.push_back(Step(repr.substr(p0)));
    } catch (Step::Invalid &e) {
	steps.clear();
	is_invalid = true;
	warnx("Invalid hotkey `%s'", repr.data());
    }
}

void Hotkey::clear()
{
    steps.clear();
    is_invalid = false;
}

bool Hotkey::isNull() const
{
    return !is_invalid && steps.empty();
}

bool Hotkey::isInvalid() const
{
    return is_invalid;
}

bool Hotkey::equals(const Hotkey &hotkey) const
{
    return !is_invalid && !hotkey.is_invalid && steps == hotkey.steps;
}

bool Hotkey::startsWith(const Hotkey &hotkey) const
{
    return !is_invalid && !hotkey.is_invalid &&
	   steps.size() >= hotkey.steps.size() &&
	   std::equal(hotkey.steps.begin(),
		      hotkey.steps.end(), steps.begin());
}

std::string Hotkey::format() const
{
    std::string result;

    for (std::vector<Step>::const_iterator i = steps.begin();
	 i != steps.end(); i++) {
	if (i != steps.begin())
	    result.push_back(' ');
	result.append((*i).format());
    }

    return result;
}

bool Toolkit::operator==(const Hotkey &h0, const Hotkey &h1)
{
    return h0.equals(h1);
}

bool Toolkit::operator!=(const Hotkey &h0, const Hotkey &h1)
{
    return !h0.equals(h1);
}
