/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <X11/Xatom.h>
#include "inputs.h"
#include "util/encoding.h"
#include "timer.h"
#include "selection.h"
#include "hooks.h"
#include "connection.h"

#define INPUT_PADDING_X 4
#define INPUT_PADDING_Y 4

#define AltMask Mod1Mask
#define NumLockMask Mod2Mask
#define AltGrMask Mod5Mask
#define IgnoreMask (LockMask | NumLockMask)

using namespace Toolkit;


class Input::CursorBlinkTimer : public Timer
{
    Input &input;
public:
    CursorBlinkTimer(Input &input) : input(input) {
    }
    void timeout() {
	input.blink();
    }
};

class Input::PasteConversion : public WideStringConversion
{
    Input &input;
public:
    PasteConversion(Input &input, Atom selection)
	: WideStringConversion(input.conn, input.hook(), selection),
	  input(input) {
    }
private:
    void notifySuccess(const std::wstring &str, Time time) {
	input.selectionNotify(str, time);
    }
};


bool is_word_character(wchar_t ch)
{
    return (ch >= '0' && ch <= '9') ||
	   (ch >= 'A' && ch <= 'Z') ||
	   (ch == '_') ||
	   (ch >= 'a' && ch <= 'z');
}


Input::Input(Connection &conn)
    : Widget(conn), scrolled(0), cursor_pos(0),
      selection_start(0), selection_length(0), is_clicked(false),
      last_event_time(0), multiple_click_time(0),
      multiple_click_count(0), multiple_click_pos(0),
      cursor_hidden(true), has_focus(false),
      timer(new CursorBlinkTimer(*this))
{
    SizeRequirements reqs;
    reqs.minWidth = INPUT_PADDING_X * 2 + 1;
    reqs.minHeight = conn.fontSet().ascent() +
		     conn.fontSet().descent() + INPUT_PADDING_Y * 2;
    reqs.maxHeight = reqs.minHeight;
    setSizeRequirements(reqs);

    setFocussables(std::vector<Focussable *>(1, this));
}

Input::~Input()
{
    delete timer;
}

std::string Input::text() const
{
    return assert_encode(_text);
}

void Input::setText(const std::string &text)
{
    _text = warn_decode(text);
    selectAll();
}

void Input::paint(const Region &region, X::Pixmap pixmap)
{
    int x = this->x() - region.x;
    int y = this->y() - region.y;

    int text_y = (height() + conn.fontSet().ascent()
			   - conn.fontSet().descent()) / 2;

    gc().setForeground(colors().inputBackground.pixel);
    pixmap.fillRectangle(gc(), x, y, width(), height());

    gc().setForeground(colors().inputText.pixel);
    pixmap.wcDrawString(conn.fontSet(), gc(),
			x + INPUT_PADDING_X - scrolled,
			y + text_y,
			_text.substr(0, selection_start));

    int pos = XwcTextEscapement(conn.fontSet(),
	const_cast<wchar_t *>(_text.c_str()), selection_start);

    if (has_focus) {
	gc().setForeground(colors().inputSelectionText.pixel);
	gc().setBackground(colors().inputSelectionBackground.pixel);
    } else {
	gc().setForeground(colors().inputSelectionUnfocussedText.pixel);
	gc().setBackground(colors().inputSelectionUnfocussedBackground.pixel);
    }
    pixmap.wcDrawImageString(conn.fontSet(), gc(),
			     x + INPUT_PADDING_X + pos - scrolled,
			     y + text_y,
			     _text.substr(selection_start, selection_length));

    int width = XwcTextEscapement(conn.fontSet(),
				  const_cast<wchar_t *>(_text.c_str())
				      + selection_start, selection_length);

    if (selection_length != 0) {
	if (has_focus)
	    gc().setForeground(colors().inputSelectionBackground.pixel);
	else
	    gc().setForeground(
		colors().inputSelectionUnfocussedBackground.pixel);

	pixmap.drawLine(gc(), x + INPUT_PADDING_X + pos - scrolled,
			      y + text_y - conn.fontSet().ascent() - 1,
			      x + INPUT_PADDING_X + pos + width - scrolled - 1,
			      y + text_y - conn.fontSet().ascent() - 1);
	pixmap.drawLine(gc(), x + INPUT_PADDING_X + pos - scrolled,
			      y + text_y + conn.fontSet().descent(),
			      x + INPUT_PADDING_X + pos + width - scrolled - 1,
			      y + text_y + conn.fontSet().descent());
    }

    gc().setForeground(colors().inputText.pixel);
    pixmap.wcDrawString(conn.fontSet(), gc(),
			x + INPUT_PADDING_X + pos + width - scrolled,
			y + text_y,
			_text.substr(selection_start + selection_length));

    if (!cursor_hidden) {
	gc().setForeground(colors().inputCursor.pixel);
	pixmap.fillRectangle(gc(),
	    x + INPUT_PADDING_X + XwcTextEscapement(conn.fontSet(),
		const_cast<wchar_t *>(_text.c_str()), cursor_pos) - scrolled,
	    y + text_y - conn.fontSet().ascent() - 1,
	    1, conn.fontSet().ascent() + conn.fontSet().descent() + 2);
    }
}

void Input::get_word_dimensions(unsigned int pos,
	unsigned int &start, unsigned int &length) const
{
    start = pos;
    while (start > 0 && is_word_character(_text[start - 1]))
	start--;
    while (pos < _text.length() && is_word_character(_text[pos]))
	pos++;
    length = pos - start;
}

void Input::motionEvent(Time time, int x, int y, int x_root, int y_root,
			unsigned int state, char is_hint)
{
    unsigned int pos, start, length;

    if (!is_clicked)
	return;

    pos = get_pos(x - this->x(), y - this->y());

    switch (multiple_click_count) {
    case 0:
	select_to(pos);
	break;
    case 1:
	get_word_dimensions(multiple_click_pos, start, length);
	if (pos < start) {
	    while (pos > 0 && pos < _text.length() &&
		   is_word_character(_text[pos - 1]) &&
		   is_word_character(_text[pos]))
		pos--;
	    length += start - pos;
	    start = pos;
	} else if (pos > start + length) {
	    while (pos > 0 && pos < _text.length() &&
		   is_word_character(_text[pos - 1]) &&
		   is_word_character(_text[pos]))
		pos++;
	    length = pos - start;
	} else
	    pos = start + length;
	setSelection(start, length, pos);
	break;
    }

    last_event_time = time;
}

void Input::buttonPressEvent(Time time, int x, int y, int x_root, int y_root,
			     unsigned int state, unsigned int button)
{
    unsigned int pos = get_pos(x - this->x(), y - this->y()), start, length;

    if (control()) {
	control()->focus(this);
	control()->enterEditor(time, x, y, x_root, y_root);
    }

    switch (button) {
    case Button1:
	if (time > multiple_click_time &&
	    time < multiple_click_time + 200 && multiple_click_pos == pos) {
	    switch (multiple_click_count++) {
	    case 0:
		get_word_dimensions(multiple_click_pos, start, length);
		setSelection(start, length, start + length, true);
		break;
	    case 1:
		selectAll();
		break;
	    }
	    reset_timer();
	    invalidate();
	} else {
	    setCursorPos(pos, true);
	    multiple_click_count = 0;
	    multiple_click_pos = pos;
	}

	is_clicked = true;
	last_event_time = time;
	multiple_click_time = time;
	break;
    case Button2:
	setCursorPos(pos, true);
	hook()->convertSelection(
	    (new PasteConversion(*this, XA_PRIMARY))->createHook(time));
	last_event_time = time;
	break;
    }
}

void Input::buttonReleaseEvent(Time time, int x, int y, int x_root, int y_root,
			       unsigned int state, unsigned int button)
{
    if (!is_clicked || button != Button1)
	return;

    last_event_time = time;
    is_clicked = false;

    if (selection_length > 0)
	hook()->select(
	    new WideStringSelectionHook(conn, XA_PRIMARY, time,
		_text.substr(selection_start, selection_length)));
}

void Input::replace_selection(const std::wstring &str)
{
    if (selection_length > 0) {
	_text = _text.substr(0, selection_start) + str +
		_text.substr(selection_start + selection_length);
	setCursorPos(selection_start + str.length());
    } else {
	_text = _text.substr(0, cursor_pos) + str +
		_text.substr(cursor_pos);
	setCursorPos(cursor_pos + str.length());
    }
}

void Input::key_motion(unsigned int pos, unsigned int state)
{
    if (pos > _text.length())
	return;

    switch (state & ShiftMask) {
    case 0:
	setCursorPos(pos, true);
	break;
    case ShiftMask:
	select_to(pos);
	break;
    }
}

void Input::keyPressEvent(Time time, int x, int y, int x_root, int y_root,
			  unsigned int state, unsigned int keycode,
			  const std::string &t, KeySym keysym)
{
    last_event_time = time;

    if (t.length() != 0 && !(t.length() == 1 && (t[0] < 32 || t[0] == 127))) {
	replace_selection(warn_decode(t));
	return;
    }

    switch (keysym) {
    case XK_BackSpace:
	switch (state & ~IgnoreMask) {
	case 0:
	    if (selection_length != 0)
		replace_selection(std::wstring());
	    else if (cursor_pos > 0) {
		_text = _text.substr(0, cursor_pos - 1) +
			_text.substr(cursor_pos);
		cursor_pos--;
		reset_timer();
		invalidate();
	    }
	    break;
	case AltMask:
	    /* backspace_word */
	    break;
	}
	break;
    case XK_Home:
    case XK_KP_Home:
	key_motion(0, state);
	break;
    case XK_Left:
    case XK_KP_Left:
	switch (state & ~IgnoreMask & ~ShiftMask) {
	case 0:
	    key_motion(cursor_pos - 1, state);
	    break;
	case ControlMask:
	    unsigned int pos = cursor_pos;
	    while (pos > 0 && !is_word_character(_text[pos - 1]))
		pos--;
	    while (pos > 0 && is_word_character(_text[pos - 1]))
		pos--;
	    key_motion(pos, state);
	    break;
	}
	break;
    case XK_Right:
    case XK_KP_Right:
	switch (state & ~IgnoreMask & ~ShiftMask) {
	case 0:
	    key_motion(cursor_pos + 1, state);
	    break;
	case ControlMask:
	    unsigned int pos = cursor_pos;
	    while (pos < _text.length() && !is_word_character(_text[pos]))
		pos++;
	    while (pos < _text.length() && is_word_character(_text[pos]))
		pos++;
	    key_motion(pos, state);
	    break;
	}
	break;
    case XK_KP_End:
    case XK_End:
	key_motion(_text.length(), state);
	break;
    case XK_Insert:
    case XK_KP_Insert:
	switch (state & ~IgnoreMask) {
	case ShiftMask:
	    hook()->convertSelection(
		(new PasteConversion(*this, conn.atoms().CLIPBOARD))
		    ->createHook(time));
	    break;
	case ControlMask:
	    if (selection_length > 0)
		hook()->select(
		    new WideStringSelectionHook(
			conn, conn.atoms().CLIPBOARD, time,
			_text.substr(selection_start, selection_length)));
	    break;
	}
	break;
    case XK_Delete:
    case XK_KP_Delete:
	switch (state & ~IgnoreMask) {
	case 0:
	    if (selection_length != 0)
		replace_selection(std::wstring());
	    else if (cursor_pos < _text.length()) {
		_text = _text.substr(0, cursor_pos) +
			_text.substr(cursor_pos + 1);
		reset_timer();
		invalidate();
	    }
	    break;
	case ShiftMask:
	    if (selection_length > 0)
		hook()->select(
		    new WideStringSelectionHook(
			conn, conn.atoms().CLIPBOARD, time,
			_text.substr(selection_start, selection_length)));
	    replace_selection(std::wstring());
	    break;
	case ControlMask:
	    replace_selection(std::wstring());
	    break;
	}
	break;
    case XK_a:
	switch (state & ~IgnoreMask) {
	case ControlMask:
	    selectAll();
	    break;
	}
	break;
    case XK_x:
	switch (state & ~IgnoreMask) {
	case ControlMask:
	    if (selection_length > 0)
		hook()->select(
		    new WideStringSelectionHook(
			conn, conn.atoms().CLIPBOARD, time,
			_text.substr(selection_start, selection_length)));
	    replace_selection(std::wstring());
	    break;
	}
	break;
    case XK_c:
	switch (state & ~IgnoreMask) {
	case ControlMask:
	    if (selection_length > 0)
		hook()->select(
		    new WideStringSelectionHook(
			conn, conn.atoms().CLIPBOARD, time,
			_text.substr(selection_start, selection_length)));
	    break;
	}
	break;
    case XK_v:
	switch (state & ~IgnoreMask) {
	case ControlMask:
	    hook()->convertSelection(
		(new PasteConversion(*this, conn.atoms().CLIPBOARD))
		    ->createHook(time));
	    break;
	}
	break;
    }
}

void Input::keyReleaseEvent(Time time, int x, int y, int x_root, int y_root,
			    unsigned int state, unsigned int keycode)
{
}

void Input::focusInEvent()
{
    has_focus = true;
    reset_timer();
    invalidate();
}

void Input::focusOutEvent()
{
    has_focus = false;
    cursor_hidden = true;
    timer->unset();
    invalidate();
}

void Input::selectionNotify(const std::wstring &str, Time time)
{
    if (time == last_event_time)
	replace_selection(str);
}

void Input::setCursorPos(unsigned int cursorPos, bool reset)
{
    setSelection(0, 0, cursorPos, reset);
}

void Input::setSelection(unsigned int selectionStart,
			 unsigned int selectionLength,
			 unsigned int cursorPos, bool reset)
{
    if (selectionStart > _text.length())
	selectionStart = _text.length();
    if (selectionStart + selectionLength > _text.length())
	selectionLength = _text.length() - selectionStart;
    if (selectionLength == 0)
	selectionStart = 0;
    if (cursorPos > _text.length())
	cursorPos = _text.length();

    if (selectionStart == selection_start &&
	selectionLength == selection_length &&
	cursorPos == cursor_pos && !reset)
	return;

    selection_start = selectionStart;
    selection_length = selectionLength;
    cursor_pos = cursorPos;
    /* ... scroll ... */

    reset_timer();
    invalidate();
}

void Input::selectAll()
{
    setSelection(0, _text.length(), _text.length(), true);
}

void Input::selectNone()
{
    setSelection(0, 0, cursor_pos, true);
}

unsigned int Input::get_pos(int x, int y) const
{
    int text_y = (height() + conn.fontSet().ascent()
			   - conn.fontSet().descent()) / 2;

    if (y < text_y - conn.fontSet().ascent())
	return 0;
    if (y > text_y + conn.fontSet().descent())
	return _text.length();

    int lb_index = 0;
    int lb_width = 0;
    int ub_index = _text.length();
    int ub_width = XwcTextEscapement(conn.fontSet(),
	const_cast<wchar_t *>(_text.c_str()), ub_index);

    x += scrolled - INPUT_PADDING_X;

    while (x >= lb_width + 1 && x < ub_width - 1 && ub_index - lb_index > 1) {
	int mb_index = (ub_index + lb_index) / 2;
	int mb_width = XwcTextEscapement(conn.fontSet(),
	    const_cast<wchar_t *>(_text.c_str()), mb_index);

	if (x >= mb_width) {
	    lb_index = mb_index;
	    lb_width = mb_width;
	} else {
	    ub_index = mb_index;
	    ub_width = mb_width;
	}
    }

    if (x < lb_width + 1)
	return lb_index;
    if (x >= ub_width - 1)
	return ub_index;

    if (x > (ub_width + lb_width) / 2)
	return ub_index;
    return lb_index;
}

void Input::select_to(unsigned int pos)
{
    if (pos == cursor_pos)
	return;

    if (selection_length == 0)
	selection_start = cursor_pos;

    if (cursor_pos == selection_start)
	if (pos >= selection_start + selection_length) {
	    selection_start += selection_length;
	    selection_length = pos - selection_start;
	} else {
	    selection_start = pos;
	    selection_length += cursor_pos - pos;
	}
    else if (cursor_pos == selection_start + selection_length)
	if (pos < selection_start) {
	    selection_length = selection_start - pos;
	    selection_start = pos;
	} else
	    selection_length += pos - cursor_pos;
    else
	if (pos < cursor_pos) {
	    selection_start = pos;
	    selection_length = cursor_pos - pos;
	} else {
	    selection_start = cursor_pos;
	    selection_length = pos - cursor_pos;
	}

    cursor_pos = pos;
    /* ... scroll ... */

    reset_timer();
    invalidate();
}

void Input::reset_timer()
{
    if (!has_focus)
	return;
    cursor_hidden = false;
    timer->set(0, 500000, true);
}

void Input::blink()
{
    cursor_hidden = !cursor_hidden;
    invalidate();
}
