/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <assert.h>
#include "menus.h"
#include "util/encoding.h"
#include "hooks.h"
#include "connection.h"
#include "global.h"
#include "icon.h"
#include "container1.h"

#define MENUITEM_PADDING_LEFT 24
#define MENUITEM_PADDING_RIGHT 9
#define MENUITEM_ICON_OFFSET 4
#define MENUITEM_PADDING_Y 4
#define MENUITEM_SPACE 20

using namespace Toolkit;


AbstractMenu::AbstractMenu(Connection &conn) : conn(conn)
{
}

AbstractMenu::~AbstractMenu()
{
    /* virtualize (not strictly necessary but will suppress some warnings) */
}


CommonMenu::CommonMenu(Connection &conn)
    : AbstractMenu(conn), highlighted_item(0), currently_open_submenu(0)
{
}

void CommonMenu::selectItem(Time time, int origin_x, int origin_y,
			    MenuItem *item)
{
    if (item == highlighted_item)
	return;

    deselectItem();

    highlighted_item = item;
    currently_open_submenu = item->action.menu();

    highlighted_item->select();
    if (currently_open_submenu && !currently_open_submenu->isVisible())
	currently_open_submenu->show(
	    time,
	    origin_x + item->x() + (isVertical() ? item->width() : 0),
	    origin_x + item->x() + (isVertical() ? 0 : item->width()), false,
	    origin_y + item->y() + (isVertical() ? 0 : item->height()),
	    origin_y + item->y() + (isVertical() ? item->height() : 0), false,
	    item->hook());
}

void CommonMenu::deselectItem()
{
    if (currently_open_submenu) {
	currently_open_submenu->close();
	currently_open_submenu = 0;
    }

    if (highlighted_item) {
	highlighted_item->deselect();
	highlighted_item = 0;
    }
}

void CommonMenu::close()
{
    deselectItem();
    hide();
}

bool CommonMenu::inChain(AbstractMenu *menu) const
{
    if (menu == this)
	return true;

    if (!currently_open_submenu)
	return false;

    return currently_open_submenu->inChain(menu);
}

/****************************************************************************/

MenuBar::MenuBar(Connection &conn) : CommonMenu(conn)
{
}

void MenuBar::hide()
{
    /* relax */
}

bool MenuBar::isVertical() const
{
    return false;
}

/****************************************************************************/

Menu::Menu(Connection &conn) : CommonMenu(conn), _widget(0), hook(0)
{
}

Menu::~Menu()
{
    hide();
}

Widget *Menu::widget() const
{
    return _widget;
}

void Menu::setWidget(Widget *widget)
{
    _widget = widget;
}

bool Menu::isVisible() const
{
    return !!hook;
}

void Menu::show(Time time, int x_root, int y_root,
		WidgetWindowHook *transient_for)
{
    show(time, x_root, x_root + 1, false,
	       y_root, y_root + 1, false, transient_for);
}

void Menu::show(Time time, int left, int right, bool prefer_right,
			   int top, int bottom, bool prefer_bottom,
		WidgetWindowHook *transient_for)
{
    if (hook) {
	hook->close();
	delete hook;
    }

    right -= widget()->sizeRequirements().minWidth + 2;
    bottom -= widget()->sizeRequirements().minHeight + 2;

    hook = create_transient_window(conn,
	(prefer_right + (right >= 0) +
	 (left + widget()->sizeRequirements().minWidth + 2
	      >= conn.screen().displayWidth()) >= 2) ? right : left,
	(prefer_bottom + (bottom >= 0) +
	 (top + widget()->sizeRequirements().minHeight + 2
	      >= conn.screen().displayHeight()) >= 2) ? bottom : top,
	1, 1, 1, conn.colors().menuBorder.pixel,
	widget(), transient_for, true);
}

void Menu::hide()
{
    if (!hook)
	return;

    hook->close();
    delete hook;
    hook = 0;
}

bool Menu::isVertical() const
{
    return true;
}

/****************************************************************************/

MenuItem::MenuItem(CommonMenu *parent, Action &action)
    : Widget(parent->conn), Action::Observer(action), parent(parent),
      text_width(0), hotkey_width(0), mnemonic_x0(0), mnemonic_x1(0),
      allow_icon(true), highlighted(false), mnemonic_visible(false)
{
    actionChanged();
}

bool MenuItem::allowIcon() const
{
    return allow_icon;
}

void MenuItem::setAllowIcon(bool allowIcon)
{
    allow_icon = allowIcon;
    update_size_requirements();
}

bool MenuItem::isMenu() const
{
    return true;
}

void MenuItem::setMnemonicVisibility(bool visible)
{
    if (visible == mnemonic_visible)
	return;

    mnemonic_visible = visible;
    invalidate();
}

void MenuItem::select()
{
    highlighted = true;
    invalidate();
}

void MenuItem::deselect()
{
    highlighted = false;
    invalidate();
}

void MenuItem::actionChanged()
{
    text_width = conn.fontSet().mbTextEscapement(action.text());
    hotkey_width = conn.fontSet().mbTextEscapement(action.hotkey().format());
    update_size_requirements();

    if (action.mnemonic() == -1)
	mnemonic_x0 = mnemonic_x1 = 0;
    else {
	std::wstring str = assert_decode(action.text());
	mnemonic_x0 = conn.fontSet().wcTextEscapement(
	    str.substr(0, action.mnemonic()));
	mnemonic_x1 = conn.fontSet().wcTextEscapement(
	    str.substr(0, action.mnemonic() + 1));
	assert(mnemonic_x1 >= mnemonic_x0);
    }

    setStatusMessage(action.disableMessage());
    invalidate();
}

void MenuItem::update_size_requirements()
{
    SizeRequirements reqs;
    reqs.minWidth = (allow_icon ? MENUITEM_PADDING_LEFT
				: MENUITEM_PADDING_RIGHT) +
	text_width + (allow_icon ? MENUITEM_SPACE : 0) +
	hotkey_width + (action.menu() ? 3 : 0) + MENUITEM_PADDING_RIGHT;
    reqs.minHeight = conn.fontSet().ascent() +
		     conn.fontSet().descent() + MENUITEM_PADDING_Y * 2;
    reqs.maxHeight = conn.fontSet().ascent() +
		     conn.fontSet().descent() + MENUITEM_PADDING_Y * 2;
    setSizeRequirements(reqs);
}

void MenuItem::paint(const Region &region, X::Pixmap pixmap)
{
    int x = this->x() - region.x;
    int y = this->y() - region.y;

    unsigned long background, text_pixel, mnemonic_pixel,
		  hotkey_pixel, arrow_pixel;

    if (action.isDisabled()) {
	background = colors().menuDisabled.pixel;
	text_pixel = colors().menuDisabledLabel.pixel;
	mnemonic_pixel = colors().menuDisabledMnemonic.pixel;
	hotkey_pixel = colors().menuDisabledHotkey.pixel;
	arrow_pixel = colors().menuDisabledArrow.pixel;
    } else if (highlighted) {
	background = colors().menuHighlight.pixel;
	text_pixel = colors().menuHighlightLabel.pixel;
	mnemonic_pixel = colors().menuHighlightMnemonic.pixel;
	hotkey_pixel = colors().menuHighlightHotkey.pixel;
	arrow_pixel = colors().menuHighlightArrow.pixel;
    } else {
	background = colors().menuBackground.pixel;
	text_pixel = colors().menuLabel.pixel;
	mnemonic_pixel = colors().menuMnemonic.pixel;
	hotkey_pixel = colors().menuHotkey.pixel;
	arrow_pixel = colors().menuArrow.pixel;
    }

    gc().setForeground(background);
    pixmap.fillRectangle(gc(), x, y, width(), height());

    int text_y = y + (height() + conn.fontSet().ascent()
			       - conn.fontSet().descent()) / 2;

    Icon *icon = action.icon();
    if (icon && allowIcon())
	X::Drawable::copyArea(icon->pixmap(conn, background,
					   action.isDisabled()),
			      pixmap, gc(), 0, 0,
			      icon->width(), icon->height(),
			      x + MENUITEM_PADDING_LEFT -
			      icon->width() - MENUITEM_ICON_OFFSET,
			      y + (height() - icon->height()) / 2);

    gc().setForeground(text_pixel);
    pixmap.mbDrawString(conn.fontSet(), gc(),
			x + (allowIcon() ? MENUITEM_PADDING_LEFT
					 : MENUITEM_PADDING_RIGHT), text_y,
			action.text());
    gc().setForeground(hotkey_pixel);
    pixmap.mbDrawString(conn.fontSet(), gc(),
			x + width() - MENUITEM_PADDING_RIGHT - hotkey_width,
			text_y, action.hotkey().format());

    if (mnemonic_visible && mnemonic_x1 != mnemonic_x0) {
	gc().setForeground(mnemonic_pixel);
	pixmap.drawLine(gc(),
	    x + (allowIcon() ? MENUITEM_PADDING_LEFT
			     : MENUITEM_PADDING_RIGHT) + mnemonic_x0,
	    text_y + conn.fontSet().descent() - 2,
	    x + (allowIcon() ? MENUITEM_PADDING_LEFT
			     : MENUITEM_PADDING_RIGHT) + mnemonic_x1 - 1,
	    text_y + conn.fontSet().descent() - 2);
    }

    if (allowIcon() && action.menu()) {
	int arrow_x = x + width() - MENUITEM_PADDING_RIGHT;
	int arrow_y = y + height() / 2;
	gc().setForeground(arrow_pixel);
	pixmap.drawLine(gc(), arrow_x, arrow_y, arrow_x, arrow_y);
	pixmap.drawLine(gc(), arrow_x - 1, arrow_y - 1,
			arrow_x - 1, arrow_y + 1);
	pixmap.drawLine(gc(), arrow_x - 2, arrow_y - 2,
			arrow_x - 2, arrow_y + 2);
    }
}

void MenuItem::enterEvent(Time time, int x_rel, int y_rel,
			  int x_root, int y_root,
			  unsigned int /*state*//*, char is_hint*/)
{
    if (!conn.rootMenu() || !conn.rootMenu()->inChain(parent))
	return;

    parent->selectItem(time, x_root - x_rel, y_root - y_rel, this);
}

void MenuItem::leaveEvent(Time /*time*/, int /*x*/, int /*y*/,
			  int /*x_root*/, int /*y_root*/,
			  unsigned int /*state*//*, char*/ /*is_hint*/)
{
    if (!conn.rootMenu() || !conn.rootMenu()->inChain(parent))
	return;

    if (action.menu() && action.menu()->isVisible())
	return;

    parent->deselectItem();
}

void MenuItem::buttonPressEvent(Time time, int x, int y,
				int x_root, int y_root,
				unsigned int state, unsigned int button)
{
    if (action.isDisabled())
	return;

    if (button == Button4 || button == Button5)
	return;

    if (!action.menu()) {
	conn.leaveMenuMode(CurrentTime);
	action.trigger();
	return;
    }

    if (conn.rootMenu() && conn.rootMenu()->inChain(parent))
	return;

    conn.enterMenuMode(time, parent);
    parent->selectItem(time, x_root - x, y_root - y, this);
}

void MenuItem::buttonReleaseEvent(Time time, int x, int y,
				  int x_root, int y_root,
				  unsigned int state, unsigned int button)
{
    if (action.isDisabled() || action.menu())
	return;

    if (button == Button4 || button == Button5)
	return;

    conn.leaveMenuMode(time);
    action.trigger();
}

/****************************************************************************/

MenuSeparator::MenuSeparator(Connection &conn) : Widget(conn)
{
    SizeRequirements reqs;
    reqs.minHeight = 7;
    reqs.maxHeight = 7;
    setSizeRequirements(reqs);
}

bool MenuSeparator::isMenu() const
{
    return true;
}

void MenuSeparator::paint(const Region &region, X::Pixmap pixmap)
{
    int x = this->x() - region.x;
    int y = this->y() - region.y;

    gc().setForeground(colors().menuSeparator.pixel);
    pixmap.drawLine(gc(), x + 7, y + 3, x + width() - 8, y + 3);
}

/****************************************************************************/

MenuBuilder::MenuBuilder(Connection &conn) : container(new VContainer(conn)),
					     _menu(new Menu(conn))
{
    container->setBackground(conn.colors().menuBackground.pixel);
    _menu->setWidget(container);
}

MenuBuilder::~MenuBuilder()
{
    delete _menu;
    delete container;
    for (std::vector<Widget *>::const_iterator i = items.begin();
	 i != items.end(); i++)
	delete *i;
}

Menu *MenuBuilder::menu() const
{
    return _menu;
}

void MenuBuilder::addAction(Action &action)
{
    MenuItem *item = new MenuItem(_menu, action);
    items.push_back(item);
    container->addWidget(item);
}

void MenuBuilder::insertAction(unsigned int pos, Action &action)
{
    MenuItem *item = new MenuItem(_menu, action);
    items.insert(items.begin() + pos, item);
    container->insertWidget(pos, item);
}

void MenuBuilder::addSeparator()
{
    MenuSeparator *separator = new MenuSeparator(_menu->conn);
    items.push_back(separator);
    container->addWidget(separator);
}

void MenuBuilder::removeWidgetAt(unsigned int pos)
{
    Widget *w = items.at(pos);
    container->removeWidget(w);
    items.erase(items.begin() + pos);
    delete w;
}

void MenuBuilder::clear()
{
    for (std::vector<Widget *>::const_iterator i = items.begin();
	 i != items.end(); i++) {
	container->removeWidget(*i);
	delete *i;
    }
    items.clear();
}

/****************************************************************************/

MenuBarBuilder::MenuBarBuilder(Connection &conn)
    : menu_bar(new MenuBar(conn)), container(new HContainer(conn))
{
    container->setBackground(conn.colors().menuBackground.pixel);
    container->setExpand(false);
}

MenuBarBuilder::~MenuBarBuilder()
{
    for (std::vector<MenuItem *>::const_iterator i = items.begin();
	 i != items.end(); i++)
	delete *i;
    for (std::vector<Action *>::const_iterator i = actions.begin();
	 i != actions.end(); i++)
	delete *i;
    delete container;
    delete menu_bar;
}

Widget *MenuBarBuilder::menuBar() const
{
    return container;
}

void MenuBarBuilder::addMenu(const std::string &text, Menu *menu)
{
    Action *action = new Action(action_set);
    action->setText(text);
    action->setMenu(menu);
    actions.push_back(action);

    MenuItem *item = new MenuItem(menu_bar, *action);
    item->setAllowIcon(false);
    items.push_back(item);
    container->addWidget(item);
}
