/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#ifndef TOOLKIT_MENUS_H
#define TOOLKIT_MENUS_H

#include "widget.h"
#include "action.h"

namespace Toolkit {

class MenuItem;
class Container;

class AbstractMenu
{
public:
    AbstractMenu(Connection &conn);
    virtual ~AbstractMenu();
    Connection &conn;
    virtual void close() = 0;
    virtual bool inChain(AbstractMenu *menu) const = 0;
};

class CommonMenu : public AbstractMenu
{
    MenuItem *highlighted_item;
    Menu *currently_open_submenu;

    virtual void hide() = 0;
    virtual bool isVertical() const = 0;
public:
    CommonMenu(Connection &conn);
    void selectItem(Time time, int origin_x, int origin_y, MenuItem *item);
    void deselectItem();
    void close();
    bool inChain(AbstractMenu *menu) const;
};

class MenuBar : public CommonMenu
{
    void hide();
    bool isVertical() const;
public:
    MenuBar(Connection &conn);
};

class Menu : public CommonMenu
{
    Widget *_widget;
    WidgetWindowHook *hook;
    void hide();
    bool isVertical() const;
public:
    Menu(Connection &conn);
    ~Menu();
    Widget *widget() const;
    void setWidget(Widget *widget);

    bool isVisible() const;
    void show(Time time, int x_root, int y_root,
	      WidgetWindowHook *transient_for);
    void show(Time time, int left, int right, bool prefer_right,
			 int top, int bottom, bool prefer_bottom,
	      WidgetWindowHook *transient_for);
};

class MenuItem : public Widget, public Action::Observer
{
    CommonMenu *const parent;
    int text_width, hotkey_width;
    int mnemonic_x0, mnemonic_x1;
    bool allow_icon;
    bool highlighted, mnemonic_visible;
    void update_size_requirements();
public:
    MenuItem(CommonMenu *parent, Action &action);
    bool allowIcon() const;
    void setAllowIcon(bool allowIcon);

    bool isMenu() const;
    void setMnemonicVisibility(bool visible);

    void select();
    void deselect();

    void actionChanged();

    void paint(const Region &region, X::Pixmap pixmap);
    void enterEvent(Time time, int x, int y, int x_root, int y_root,
		    unsigned int state/*, char is_hint*/);
    void leaveEvent(Time time, int x, int y, int x_root, int y_root,
		    unsigned int state/*, char is_hint*/);
    void buttonPressEvent(Time time, int x, int y,
			  int x_root, int y_root,
			  unsigned int state, unsigned int button);
    void buttonReleaseEvent(Time time, int x, int y,
			    int x_root, int y_root,
			    unsigned int state, unsigned int button);
};

class MenuSeparator : public Widget
{
public:
    MenuSeparator(Connection &conn);
    bool isMenu() const;
    void paint(const Region &region, X::Pixmap pixmap);
};

class MenuBuilder
{
    Container *const container;
    Menu *const _menu;
    std::vector<Widget *> items;
public:
    MenuBuilder(Connection &conn);
    ~MenuBuilder();
    Menu *menu() const;
    void addAction(Action &action);
    void insertAction(unsigned int pos, Action &action);
    void addSeparator();
    void removeWidgetAt(unsigned int pos);
    void clear();
};

class MenuBarBuilder
{
    MenuBar *const menu_bar;
    Container *const container;
    std::vector<MenuItem *> items;
    std::vector<Action *> actions;
    ActionSet action_set;
public:
    MenuBarBuilder(Connection &conn);
    ~MenuBarBuilder();
    Widget *menuBar() const;
    void addMenu(const std::string &text, Menu *menu);
};

} /* namespace Toolkit */

#endif
