/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <assert.h>
#include <algorithm>
#include "model_view.h"
#include "util/encoding.h"


Item::Item() : _interior(0), _icon(0), is_bold(false), is_italic(false),
	       _editor(0)
{
}

Item::~Item()
{
    assert(shadows.empty());
}

ItemInterior *Item::interior() const
{
    return _interior;
}

void Item::setInterior(ItemInterior *interior)
{
    if (interior == _interior)
	return;

    _interior = interior;

    std::vector<Shadow *> cached_shadows = shadows;
    for (std::vector<Shadow *>::const_iterator i = cached_shadows.begin();
	 i != cached_shadows.end(); i++)
	(*i)->update_interior();
}

std::wstring Item::caption() const
{
    return _caption;
}

void Item::setCaption(const std::string &caption)
{
    std::wstring str = warn_decode(caption);

    if (str == _caption)
	return;

    _caption = str;

    std::vector<Shadow *> cached_shadows = shadows;
    for (std::vector<Shadow *>::const_iterator i = cached_shadows.begin();
	 i != cached_shadows.end(); i++)
	(*i)->updateItem();
}

std::wstring Item::hotkey() const
{
    return _hotkey;
}

void Item::setHotkey(const std::string &hotkey)
{
    std::wstring str = warn_decode(hotkey);

    if (str == _hotkey)
	return;

    _hotkey = str;

    std::vector<Shadow *> cached_shadows = shadows;
    for (std::vector<Shadow *>::const_iterator i = cached_shadows.begin();
	 i != cached_shadows.end(); i++)
	(*i)->updateItem();
}

Icon *Item::icon() const
{
    return _icon;
}

void Item::setIcon(Icon *icon)
{
    if (icon == _icon)
	return;

    _icon = icon;

    std::vector<Shadow *> cached_shadows = shadows;
    for (std::vector<Shadow *>::const_iterator i = cached_shadows.begin();
	 i != cached_shadows.end(); i++)
	(*i)->updateItem();
}

bool Item::isBold() const
{
    return is_bold;
}

void Item::setBold(bool bold)
{
    if (bold == is_bold)
	return;

    is_bold = bold;

    std::vector<Shadow *> cached_shadows = shadows;
    for (std::vector<Shadow *>::const_iterator i = cached_shadows.begin();
	 i != cached_shadows.end(); i++)
	(*i)->updateItem();
}

bool Item::isItalic() const
{
    return is_italic;
}

void Item::setItalic(bool italic)
{
    if (italic == is_italic)
	return;

    is_italic = italic;

    std::vector<Shadow *> cached_shadows = shadows;
    for (std::vector<Shadow *>::const_iterator i = cached_shadows.begin();
	 i != cached_shadows.end(); i++)
	(*i)->updateItem();
}

ItemEditor *Item::editor() const
{
    return _editor;
}

void Item::setEditor(ItemEditor *editor)
{
    if (editor == _editor)
	return;

    _editor = editor;

    std::vector<Shadow *> cached_shadows = shadows;
    for (std::vector<Shadow *>::const_iterator i = cached_shadows.begin();
	 i != cached_shadows.end(); i++)
	(*i)->updateItem();
}

void Item::add_shadow(Shadow *shadow)
{
    assert(std::find(shadows.begin(), shadows.end(), shadow) == shadows.end());
    shadows.push_back(shadow);
}

void Item::remove_shadow(Shadow *shadow)
{
    std::vector<Shadow *>::iterator i =
	std::find(shadows.begin(), shadows.end(), shadow);
    assert(i != shadows.end());
    shadows.erase(i);
}

/****************************************************************************/

ItemInterior::~ItemInterior()
{
    assert(shadows.empty());
}

Item *ItemInterior::cachedItem(unsigned int index) const
{
    assert(index < item_cache.size());
    Item *&item = item_cache[index];

    if (!item) {
	item = this->item(index);
	assert(item);
    }
    return item;
}

unsigned int ItemInterior::cachedIndexOf(Item *item) const
{
    if (!item)
	return NPOS;

    std::vector<Item *>::const_iterator i = std::find(item_cache.begin(),
						      item_cache.end(), item);
    if (i != item_cache.end())
	return i - item_cache.begin();

    unsigned int index = indexOf(item);
    if (index != NPOS)
	assert(cachedItem(index) == item);
    return index;
}

unsigned int ItemInterior::count() const
{
    return item_cache.size();
}

void ItemInterior::replace(unsigned int index, unsigned int old_count,
					       unsigned int new_count)
{
    assert(item_cache.size() >= index + old_count);
    if (old_count == 0 && new_count == 0)
	return;

    std::vector<Item *> zero(new_count, 0);
    item_cache.insert(item_cache.erase(item_cache.begin() + index,
				       item_cache.begin() + index + old_count),
		      zero.begin(), zero.end());

    std::vector<Shadow *> cached_shadows = shadows;
    for (std::vector<Shadow *>::const_iterator i = cached_shadows.begin();
	 i != cached_shadows.end(); i++)
	(*i)->update_items(index, old_count, new_count);
}

void ItemInterior::add_shadow(Shadow *shadow)
{
    assert(std::find(shadows.begin(), shadows.end(), shadow) == shadows.end());
    shadows.push_back(shadow);
}

void ItemInterior::remove_shadow(Shadow *shadow)
{
    std::vector<Shadow *>::iterator i =
	std::find(shadows.begin(), shadows.end(), shadow);
    assert(i != shadows.end());
    shadows.erase(i);
}

/****************************************************************************/

Shadow::Shadow(ItemInterior *interior) : _item(0), _parent(0), _interior(0)
{
    set_interior(interior);
}

Shadow::Shadow(Item *item, Shadow *parent)
    : _item(item), _parent(parent), _interior(0)
{
    assert(parent);
    _item->add_shadow(this);
    set_interior(_item->interior());
}

Shadow::~Shadow()
{
    if (_item)
	_item->remove_shadow(this);

    set_interior(0);
}

void Shadow::set_interior(ItemInterior *interior)
{
    if (interior == _interior)
	return;

    if (_interior)
	_interior->remove_shadow(this);
    for (std::vector<Shadow *>::const_iterator i = shadow_cache.begin();
	 i != shadow_cache.end(); i++)
	delete *i;

    _interior = interior;

    if (_interior) {
	shadow_cache = std::vector<Shadow *>(interior->count(), 0);
	_interior->add_shadow(this);
    } else
	shadow_cache.clear();
}

void Shadow::update_interior()
{
    if (item()->interior() == interior())
	return;

    int old_count = interior() ? interior()->count() : 0;
    int new_count = item()->interior() ? item()->interior()->count() : 0;

    set_interior(item()->interior());
    updateItems(0, old_count, new_count);
}

void Shadow::update_items(unsigned int index, unsigned int old_count,
					      unsigned int new_count)
{
    std::vector<Shadow *>::iterator begin = shadow_cache.begin() + index;
    std::vector<Shadow *>::iterator end = begin + old_count;
    for (std::vector<Shadow *>::const_iterator i = begin; i != end; i++)
	delete *i;

    std::vector<Shadow *> zero(new_count, 0);
    shadow_cache.insert(shadow_cache.erase(begin, end),
			zero.begin(), zero.end());

    updateItems(index, old_count, new_count);
}

Item *Shadow::item() const
{
    return _item;
}

Shadow *Shadow::parent() const
{
    return _parent;
}

ItemInterior *Shadow::interior() const
{
    return _interior;
}

unsigned int Shadow::indexOf(Shadow *shadow) const
{
    std::vector<Shadow *>::const_iterator i =
	std::find(shadow_cache.begin(), shadow_cache.end(), shadow);
    return i == shadow_cache.end() ? NPOS : i - shadow_cache.begin();
}

Shadow *Shadow::shadow(unsigned int index) const
{
    assert(index < shadow_cache.size());
    Shadow *&shadow = shadow_cache[index];

    if (!shadow)
	shadow = buildShadow(interior()->cachedItem(index));
    return shadow;
}

unsigned int Shadow::count() const
{
    return shadow_cache.size();
}

Shadow *Shadow::lookup(const std::vector<Item *> &items)
{
    if (items.empty())
	return this;
    if (!interior())
	return 0;
    unsigned int index = interior()->cachedIndexOf(items[0]);
    if (index == NPOS)
	return 0;
    return shadow(index)->lookup(
	std::vector<Item *>(items.begin() + 1, items.end()));
}
