/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "navigator.h"
#include "global.h"
#include "hooks.h"
#include "connection.h"

#define NAVIGATOR_SIZE 16

using namespace Toolkit;


NavigatorData::NavigatorData()
    : view_x(0), view_y(0), view_width(0), view_height(0),
      content_width(0), content_height(0)
{
}

NavigatorData::NavigatorData(int view_x, int view_y,
			     int view_width, int view_height,
			     int content_width, int content_height,
			     X::Pixmap pixmap)
    : view_x(view_x), view_y(view_y),
      view_width(view_width), view_height(view_height),
      content_width(content_width), content_height(content_height),
      _pixmap(pixmap)
{
    if (view_width < 1)
	view_width = 1;
    if (view_height < 1)
	view_height = 1;

    if (content_width < view_width)
	content_width = view_width;
    if (content_height < view_height)
	content_height = view_height;

    apply_constraints();
}

void NavigatorData::apply_constraints()
{
    if (view_x < 0)
	view_x = 0;
    if (view_x > content_width - view_width)
	view_x = content_width - view_width;
    if (view_y < 0)
	view_y = 0;
    if (view_y > content_height - view_height)
	view_y = content_height - view_height;
}

int NavigatorData::viewX() const
{
    return view_x;
}

int NavigatorData::viewY() const
{
    return view_y;
}

int NavigatorData::viewWidth() const
{
    return view_width;
}

int NavigatorData::viewHeight() const
{
    return view_height;
}

int NavigatorData::contentWidth() const
{
    return content_width;
}

int NavigatorData::contentHeight() const
{
    return content_height;
}

X::Pixmap NavigatorData::pixmap() const
{
    return _pixmap;
}

void NavigatorData::setPosition(int x, int y)
{
    view_x = x;
    view_y = y;
    apply_constraints();
}

/****************************************************************************/

class Navigator::Operation
{
    class Pane : public Widget
    {
	NavigatorData const &data;
    public:
	Pane(Connection &conn, NavigatorData const &data)
	      : Widget(conn), data(data) {
	    SizeRequirements reqs;
	    reqs.minWidth = data.contentWidth();
	    reqs.minHeight = data.contentHeight();
	    setSizeRequirements(reqs);
	}
	void paint(const Region &region, X::Pixmap pixmap) {
	    int x = this->x() - region.x;
	    int y = this->y() - region.y;

	    X::Drawable::copyArea(data.pixmap(), pixmap, gc(), 0, 0,
				  data.contentWidth(),
				  data.contentHeight(), x, y);

	    gc().setFunction(GXinvert);
	    pixmap.drawRectangle(gc(), x + data.viewX(),
				       y + data.viewY(),
				       data.viewWidth() - 1,
				       data.viewHeight() - 1);
	    gc().setFunction(GXcopy);
	}
    };

    Navigator &navigator;
    NavigatorData data;
    Pane *const pane;
    int const origin_x, origin_y;
    WidgetWindowHook *const hook;
public:
    Operation(Navigator &navigator, NavigatorData const &data,
	      int x_root, int y_root, unsigned long border_pixel,
	      WidgetWindowHook *transient_for)
	: navigator(navigator), data(data),
	  pane(new Pane(navigator.conn, this->data)),
	  origin_x(x_root - data.viewX()),
	  origin_y(y_root - data.viewY()),
	  hook(create_transient_window(
		   navigator.conn, origin_x - data.viewWidth() / 2 - 1,
				   origin_y - data.viewHeight() / 2 - 1, 0, 0,
		   1, border_pixel, pane, transient_for, false)) {
    }
    ~Operation() {
	hook->close();
	delete hook;

	data.pixmap().free();
	delete pane;
    }
    void motionEvent(int x_root, int y_root) {
	data.setPosition(x_root - origin_x, y_root - origin_y);
	navigator.navigateTo(data.viewX(), data.viewY());
	pane->invalidate();
    }
};

/****************************************************************************/

Navigator::Navigator(Connection &conn) : Widget(conn), op(0)
{
    SizeRequirements reqs;
    reqs.minWidth = NAVIGATOR_SIZE;
    reqs.minHeight = NAVIGATOR_SIZE;
    setSizeRequirements(reqs);
}

Navigator::~Navigator()
{
    delete op;
}

void Navigator::paint(const Region &region, X::Pixmap pixmap)
{
    int x = this->x() - region.x;
    int y = this->y() - region.y;

    gc().setForeground(colors().navigatorBackground.pixel);
    pixmap.fillRectangle(gc(), x, y, width(), height());

    x += width() / 2;
    y += height() / 2;

    gc().setForeground(colors().navigatorSymbol.pixel);
    pixmap.drawLine(gc(), x - 1, y - 7, x - 1, y + 6);
    pixmap.drawLine(gc(), x,     y - 7, x,     y + 6);
    pixmap.drawLine(gc(), x - 2, y - 6, x + 1, y - 6);
    pixmap.drawLine(gc(), x - 3, y - 5, x + 2, y - 5);
    pixmap.drawLine(gc(), x - 3, y + 4, x + 2, y + 4);
    pixmap.drawLine(gc(), x - 2, y + 5, x + 1, y + 5);

    pixmap.drawLine(gc(), x - 7, y - 1, x + 6, y - 1);
    pixmap.drawLine(gc(), x - 7, y,     x + 6, y);
    pixmap.drawLine(gc(), x - 6, y - 2, x - 6, y + 1);
    pixmap.drawLine(gc(), x - 5, y - 3, x - 5, y + 2);
    pixmap.drawLine(gc(), x + 4, y - 3, x + 4, y + 2);
    pixmap.drawLine(gc(), x + 5, y - 2, x + 5, y + 1);
}

void Navigator::motionEvent(Time time, int x, int y, int x_root, int y_root,
			    unsigned int state, char is_hint)
{
    if (op)
	op->motionEvent(x_root, y_root);
}

void Navigator::buttonPressEvent(Time time, int x, int y,
				 int x_root, int y_root,
				 unsigned int state, unsigned int button)
{
    if (op)
	return;

    NavigatorData data = getNavigatorData();
    if (data.viewWidth() >= data.contentWidth() ||
	data.viewHeight() >= data.contentHeight())
	return;

    op = new Operation(*this, data, x_root, y_root,
		       colors().navigatorBorder.pixel, hook());
}

void Navigator::buttonReleaseEvent(Time time, int x, int y,
				   int x_root, int y_root,
				   unsigned int state, unsigned int button)
{
    delete op;
    op = 0;
}
