/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <err.h>
#include <X11/Xatom.h>
#include "selection.h"
#include "hooks.h"
#include "connection.h"

using namespace Toolkit;


SelectionHook::SelectionHook(Atom selection, Time time)
    : selection(selection), time(time)
{
}

SelectionHook::~SelectionHook()
{
    /* virtualize */
}

bool SelectionHook::request(
    X::Window requestor, Atom target, Atom property, Time time)
{
    return false;
}

void SelectionHook::clear(Time time)
{
}

/****************************************************************************/

ConversionHook::ConversionHook(Atom selection, Atom target, Time time)
    : selection(selection), target(target), time(time)
{
}

ConversionHook::~ConversionHook()
{
    /* virtualize */
}

void ConversionHook::notify(Atom property, Time time)
{
}

void ConversionHook::cancel(Time time)
{
}

/****************************************************************************/

class ComplexConversion::Hook : public ConversionHook
{
    ComplexConversion *const conversion;
public:
    Hook(Atom selection, Atom target, Time time, ComplexConversion *conversion)
	: ConversionHook(selection, target, time), conversion(conversion) {
    }

    void notify(Atom property, Time time) {
	if (property != None) {
	    conversion->notify(property, time);
	    delete conversion;
	    return;
	}

	ConversionHook *hook = conversion->createHook(time);

	if (hook != NULL) {
	    conversion->window()->convertSelection(hook);
	    return;
	}

	conversion->notify(None, time);
	delete conversion;
    }
    void cancel(Time time) {
	conversion->cancel(time);
	delete conversion;
    }
};

/****************************************************************************/

ComplexConversion::ComplexConversion(
    SelectionWindowHook *window, Atom selection,
    const std::vector<Atom> &targets)
    : _window(window), _selection(selection), targets(targets), target_index(0)
{
}

ComplexConversion::~ComplexConversion()
{
    /* virtualize */
}

ConversionHook *ComplexConversion::createHook(Time time)
{
    if (target_index >= targets.size())
	return NULL;

    return new Hook(selection(), targets[target_index++], time, this);
}

void ComplexConversion::notify(Atom property, Time time)
{
    /* chill */
}

void ComplexConversion::cancel(Time time)
{
    /* chill */
}

SelectionWindowHook *ComplexConversion::window() const
{
    return _window;
}

Atom ComplexConversion::selection() const
{
    return _selection;
}

Atom ComplexConversion::target() const
{
    if (target_index == 0)
	return None;

    return targets[target_index - 1];
}

/****************************************************************************/

StringSelectionHook::StringSelectionHook(
    Connection &conn, Atom selection, Time time, const std::string &str)
    : SelectionHook(selection, time), conn(conn), str(str)
{
}

bool StringSelectionHook::request(X::Window requestor, Atom target,
				  Atom property, Time time)
{
    char *list = const_cast<char *>(str.c_str());
    XTextProperty text_property;
    XICCEncodingStyle style;

    if (target == conn.atoms().UTF8_STRING)
	style = XUTF8StringStyle;
    else if (target == conn.atoms().TEXT)
	style = XTextStyle;
    else if (target == conn.atoms().COMPOUND_TEXT)
	style = XCompoundTextStyle;
    else if (target == XA_STRING)
	style = XStringStyle;
    else
	return false;

    if (conn.display().utf8TextListToTextProperty(
	    &list, 1, style, &text_property) != 0) {
	warnx("Xutf8TextListToTextProperty: Could not convert "
	      "'%s' to %ld ('%s')", str.c_str(), target,
	      conn.display().getAtomName(target));
	return false;
    }

    requestor.setTextProperty(&text_property, property);
    XFree(text_property.value);
    return true;
}

/****************************************************************************/

static std::vector<Atom> string_targets(Connection &conn)
{
    std::vector<Atom> v;
    v.push_back(conn.atoms().UTF8_STRING);
    v.push_back(conn.atoms().TEXT);
    v.push_back(conn.atoms().COMPOUND_TEXT);
    v.push_back(XA_STRING);
    return v;
}

StringConversion::StringConversion(
    Connection &conn, SelectionWindowHook *window, Atom selection)
    : ComplexConversion(window, selection, string_targets(conn)), conn(conn)
{
}

void StringConversion::notify(Atom property, Time time)
{
    XTextProperty text_prop;
    char **list_return;
    int count_return;

    if (property == None) {
	notifyFailure(time);
	return;
    }

    if (window()->window().getTextProperty(&text_prop, property) == 0) {
	warnx("XGetTextProperty: Unable to read and store selection data");
	notifyFailure(time);
    } else {
	if (conn.display().utf8TextPropertyToTextList(
		&text_prop, &list_return, &count_return) != Success) {
	    warnx("Xutf8TextPropertyToTextList: Cannot fully convert "
		  "selection data to the encoding of the current locale");
	    notifyFailure(time);
	} else if (count_return != 1) {
	    warnx("Selection contains %d rather than one string",
		  count_return);
	    notifyFailure(time);
	} else
	    notifySuccess(list_return[0], time);

	XFreeStringList(list_return);
	XFree(text_prop.value);
    }

    window()->window().deleteProperty(property);
}

void StringConversion::notifySuccess(const std::string &str, Time time)
{
    /* chill */
}

void StringConversion::notifyFailure(Time time)
{
    /* chill */
}

/****************************************************************************/

WideStringSelectionHook::WideStringSelectionHook(
    Connection &conn, Atom selection, Time time, const std::wstring &str)
    : SelectionHook(selection, time), conn(conn), str(str)
{
}

bool WideStringSelectionHook::request(X::Window requestor, Atom target,
				      Atom property, Time time)
{
    wchar_t *list = const_cast<wchar_t *>(str.c_str());
    XTextProperty text_property;
    XICCEncodingStyle style;

    if (target == conn.atoms().UTF8_STRING)
	style = XUTF8StringStyle;
    else if (target == conn.atoms().TEXT)
	style = XTextStyle;
    else if (target == conn.atoms().COMPOUND_TEXT)
	style = XCompoundTextStyle;
    else if (target == XA_STRING)
	style = XStringStyle;
    else
	return false;

    if (conn.display().wcTextListToTextProperty(
	    &list, 1, style, &text_property) != 0) {
	warnx("XwcTextListToTextProperty: Could not convert to %ld ('%s')",
	      target, conn.display().getAtomName(target));
	return false;
    }

    requestor.setTextProperty(&text_property, property);
    XFree(text_property.value);
    return true;
}

/****************************************************************************/

WideStringConversion::WideStringConversion(
    Connection &conn, SelectionWindowHook *window, Atom selection)
    : ComplexConversion(window, selection, string_targets(conn)), conn(conn)
{
}

void WideStringConversion::notify(Atom property, Time time)
{
    XTextProperty text_prop;
    wchar_t **list_return;
    int count_return;

    if (property == None) {
	notifyFailure(time);
	return;
    }

    if (window()->window().getTextProperty(&text_prop, property) == 0) {
	warnx("XGetTextProperty: Unable to read and store selection data");
	notifyFailure(time);
    } else {
	if (conn.display().wcTextPropertyToTextList(
		&text_prop, &list_return, &count_return) != Success) {
	    warnx("XwcTextPropertyToTextList: Cannot fully convert "
		  "selection data to the encoding of the current locale");
	    notifyFailure(time);
	} else if (count_return != 1) {
	    warnx("Selection contains %d rather than one string",
		  count_return);
	    notifyFailure(time);
	} else
	    notifySuccess(list_return[0], time);

	XwcFreeStringList(list_return);
	XFree(text_prop.value);
    }

    window()->window().deleteProperty(property);
}

void WideStringConversion::notifySuccess(const std::wstring &str, Time time)
{
    /* chill */
}

void WideStringConversion::notifyFailure(Time time)
{
    /* chill */
}
