/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "selector.h"
#include "menus.h"
#include "connection.h"
#include "hooks.h"
#include "global.h"
#include "views/dropdown.h"
#include "scrolling.h"
#include "icon.h"

#define SELECTOR_PADDING_X 4
#define SELECTOR_PADDING_Y 4

using namespace Toolkit;


class Toolkit::SelectorMenu : public AbstractMenu
{
    Widget *const widget;
    Selector &selector;
    WidgetWindowHook *hook;
public:
    SelectorMenu(Widget *widget, Selector &selector)
	: AbstractMenu(selector.conn),
	  widget(widget), selector(selector), hook(0) {
    }
    ~SelectorMenu() {
	if (conn.rootMenu() == this)
	    conn.leaveMenuMode(CurrentTime);
	/* no need to manually delete the hook
	   since this calls close() for us */
    }

    void open(Time time, int x_root, int y_root,
	      int width, int height, WidgetWindowHook *transient_for) {
	if (hook) {
	    hook->close();
	    delete hook;
	}

	hook = create_transient_window(conn, x_root, y_root, width, height,
				       1, selector.colors().menuBorder.pixel,
				       widget, transient_for, true);
    }
    void close() {
	if (hook) {
	    hook->close();
	    delete hook;
	}

	hook = 0;
	if (selector.control())
	    selector.control()->leaveEditor(false);
    }
    bool inChain(AbstractMenu *menu) const {
	return menu == this;
    }
    bool isOpen() const {
	return !!hook;
    }
};

/****************************************************************************/

Selector::Selector(ItemInterior *choices, Connection &conn)
    : Widget(conn), selected_item(0),
      view(new DropdownView(*this)), pane(new ScrollPane(conn)),
      menu(new SelectorMenu(pane, *this)), has_focus(false)
{
    view->setRootInterior(choices);

    pane->mainControl()->setWidget(view);
    pane->setBackground(colors().background.pixel);

    SizeRequirements reqs;
    reqs.minWidth = SELECTOR_PADDING_X * 2;
    reqs.minHeight = conn.fontSet().ascent() +
		     conn.fontSet().descent() + SELECTOR_PADDING_Y * 2;
    reqs.maxHeight = reqs.minHeight;
    setSizeRequirements(reqs);

    setFocussables(std::vector<Focussable *>(1, this));
}

Selector::~Selector()
{
    delete menu;
    delete pane;
    delete view;
}

Item *Selector::selectedItem() const
{
    return selected_item;
}

void Selector::setSelectedItem(Item *item)
{
    selected_item = item;
    invalidate();
}

void Selector::openMenu(Time time, int global_x, int global_y)
{
    if (menu->isOpen())
	return;

    unsigned int index = view->rootInterior()->cachedIndexOf(selected_item);
    view->setSelected(index == NPOS ? 0 : view->rootShadow()->shadow(index));

    conn.enterMenuMode(time, menu);
    menu->open(time, global_x, global_y + height() + 1,
	       width() - 2, 120, hook());
}

void Selector::closeMenu(Time time, bool select_item)
{
    if (!menu->isOpen())
	return;

    if (select_item)
	setSelectedItem(view->selected() ? view->selected()->item() : 0);
    conn.leaveMenuMode(time);
}

void Selector::paint(const Region &region, X::Pixmap pixmap)
{
    int x = this->x() - region.x;
    int y = this->y() - region.y;

    gc().setForeground(has_focus ? colors().selectorFocussedBackground.pixel
				 : colors().selectorBackground.pixel);
    pixmap.fillRectangle(gc(), x, y, width(), height());

    if (selected_item) {
	Icon *icon = selected_item->icon();

	gc().setForeground(has_focus ? colors().selectorFocussedText.pixel
				     : colors().selectorText.pixel);
	pixmap.wcDrawString(conn.fontSet(), gc(),
			    x + SELECTOR_PADDING_X +
				icon->width() + SELECTOR_PADDING_X,
			    y + (height() + conn.fontSet().ascent()
					  - conn.fontSet().descent()) / 2,
			    selected_item->caption());

	X::Drawable::copyArea(icon->pixmap(conn, has_focus
				  ? colors().selectorFocussedBackground.pixel
				  : colors().selectorBackground.pixel),
			      pixmap, gc(), 0, 0,
			      icon->width(), icon->height(),
			      x + SELECTOR_PADDING_X,
			      y + (height() - icon->height()) / 2);
    }

    x += width() - SELECTOR_PADDING_X - 8;
    y += height() / 2 - 4;

    gc().setForeground(has_focus ? colors().selectorFocussedArrow.pixel
				 : colors().selectorArrow.pixel);
    pixmap.drawLine(gc(), x,     y + 2, x + 6, y + 2);
    pixmap.drawLine(gc(), x + 1, y + 3, x + 5, y + 3);
    pixmap.drawLine(gc(), x + 2, y + 4, x + 4, y + 4);
    pixmap.drawLine(gc(), x + 3, y + 5, x + 3, y + 5);
}

void Selector::buttonPressEvent(Time time, int x, int y,
				int x_root, int y_root,
				unsigned int state, unsigned int button)
{
    if (control()) {
	control()->focus(this);
	control()->enterEditor(time, x, y, x_root, y_root);
    }

    openMenu(time, x_root - x + this->x(),
		   y_root - y + this->y());
}

void Selector::keyPressEvent(Time time, int x, int y,
			     int x_root, int y_root,
			     unsigned int state, unsigned int keycode,
			     const std::string &text, KeySym keysym)
{
    switch (keysym) {
    case XK_space:
    case XK_Return:
	if (menu->isOpen())
	    closeMenu(time, true);
	else
	    openMenu(time, x_root - x + this->x(),
			   y_root - y + this->y());
	break;
    case XK_Escape:
	closeMenu(time, false);
	break;
    default:
	if (menu->isOpen())
	    view->keyPressEvent(time, x, y, x_root, y_root,
				state, keycode, text, keysym);
    }
}

void Selector::focusInEvent()
{
    has_focus = true;
    invalidate();
}

void Selector::focusOutEvent()
{
    closeMenu(CurrentTime, false);
    has_focus = false;
    invalidate();
}
