/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <sys/time.h>
#include <stdlib.h>
#include <err.h>
#include "timer.h"

using namespace Toolkit;


static struct timeval tv_add(const struct timeval &tv0,
			     const struct timeval &tv1)
{
    struct timeval tv;
    if (tv0.tv_usec + tv1.tv_usec > 1000000) {
	tv.tv_sec = tv0.tv_sec + tv1.tv_sec + 1;
	tv.tv_usec = tv0.tv_usec + tv1.tv_usec - 1000000;
    } else {
	tv.tv_sec = tv0.tv_sec + tv1.tv_sec;
	tv.tv_usec = tv0.tv_usec + tv1.tv_usec;
    }
    return tv;
}

static struct timeval tv_sub(const struct timeval &tv0,
			     const struct timeval &tv1)
{
    struct timeval tv;
    if (tv0.tv_usec < tv1.tv_usec) {
	tv.tv_sec = tv0.tv_sec - tv1.tv_sec - 1;
	tv.tv_usec = tv0.tv_usec - tv1.tv_usec + 1000000;
    } else {
	tv.tv_sec = tv0.tv_sec - tv1.tv_sec;
	tv.tv_usec = tv0.tv_usec - tv1.tv_usec;
    }
    return tv;
}

static short tv_comp(const struct timeval &tv0, const struct timeval &tv1)
{
    if (tv1.tv_sec < tv0.tv_sec)
	return -1;
    if (tv1.tv_sec > tv0.tv_sec)
	return 1;
    if (tv1.tv_usec < tv0.tv_usec)
	return -1;
    if (tv1.tv_usec > tv0.tv_usec)
	return 1;
    return 0;
}

static struct timespec tv_timespec(const struct timeval &tv)
{
    struct timespec timeout;
    timeout.tv_sec = tv.tv_sec;
    timeout.tv_nsec = tv.tv_usec * 1000;
    return timeout;
}

static bool tv_is_null(const struct timeval &tv)
{
    return tv.tv_sec == 0 && tv.tv_usec == 0;
}

static struct timeval current_time()
{
    struct timeval tv;
    if (gettimeofday(&tv, NULL) == -1)
	err(EXIT_FAILURE, "gettimeofday");
    return tv;
}

/****************************************************************************/

class TimerRecord
{
public:
    TimerRecord(const struct timeval &timeout,
		const struct timeval &interval,
		Timer *timer, TimerRecord *next)
	: timeout(timeout), interval(interval), timer(timer), next(next) {
    }
    struct timeval const timeout, interval;
    Timer *const timer;
    TimerRecord *next;
};

static TimerRecord *next_timer_record = 0;


static void set_timer(const struct timeval &timeout,
		      const struct timeval &interval, Timer *timer)
{
    TimerRecord **ptr = &next_timer_record;

    while (*ptr && tv_comp((*ptr)->timeout, timeout) != -1)
	ptr = &(*ptr)->next;

    *ptr = new TimerRecord(timeout, interval, timer, *ptr);
}

static void unset_timer(Timer *timer)
{
    TimerRecord **ptr = &next_timer_record;

    while (*ptr)
	if ((*ptr)->timer == timer) {
	    TimerRecord *next = (*ptr)->next;
	    delete *ptr;
	    *ptr = next;
	} else
	    ptr = &(*ptr)->next;
}

static bool has_timer(Timer const *timer)
{
    TimerRecord **ptr = &next_timer_record;

    while (*ptr) {
	if ((*ptr)->timer == timer)
	    return true;
	ptr = &(*ptr)->next;
    }
    return false;
}

const struct timespec *Toolkit::process_timers()
{
    struct timeval now = current_time();
    static struct timespec spec;

    while (next_timer_record &&
	   tv_comp(now, next_timer_record->timeout) != 1) {
	TimerRecord *rec = next_timer_record;
	next_timer_record = rec->next;
	if (!tv_is_null(rec->interval))
	    set_timer(tv_add(rec->timeout, rec->interval),
		      rec->interval, rec->timer);
	rec->timer->timeout();
	delete rec;
    }

    if (!next_timer_record)
	return NULL;

    spec = tv_timespec(tv_sub(next_timer_record->timeout, now));
    return &spec;
}

/****************************************************************************/

Timer::~Timer()
{
    unset();
}

void Timer::set(long sec, long usec, bool repeat)
{
    struct timeval interval;
    interval.tv_sec = sec;
    interval.tv_usec = usec;

    unset_timer(this);

    if (repeat)
	set_timer(tv_add(current_time(), interval), interval, this);
    else {
	struct timeval tv_null;
	tv_null.tv_sec = 0;
	tv_null.tv_usec = 0;
	set_timer(tv_add(current_time(), interval), tv_null, this);
    }
}

void Timer::unset()
{
    unset_timer(this);
}

bool Timer::isActive() const
{
    return has_timer(this);
}
