/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "toolbar.h"
#include "connection.h"
#include "icon.h"

#define TOOLBAR_PADDING_X 4
#define TOOLBAR_PADDING_Y 2
#define TOOLBAR_BUTTON_WIDTH 26
#define TOOLBAR_BUTTON_HEIGHT 26
#define TOOLBAR_BUTTON_MENU_WIDTH 9
#define TOOLBAR_SEPARATOR_PADDING_X 4
#define TOOLBAR_SEPARATOR_PADDING_Y 1
#define TOOLBAR_OVERFLOW_WIDTH 7

using namespace Toolkit;


class Toolkit::ToolbarButtonMenu : public AbstractMenu
{
    ToolbarButton &button;
    AbstractMenu *_menu;
public:
    ToolbarButtonMenu(ToolbarButton &button)
	: AbstractMenu(button.conn), button(button), _menu(0) {
    }
    AbstractMenu *menu() const {
	return _menu;
    }
    void setMenu(AbstractMenu *menu) {
	_menu = menu;
    }
    void close() {
	if (_menu)
	    _menu->close();
	button.invalidate();
    }
    bool inChain(AbstractMenu *menu) const {
	return _menu && _menu->inChain(menu);
    }
};


ToolbarButton::ToolbarButton(Connection &conn, Action &action)
    : Widget(conn), Action::Observer(action), highlighted(false),
      pressed(false), menu(new ToolbarButtonMenu(*this))
{
    actionChanged();
}

ToolbarButton::~ToolbarButton()
{
    delete menu;
}

void ToolbarButton::actionChanged()
{
    SizeRequirements reqs;
    reqs.minWidth = TOOLBAR_BUTTON_WIDTH;
    reqs.maxWidth = TOOLBAR_BUTTON_WIDTH;
    if (action.menu()) {
	reqs.minWidth += TOOLBAR_BUTTON_MENU_WIDTH;
	reqs.maxWidth += TOOLBAR_BUTTON_MENU_WIDTH;
    }
    reqs.minHeight = TOOLBAR_BUTTON_HEIGHT;
    setSizeRequirements(reqs);

    menu->setMenu(action.menu());
    invalidate();

    setStatusMessage(action.disableMessage());
}

void ToolbarButton::paint(const Region &region, X::Pixmap pixmap)
{
    int x = this->x() - region.x;
    int y = this->y() - region.y;

    unsigned long background, border_pixel, arrow_pixel;

    if ((pressed && highlighted) ||
	(action.menu() && action.menu()->isVisible())) {
	background = colors().toolbarPressed.pixel;
	border_pixel = colors().toolbarPressedBorder.pixel;
	arrow_pixel = colors().toolbarPressedArrow.pixel;
    } else if (action.isDisabled()) {
	if (action.isActive()) {
	    background = colors().toolbarDisabledActive.pixel;
	    border_pixel = colors().toolbarDisabledActiveBorder.pixel;
	    arrow_pixel = colors().toolbarDisabledActiveArrow.pixel;
	} else {
	    background = colors().toolbarDisabled.pixel;
	    border_pixel = colors().toolbarDisabledBorder.pixel;
	    arrow_pixel = colors().toolbarDisabledArrow.pixel;
	}
    } else if (highlighted) {
	if (action.isActive()) {
	    background = colors().toolbarHighlightActive.pixel;
	    border_pixel = colors().toolbarHighlightActiveBorder.pixel;
	    arrow_pixel = colors().toolbarHighlightActiveArrow.pixel;
	} else {
	    background = colors().toolbarHighlight.pixel;
	    border_pixel = colors().toolbarHighlightBorder.pixel;
	    arrow_pixel = colors().toolbarHighlightArrow.pixel;
	}
    } else {
	if (action.isActive()) {
	    background = colors().toolbarActive.pixel;
	    border_pixel = colors().toolbarActiveBorder.pixel;
	    arrow_pixel = colors().toolbarActiveArrow.pixel;
	} else {
	    background = colors().toolbarButton.pixel;
	    border_pixel = colors().toolbarButtonBorder.pixel;
	    arrow_pixel = colors().toolbarButtonArrow.pixel;
	}
    }

    gc().setForeground(border_pixel);
    pixmap.drawRectangle(gc(), x, y, width() - 1, height() - 1);
    gc().setForeground(background);
    pixmap.fillRectangle(gc(), x + 1, y + 1, width() - 2, height() - 2);

    X::Drawable::copyArea(action.icon()->pixmap(conn, background,
						action.isDisabled()),
			  pixmap, gc(), 0, 0, action.icon()->width(),
					      action.icon()->height(),
			  x + (width() - action.icon()->width() -
			       (action.menu() ? TOOLBAR_BUTTON_MENU_WIDTH
					      : 0)) / 2,
			  y + (height() - action.icon()->height()) / 2);

    if (!action.menu())
	return;

    x += width() - 8;
    y += height() / 2 - 1;

    gc().setForeground(arrow_pixel);
    pixmap.drawLine(gc(), x - 2, y - 1, x + 2, y - 1);
    pixmap.drawLine(gc(), x - 1, y,     x + 1, y);
    pixmap.drawLine(gc(), x,     y + 1, x,     y + 1);
}

void ToolbarButton::enterEvent(Time time, int x, int y, int x_root, int y_root,
			       unsigned int state)
{
    highlighted = true;
    invalidate();
}

void ToolbarButton::leaveEvent(Time time, int x, int y, int x_root, int y_root,
			       unsigned int state)
{
    highlighted = false;
    invalidate();
}

void ToolbarButton::buttonPressEvent(Time time, int x, int y,
				     int x_root, int y_root,
				     unsigned int state, unsigned int button)
{
    if (button != Button1)
	return;

    if (action.isDisabled())
	return;

    if (!action.menu()) {
	pressed = true;
	invalidate();
	return;
    }

    if (action.menu()->isVisible()) {
	conn.leaveMenuMode(time);
	return;
    }

    conn.enterMenuMode(time, menu);
    action.menu()->show(time, x_root - x + this->x(),
			      x_root - x + this->x() + width(), false,
			      y_root - y + this->y() + height(),
			      y_root - y + this->y(), false, hook());
}

void ToolbarButton::buttonReleaseEvent(Time time, int x, int y,
				       int x_root, int y_root,
				       unsigned int state, unsigned int button)
{
    if (!pressed || button != Button1)
	return;

    pressed = false;
    invalidate();

    if (highlighted && !(state & ~(1 << (button + 7)) & ButtonMask))
	action.trigger();
}

/****************************************************************************/

ToolbarSeparator::ToolbarSeparator(Connection &conn) : Widget(conn)
{
    SizeRequirements reqs;
    reqs.minWidth = 1 + TOOLBAR_SEPARATOR_PADDING_X * 2;
    reqs.maxWidth = 1 + TOOLBAR_SEPARATOR_PADDING_X * 2;
    reqs.minHeight = TOOLBAR_SEPARATOR_PADDING_Y * 2;
    setSizeRequirements(reqs);
}

void ToolbarSeparator::paint(const Region &region, X::Pixmap pixmap)
{
    int x = this->x() - region.x;
    int y = this->y() - region.y;

    gc().setForeground(colors().toolbarSeparator.pixel);
    pixmap.drawLine(
	gc(), x + width() / 2, y + TOOLBAR_SEPARATOR_PADDING_Y,
	x + width() / 2, y + height() - TOOLBAR_SEPARATOR_PADDING_Y - 1);
}

/****************************************************************************/

Toolbar::Toolbar(Connection &conn)
    : ContainerWidget(conn), menu_container(conn),
      full_width(0), expanding(0), menu(conn)
{
    menu_container.setBackground(conn.colors().menuBackground.pixel);
    menu.setWidget(&menu_container);
}

Toolbar::~Toolbar()
{
    std::vector<Control *> c = registeredControls();

    for (std::vector<Control *>::const_iterator i = c.begin();
	 i != c.end(); i++)
	delete *i;
}

void Toolbar::addWidgets(Widget *widget, Widget *menuWidget, bool isSeparator)
{
    Control *control = new Control(*this);
    control->setWidget(widget);

    update_size_requirements();
    requestLayout();

    menu_container.addWidget(menuWidget);
    menu_widget[control] = menuWidget;
    is_separator[control] = isSeparator;
}

void Toolbar::paint(const Region &region, X::Pixmap pixmap)
{
    int x = this->x() - region.x;
    int y = this->y() - region.y;

    gc().setForeground(colors().toolbar.pixel);
    pixmap.fillRectangle(gc(), x, y, width(), height());

    if (width() < full_width) {
	int overflow_x = x + width() - TOOLBAR_PADDING_X
	    - TOOLBAR_OVERFLOW_WIDTH + 2;
	int overflow_y = y + height() / 2 - 1;

	gc().setForeground(colors().toolbarOverflow.pixel);
	for (int i = -1; i < 2; i++) {
	    pixmap.drawLine(gc(), overflow_x - 2, overflow_y + i * 6 - 1,
				  overflow_x + 2, overflow_y + i * 6 - 1);
	    pixmap.drawLine(gc(), overflow_x - 1, overflow_y + i * 6,
				  overflow_x + 1, overflow_y + i * 6);
	    pixmap.drawLine(gc(), overflow_x,     overflow_y + i * 6 + 1,
				  overflow_x,     overflow_y + i * 6 + 1);
	}
    }

    ContainerWidget::paint(region, pixmap);
}

void Toolbar::sizeRequirementsChanged(WidgetControl *control)
{
    update_size_requirements();
    requestLayout();
}

void Toolbar::update_size_requirements()
{
    std::vector<Control *> controls = this->registeredControls();
    if (controls.empty())
	return;

    full_width = 0;
    expanding = 0;
    int height = 0;

    for (std::vector<Control *>::const_iterator i = controls.begin();
	 i != controls.end(); i++) {
	SizeRequirements reqs = (*i)->sizeRequirements();

	full_width += reqs.minWidth;
	if (reqs.minHeight > height)
	    height = reqs.minHeight;
	if (reqs.maxWidth != reqs.minWidth)
	    expanding++;
    }

    full_width += TOOLBAR_PADDING_X * 2;
    height += TOOLBAR_PADDING_Y * 2;

    SizeRequirements reqs;
    reqs.minWidth = TOOLBAR_OVERFLOW_WIDTH + TOOLBAR_PADDING_X * 3;
    reqs.minHeight = height;
    reqs.maxHeight = height;
    setSizeRequirements(reqs);
}

void Toolbar::updateLayout()
{
    std::vector<Control *> controls = this->registeredControls();
    int used = 0;
    int available = width() - full_width;
    int interested = expanding;

    int available_width = width() - TOOLBAR_PADDING_X * 2;
    if (width() < full_width)
	available_width -= TOOLBAR_OVERFLOW_WIDTH + TOOLBAR_PADDING_X;

    for (std::vector<Control *>::const_iterator i = controls.begin();
	 i != controls.end(); i++) {
	SizeRequirements reqs = (*i)->sizeRequirements();
	Region dimensions(x() + TOOLBAR_PADDING_X + used,
			  y() + TOOLBAR_PADDING_Y,
			  reqs.minWidth,
			  height() - TOOLBAR_PADDING_Y * 2);

	if (available_width - used < reqs.minWidth) {
	    std::vector<Control *>::const_iterator j = i;
	    while (j != controls.begin() && is_separator[*--j])
		(*j)->setEnabled(false);

	    bool has_seen_item = false;
	    while (i != controls.end()) {
		(*i)->setEnabled(false);
		if (!is_separator[*i])
		    has_seen_item = true;
		menu_container.setWidgetEnabled(has_seen_item,
						menu_widget[*i]);
		i++;
	    }
	    break;
	}

	if (reqs.maxWidth != reqs.minWidth && width() >= full_width) {
	    int extra = (available - 1) / interested + 1;
	    dimensions.width += extra;
	    available -= extra;
	    interested--;
	}

	(*i)->setEnabled(true);
	menu_container.setWidgetEnabled(false, menu_widget[*i]);
	(*i)->setDimensions(dimensions);
	used += dimensions.width;
    }
}

void Toolbar::buttonPressEvent(Time time, int x, int y, int x_root, int y_root,
			       unsigned int state, unsigned int button)
{
    if (button != Button1 || width() >= full_width ||
	x < this->x() + this->width() - TOOLBAR_OVERFLOW_WIDTH
				      - TOOLBAR_PADDING_X * 2)
	return;

    if (menu.isVisible()) {
	conn.leaveMenuMode(time);
	return;
    }

    conn.enterMenuMode(time, &menu);
    menu.show(time, x_root - x + this->x(),
		    x_root - x + this->x() + width(), true,
		    y_root - y + this->y() + height(),
		    y_root - y + this->y(), false, hook());
}

/****************************************************************************/

ToolbarBuilder::ToolbarBuilder(Connection &conn) : toolbar(conn)
{
}

ToolbarBuilder::~ToolbarBuilder()
{
    for (std::vector<Widget *>::const_iterator i = menu_widgets.begin();
	 i != menu_widgets.end(); i++)
	delete *i;
    for (std::vector<Widget *>::const_iterator i = toolbar_widgets.begin();
	 i != toolbar_widgets.end(); i++)
	delete *i;
}

void ToolbarBuilder::addAction(Action &action)
{
    ToolbarButton *button = new ToolbarButton(toolbar.conn, action);
    toolbar_widgets.push_back(button);

    MenuItem *item = new MenuItem(&toolbar.menu, action);
    menu_widgets.push_back(item);

    toolbar.addWidgets(button, item, false);
}

void ToolbarBuilder::addSeparator()
{
    ToolbarSeparator *toolbar_separator = new ToolbarSeparator(toolbar.conn);
    toolbar_widgets.push_back(toolbar_separator);

    MenuSeparator *menu_separator = new MenuSeparator(toolbar.conn);
    menu_widgets.push_back(menu_separator);

    toolbar.addWidgets(toolbar_separator, menu_separator, true);
}

void ToolbarBuilder::addSpacer()
{
    Widget *spacer = new Widget(toolbar.conn);
    toolbar_widgets.push_back(spacer);

    MenuSeparator *separator = new MenuSeparator(toolbar.conn);
    menu_widgets.push_back(separator);

    toolbar.addWidgets(spacer, separator, true);
}
