/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "buttonview.h"
#include "../connection.h"
#include "../icon.h"

#define BUTTONVIEW_PADDING_X 4
#define BUTTONVIEW_PADDING_Y 3
#define BUTTONVIEW_ITEM_PADDING_X 18
#define BUTTONVIEW_ITEM_PADDING_Y 4
#define BUTTONVIEW_ITEM_EXTRA_YSPACE 1
#define BUTTONVIEW_ITEM_ICON_PADDING_X 5
#define BUTTONVIEW_ARROW_OFFSET 8
#define BUTTONVIEW_HOTKEY_OFFSET 24

using namespace Toolkit;


ButtonView::ButtonView(Connection &conn) : View(BUTTONVIEW_PADDING_X,
						BUTTONVIEW_PADDING_Y, conn)
{
    setFocussables(std::vector<Focussable *>(1, this));
}

void ButtonView::buttonPressEvent(Time time, int x, int y,
				  int x_root, int y_root,
				  unsigned int state, unsigned int button)
{
    if (button == Button4 || button == Button5) {
	View::buttonPressEvent(time, x, y, x_root, y_root, state, button);
	return;
    }

    if (control())
	control()->focus(this);

    ViewShadow *shadow = findShadow(x, y, ViewShadow::SensitiveRegion);
    if (!shadow)
	return;

    if (shadow->count() != 0)
	shadow->setExpanded(!shadow->isExpanded());
    else
	setSelected(shadow);
}

ViewShadowLayout ButtonView::layoutShadow(ViewShadow *shadow)
{
    int anchor_x = shadow->anchorX();
    int anchor_y = shadow->anchorY();

    int height = conn.fontSet().ascent() +
		 conn.fontSet().descent() + BUTTONVIEW_ITEM_PADDING_Y * 2;

    ViewShadowLayout layout;

    layout.sensitiveRegion =
	Region(anchor_x, anchor_y + BUTTONVIEW_ITEM_EXTRA_YSPACE,
	       width() - anchor_x - BUTTONVIEW_PADDING_X * 2, height);
    layout.height = height + BUTTONVIEW_ITEM_EXTRA_YSPACE * 2;

    return layout;
}

void ButtonView::paintShadow(ViewShadow *shadow,
			     int origin_x, int origin_y, X::Pixmap pixmap)
{
    Region region = shadow->layout().sensitiveRegion;

    unsigned long background, border, text, hotkey_text;

    if (isSelected(shadow)) {
	text = colors().btnviewSelectedText.pixel;
	hotkey_text = colors().btnviewSelectedHotkey.pixel;
	border = colors().btnviewSelectedBorder.pixel;
	background = colors().btnviewSelectedBackground.pixel;
    } else if (shadow->count() != 0) {
	if (shadow == highlightedShadow()) {
	    text = colors().btnviewCatHghlghtText.pixel;
	    hotkey_text = colors().btnviewCatHghlghtHotkey.pixel;
	    border = colors().btnviewCatHghlghtBorder.pixel;
	    background = colors().btnviewCatHghlghtBackground.pixel;
	} else {
	    text = colors().btnviewCategoryText.pixel;
	    hotkey_text = colors().btnviewCategoryHotkey.pixel;
	    border = colors().btnviewCategoryBorder.pixel;
	    background = colors().btnviewCategoryBackground.pixel;
	}
    } else {
	if (shadow == highlightedShadow()) {
	    text = colors().btnviewHighlightText.pixel;
	    hotkey_text = colors().btnviewHighlightHotkey.pixel;
	    border = colors().btnviewHighlightBorder.pixel;
	    background = colors().btnviewHighlightBackground.pixel;
	} else {
	    text = colors().btnviewItemText.pixel;
	    hotkey_text = colors().btnviewItemHotkey.pixel;
	    border = colors().btnviewItemBorder.pixel;
	    background = colors().btnviewItemBackground.pixel;
	}
    }

    gc().setForeground(background);
    pixmap.fillRectangle(gc(), origin_x + region.x + 1,
			       origin_y + region.y + 1,
			       region.width - 2, region.height - 2);
    gc().setForeground(border);
    pixmap.drawRectangle(gc(), origin_x + region.x, origin_y + region.y,
			       region.width - 1, region.height - 1);

    Icon *icon = shadow->item()->icon();
    int icon_width = icon ? icon->width() + BUTTONVIEW_ITEM_ICON_PADDING_X : 0;

    X::FontSet font_set = shadow->count() != 0 ? conn.fontSetBold()
					       : conn.fontSet();

    gc().setForeground(text);
    pixmap.wcDrawString(font_set, gc(),
			origin_x + region.x +
			    BUTTONVIEW_ITEM_PADDING_X + icon_width,
			origin_y + region.y +
			    (region.height + conn.fontSet().ascent()
					   - conn.fontSet().descent()) / 2,
			shadow->item()->caption());

    gc().setForeground(hotkey_text);
    pixmap.wcDrawString(font_set, gc(),
			origin_x + region.x + region.width
			    - BUTTONVIEW_HOTKEY_OFFSET
			    - conn.cachedEscapement(
				conn.fontSet(), shadow->item()->hotkey()) / 2,
			origin_y + region.y +
			    (region.height + conn.fontSet().ascent()
					   - conn.fontSet().descent()) / 2,
			shadow->item()->hotkey());

    if (icon)
	X::Drawable::copyArea(icon->pixmap(conn, background), pixmap, gc(),
			      0, 0, icon->width(), icon->height(),
			      origin_x + region.x + BUTTONVIEW_ITEM_PADDING_X,
			      origin_y + region.y +
				  (region.height - icon->height()) / 2);

    if (shadow->count() == 0)
	return;

    if (shadow == highlightedShadow())
	gc().setForeground(colors().btnviewCatHghlghtArrow.pixel);
    else
	gc().setForeground(colors().btnviewCategoryArrow.pixel);

    int arr_x = origin_x + region.x + BUTTONVIEW_ARROW_OFFSET;
    int arr_y = origin_y + region.y + region.height / 2 - 1;

    if (shadow->isExpanded()) {
	pixmap.drawLine(gc(), arr_x - 3, arr_y - 1, arr_x + 3, arr_y - 1);
	pixmap.drawLine(gc(), arr_x - 2, arr_y,     arr_x + 2, arr_y);
	pixmap.drawLine(gc(), arr_x - 1, arr_y + 1, arr_x + 1, arr_y + 1);
	pixmap.drawLine(gc(), arr_x,     arr_y + 2, arr_x,     arr_y + 2);
    } else {
	pixmap.drawLine(gc(), arr_x - 1, arr_y - 3, arr_x - 1, arr_y + 3);
	pixmap.drawLine(gc(), arr_x,     arr_y - 2, arr_x,     arr_y + 2);
	pixmap.drawLine(gc(), arr_x + 1, arr_y - 1, arr_x + 1, arr_y + 1);
	pixmap.drawLine(gc(), arr_x + 2, arr_y,     arr_x + 2, arr_y);
    }
}
