/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "directoryview.h"
#include "../connection.h"
#include "../icon.h"

#define TREEVIEW_ITEM_PADDING_X 4
#define TREEVIEW_ITEM_PADDING_Y 2
#define TREEVIEW_PRE_V 12
#define TREEVIEW_POST_V 12
#define TREEVIEW_PLUP_SIZE 9
#define TREEVIEW_PADDING_X 4
#define TREEVIEW_PADDING_Y 4
#define TREEVIEW_INDENT 24

using namespace Toolkit;


DirectoryView::DirectoryView(Connection &conn) : View(TREEVIEW_PADDING_X,
						      TREEVIEW_PADDING_Y, conn)
{
    setFocussables(std::vector<Focussable *>(1, this));
}

void DirectoryView::buttonPressEvent(Time time, int x, int y,
				     int x_root, int y_root,
				     unsigned int state, unsigned int button)
{
    if (button == Button4 || button == Button5) {
	View::buttonPressEvent(time, x, y, x_root, y_root, state, button);
	return;
    }

    if (control())
	control()->focus(this);

    ViewShadow *plup_shadow = findShadow(x, y, ViewShadow::PlupRegion);
    if (plup_shadow) {
	plup_shadow->setExpanded(!plup_shadow->isExpanded());
	return;
    }

    setSelected(findShadow(x, y, ViewShadow::SensitiveRegion),
		state & ControlMask);
}

ViewShadowLayout DirectoryView::layoutShadow(ViewShadow *shadow)
{
    int anchor_x = shadow->anchorX();
    int anchor_y = shadow->anchorY();

    int height = conn.fontSet().ascent() +
		 conn.fontSet().descent() + TREEVIEW_ITEM_PADDING_Y * 2;

    ViewShadowLayout layout;

    layout.sensitiveRegion =
	Region(anchor_x, anchor_y,
	       width() - anchor_x - TREEVIEW_PADDING_X * 2, height);
    layout.plupRegion =
	Region(anchor_x - TREEVIEW_POST_V - TREEVIEW_PLUP_SIZE / 2,
	       anchor_y + height / 2 - TREEVIEW_PLUP_SIZE / 2,
	       TREEVIEW_PLUP_SIZE, TREEVIEW_PLUP_SIZE);
    layout.indent = TREEVIEW_INDENT;
    layout.height = height;

    return layout;
}

void DirectoryView::paintShadow(ViewShadow *shadow,
				int origin_x, int origin_y, X::Pixmap pixmap)
{
    int anchor_x = shadow->anchorX();
    int anchor_y = shadow->anchorY();

    Region region = shadow->layout().sensitiveRegion;

    int plup_x = anchor_x - TREEVIEW_POST_V;
    int plup_y = anchor_y + region.height / 2;

    unsigned long background;

    if (isSelected(shadow)) {
	background = colors().viewSelected.pixel;
	gc().setForeground(background);
	pixmap.fillRectangle(gc(),
			     origin_x + region.x,
			     origin_y + region.y,
			     region.width,
			     region.height);
	gc().setForeground(colors().viewSelectedText.pixel);
    } else if (shadow == highlightedShadow()) {
	background = colors().viewHighlight.pixel;
	gc().setForeground(background);
	pixmap.fillRectangle(gc(),
			     origin_x + region.x,
			     origin_y + region.y,
			     region.width,
			     region.height);
	gc().setForeground(colors().viewHighlightText.pixel);
    } else {
	background = colors().viewBackground.pixel;
	gc().setForeground(background);
	pixmap.fillRectangle(gc(),
			     origin_x + region.x,
			     origin_y + region.y,
			     region.width,
			     region.height);
	gc().setForeground(colors().viewItemText.pixel);
    }

    Icon *icon = shadow->item()->icon();

    pixmap.wcDrawString(shadow->item()->isBold() ? conn.fontSetBold()
						 : conn.fontSet(), gc(),
			origin_x + anchor_x + TREEVIEW_ITEM_PADDING_X +
			    icon->width() + TREEVIEW_ITEM_PADDING_X,
			origin_y + anchor_y +
			    (region.height + conn.fontSet().ascent()
					   - conn.fontSet().descent()) / 2,
			shadow->item()->caption());

    X::Drawable::copyArea(icon->pixmap(conn, background),
			  pixmap, gc(), 0, 0, icon->width(), icon->height(),
			  origin_x + anchor_x + TREEVIEW_ITEM_PADDING_X,
			  origin_y + anchor_y +
			      (region.height - icon->height()) / 2);

    XGCValues values;

    if (shadow->parent()->parent()) {
	values.foreground = colors().viewLines.pixel;
	values.line_style = LineOnOffDash;
	values.dash_offset = plup_x % 2;
	values.dashes = 1;
	gc().change(GCForeground | GCLineStyle | GCDashOffset | GCDashList,
		    &values);

	pixmap.drawLine(
	    gc(),
	    origin_x + plup_x,
	    origin_y + plup_y,
	    origin_x + anchor_x - 1,
	    origin_y + plup_y);
    }

    if (shadow->isExpanded() && shadow->count() != 0) {
	values.foreground = colors().viewLines.pixel;
	values.line_style = LineOnOffDash;
	values.dash_offset = !((anchor_y + region.height) % 2);
	values.dashes = 1;
	gc().change(GCForeground | GCLineStyle | GCDashOffset | GCDashList,
		    &values);

	pixmap.drawLine(
	    gc(),
	    origin_x + anchor_x + TREEVIEW_PRE_V,
	    origin_y + anchor_y + region.height,
	    origin_x + anchor_x + TREEVIEW_PRE_V,
	    origin_y + shadow->last()->anchorY()
		     + shadow->last()->layout().height / 2);
    }

    values.line_style = LineSolid;
    gc().change(GCLineStyle, &values);

    if (shadow->parent()->parent() && shadow->count() != 0) {
	gc().setForeground(colors().viewPlupBackground.pixel);
	pixmap.fillRectangle(
	    gc(),
	    origin_x + plup_x - (TREEVIEW_PLUP_SIZE - 1) / 2 + 1,
	    origin_y + plup_y - (TREEVIEW_PLUP_SIZE - 1) / 2 + 1,
	    TREEVIEW_PLUP_SIZE - 2,
	    TREEVIEW_PLUP_SIZE - 2);
	gc().setForeground(colors().viewPlupBorder.pixel);
	pixmap.drawRectangle(
	    gc(),
	    origin_x + plup_x - (TREEVIEW_PLUP_SIZE - 1) / 2,
	    origin_y + plup_y - (TREEVIEW_PLUP_SIZE - 1) / 2,
	    TREEVIEW_PLUP_SIZE - 1,
	    TREEVIEW_PLUP_SIZE - 1);

	gc().setForeground(colors().viewPlupForeground.pixel);
	pixmap.drawLine(
	    gc(),
	    origin_x + plup_x - 2, origin_y + plup_y,
	    origin_x + plup_x + 2, origin_y + plup_y);
	if (!shadow->isExpanded())
	    pixmap.drawLine(
		gc(),
		origin_x + plup_x, origin_y + plup_y - 2,
		origin_x + plup_x, origin_y + plup_y + 2);
    }
}
