/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "dropdown.h"
#include "../selector.h"
#include "../connection.h"
#include "../icon.h"

#define DROPDOWNVIEW_ITEM_PADDING_X 4
#define DROPDOWNVIEW_ITEM_PADDING_Y 2
#define DROPDOWNVIEW_INDENT 12

using namespace Toolkit;


DropdownView::DropdownView(Selector &selector)
    : View(0, 0, selector.conn), selector(selector), _selected(0)
{
    setFocussables(std::vector<Focussable *>(1, this));
}

ViewShadow *DropdownView::selected() const
{
    return _selected;
}

void DropdownView::setSelected(ViewShadow *shadow)
{
    if (shadow == _selected)
	return;

    _selected = shadow;
    invalidate();
}

bool DropdownView::isMenu() const
{
    return true;
}

void DropdownView::buttonPressEvent(Time time, int x, int y,
				    int x_root, int y_root,
				    unsigned int state, unsigned int button)
{
    if (button == Button4 || button == Button5) {
	View::buttonPressEvent(time, x, y, x_root, y_root, state, button);
	return;
    }

    if (control())
	control()->focus(this);

    if (button != Button1)
	return;

    ViewShadow *shadow = findShadow(x, y, ViewShadow::SensitiveRegion);
    if (!shadow)
	return;

    setSelected(shadow);
    selector.closeMenu(time, true);
}

void DropdownView::buttonReleaseEvent(Time time, int x, int y,
				      int x_root, int y_root,
				      unsigned int state, unsigned int button)
{
    if (button != Button1)
	return;

    ViewShadow *shadow = findShadow(x, y, ViewShadow::SensitiveRegion);
    if (!shadow)
	return;

    setSelected(shadow);
    selector.closeMenu(time, true);
}

void DropdownView::keyPressEvent(Time time, int x, int y,
				 int x_root, int y_root,
				 unsigned int state, unsigned int keycode,
				 const std::string &text, KeySym keysym)
{
    ViewShadow *shadow;

    switch (keysym) {
    case XK_Down:
	if (selected() && (shadow = selected()->next()))
	    setSelected(shadow);
	else
	    setSelected(rootShadow()->shadow(0));
	break;
    case XK_Up:
	if (selected() && (shadow = selected()->previous()))
	    setSelected(shadow);
	else
	    setSelected(rootShadow()->lastDescendant());
	break;
    }
}

ViewShadowLayout DropdownView::layoutShadow(ViewShadow *shadow)
{
    int height = conn.fontSet().ascent() +
		 conn.fontSet().descent() + DROPDOWNVIEW_ITEM_PADDING_Y * 2;

    ViewShadowLayout layout;

    layout.sensitiveRegion = Region(0, shadow->anchorY(), width(), height);
    layout.indent = DROPDOWNVIEW_INDENT;
    layout.height = height;

    return layout;
}

void DropdownView::paintShadow(ViewShadow *shadow,
			       int origin_x, int origin_y, X::Pixmap pixmap)
{
    int anchor_x = shadow->anchorX();
    int anchor_y = shadow->anchorY();

    Region region = shadow->layout().sensitiveRegion;

    if (shadow == selected()) {
	gc().setForeground(colors().viewSelected.pixel);
	pixmap.fillRectangle(gc(),
			     origin_x + region.x,
			     origin_y + region.y,
			     region.width,
			     region.height);
	gc().setForeground(colors().viewSelectedText.pixel);
    } else {
	gc().setForeground(colors().viewBackground.pixel);
	pixmap.fillRectangle(gc(),
			     origin_x + region.x,
			     origin_y + region.y,
			     region.width,
			     region.height);
	gc().setForeground(colors().viewItemText.pixel);
    }

    Icon *icon = shadow->item()->icon();

    pixmap.wcDrawString(conn.fontSet(), gc(),
			origin_x + anchor_x + DROPDOWNVIEW_ITEM_PADDING_X +
			    icon->width() + DROPDOWNVIEW_ITEM_PADDING_X,
			origin_y + anchor_y +
			    (region.height + conn.fontSet().ascent()
					   - conn.fontSet().descent()) / 2,
			shadow->item()->caption());

    X::Drawable::copyArea(icon->pixmap(conn, shadow == selected()
			      ? colors().viewSelected.pixel
			      : colors().viewBackground.pixel),
			  pixmap, gc(), 0, 0, icon->width(), icon->height(),
			  origin_x + anchor_x + DROPDOWNVIEW_ITEM_PADDING_X,
			  origin_y + anchor_y +
			      (region.height - icon->height()) / 2);
}

void DropdownView::highlightedShadowChanged()
{
    if (!highlightedShadow())
	return;

    setSelected(highlightedShadow());
}
