/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "propertiesview.h"
#include "../editors.h"
#include "../connection.h"

#define PROPVIEW_FIXED 17
#define PROPVIEW_PLUP_SIZE 9
#define PROPVIEW_ITEM_PADDING_X 4
#define PROPVIEW_ITEM_PADDING_Y 2
#define PROPVIEW_INDENT 16

using namespace Toolkit;


class PropertiesView::Control : public WidgetControl
{
    PropertiesView &view;
public:
    Control(PropertiesView &view) : WidgetControl(view.conn), view(view) {
    }

    X::Window window() const {
	return view.control() ? view.control()->window() : X::Window();
    }
    X::GC gc() const {
	return view.control() ? view.control()->gc() : X::GC();
    }
    WidgetWindowHook *hook() const {
	return view.control() ? view.control()->hook() : 0;
    }

    void focus(Focussable *focussable) {
	if (view.control())
	    view.control()->focus(&view);
    }
    void focussablesChanged() {
	/* ignore */
    }
    void enterEditor(Time time, int x, int y, int x_root, int y_root) {
	view.enterEditor(time, x, y, x_root, y_root);
    }
    void leaveEditor(bool cancel) {
	view.leaveEditor(cancel);
    }

    void setDimensions(const Region &dimensions) {
	WidgetControl::setDimensions(dimensions);
    }
private:
    void sizeRequirementsChanged() {
	/* ignore */
    }
    void propagateChildLayoutRequest() {
	if (view.control())
	    view.control()->requestChildLayout();
    }
};

/****************************************************************************/

PropertiesView::PropertiesView(Connection &conn)
    : View(0, 0, conn), _selected(0), editor(0),
      editor_control(new Control(*this)),
      has_focus(false), editor_has_focus(false), origin_x(0), origin_y(0)
{
    setFocussables(std::vector<Focussable *>(1, this));
}

PropertiesView::~PropertiesView()
{
    hide_editor_widget();
    delete editor_control;
}

void PropertiesView::show_editor_widget()
{
    if (!has_focus || editor_control->widget() != 0 || !editor)
	return;

    editor->createWidget(conn);
    editor_control->setWidget(editor->widget());
    invalidate();
}

void PropertiesView::hide_editor_widget()
{
    if (!editor_control->widget())
	return;

    if (editor_has_focus)
	leaveEditor(false);

    editor->deleteWidget();
    editor_control->setWidget(0);
    invalidate();
}

void PropertiesView::enterEditor(Time time, int x, int y,
				 int x_root, int y_root)
{
    if (!has_focus || editor_has_focus || !editor)
	return;

    show_editor_widget();

    editor_control->widget()->focussables().front()->focusInEvent();
    editor->invoke(time, x, y, x_root, y_root);

    editor_has_focus = true;
}

void PropertiesView::leaveEditor(bool cancel)
{
    if (!editor_has_focus)
	return;

    if (cancel)
	editor->reset();
    else
	editor->accept();
    editor_control->widget()->focussables().front()->focusOutEvent();

    editor_has_focus = false;
}

ViewShadow *PropertiesView::selected() const
{
    return _selected;
}

void PropertiesView::setSelected(ViewShadow *shadow)
{
    if (shadow == _selected)
	return;

    hide_editor_widget();
    _selected = shadow;

    if (!shadow)
	editor = 0;
    else {
	Region region = shadow->layout().editorRegion;
	region.x += this->x() + originX();
	region.y += this->y() + originY();
	editor_control->setDimensions(region);

	editor = shadow->item()->editor();
	if (has_focus)
	    show_editor_widget();
    }
    invalidate();
}

std::vector<Widget *> PropertiesView::children() const
{
    std::vector<Widget *> widgets;
    if (editor_control->widget() != 0)
	widgets.push_back(editor_control->widget());
    return widgets;
}

void PropertiesView::recursivelyUpdateLayout()
{
    Widget::recursivelyUpdateLayout();
    editor_control->updateChildLayout();
}

void PropertiesView::updateLayout()
{
    View::updateLayout();

    if (!selected())
	return;

    ViewShadow *shadow = selected();
    do {
	ViewShadow *parent = static_cast<ViewShadow *>(shadow->parent());
	if (!parent || !parent->isExpanded() ||
	    parent->indexOf(shadow) == NPOS) {
	    setSelected(0);
	    return;
	}
	shadow = parent;
    } while (shadow != rootShadow());

    Region r = selected()->layout().editorRegion;
    r.x += this->x() + originX();
    r.y += this->y() + originY();
    editor_control->setDimensions(r);
}

void PropertiesView::setMnemonicVisibility(bool visible)
{
    if (editor_control->widget())
	editor_control->widget()->setMnemonicVisibility(visible);
}

void PropertiesView::windowPositionChanged()
{
    if (editor_control->widget())
	editor_control->widget()->windowPositionChanged();
}

void PropertiesView::paint(const Region &region, X::Pixmap pixmap)
{
    View::paint(region, pixmap);

    if (editor_control->widget())
	editor_control->widget()->paint(region, pixmap);
}

ViewShadow *PropertiesView::next_valid_shadow(ViewShadow *shadow)
{
    if (!shadow)
	shadow = rootShadow();

    while (shadow->isExpanded() && shadow->count() != 0 &&
	   !(shadow->item() && shadow->item()->editor())) {
	shadow = shadow->next();
	if (!shadow)
	    shadow = rootShadow();
    }
    return shadow == rootShadow() ? 0 : shadow;
}

ViewShadow *PropertiesView::previous_valid_shadow(ViewShadow *shadow)
{
    if (!shadow)
	shadow = rootShadow()->lastDescendant();

    while (shadow->isExpanded() && shadow->count() != 0 &&
	   !(shadow->item() && shadow->item()->editor())) {
	shadow = shadow->previous();
	if (!shadow)
	    shadow = rootShadow()->lastDescendant();
    }
    return shadow;
}

void PropertiesView::focusThis()
{
    if (control())
	control()->focus(this);
}

void PropertiesView::buttonPressEvent(Time time, int x, int y,
				      int x_root, int y_root,
				      unsigned int state, unsigned int button)
{
    if (button == Button4 || button == Button5) {
	View::buttonPressEvent(time, x, y, x_root, y_root, state, button);
	return;
    }

    if (control())
	control()->focus(this);

    ViewShadow *shadow = findShadow(x, y, ViewShadow::PlupRegion);
    if (shadow) {
	shadow->setExpanded(!shadow->isExpanded());
	return;
    }

    shadow = findShadow(x, y, ViewShadow::EditorRegion);
    if (shadow) {
	setSelected(shadow);
	enterEditor(time, x, y, x_root, y_root);
	return;
    }
    leaveEditor(false);

    shadow = findShadow(x, y, ViewShadow::SensitiveRegion);
    if (shadow) {
	if (shadow->count() != 0 && !shadow->item()->editor())
	    shadow->setExpanded(true);
	setSelected(next_valid_shadow(shadow));
    }
}

void PropertiesView::keyPressEvent(Time time, int x, int y,
				   int x_root, int y_root,
				   unsigned int state, unsigned int keycode,
				   const std::string &text, KeySym keysym)
{
    if (editor_has_focus) {
	switch (keysym) {
	case XK_Return:
	    leaveEditor(false);
	    break;
	case XK_Escape:
	    leaveEditor(true);
	    break;
	default:
	    editor_control->widget()->focussables().front()->keyPressEvent(
		time, x, y, x_root, y_root, state, keycode, text, keysym);
	}
	return;
    }

    switch (keysym) {
    case XK_Down:
	if (!selected())
	    setSelected(next_valid_shadow(rootShadow()));
	else
	    setSelected(next_valid_shadow(selected()->next()));
	break;
    case XK_Up:
	if (!selected())
	    setSelected(previous_valid_shadow(rootShadow()));
	else
	    setSelected(previous_valid_shadow(selected()->previous()));
	break;
    case XK_Right:
	if (selected() && selected()->count() != 0) {
	    selected()->setExpanded(true);
	    setSelected(next_valid_shadow(selected()));
	    invalidate();
	}
	break;
    case XK_Left:
	if (selected() && selected()->parent()->parent()) {
	    selected()->parent()->setExpanded(false);
	    setSelected(selected()->parent());
	    invalidate();
	}
	break;
    case XK_Return:
    case XK_space:
	enterEditor(time, x, y, x_root, y_root);
	return;
    }
}

void PropertiesView::keyReleaseEvent(Time time, int x, int y,
				     int x_root, int y_root,
				     unsigned int state, unsigned int keycode)
{
    if (editor_has_focus)
	editor_control->widget()->focussables().front()->keyReleaseEvent(
	    time, x, y, x_root, y_root, state, keycode);
}

void PropertiesView::focusInEvent()
{
    has_focus = true;

    if (selected())
	show_editor_widget();
    else
	setSelected(next_valid_shadow(rootShadow()));

    invalidate();
}

void PropertiesView::focusOutEvent()
{
    hide_editor_widget();
    has_focus = false;
    invalidate();
}

ViewShadowLayout PropertiesView::layoutShadow(ViewShadow *shadow)
{
    int anchor_y = shadow->anchorY();
    int height = conn.fontSet().ascent() +
		 conn.fontSet().descent() + PROPVIEW_ITEM_PADDING_Y * 2;

    ViewShadowLayout layout;

    layout.sensitiveRegion =
	Region(PROPVIEW_FIXED, anchor_y + 1,
	       width() - PROPVIEW_FIXED, height);
    layout.plupRegion =
	Region(PROPVIEW_FIXED / 2 - (PROPVIEW_PLUP_SIZE - 1) / 2,
	       anchor_y + height / 2 - (PROPVIEW_PLUP_SIZE - 1) / 2,
	       PROPVIEW_PLUP_SIZE, PROPVIEW_PLUP_SIZE);
    if (shadow->item()->editor())
	layout.editorRegion =
	    Region((width() + PROPVIEW_FIXED) / 2 + 1, anchor_y + 1,
		   (width() - PROPVIEW_FIXED - 1) / 2, height);

    if (!shadow->parent()->parent())
	layout.indent = PROPVIEW_FIXED;
    else
	layout.indent = PROPVIEW_INDENT;
    layout.height = height + 1;

    return layout;
}

void PropertiesView::paintShadow(ViewShadow *shadow,
				 int origin_x, int origin_y, X::Pixmap pixmap)
{
    int anchor_x = shadow->anchorX();
    Region region = shadow->layout().sensitiveRegion;

    unsigned long background;

    if (!shadow->parent()->parent()) {
	if (has_focus && shadow == selected()) {
	    background = colors().propertiesSelectedFixed.pixel;
	    gc().setForeground(background);
	    pixmap.fillRectangle(gc(),
				 origin_x,
				 origin_y + region.y - 1,
				 width(),
				 region.height + 1);
	    gc().setForeground(colors().propertiesSelectedFixedText.pixel);
	} else {
	    background = colors().propertiesFixed.pixel;
	    gc().setForeground(background);
	    pixmap.fillRectangle(gc(),
				 origin_x,
				 origin_y + region.y - 1,
				 width(),
				 region.height + 1);
	    gc().setForeground(colors().propertiesFixedText.pixel);
	}

	pixmap.wcDrawString(conn.fontSetBold(), gc(),
			    origin_x + PROPVIEW_FIXED +
				PROPVIEW_ITEM_PADDING_X,
			    origin_y + region.y +
				(region.height
				    + conn.fontSet().ascent()
				    - conn.fontSet().descent()) / 2,
			    shadow->item()->caption());
    } else {
	gc().setForeground(colors().propertiesFixed.pixel);
	pixmap.drawLine(gc(), origin_x, origin_y + region.y - 1,
			      origin_x + width() - 1, origin_y + region.y - 1);
	pixmap.fillRectangle(gc(),
			     origin_x,
			     origin_y + region.y,
			     PROPVIEW_FIXED,
			     region.height);

	if (has_focus && shadow == selected()) {
	    background = colors().propertiesSelected.pixel;
	    gc().setForeground(background);
	    pixmap.fillRectangle(gc(),
				 origin_x + region.x,
				 origin_y + region.y,
				 region.width,
				 region.height);
	    gc().setForeground(colors().propertiesSelectedText.pixel);
	} else if (shadow == highlightedShadow()) {
	    background = colors().viewHighlight.pixel;
	    gc().setForeground(background);
	    pixmap.fillRectangle(gc(),
				 origin_x + region.x,
				 origin_y + region.y,
				 region.width,
				 region.height);
	    gc().setForeground(colors().viewHighlightText.pixel);
	} else {
	    background = colors().viewBackground.pixel;
	    gc().setForeground(background);
	    pixmap.fillRectangle(gc(),
				 origin_x + region.x,
				 origin_y + region.y,
				 region.width,
				 region.height);
	    gc().setForeground(colors().viewItemText.pixel);
	}

	pixmap.wcDrawString(conn.fontSet(), gc(),
			    origin_x + anchor_x + PROPVIEW_ITEM_PADDING_X,
			    origin_y + region.y +
				(region.height
				    + conn.fontSet().ascent()
				    - conn.fontSet().descent()) / 2,
			    conn.ellipsis(
				conn.fontSet(), shadow->item()->caption(),
				(width() + PROPVIEW_FIXED) / 2 - anchor_x -
				    PROPVIEW_ITEM_PADDING_X * 2));
    }

    ItemEditor *item_editor = shadow->item()->editor();

    if (item_editor) {
	gc().setForeground(colors().propertiesFixed.pixel);
	pixmap.drawLine(gc(), origin_x + (width() + PROPVIEW_FIXED) / 2,
			      origin_y + region.y,
			      origin_x + (width() + PROPVIEW_FIXED) / 2,
			      origin_y + region.y + region.height - 1);


	if ((!has_focus || shadow != selected()) && item_editor) {
	    Region editor_region = shadow->layout().editorRegion;
	    editor_region.x += origin_x;
	    editor_region.y += origin_y;
	    item_editor->paint(conn, gc(), editor_region,
			       background, pixmap);
	}
    }

    if (!shadow->count() != 0)
	return;

    region = shadow->layout().plupRegion;

    gc().setForeground(colors().viewPlupBackground.pixel);
    pixmap.fillRectangle(gc(),
			 origin_x + region.x,
			 origin_y + region.y,
			 region.width,
			 region.height);
    gc().setForeground(colors().viewPlupBorder.pixel);
    pixmap.drawRectangle(gc(),
			 origin_x + region.x,
			 origin_y + region.y,
			 region.width - 1,
			 region.height - 1);

    int plup_x = region.x + (region.width - 1) / 2;
    int plup_y = region.y + (region.height - 1) / 2;

    gc().setForeground(colors().viewPlupForeground.pixel);
    pixmap.drawLine(
	gc(),
	origin_x + plup_x - 2, origin_y + plup_y,
	origin_x + plup_x + 2, origin_y + plup_y);
    if (!shadow->isExpanded())
	pixmap.drawLine(
	    gc(),
	    origin_x + plup_x, origin_y + plup_y - 2,
	    origin_x + plup_x, origin_y + plup_y + 2);
}
