/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <assert.h>
#include "widget.h"
#include "hooks.h"
#include "connection.h"

using namespace Toolkit;


WidgetScroller::WidgetScroller(Widget &widget)
    : widget(widget), _range(0), page_size(0), _position(0)
{
}

WidgetScroller::~WidgetScroller()
{
    /* virtualize (not strictly necessary but will suppress some warnings) */
}

int WidgetScroller::range() const
{
    return _range;
}

int WidgetScroller::pageSize() const
{
    return page_size;
}

int WidgetScroller::position() const
{
    return _position;
}

void WidgetScroller::setRange(int range)
{
    if (range < 0)
	range = 0;

    if (range == _range)
	return;

    _range = range;

    if (pageSize() > range)
	setPageSize(range);

    if (position() > range - pageSize())
	setPosition(range - pageSize());

    if (control_scroller())
	control_scroller()->updateRange();
}

void WidgetScroller::setPageSize(int pageSize)
{
    if (pageSize < 1)
	pageSize = 1;

    if (pageSize > range())
	pageSize = range();

    if (pageSize == page_size)
	return;

    page_size = pageSize;

    if (position() > range() - pageSize)
	setPosition(range() - pageSize);

    if (control_scroller())
	control_scroller()->updatePageSize();
}

void WidgetScroller::setPosition(int position)
{
    if (position < 0)
	position = 0;

    if (position > range() - pageSize())
	position = range() - pageSize();

    if (position == _position)
	return;

    _position = position;

    updatePosition();
    if (control_scroller())
	control_scroller()->updatePosition();
}

WidgetControlScroller *WidgetScroller::control_scroller() const
{
    if (widget.horzScroller() == this && widget.control())
	return widget.control()->horzScroller();
    if (widget.vertScroller() == this && widget.control())
	return widget.control()->vertScroller();
    return 0;
}

WidgetControlScroller::~WidgetControlScroller()
{
    /* virtualize */
}

WidgetControlScroller::WidgetControlScroller(WidgetControl &control)
    : control(control)
{
}

/****************************************************************************/

void Focussable::keyPressEvent(
    Time /*time*/, int /*x*/, int /*y*/, int /*x_root*/, int /*y_root*/,
    unsigned int /*state*/, unsigned int /*keycode*/,
    const std::string &/*text*/, KeySym /*keysym*/)
{
    /* relax */
}

void Focussable::keyReleaseEvent(
    Time /*time*/, int /*x*/, int /*y*/, int /*x_root*/, int /*y_root*/,
    unsigned int /*state*/, unsigned int /*keycode*/)
{
    /* relax */
}

void Focussable::focusInEvent()
{
    /* relax */
}

void Focussable::focusOutEvent()
{
    /* relax */
}

/****************************************************************************/

WidgetControl::WidgetControl(Connection &conn)
    : _widget(0), needs_child_layout(false), conn(conn)
{
}

WidgetControl::~WidgetControl()
{
    if (!_widget)
	return;

    Widget *w = _widget;
    _widget = 0;
    w->setControl(0);
}

Widget *WidgetControl::widget() const
{
    return _widget;
}

void WidgetControl::setWidget(Widget *widget)
{
    /* May be called by a destructing widget or on a destructing control with
       widget == 0, so don't call a virtual method on either us or _widget in
       that case, especially not hook() or widgetChanged() -.- */

    if (widget == _widget)
	return;
    if (widget)
	assert(&widget->conn == &conn);

    bool had_focussables = _widget && !_widget->focussables().empty();

    if (_widget) {
	Widget *w = _widget;
	_widget = 0;
	w->setControl(0);
	if (hook())
	    hook()->withdrawWidget(w);
    }

    _widget = widget;
    if (widget)
	widget->setControl(this);

    widgetChanged();
    if (widget)
	setSizeRequirements(widget->sizeRequirements());

    if (widget && !widget->focussables().empty() || had_focussables)
	focussablesChanged();
}

SizeRequirements WidgetControl::sizeRequirements() const
{
    return size_requirements;
}

void WidgetControl::setSizeRequirements(const SizeRequirements &reqs)
{
    if (reqs == size_requirements)
	return;

    size_requirements = reqs;
    sizeRequirementsChanged();
}

void WidgetControl::requestChildLayout()
{
    if (needs_child_layout)
	return;

    needs_child_layout = true;
    propagateChildLayoutRequest();
}

void WidgetControl::updateChildLayout()
{
    if (!needs_child_layout)
	return;

    needs_child_layout = false;
    if (widget())
	widget()->recursivelyUpdateLayout();
}

Toolkit::Region WidgetControl::dimensions() const
{
    return _dimensions;
}

Toolkit::Region WidgetControl::screenDimensions() const
{
    if (!hook())
	return _dimensions;

    Region d(hook()->innerDimensions());
    return Region(_dimensions.x + d.x, _dimensions.y + d.y,
		  _dimensions.width, _dimensions.height);
}

void WidgetControl::setDimensions(const Region &dimensions)
{
    if (dimensions == _dimensions)
	return;

    _dimensions = dimensions;
    if (widget())
	widget()->requestLayout();
}

WidgetControlScroller *WidgetControl::horzScroller() const
{
    return 0;
}

WidgetControlScroller *WidgetControl::vertScroller() const
{
    return 0;
}

void WidgetControl::enterEditor(Time time, int x, int y,
				int x_root, int y_root)
{
    /* relax */
}

void WidgetControl::leaveEditor(bool cancel)
{
    /* relax */
}

void WidgetControl::widgetChanged()
{
    /* relax */
}

/****************************************************************************/

Widget::Widget(Connection &conn, WidgetScroller *horzScroller,
				 WidgetScroller *vertScroller)
    : widget_control(0), needs_layout(false),
      horz_scroller(horzScroller), vert_scroller(vertScroller), conn(conn)
{
    if (horzScroller)
	assert(&horzScroller->widget == this);
    if (vertScroller)
	assert(&vertScroller->widget == this);
}

Widget::~Widget()
{
    if (!widget_control)
	return;

    WidgetControl *c = widget_control;
    widget_control = 0;
    c->setWidget(0);
}

ConnectionColors const &Widget::colors() const
{
    return conn.colors();
}

WidgetControl *Widget::control() const
{
    return widget_control;
}

void Widget::setControl(WidgetControl *control)
{
    /* May be called by a destructing control or on a destructing widget with
       control == 0, so don't call a virtual method on either us or
       widget_control in that case */

    if (control == widget_control)
	return;

    Region old_wpos, new_wpos;

    if (control) {
	assert(&control->conn == &conn);
	if (hook())
	    old_wpos = hook()->innerDimensions();
    }

    if (widget_control) {
	WidgetControl *c = widget_control;
	widget_control = 0;
	c->setWidget(0);
    }

    widget_control = control;
    needs_layout = true;

    if (control) {
	control->setWidget(this);
	control->requestChildLayout();

	if (hook())
	    new_wpos = hook()->innerDimensions();
	if (new_wpos.x != old_wpos.x || new_wpos.y != old_wpos.y)
	    windowPositionChanged();
    }
}

std::vector<Widget *> Widget::children() const
{
    return std::vector<Widget *>();
}

std::vector<Focussable *> Widget::focussables()
{
    return _focussables;
}

void Widget::setFocussables(const std::vector<Focussable *> &focussables)
{
    if (focussables == _focussables)
	return;

    _focussables = focussables;
    if (control())
	control()->focussablesChanged();
}

std::string Widget::statusMessage() const
{
    return status_message;
}

void Widget::setStatusMessage(const std::string &message)
{
    if (message == status_message)
	return;

    status_message = message;
    if (control() && control()->hook())
	control()->hook()->updateStatusMessage();
}

void Widget::paint(const Region &region, X::Pixmap pixmap)
{
}

void Widget::invalidate()
{
    if (!control() || !control()->hook())
	return;

    control()->hook()->invalidate(Region(x(), y(), width(), height()));
}

void Widget::enterEvent(Time /*time*/, int /*x*/, int /*y*/,
			int /*x_root*/, int /*y_root*/,
			unsigned int /*state*//*, char*/ /*is_hint*/)
{
    /* relax */
}

void Widget::leaveEvent(Time /*time*/, int /*x*/, int /*y*/,
			int /*x_root*/, int /*y_root*/,
			unsigned int /*state*//*, char*/ /*is_hint*/)
{
    /* relax */
}

void Widget::motionEvent(Time /*time*/, int /*x*/, int /*y*/,
			int /*x_root*/, int /*y_root*/,
			unsigned int /*state*/, char /*is_hint*/)
{
    /* relax */
}

void Widget::buttonPressEvent(
    Time /*time*/, int /*x*/, int /*y*/,
    int /*x_root*/, int /*y_root*/,
    unsigned int /*state*/, unsigned int /*button*/)
{
    /* relax */
}

void Widget::buttonReleaseEvent(
    Time /*time*/, int /*x*/, int /*y*/,
    int /*x_root*/, int /*y_root*/,
    unsigned int /*state*/, unsigned int /*button*/)
{
    /* relax */
}

int Widget::x() const
{
    if (!control())
	return 0;

    return control()->dimensions().x;
}

int Widget::y() const
{
    if (!control())
	return 0;

    return control()->dimensions().y;
}

int Widget::width() const
{
    if (!control())
	return 0;

    return control()->dimensions().width;
}

int Widget::height() const
{
    if (!control())
	return 0;

    return control()->dimensions().height;
}

X::Window Widget::window() const
{
    if (!control())
	return X::Window();

    return control()->window();
}

X::GC Widget::gc() const
{
    if (!control())
	return X::GC();

    return control()->gc();
}

WidgetWindowHook *Widget::hook() const
{
    if (!control())
	return 0;

    return control()->hook();
}

Widget *Widget::widgetAt(int x, int y)
{
    std::vector<Widget *> cached_children = children();

    for (std::vector<Widget *>::const_iterator i = cached_children.begin();
	 i != cached_children.end(); i++)
	if (x >= (*i)->x() && x < (*i)->x() + (*i)->width() &&
	    y >= (*i)->y() && y < (*i)->y() + (*i)->height())
	    return (*i)->widgetAt(x, y);

    if (x >= this->x() && x < this->x() + this->width() &&
	y >= this->y() && y < this->y() + this->height())
	return this;

    return 0;
}

SizeRequirements Widget::sizeRequirements() const
{
    return size_requirements;
}

void Widget::setSizeRequirements(const SizeRequirements &reqs)
{
    if (reqs == size_requirements)
	return;

    size_requirements = reqs;
    if (control())
	control()->setSizeRequirements(reqs);
}

void Widget::requestLayout()
{
    needs_layout = true;
    invalidate();
    /* It may not be necessary to invalidate the entire widget on every layout
       change, but it won't hurt much either, and we're on the safe side. */

    if (control())
	control()->requestChildLayout();
}

void Widget::recursivelyUpdateLayout()
{
    if (needs_layout)
	updateLayout();
    needs_layout = false;
}

void Widget::updateLayout()
{
    /* relax */
}

bool Widget::isMenu() const
{
    return false;
}

void Widget::setMnemonicVisibility(bool visible)
{
    /* relax */
}

void Widget::windowPositionChanged()
{
    /* relax */
}

WidgetScroller *Widget::horzScroller() const
{
    return horz_scroller;
}

WidgetScroller *Widget::vertScroller() const
{
    return vert_scroller;
}
