/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#ifndef TOOLKIT_WIDGET_H
#define TOOLKIT_WIDGET_H

#include "util/region.h"
#include "util/size_reqs.h"
#include "x_abstraction.h"

namespace Toolkit {

class Widget;
class WidgetControlScroller;
class WidgetControl;
class Connection;
class WidgetWindowHook;
class ConnectionColors;

class WidgetScroller
{
public:
    WidgetScroller(Widget &widget);
    virtual ~WidgetScroller();
    Widget &widget;

    int range() const;
    int pageSize() const;
    int position() const;
    void setPosition(int position);

    virtual int lineSize() const = 0;

protected:
    void setRange(int range);
    void setPageSize(int pageSize);

private:
    int _range, page_size, _position;
    WidgetControlScroller *control_scroller() const;

    virtual void updatePosition() = 0;
};

class WidgetControlScroller
{
public:
    WidgetControlScroller(WidgetControl &control);
    virtual ~WidgetControlScroller();
    WidgetControl &control;

    virtual void updateRange() = 0;
    virtual void updatePageSize() = 0;
    virtual void updatePosition() = 0;
};

class Focussable
{
public:
    virtual void keyPressEvent(Time time, int x, int y,
			       int x_root, int y_root,
			       unsigned int state, unsigned int keycode,
			       const std::string &text, KeySym keysym);
    virtual void keyReleaseEvent(Time time, int x, int y,
				 int x_root, int y_root,
				 unsigned int state, unsigned int keycode);
    virtual void focusInEvent();
    virtual void focusOutEvent();
};

class WidgetControl
{
    Widget *_widget;
    Region _dimensions;
    SizeRequirements size_requirements;
    bool needs_child_layout;
public:
    WidgetControl(Connection &conn);
    virtual ~WidgetControl();
    Connection &conn;
    Widget *widget() const;
    void setWidget(Widget *widget);

    SizeRequirements sizeRequirements() const;
    void setSizeRequirements(const SizeRequirements &reqs);
    void requestChildLayout();
    void updateChildLayout();

    virtual X::Window window() const = 0;
    virtual X::GC gc() const = 0;
    virtual WidgetWindowHook *hook() const = 0;

    Region dimensions() const;
    Region screenDimensions() const;

    virtual WidgetControlScroller *horzScroller() const;
    virtual WidgetControlScroller *vertScroller() const;

    virtual void focus(Focussable *focussable) = 0;
    virtual void focussablesChanged() = 0;

    virtual void enterEditor(Time time, int x, int y,
			     int x_root, int y_root);
    virtual void leaveEditor(bool cancel);

protected:
    void setDimensions(const Region &dimensions);

private:
    virtual void sizeRequirementsChanged() = 0;
    virtual void propagateChildLayoutRequest() = 0;
    virtual void widgetChanged();
};

class Widget
{
    WidgetControl *widget_control;
    SizeRequirements size_requirements;
    std::vector<Focussable *> _focussables;
    std::string status_message;
    bool needs_layout;
    WidgetScroller *const horz_scroller, *const vert_scroller;
public:
    Widget(Connection &conn, WidgetScroller *horzScroller = 0,
			     WidgetScroller *vertScroller = 0);
    virtual ~Widget();
    Connection &conn;
    ConnectionColors const &colors() const;
    WidgetControl *control() const;
    void setControl(WidgetControl *control);

    int x() const;
    int y() const;
    int width() const;
    int height() const;
    X::Window window() const;
    X::GC gc() const;
    WidgetWindowHook *hook() const;

    Widget *widgetAt(int x, int y);

    SizeRequirements sizeRequirements() const;
    std::vector<Focussable *> focussables();
    std::string statusMessage() const;

    void requestLayout();
    virtual void recursivelyUpdateLayout();

    virtual bool isMenu() const;
    virtual void setMnemonicVisibility(bool visible);
    virtual void windowPositionChanged();

    WidgetScroller *horzScroller() const;
    WidgetScroller *vertScroller() const;

    virtual void paint(const Region &region, X::Pixmap pixmap);
    void invalidate();

    virtual void enterEvent(Time time, int x, int y, int x_root, int y_root,
			    unsigned int state/*, char is_hint*/);
    virtual void leaveEvent(Time time, int x, int y, int x_root, int y_root,
			    unsigned int state/*, char is_hint*/);
    virtual void motionEvent(Time time, int x, int y, int x_root, int y_root,
			     unsigned int state, char is_hint);
    virtual void buttonPressEvent(Time time, int x, int y,
				  int x_root, int y_root,
				  unsigned int state, unsigned int button);
    virtual void buttonReleaseEvent(Time time, int x, int y,
				    int x_root, int y_root,
				    unsigned int state, unsigned int button);

protected:
    void setSizeRequirements(const SizeRequirements &reqs);
    void setFocussables(const std::vector<Focussable *> &focussables);
    void setStatusMessage(const std::string &message);

private:
    virtual std::vector<Widget *> children() const;
    virtual void updateLayout();
};

} /* namespace Toolkit */

#endif
