/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#ifndef X_ABSTRACTION_H
#define X_ABSTRACTION_H

#include <string>
#include <vector>
#include <X11/Xutil.h>

#define ButtonMask (Button1Mask | Button2Mask | Button3Mask \
		      | Button4Mask | Button5Mask)

namespace X {

class Display;
class Screen;
class GC;
class FontSet;
class InputMethod;
class InputContext;
class Font;
class Colormap;
class Drawable;
class Window;
class Pixmap;
class Cursor;

class Display
{
    ::Display *display;
public:
    static Display open(const char *display_name);

    Display();
    Display(::Display *display);
    operator ::Display *() const;
    bool isNull() const;

    int close();

    int connectionNumber();
    Screen defaultScreen();
    Window defaultRootWindow();

    int nextEvent(XEvent *event_return);
    int peekEvent(XEvent *event_return);
    int maskEvent(long event_mask, XEvent *event_return);
    bool checkMaskEvent(long event_mask, XEvent *event_return);
    bool checkTypedEvent(int event_type, XEvent *event_return);

    int flush();
    int sync(bool discard);
    int eventsQueued(int mode);
    int pending();

    Cursor createFontCursor(unsigned int shape);
    Cursor createPixmapCursor(
	::Pixmap source, ::Pixmap mask,
	XColor *foreground_color, XColor *background_color,
	unsigned int x, unsigned int y);
    Cursor createGlyphCursor(
	::Font source_font, ::Font mask_font,
	unsigned int source_char, unsigned int mask_char,
	XColor *foreground_color, XColor *background_color);

    Font loadFont(const char *name);
    XFontStruct *queryFont(XID font_ID);
    XFontStruct *loadQueryFont(const char *name);
    int freeFont(XFontStruct *font_struct);

    int ungrabPointer(Time time);
    int changeActivePointerGrab(unsigned int event_mask,
				::Cursor cursor, Time time);

    FontSet createFontSet(const char *base_font_name_list,
			  char ***missing_charset_list_return,
			  int *missing_charset_count_return,
			  char **def_string_return);
    InputMethod openIM(struct _XrmHashBucketRec *db,
		       char *res_name, char *res_class);

    Atom internAtom(const char *atom_name, Bool only_if_exists);
    Status internAtoms(char **names, int count, Bool only_if_exists,
		       Atom *atoms_return);
    char *getAtomName(Atom atom);
    Status getAtomNames(Atom *atoms, int count, char **names_return);

    int mbTextListToTextProperty(char **list, int count,
	XICCEncodingStyle style, XTextProperty *text_prop_return);
    int wcTextListToTextProperty(wchar_t **list, int count,
	XICCEncodingStyle style, XTextProperty *text_prop_return);
    int utf8TextListToTextProperty(char **list, int count,
	XICCEncodingStyle style, XTextProperty *text_prop_return);
    int mbTextPropertyToTextList(XTextProperty *text_prop,
	char ***list_return, int *count_return);
    int wcTextPropertyToTextList(XTextProperty *text_prop,
	wchar_t ***list_return, int *count_return);
    int utf8TextPropertyToTextList(XTextProperty *text_prop,
	char ***list_return, int *count_return);

    int setSelectionOwner(Atom selection, ::Window owner, Time time);
    Window getSelectionOwner(Atom selection);
    int convertSelection(Atom selection, Atom target,
			 Atom property, ::Window requestor, Time time);

    std::vector<XVisualInfo> getVisualInfo(long vinfo_mask,
					   XVisualInfo *vinfo_template);

    std::vector<XPixmapFormatValues> listPixmapFormats();
    int imageByteOrder();
    int bitmapBitOrder();
    int bitmapPad();
    int bitmapUnit();
};

class Screen
{
    ::Display *display;
    int screen_number;
public:
    Screen();
    Screen(::Display *display, int screen_number);
    operator int() const;

    unsigned long blackPixel();
    unsigned long whitePixel();
    Window rootWindow();
    int defaultDepth();
    Visual *defaultVisual();
    Colormap defaultColormap();

    Status matchVisualInfo(int depth, int c_class, XVisualInfo *vinfo_return);

    int displayHeight();
    int displayHeightMM();
    int displayWidth();
    int displayWidthMM();
};

class GC
{
    ::Display *display;
    ::GC gc;
public:
    GC();
    GC(::Display *display, ::GC gc);
    operator ::GC() const;
    bool isNull() const;

    int setState(unsigned long foreground, unsigned int background,
		 int function, unsigned long plane_mask);
    int setFunction(int function);
    int setPlaneMask(unsigned long plane_mask);
    int setForeground(unsigned long foreground);
    int setBackground(unsigned long background);

    static int copy(GC src, GC dest, unsigned long valuemask);
    int change(unsigned long valuemask, XGCValues *values);
    Status getValues(unsigned long valuemask, XGCValues *values_return);
    int free();
    GContext gContext();

    int setFont(::Font font);

    int setClipOrigin(int clip_x_origin, int clip_y_origin);
    int setClipMask(::Pixmap pixmap);
    int setClipRectangles(int clip_x_origin, int clip_y_origin,
			  XRectangle rectangles[], int n, int ordering);
};

class FontSet
{
    ::Display *display;
    XFontSet font_set;
    XFontSetExtents *_extents;
public:
    FontSet();
    FontSet(::Display *display, XFontSet font_set);
    operator XFontSet() const;
    bool isNull() const;

    int fonts(XFontStruct ***font_struct_list_return,
	      char ***font_name_list_return);
    char *baseFontNameList();
    char *locale();
    bool contextDependentDrawing();
    bool contextualDrawing();
    bool directionalDependentDrawing();

    XFontSetExtents *extents() const;

    int mbTextEscapement(const std::string &string);
    int wcTextEscapement(const std::wstring &string);
    int utf8TextEscapement(const std::string &string);

    int mbTextExtents(const std::string &string,
		      XRectangle *overall_ink_return,
		      XRectangle *overall_logical_return);
    int wcTextExtents(const std::wstring &string,
		      XRectangle *overall_ink_return,
		      XRectangle *overall_logical_return);
    int utf8TextExtents(const std::string &string,
			XRectangle *overall_ink_return,
			XRectangle *overall_logical_return);

    Status mbTextPerCharExtents(const std::string &string,
				XRectangle *ink_array_return,
				XRectangle *logical_array_return,
				int array_size, int *num_chars_return,
				XRectangle *overall_ink_return,
				XRectangle *overall_logical_return);
    Status wcTextPerCharExtents(const std::wstring &string,
				XRectangle *ink_array_return,
				XRectangle *logical_array_return,
				int array_size, int *num_chars_return,
				XRectangle *overall_ink_return,
				XRectangle *overall_logical_return);
    Status utf8TextPerCharExtents(const std::string &string,
				  XRectangle *ink_array_return,
				  XRectangle *logical_array_return,
				  int array_size, int *num_chars_return,
				  XRectangle *overall_ink_return,
				  XRectangle *overall_logical_return);

    void free();

    /* convenience functions */
    int ascent() const;
    int descent() const;
};

class InputMethod
{
    XIM im;
public:
    InputMethod();
    InputMethod(XIM im);
    operator XIM() const;
    bool isNull() const;

    Status close();
    // char *setValues(...);
    // char *getValues(...);
    ::Display *display();
    char *locale();

    // InputContext createIC(...);
};

class InputContext
{
    XIC ic;
public:
    InputContext();
    InputContext(XIC ic);
    operator XIC() const;
    bool isNull() const;

    void destroy();
    InputMethod IM();

    void setFocus();
    void unsetFocus();

    std::string mbReset();
    std::wstring wcReset();
    std::string utf8Reset();

    // char *setValues(...);
    // char *getValues(...);

    std::string mbLookupString(const XKeyPressedEvent *event,
			       KeySym *keysym_return, Status *status_return);
    std::wstring wcLookupString(const XKeyPressedEvent *event,
				KeySym *keysym_return, Status *status_return);
    std::string utf8LookupString(const XKeyPressedEvent *event,
				 KeySym *keysym_return, Status *status_return);
};

class Font
{
    ::Display *display;
    ::Font font;
public:
    Font();
    Font(::Display *display, ::Font font);
    operator ::Font() const;

    int unload();
};

class Colormap
{
    ::Display *display;
    ::Colormap colormap;
public:
    Colormap();
    Colormap(::Display *display, ::Colormap colormap);
    operator ::Colormap() const;

    Colormap copyAndFree();
    int free();

    Status allocColor(XColor *screen_in_out);
    Status allocNamedColor(const char *color_name, XColor *screen_def_return,
			   XColor *exact_def_return);
    Status allocColorCells(Bool contig,
			   unsigned long longplane_masks_return[],
			   unsigned int nplanes,
			   unsigned long pixels_return[],
			   unsigned int npixels);
    Status allocColorPlanes(Bool contig, unsigned long pixels_return[],
			    int ncolors, int nreds, int ngreens, int nblues,
			    unsigned long *rmask_return,
			    unsigned long *gmask_return,
			    unsigned long *bmask_return);
    int freeColors(unsigned long pixels[], int npixels, unsigned long planes);
};

class Drawable
{
    ::Display *display;
    ::Drawable d;
public:
    Drawable();
    Drawable(::Display *display, ::Drawable d);
    operator ::Drawable() const;
    bool isNull() const;

    GC createGC(unsigned long valuemask, XGCValues *values);

    int drawArc(::GC gc, int x, int y, unsigned int width, unsigned int height,
		int angle1, int angle2);
    int drawArcs(::GC gc, XArc *arcs, int narcs);
    int drawImageString(::GC gc, int x, int y, const std::string &string);
    int drawImageString16(::GC gc, int x, int y, XChar2b *string, int length);
    void mbDrawImageString(XFontSet font_set, ::GC gc,
			   int x, int y, const std::string &string);
    void wcDrawImageString(XFontSet font_set, ::GC gc,
			   int x, int y, const std::wstring &string);
    void utf8DrawImageString(XFontSet font_set, ::GC gc,
			     int x, int y, const std::string &string);
    int drawLine(::GC gc, int x1, int y1, int x2, int y2);
    int drawLines(::GC gc, XPoint *points, int npoints, int mode);
    int drawPoint(::GC gc, int x, int y);
    int drawPoints(::GC gc, XPoint *points, int npoints, int mode);
    int drawRectangle(::GC gc, int x, int y,
		      unsigned int width, unsigned int height);
    int drawRectangles(::GC gc, XRectangle *rectangles, int nrectangles);
    int drawSegments(::GC gc, XSegment *segments, int nsegments);
    int drawString(::GC gc, int x, int y, const std::string &string);
    int drawString16(::GC gc, int x, int y, XChar2b *string, int length);
    void mbDrawString(XFontSet font_set, ::GC gc,
		      int x, int y, const std::string &string);
    void wcDrawString(XFontSet font_set, ::GC gc,
		      int x, int y, const std::wstring &string);
    void utf8DrawString(XFontSet font_set, ::GC gc,
			int x, int y, const std::string &string);
    int drawText(::GC gc, int x, int y, XTextItem *items, int nitems);
    int drawText16(::GC gc, int x, int y, XTextItem16 *items, int nitems);

    int fillArc(::GC gc, int x, int y, unsigned int width, unsigned int height,
		int angle1, int angle2);
    int fillArcs(::GC gc, XArc *arcs, int narcs);
    int fillPolygon(::GC gc, XPoint *points, int npoints, int shape, int mode);
    int fillRectangle(::GC gc, int x, int y,
		      unsigned int width, unsigned int height);
    int fillRectangles(::GC gc, XRectangle *rectangles, int nrectangles);

    int readBitmapFile(
	const char *filename,
	unsigned int *width_return, unsigned int *height_return,
	Pixmap *bitmap_return, int *x_hot_return, int *y_hot_return);
    Pixmap createPixmapFromBitmapData(
	char *data, unsigned int width, unsigned int height,
	unsigned long fg, unsigned int bg, unsigned int depth);
    Pixmap createBitmapFromData(
	const char *data, unsigned int width, unsigned int height);

    static int copyArea(
	Drawable src, Drawable dest, ::GC gc, int src_x, int src_y,
	unsigned int width, unsigned int height, int dest_x, int dest_y);

    static int copyPlane(
	Drawable src, Drawable dest, ::GC gc, int src_x, int src_y,
	unsigned int width, unsigned int height, int dest_x, int dest_y,
	unsigned long plane);

    Pixmap createPixmap(unsigned int width, unsigned int height,
			unsigned int depth);

    Status queryBestCursor(unsigned int width, unsigned int height,
			   unsigned int *width_return,
			   unsigned int *height_return);

    int putImage(::GC gc, XImage *image, int src_x, int src_y, int dest_x,
		 int dest_y, unsigned int width, unsigned int height);
    XImage *getImage(int x, int y, unsigned int width, unsigned int height,
		     unsigned long plane_mask, int format);
    XImage *getSubImage(int x, int y, unsigned int width, unsigned int height,
			unsigned long plane_mask, int format,
			XImage *dest_image, int dest_x, int dest_y);
};

class Window : public Drawable
{
    ::Display *display;
    ::Window w;
public:
    Window();
    Window(::Display *display, ::Window w);
    operator ::Window() const;

    int changeAttributes(unsigned long valuemask,
			 XSetWindowAttributes *attributes);
    int setBackground(unsigned long background_pixel);
    int setBackgroundPixmap(::Pixmap background_pixmap);
    int setBorder(unsigned long border_pixel);
    int setBorderPixmap(::Pixmap border_pixmap);
    int setColormap(::Colormap colormap);

    int clearArea(int x, int y, unsigned width, unsigned height,
		  bool exposures);
    int clear();

    Window createWindow(
	int x, int y, unsigned int width, unsigned int height,
	unsigned int border_width, int depth,
	unsigned int _class, Visual *visual,
	unsigned long valuemask, XSetWindowAttributes *attributes);
    Window createSimpleWindow(
	int x, int y, unsigned int width, unsigned int height,
	unsigned int border_width, unsigned long border,
	unsigned long background);

    int selectInput(long event_mask);

    int map();
    int mapRaised();
    int mapSubwindows();

    int unmap();
    int unmapSubwindows();

    int event(long event_mask, XEvent *event_return);
    bool checkEvent(long event_mask, XEvent *event_return);
    bool checkTypedEvent(int event_type, XEvent *event_return);

    Colormap createColormap(Visual *visual, int alloc);

    int defineCursor(::Cursor cursor);
    int undefineCursor();

    int configure(unsigned value_mask, XWindowChanges *changes);
    int move(int x, int y);
    int resize(unsigned width, unsigned height);
    int moveResize(int x, int y, unsigned width, unsigned height);
    int setBorderWidth(unsigned width);

    int destroy();
    int destroySubwindows();

    int grabPointer(bool owner_events, unsigned int event_mask,
		    int pointer_mode, int keyboard_mode,
		    ::Window confine_to, ::Cursor cursor, Time time);

    int getProperty(Atom property, long long_offset, long long_length,
		    Bool _delete, Atom req_type, Atom *actual_type_return,
		    int *actual_format_return, unsigned long *nitems_return,
		    unsigned long *bytes_after_return,
		    unsigned char **prop_return);
    Atom *listProperties(int *num_prop_return);
    int changeProperty(Atom property, Atom type, int format, int mode,
		       const unsigned char *data, int nelements);
    int rotateProperties(Atom properties[], int num_prop, int npositions);
    int deleteProperty(Atom property);

    void setTextProperty(XTextProperty *text_prop, Atom property);
    Status getTextProperty(XTextProperty *text_prop_return, Atom property);

    void setWMProperties(XTextProperty *window_name, XTextProperty *icon_name,
			 char **argv, int argc,
			 XSizeHints *normal_hints, XWMHints *wm_hints,
			 XClassHint *class_hints);
    void mbSetWMProperties(const std::string &window_name,
			   const std::string &icon_name,
			   char *argv[], int argc,
			   XSizeHints *normal_hints, XWMHints *wm_hints,
			   XClassHint *class_hints);
    void utf8SetWMProperties(const std::string &window_name,
			     const std::string &icon_name,
			     char *argv[], int argc,
			     XSizeHints *normal_hints, XWMHints *wm_hints,
			     XClassHint *class_hints);

    void setWMName(XTextProperty *text_prop);
    Status getWMName(XTextProperty *text_prop_return);
    int storeName(const char *window_name);
    Status fetchName(char **window_name_return);

    void setWMIconName(XTextProperty *text_prop);
    Status getWMIconName(XTextProperty *text_prop_return);
    int setIconName(const char *icon_name);
    Status getIconName(char **icon_name_return);

    void setWMNormalHints(XSizeHints *hints);
    Status getWMNormalHints(XSizeHints *hints_return, long *supplied_return);
    void setWMSizeHints(XSizeHints *hints, Atom property);
    Status getWMSizeHints(XSizeHints *hints_return, long *supplied_return,
			  Atom property);

    int setWMHints(XWMHints *wmhints);
    XWMHints *getWMHints();

    int setClassHint(XClassHint *class_hints);
    Status getClassHint(XClassHint *class_hints_return);

    int setTransientForHint(::Window prop_window);
    Status getTransientForHint(::Window *prop_window_return);

    Status setWMProtocols(Atom *protocols, int count);
    Status getWMProtocols(Atom **protocols_return, int *count_return);

    Status sendEvent(Bool propagate, long event_mask, XEvent *event_send);
};

class Pixmap : public Drawable
{
    ::Display *display;
    ::Pixmap pixmap;

public:
    Pixmap();
    Pixmap(::Display *display, ::Pixmap pixmap);
    operator ::Pixmap() const;

    int writeBitmapFile(const char *filename,
			unsigned int width, unsigned int height,
			int x_hot, int y_hot);
    int free();
};

class Cursor : public Drawable
{
    ::Display *display;
    ::Cursor cursor;
public:
    Cursor();
    Cursor(::Display *display, ::Cursor cursor);
    operator ::Cursor() const;

    int recolor(XColor *foreground_color, XColor *background_color);
    int free();
};

} /* namespace X */

#endif
